/*******************************************************************************
 * Copyright (c) 2004-2008 Andras Schmidt, Andras Balogh, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Andras Schmidt, Andras Balogh, Istvan Rath - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.core.notification;

import java.util.Collection;

import org.eclipse.viatra2.core.ICoreNotificationListener;
import org.eclipse.viatra2.core.IModelElement;

/**
 * This is an interface of a notification object. A notification object is sent,
 * whenever the modelspace changes. It contains string constants which identify
 * the various notification types supported by the Viatra Core, as well as some
 * utility getter functions.
 * 
 * @author Andras Schmidt
 * 
 *         Modified on 2006.08.08 by Istvan Rath: - added transaction-specific
 *         ID's (TA prefix) - cleaned up javadoc
 */
public interface ICoreNotificationObject {
	/**
	 * Signals that a "user mark" (checkpoint) has been added to the undo stack.
	 * This is only used by the undo manager.
	 */
	public static final String USER_MARK = "user mark";

	/**
	 * Means that one atomic step is finished. This defines a consistent point,
	 * where undo can be stopped.
	 */
	public static final String ACTION_ATOMIC_STEP_READY = "atomic step ready";

	/**
	 * Means that an entity was created. Parameters: IModelElement container
	 * entity (may be null, then containment add will come in an other
	 * notification) IModelElement of newly created entity String name of newly
	 * created entity String value of newly created entity
	 */
	public static final String ACTION_CREATE_ENTITY = "create entity";

	/**
	 * Means that an entity was deleted Parameters: String ID of deleted entity
	 * String name of deleted entity String value of deleted entity String ID of
	 * parent
	 */
	public static final String ACTION_DELETE_ENTITY = "delete entity";

	/**
	 * Means that a function was created Parameters: String ID of container
	 * entity (may be null, then containment add will come in an other
	 * notification) String ID of from entity String ID of to entity String ID
	 * of newly created function String name of newly created function
	 */
	public static final String ACTION_CREATE_RELATION = "create relation";

	/**
	 * Means that an entity was deleted Parameters: String ID of deleted entity
	 * String name of deleted entity String value of deleted entity String ID of
	 * parent
	 */
	public static final String ACTION_DELETE_RELATION = "delete relation";

	/**
	 * Means that a relation's source model element was changed.
	 */
	public static final String ACTION_SET_RELATION_FROM = "set relation from";

	/**
	 * Means that a relation's target model element was changed.
	 */
	public static final String ACTION_SET_RELATION_TO = "set relation to";

	/**
	 * Means that an instanceOf implicit relation was created.
	 */
	public static final String ACTION_CREATE_INSTANCEOF = "create instanceof";

	/**
	 * Means that an instanceOf implicit relation was deleted.
	 */
	public static final String ACTION_DELETE_INSTANCEOF = "delete instanceof";

	/**
	 * Means that an supertypeOf implicit relation was created.
	 */
	public static final String ACTION_CREATE_SUPERTYPEOF = "create supertypeof";

	/**
	 * Means that an supertypeOf implicit relation was deleted.
	 */
	public static final String ACTION_DELETE_SUPERTYPEOF = "delete supertypeof";

	/**
	 * Means that a containment relation was deleted. Note: the state of the
	 * modelspace is not consistent!
	 */
	public static final String ACTION_DELETE_CONTAINMENT = "delete containment";

	/**
	 * Means that the value of an entity was changed.
	 */
	public static final String ACTION_SET_VALUE = "set value";

	/**
	 * Means that the view info of an element was changed.
	 */
	public static final String ACTION_SET_VIEW_INFO = "set view info";

	/**
	 * Means that the name of an element was changed.
	 */
	public static final String ACTION_SET_NAME = "set name";

	/**
	 * Means that the isFinalType attribute of an element was changed.
	 */
	public static final String ACTION_SET_ISFINALTYPE = "set isfinaltype";

	/**
	 * Means that the isAny attribute of an element was changed.
	 */
	public static final String ACTION_SET_ISANY = "set isany";

	/**
	 * Means that an element was moved to an entity, with removing from all
	 * others.
	 */
	public static final String ACTION_MOVE_ELEMENT_TO = "move element to";

	/**
	 * Means that an undoable transaction has begun.
	 */
	public static final String TA_UNDOABLE_TRANSACTION_BEGIN = "begin undoable transaction";

	/**
	 * Means that a non-undoable transaction has begun.
	 */
	public static final String TA_TRANSACTION_BEGIN = "begin transaction";

	/**
	 * Means that a sub-transaction has begun.
	 */
	public static final String TA_SUBTRANSACTION_BEGIN = "begin subtransaction";

	/**
	 * Signals the end of a sub-transaction.
	 */
	public static final String TA_SUBTRANSACTION_END = "end subtransaction";

	/**
	 * Signals the end of a transaction.
	 */
	public static final String TA_TRANSACTION_END = "end transaction";

	/**
	 * Signals that an undo operation has begun.
	 */
	public static final String TA_UNDO_BEGIN = "begin undo";

	/**
	 * Signals that an undo operation has ended.
	 */
	public static final String TA_UNDO_END = "end undo";

	/**
	 * Signals that a relation's inverse has been set.
	 */
	public static final String ACTION_SET_RELATION_INVERSE = "set inverse";
	
	/**
	 * The relation's isAggregation flag has been set.
	 */
	public static final String ACTION_SET_RELATION_ISAGGREGATION = "set isAggregation";
	
	/**
	 * The relation's multiplicity has been set.
	 */
	public static final String ACTION_SET_RELATION_MULTIPLICITY = "set multiplicity";

	/**
	 * Signals that the transaction has been aborted.
	 */
	public static final String TA_TRANSACTION_ABORT = "transaction abort";

	/**
	 * Signals that the subtransaction has been aborted.
	 */
	public static final String TA_SUBTRANSACTION_ABORT = "subtransaction abort";

	/**
	 * Converts this notification to a human readable string representation.
	 */
	public String toString();

	/**
	 * Get the type of this action. The returned string constant will always be
	 * one of the types defined in the interface.
	 */
	public String getActionType();

	/**
	 * Get the objects receiving this notification instance.
	 */
	public Collection<ICoreNotificationListener> getListeners();

	/**
	 * Get a collection of model elements which were involved in the
	 * modification which this notification object is signalling.
	 */
	public Collection<IModelElement> getNotifiedObjects();
}
