/*******************************************************************************
 * Copyright (c) 2004-2008 Andras Schmidt, Andras Balogh, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Andras Schmidt, Andras Balogh, Istvan Rath - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.core.simple;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.locks.ReadWriteLock;

import org.eclipse.viatra2.copier.ModelCopy;
import org.eclipse.viatra2.copier.ModelCopyException;
import org.eclipse.viatra2.core.EDeleteSemantics;
import org.eclipse.viatra2.core.EModelElementProperty;
import org.eclipse.viatra2.core.EMultiplicityKind;
import org.eclipse.viatra2.core.ICoreNotificationListener;
import org.eclipse.viatra2.core.IEntity;
import org.eclipse.viatra2.core.IModelElement;
import org.eclipse.viatra2.core.IModelManager;
import org.eclipse.viatra2.core.IModelSpace;
import org.eclipse.viatra2.core.IRelation;
import org.eclipse.viatra2.core.constraint.EConstraint;
import org.eclipse.viatra2.core.notification.ICoreNotificationObject;
import org.eclipse.viatra2.core.simple.constraint.ConstraintTranslator;
import org.eclipse.viatra2.core.simple.notification.NotificationObject;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectAtomicStepReady;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectCreateEntity;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectCreateInstanceOf;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectCreateRelation;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectCreateSupertypeOf;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectDeleteContainment;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectDeleteEntity;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectDeleteInstanceOf;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectDeleteRelation;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectDeleteSupertypeOf;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectMoreAtomicsToOneAtomEnd;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectMoreAtomicsToOneAtomStart;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectMoveTo;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectSetIsAny;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectSetIsFinalType;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectSetName;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectSetRelationFrom;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectSetRelationInverse;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectSetRelationIsAggregation;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectSetRelationMultiplicity;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectSetRelationTo;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectSetValue;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectSetViewInfo;
import org.eclipse.viatra2.errors.VPMCoreException;
import org.eclipse.viatra2.errors.VPMCoreNullParameterException;
import org.eclipse.viatra2.errors.VPMCoreRuntimeException;
import org.eclipse.viatra2.errors.VPMRuntimeException;
import org.eclipse.viatra2.logger.Logger;

/**
 * Class that manages the SimpleCore modelSpace.
 * 
 * @author Andras Schmidt, Istvan Rath
 * 
 */
public class SimpleModelManager implements IModelManager {
	private SimpleModelSpace modelSpace;

	private SimpleNotificationManager nManager;

	private Logger logger;

	/**
	 * The lock that must be locked when reading or updating the model
	 */
	private ReadWriteLock lock = null;

	// public static final String topEntityName="vpm.entity";
	// public static final String topRelationName="vpm.entity.relation";
	// public static final String topFunctionName="vpm.entity.function";
	// / Exception messages
	public static final String EXCEPTION_TYPE_BETWEEN_DIFFERENT_ELEMENTS = "type between different elements cannot be added: ";

	public static final String EXCEPTION_INSTANCE_BETWEEN_DIFFERENT_ELEMENTS = "instance between different elements cannot be added: ";

	public static final String EXCEPTION_ALREADY_SUPERTYPE = "new supertype: already supertype (may be recursive): ";

	public static final String EXCEPTION_ALREADY_INSTANCE = "new instanceof: already instance (may be recursive): ";

	public static final String EXCEPTION_NEW_SUPERTYPE_CAUSES_CIRCLE = "new supertype: relation would cause circle: ";

	public static final String EXCEPTION_DELETE_SUPERTYPE_EDITOR_NO_RELATION = "deleteSupetypeOf(editor semantics): no direct supertypeOf relation is found: ";

	public static final String EXCEPTION_DELETE_INSTANCEOF_NO_RELATION = "deleteInstanceOf(editor semantics): no direct instanceOf relation is found: ";

	public static final String EXCEPTION_NEW_SUPERTYPE_MAKES_EXISTING_RELATION_IMPLICIT = "new supertype(editor semantics): relation would cause existing relation implicit: ";

	public static final String EXCEPTION_NEW_INSTANCE_MAKES_EXISTING_RELATION_IMPLICIT = "new instance(editor semantics): relation would cause existing relation implicit: ";

	private static final String parameterNullNotAllowed = "Parameter null not allowed";

	private static final String WARNING_DELETE_NOT_SUPERTYPE = "deleteSupertypeOf(machine semantics): no supertypeof between elements: ";

	/**
	 * Check the string if it can be the name of an element. The name of an
	 * element must not contain '.'
	 * 
	 * @param name
	 *            the requested name for new element
	 * @throws VPMCoreException
	 *             if the name contains '.'
	 */
	private void checkElemName(String name) throws VPMCoreException {
		if (name.indexOf('.') >= 0)
			throw new VPMCoreException(
					"Name of element must not contain character dot('.')");
	}

	/**
	 * Initializes the model manager.
	 * 
	 * @param l
	 *            - the framework logger
	 * @param m
	 *            - the owning modelspace
	 * @throws VPMRuntimeException
	 */
	public void init(Logger l, IModelSpace m) throws VPMRuntimeException {
		modelSpace = (SimpleModelSpace) m;
		lock = modelSpace.lock;
		nManager = (SimpleNotificationManager) modelSpace
				.getNotificationManager();
		logger = l;
	}

	/**
	 * Retrieves the root element of the modelspace.
	 * 
	 * @return root entity
	 */
	public IEntity getRoot() {
		return modelSpace.rootEntity;
	}

	/**
	 * Delete an element and all elements in its namespace
	 * 
	 * @param root
	 * @throws VPMCoreException
	 */
	private void deleteBranch(IModelElement root) throws VPMCoreException {
		// Delete all elements in namespace (recursive)
		while (root.getElementsInNamespace().size() > 0) {
			IModelElement me = root.getElementsInNamespace().iterator().next();
			deleteBranch(me);
			// If the element is a relation then deleting the target element
			// deletes the relation as well, so at this point the root can be
			// deleted from model
			if (root.isDeleted())
				return;
		}
		// delete this element
		deleteElement(root, EDeleteSemantics.DELETE_SEMANTICS_FORCE);
	}

	/**
	 * Deletes the entity from the modelspace according to a specified
	 * semantics.
	 * 
	 * @param e
	 *            entity to be deleted
	 * @param semantics
	 *            semantics id
	 */
	public void deleteEntity(IEntity e, EDeleteSemantics semantics)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			// checl parameters
			if (e == null) {
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			}
			if (e == modelSpace.ancestorEntity
					|| e == modelSpace.ancestorContainer) {
				throw new VPMCoreException(
						"ancestor entity can not be deleted.");
			}
			if (!modelSpace.entitiesSet.contains(e)) {
				throw new VPMCoreException(
						"Entity to be deleted not in modelSpace.");
			}
			if (EDeleteSemantics.DELETE_SEMANTICS_BRANCH_FORCE == semantics) {
				deleteBranch(e);
				return;
			}
			if (semantics == EDeleteSemantics.DELETE_SEMANTICS_FORCE) {
				if (e.equals(getRoot())) {
					throw new VPMCoreException("Root entity cannot be deleted.");
				}
				sendNotification(new NotificationObjectMoreAtomicsToOneAtomStart());
				// Delete all relations that have this element as source or
				// target
				deleteConnections(e, semantics); // moved here by Istvan Rath:
													// IMPORTANT FOR DSM fw!!!
				// Delete all instanceof and supertype relations from and to
				// this element
				deleteFromTypeHierarchy(e);
				IEntity parent = e.getParent();
				// Remove the element from its container: now it has no parent
				deleteFromContainer(e);
				// Move all contained entities to the root
				deleteFromContainerHierarchy(e);
				// notify the change
				NotificationObject note = new NotificationObjectDeleteEntity(
						(SimpleEntity) e, (SimpleEntity) parent);
				// remove the object from the model
				modelSpace.entities
						.remove(((SimpleModelElement) e).getLongID());
				modelSpace.entitiesSet.remove(e);
				// set the element deleted
				((SimpleEntity) e).isDeleted = true;
				sendNotification(note);
				sendNotification(new NotificationObjectMoreAtomicsToOneAtomEnd());
				sendNotification(new NotificationObjectAtomicStepReady());
				return;
			}
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Deletes a relation (or function) from the modelspace.
	 * 
	 * @param r
	 *            the relation to be deleted.
	 * @throws VPMCoreNullParameterException
	 */
	public void deleteRelation(IRelation r, EDeleteSemantics semantics)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (r == null)
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			if (r == modelSpace.ancestorRelation)
				throw new VPMCoreException(
						"ancestor relation can not be deleted");
			_deleteRelation(r, semantics);
			return;
		} finally {
			lock.writeLock().unlock();
		}
	}

	private void _deleteRelation(IRelation r, EDeleteSemantics semantics)
			throws VPMCoreException {
		if (EDeleteSemantics.DELETE_SEMANTICS_BRANCH_FORCE == (semantics)) {
			deleteBranch(r);
			return;
		}
		if (semantics == (EDeleteSemantics.DELETE_SEMANTICS_FORCE)) {
			SimpleRelation rel = (SimpleRelation) r;
			sendNotification(new NotificationObjectMoreAtomicsToOneAtomStart());
			// delete all type and subtype relations
			deleteFromTypeHierarchy(rel);
			NotificationObject note;
			note = new NotificationObjectDeleteRelation(rel,
					(SimpleModelElement) rel.getFrom(),
					(SimpleModelElement) rel.getTo());
			setRelFrom(rel, null);
			setRelTo(rel, null);
			// delete all relations from and to this element
			deleteConnections(rel, semantics);
			// remove the relation from model
			modelSpace.relations.remove(rel.getLongID());
			modelSpace.relationsSet.remove(rel);
			rel.isDeleted = true;
			sendNotification(note);
			sendNotification(new NotificationObjectMoreAtomicsToOneAtomEnd());
			sendNotification(new NotificationObjectAtomicStepReady());
			return;
		}
		throw new VPMCoreRuntimeException("No semantics defined for delete");
	}

	/**
	 * Delete all instanceOf and supertypeOf links from this element
	 * 
	 * @param _me
	 */
	private void deleteFromTypeHierarchy(IModelElement _me)
			throws VPMCoreException {
		SimpleModelElement me = (SimpleModelElement) _me;
		ArrayList<IModelElement> l = new ArrayList<IModelElement>(me.supertype);
		for (int i = 0; i < l.size(); ++i) {
			deleteSupertypeOfDirect((SimpleModelElement) l.get(i), me);
		}
		l = new ArrayList<IModelElement>(me.subtypes);
		for (int i = 0; i < l.size(); ++i) {
			deleteSupertypeOfDirect(me, (SimpleModelElement) l.get(i));
		}

		l = new ArrayList<IModelElement>(me.type);
		for (int i = 0; i < l.size(); ++i) {
			deleteInstanceOfDirect((SimpleModelElement) l.get(i), me);
		}
		l = new ArrayList<IModelElement>(me.instances);
		for (int i = 0; i < l.size(); ++i) {
			deleteInstanceOfDirect(me, (SimpleModelElement) l.get(i));
		}
	}

	private void deleteFromContainer(IEntity _me)
			throws VPMCoreNullParameterException, VPMCoreException {
		SimpleEntity me = (SimpleEntity) _me;
		SimpleEntity par = me.parent;
		par.containment.remove(me.getLongID());
		removeFromNamespace(par, _me);
		me.parent = null;
		NotificationObject note = new NotificationObjectDeleteContainment(par,
				me);
		sendNotification(note);
	}

	private void deleteFromContainerHierarchy(IEntity _me)
			throws VPMCoreNullParameterException, VPMCoreException {
		SimpleEntity ent = (SimpleEntity) _me;
		ArrayList<IModelElement> children = new ArrayList<IModelElement>(
				ent.containment.values());
		for (int i = 0; i < children.size(); ++i) {
			SimpleEntity child = (SimpleEntity) children.get(i);
			moveEntityTo(child, getRoot());
		}
	}

	private void deleteConnections(IModelElement _ent,
			EDeleteSemantics semantics) {
		SimpleModelElement ent = (SimpleModelElement) _ent;
		// Iterator it=ent.getConnections().iterator();
		Iterator it = ent.getRelations().iterator();
		// TODO Self connections must be contained only once in the collection.
		// So the collection is a Set
		while (it.hasNext()) {
			SimpleRelation r = (SimpleRelation) it.next();
			try {
				// TODO elvileg nem lehet exception
				deleteRelation(r, semantics);
			} catch (VPMCoreException e) {
			}
		}
	}

	/**
	 * Retrives the model element by its name
	 * 
	 * @param n
	 *            the fully qualified model element name
	 * @return element if found, else null
	 */
	public IModelElement getElementByName(String n) {
		try {
			lock.readLock().lock();
			IModelElement weAreHere = getRoot();
			while (n != null && !n.equals("") && weAreHere != null) {
				int k;
				int l = n.indexOf('.');
				k = l + 1;
				if (l == -1) {
					l = n.length();
					k = l;
				}
				String name = n.substring(0, l);
				weAreHere = weAreHere.getElementInNamespaceByName(name);
				n = n.substring(k);
			}
			return weAreHere;
		} finally {
			lock.readLock().unlock();
		}
	}

	/**
	 * Retrieves an entity by its name
	 * 
	 * @param n
	 *            fully qualified name of entity
	 * @return entity if found, else null
	 */
	public IEntity getEntityByName(String n) {
		IModelElement me = getElementByName(n);
		if (me instanceof IEntity) {
			return (IEntity) me;
		} else {
			return null;
		}
	}

	/**
	 * Retrieves a relation by its name
	 * 
	 * @param n
	 *            fully qualified name of relation
	 * @return relation if found, else null
	 */
	public IRelation getRelationByName(String n) {
		IModelElement me = getElementByName(n);
		if (me instanceof IRelation) {
			return (IRelation) me;
		}
		return null;
	}

	/**
	 * Retrives the model element by its ID
	 * 
	 * @param n
	 *            model element ID
	 * @return element if found, else null
	 */
	public IModelElement getElementByID(String n) {
		if (n == null)
			return null; // added by Istvan Rath.
		try {
			lock.readLock().lock();
			long reqID = 0;
			try {
				reqID = Long.parseLong(n);
			} catch (NumberFormatException e) {
				// if the string represents no number, then it is no ID of any
				// elements in modelspace
				return null;
			}
			if (!n.equals(Long.toString(reqID)))
				return null;
			IModelElement ret = modelSpace.entities.get(reqID);
			if (ret == null)
				ret = modelSpace.relations.get(reqID);
			return ret;
		} finally {
			lock.readLock().unlock();
		}
	}

	/**
	 * Retrieves an entity by its ID
	 * 
	 * @param n
	 *            ID of entity
	 * @return entity if found, else null
	 */
	public IEntity getEntityByID(String n) {
		IModelElement ret = getElementByID(n);
		if (ret instanceof IEntity)
			return (IEntity) ret;
		return null;
	}

	/**
	 * Retrieves a relation by its ID
	 * 
	 * @param n
	 *            ID of relation
	 * @return relation if found, else null
	 */
	public IRelation getRelationByID(String n) {
		IModelElement ret = getElementByID(n);
		if (ret instanceof IRelation)
			return (IRelation) ret;
		return null;
	}

	/**
	 * Retrives all entities currently in the modelspace.
	 * 
	 * @return collection of entities
	 */
	public Set<? extends IEntity> getEntities() {
		try {
			lock.readLock().lock();
			// return new ConstantSet<IEntity>(modelSpace.entitiesSet);
			return modelSpace.entitiesSet;
		} finally {
			lock.readLock().unlock();
		}
	}

	/**
	 * Retrieves all relations currently in the modelspace.
	 * 
	 * @return collection of relations
	 */
	public Set<? extends IRelation> getRelations() {
		try {
			lock.readLock().lock();
			// return new ConstantSet<IRelation>(modelSpace.relationsSet);
			return modelSpace.relationsSet;
		} finally {
			lock.readLock().unlock();
		}
	}

	private void setElemName(SimpleModelElement ent, String name) {
		ent.name = name;
	}

	private void setRelFrom(SimpleRelation rel, SimpleModelElement from)
			throws VPMCoreException {
		SimpleModelElement oldRelFrom = (SimpleModelElement) rel.getFrom();
		if (oldRelFrom != null) {
			oldRelFrom.relationsFrom.remove(rel.getLongID());
			removeFromNamespace(rel.from, rel);
		}
		rel.from = from;
		if (from != null) {
			from.relationsFrom.put(rel.getLongID(), rel);
			addToNamespace(from, rel);
		}
	}

	private void setRelTo(SimpleRelation rel, SimpleModelElement to) {
		SimpleModelElement oldRelTo = (SimpleModelElement) rel.getTo();
		if (oldRelTo != null) {
			oldRelTo.relationsTo.remove(rel);
		}
		rel.to = to;
		if (to != null) {
			to.relationsTo.add(rel);
		}
	}

	private SimpleRelation createRelation(IModelElement from, IModelElement to,
			String name) throws VPMCoreException {
		checkElemName(name);
		sendNotification(new NotificationObjectMoreAtomicsToOneAtomStart());
		SimpleRelation newRel = new SimpleRelation(modelSpace.newId(),
				modelSpace);
		modelSpace.relations.put(newRel.getLongID(), newRel);
		modelSpace.relationsSet.add(newRel);
		setElemName(newRel, name);
		setRelFrom(newRel, (SimpleModelElement) from);
		setRelTo(newRel, (SimpleModelElement) to);
		NotificationObject note = new NotificationObjectCreateRelation(newRel,
				(SimpleModelElement) from, (SimpleModelElement) to);
		sendNotification(note);
		sendNotification(new NotificationObjectMoreAtomicsToOneAtomEnd());
		return newRel;
	}

	// protected SimpleRelation undoCreateRelation(SimpleRelation rel,
	// SimpleEntity from, SimpleEntity to) throws VPMCoreException
	protected SimpleRelation undoCreateRelation(SimpleRelation rel,
			SimpleModelElement from, SimpleModelElement to)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			modelSpace.relations.put(rel.getLongID(), rel);
			modelSpace.relationsSet.add(rel);
			setRelFrom(rel, from);
			setRelTo(rel, to);
			rel.isDeleted = false;
			NotificationObject note = new NotificationObjectCreateRelation(rel,
					from, to);
			sendNotification(note);
			return rel;
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Creates a new entity in modelSpace
	 * 
	 * @param container
	 *            the entity's initial container
	 * @param name
	 *            the entity's initial name
	 * @param value
	 *            the entity's initial value
	 * @return the entity
	 * @throws VPMCoreException
	 */
	private SimpleEntity createEntity(SimpleEntity container, String name,
			String value) throws VPMCoreException {
		checkElemName(name);
		sendNotification(new NotificationObjectMoreAtomicsToOneAtomStart());

		SimpleEntity newEnt = new SimpleEntity(modelSpace.newId(), modelSpace);
		modelSpace.entities.put(newEnt.getLongID(), newEnt);
		modelSpace.entitiesSet.add(newEnt);
		setElemName(newEnt, name);
		newEnt.value = value;
		firstContainment(container, newEnt);
		NotificationObject notification = new NotificationObjectCreateEntity(
				newEnt, container);
		sendNotification(notification);
		sendNotification(new NotificationObjectMoreAtomicsToOneAtomEnd());
		return newEnt;
	}

	/**
	 * Put me into container.
	 * 
	 * @param container
	 *            the container
	 * @param me
	 *            the modelelement to put into container
	 */
	protected void undoCreateContainment(SimpleEntity container, SimpleEntity me)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			container.containment.put(me.getLongID(), me);
			me.parent = container;
			// following added by Istvan Rath
			addToNamespace(container, me);
			NotificationObject note = new NotificationObjectMoveTo(me,
					container, (SimpleEntity) me.getParent());
			sendNotification(note);
			// end added by Istvan Rath
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Put an entity into modelspace when undo.
	 * 
	 * @param entity
	 *            the entity to put back into modelSpace
	 * @return the entity
	 */
	protected SimpleEntity undoCreateEntity(SimpleEntity entity,
			SimpleEntity parent) throws VPMCoreException {
		try {
			lock.writeLock().lock();
			entity.notificationListeners = new ArrayList<ICoreNotificationListener>();
			modelSpace.entities.put(entity.getLongID(), entity);
			modelSpace.entitiesSet.add(entity);
			entity.isDeleted = false;
			NotificationObject notification = new NotificationObjectCreateEntity(
					entity, parent);
			sendNotification(notification);
			return entity;
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Creates a new entity with a random and unique name in the modelspace root
	 * 
	 * @return reference to the entity
	 * @throws VPMCoreException
	 */
	public IEntity newEntity() {
		try {
			return newEntity(modelSpace.uniqueName(), "", getRoot());
		} catch (VPMCoreException e) {
			return null;
		}
	}

	/**
	 * Creates a new entity with a random and unique name in the specified
	 * container
	 * 
	 * @param container
	 *            the container
	 * @return reference to the entity
	 * @throws VPMCoreException
	 * @throws VPMCoreException
	 */
	public IEntity newEntity(IEntity container) throws VPMCoreException {
		return newEntity(modelSpace.uniqueName(), "", container);
	}

	/**
	 * Creates a new entity with the given name in the modelspace root
	 * 
	 * @param name
	 *            the name of the entity
	 * @return reference to the entity
	 * @throws VPMCoreException
	 */
	public IEntity newEntity(String name) throws VPMCoreException {
		return newEntity(name, "", getRoot());
	}

	/**
	 * Creates a new entity with the given name and container
	 * 
	 * @param name
	 *            the name of the entity
	 * @param container
	 *            the container
	 * @return reference to the entity
	 * @throws VPMCoreException
	 * @throws VPMCoreException
	 */
	public IEntity newEntity(String name, IEntity container)
			throws VPMCoreException {
		return newEntity(name, "", container);
	}

	/**
	 * Creates a new entity with the given name and value in the modelspace root
	 * 
	 * @param name
	 *            the name of the entity
	 * @param value
	 *            the value of the entity
	 * @return reference to the entity
	 * @throws VPMCoreException
	 */
	public IEntity newEntity(String name, String value) throws VPMCoreException {
		return newEntity(name, value, getRoot());
	}

	/**
	 * Creates a new entity with the given name, value and container
	 * 
	 * @param name
	 *            the name of the entity
	 * @param value
	 *            the value of the entity
	 * @param container
	 *            the container
	 * @return reference to the entity
	 * @throws VPMCoreException
	 */
	public IEntity newEntity(String name, String value, IEntity container)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (name == null || container == null || value == null) {
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			}
			SimpleEntity newEnt = createEntity((SimpleEntity) container, name,
					value);
			sendNotification(new NotificationObjectAtomicStepReady());
			return newEnt;
		} finally {
			lock.writeLock().unlock();
		}
	}
	
	/**
	 * Creates a new entity with the given name, value container and a single type. 
	 * Each of the parameters can be left null.
	 * 
	 * @param name
	 *            the name of the entity
	 * @param value
	 *            the value of the entity
	 * @param container
	 *            the container
	 * @param type
	 *            the single initial type of the entity   
	 * @return reference to the entity
	 * @throws VPMCoreException
	 */
	public IEntity newEntity(String name, String value, IEntity container, IEntity type)
			throws VPMCoreException 
	{
		try {
			lock.writeLock().lock();
			if (name==null) name = modelSpace.uniqueName();
			if (value==null) value = "";
			if (container==null) container = getRoot();
			IEntity newEnt = newEntity(name, value, container);
			if (type != null) newInstanceOf(type, newEnt);
			return newEnt;
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Creates a new relation with a random and unique name, between the given model elements.
	 * 
	 * @param from
	 *            source entity
	 * @param to
	 *            target entity
	 * @return reference to the new relation
	 * @throws VPMCoreException
	 */
	public IRelation newRelation(IModelElement from, IModelElement to)
			throws VPMCoreException {
		return newRelation(modelSpace.uniqueName(), from, to);
	}

	public IRelation newRelation(IModelElement from, IModelElement to,
			EMultiplicityKind multiplicity, boolean isaggregation,
			IRelation inverse) throws VPMCoreException {
		return newRelation(modelSpace.uniqueName(), from, to, multiplicity,
				isaggregation, inverse);
	}

	public IRelation newRelation(String name, IModelElement from,
			IModelElement to, EMultiplicityKind multiplicity,
			boolean isaggregation, IRelation inverse) throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (from == null || to == null) {
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			}
			SimpleRelation newRel = createRelation(from, to, name);
			newRel.multiplicity = multiplicity;
			newRel.inverse = inverse;
			newRel.isAggregation = isaggregation;
			sendNotification(new NotificationObjectAtomicStepReady());
			return newRel;
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Creates a new relation with the given name, between the given model elements.
	 * 
	 * @param name
	 *            name of relation
	 * @param from
	 *            source entity
	 * @param to
	 *            target entity
	 * @return reference to the new relation
	 * @throws VPMCoreException
	 */
	public IRelation newRelation(String name, IModelElement from,
			IModelElement to) throws VPMCoreException {
		return newRelation(name, from, to, EMultiplicityKind.MANY_TO_MANY,
				false, null);
	}
	
	/**
	 * Creates a new relation with the given name, between the given model elements, with the given single type.
	 * Name and type can be left null.
	 * 
	 * @param name
	 *            name of relation
	 * @param from
	 *            source entity
	 * @param to
	 *            target entity
	 * @param type
	 *            the single initial type of the relation       
	 * @return reference to the new relation
	 * @throws VPMCoreException
	 */
	public IRelation newRelation(String name, IModelElement from,
			IModelElement to, IRelation type) throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (name==null) name = modelSpace.uniqueName();
			IRelation newRel = newRelation(name, from, to);
			if (type!=null) newInstanceOf(type, newRel);
			return newRel;
		} finally {
			lock.writeLock().unlock();
		}
	}
	
	/**
	 * returns the type of modelelement
	 * 
	 * @param e
	 * @return 0: entity 1: function 2: relation -1:not modelelement
	 */
	int getElementType(IModelElement e) {
		if (e instanceof IEntity)
			return 0;
		if (e instanceof IRelation)
			return 2;
		return -1;
	}

	/**
	 * Creates an instanceOf relation between the two model elements
	 * 
	 * @param type
	 *            the type element
	 * @param inst
	 *            the instance
	 */
	public void newInstanceOf(IModelElement type, IModelElement inst)
			throws VPMCoreException {
		newInstanceOfMachine(type, inst);
	}

	/**
	 * Creates a new supertypeOf relation between the two elements
	 * 
	 * @param sup
	 *            the supertype
	 * @param sub
	 *            the subtype
	 * @throws VPMCoreNullParameterException
	 */
	public void newSupertypeOf(IModelElement sup, IModelElement sub)
			throws VPMCoreException {
		newSupertypeOfMachine(sup, sub);
	}

	/**
	 * Creates a new supertypeOf relation between the two elements with editor
	 * semantics: only the minimal requied supertypeOf relations can be exactly
	 * added
	 * 
	 * @param sup
	 *            the supertype
	 * @param sub
	 *            the subtype
	 * @throws VPMCoreNullParameterException
	 */
	public void newSupertypeOfEditor(IModelElement sup, IModelElement sub)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (sup == null || sub == null) {
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			}
			SimpleModelElement _sup = (SimpleModelElement) sup;
			SimpleModelElement _sub = (SimpleModelElement) sub;
			if (getElementType(sup) != getElementType(sub))
				throw new VPMCoreException(
						EXCEPTION_TYPE_BETWEEN_DIFFERENT_ELEMENTS
								+ sup.getFullyQualifiedName() + ","
								+ sub.getFullyQualifiedName());
			if (_sub.getAllSupertypes().contains(_sup))
				throw new VPMCoreException(EXCEPTION_ALREADY_SUPERTYPE
						+ sup.getFullyQualifiedName() + ","
						+ sub.getFullyQualifiedName());
			if (_sup.getAllSupertypes().contains(_sub))
				throw new VPMCoreException(
						EXCEPTION_NEW_SUPERTYPE_CAUSES_CIRCLE
								+ sup.getFullyQualifiedName() + ","
								+ sub.getFullyQualifiedName());
			if (!CollectionsDisjunktTester.isDisjunct(_sup.getAllSupertypes(),
					_sub.supertype)
					|| !CollectionsDisjunktTester.isDisjunct(_sub
							.getAllSubtypes(), _sup.subtypes))
				throw new VPMCoreException(
						EXCEPTION_NEW_SUPERTYPE_MAKES_EXISTING_RELATION_IMPLICIT
								+ sup.getFullyQualifiedName() + ","
								+ sub.getFullyQualifiedName());
			newSupertypeOfDirect(_sup, _sub);
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Creates a new instanceOf relation between the two elements with editor
	 * semantics: only the minimal requied instanceOf relations can be exactly
	 * added
	 * 
	 * @param sup
	 *            the supertype
	 * @param inst
	 *            the subtype
	 * @throws VPMCoreNullParameterException
	 */
	public void newInstanceOfEditor(IModelElement type, IModelElement inst)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (type == null || inst == null) {
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			}
			SimpleModelElement _type = (SimpleModelElement) type;
			SimpleModelElement _inst = (SimpleModelElement) inst;
			if (getElementType(type) != getElementType(inst))
				throw new VPMCoreException(
						EXCEPTION_INSTANCE_BETWEEN_DIFFERENT_ELEMENTS
								+ type.getFullyQualifiedName() + ","
								+ inst.getFullyQualifiedName());
			if (_inst.getAllTypes().contains(_type))
				throw new VPMCoreException(EXCEPTION_ALREADY_INSTANCE
						+ type.getFullyQualifiedName() + ","
						+ inst.getFullyQualifiedName());
			if (!CollectionsDisjunktTester.isDisjunct(_type.getAllSupertypes(),
					_inst.type))
				throw new VPMCoreException(
						EXCEPTION_NEW_INSTANCE_MAKES_EXISTING_RELATION_IMPLICIT
								+ type.getFullyQualifiedName() + ","
								+ inst.getFullyQualifiedName());
			newInstanceOfDirect(_type, _inst);
			if (inst instanceof IRelation) {
				checkIsAggregationSemantics((SimpleRelation) inst);
			}
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Creates a new instanceOf relation between the two elements with machine
	 * semantics: only the minimal requied instanceOf relations can be exactly
	 * added
	 * 
	 * @param sup
	 *            the supertype
	 * @param inst
	 *            the subtype
	 * @throws VPMCoreNullParameterException
	 */
	public void newInstanceOfMachine(IModelElement type, IModelElement inst)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (type == null || inst == null) {
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			}
			SimpleModelElement _type = (SimpleModelElement) type;
			SimpleModelElement _inst = (SimpleModelElement) inst;
			if (getElementType(type) != getElementType(inst))
				throw new VPMCoreException(
						EXCEPTION_INSTANCE_BETWEEN_DIFFERENT_ELEMENTS
								+ type.getFullyQualifiedName() + ","
								+ inst.getFullyQualifiedName());
			if (_inst.getAllTypes().contains(_type))
				return;
			newInstanceOfDirect(_type, _inst);
			// Delete direct arcs made implicit
			Collection col = CollectionsDisjunktTester.sharedElements(_type
					.getAllSupertypes(), _inst.type);
			Iterator it = col.iterator();
			while (it.hasNext()) {
				SimpleModelElement me = (SimpleModelElement) it.next();
				deleteInstanceOfDirect(me, _inst);
			}
			if (inst instanceof IRelation) {
				checkIsAggregationSemantics((SimpleRelation) inst);
			}
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Creates a new supertypeOf relation between the two elements with editor
	 * semantics: relation will be added if not present already
	 * 
	 * @param sup
	 *            the supertype
	 * @param sub
	 *            the subtype
	 * @throws VPMCoreNullParameterException
	 */
	public void newSupertypeOfMachine(IModelElement _sup, IModelElement _sub)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (_sup == null || _sub == null) {
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			}
			SimpleModelElement sup = (SimpleModelElement) _sup;
			SimpleModelElement sub = (SimpleModelElement) _sub;
			if (getElementType(sup) != getElementType(sub))
				throw new VPMCoreException(
						EXCEPTION_TYPE_BETWEEN_DIFFERENT_ELEMENTS
								+ sup.getFullyQualifiedName() + ","
								+ sub.getFullyQualifiedName());
			if (sup == sub || _sub.getAllSupertypes().contains(_sup)) {
				// TODO kell-e warning??
				// log_warn(WARNING_ALREADY_SUPERTYPE,sup, sub);
				return;
			}
			if (_sup.getAllSupertypes().contains(_sub))
				throw new VPMCoreException(
						EXCEPTION_NEW_SUPERTYPE_CAUSES_CIRCLE
								+ sup.getFullyQualifiedName() + ","
								+ sub.getFullyQualifiedName());
			// Add this relation
			newSupertypeOfDirect(sup, sub);
			// Delete all relations made implicit by this relation
			Collection madeImplicit = CollectionsDisjunktTester.sharedElements(
					sub.supertype, sup.getAllSupertypes());
			Iterator it = madeImplicit.iterator();
			while (it.hasNext()) {
				// implicit ancestor of sub
				SimpleModelElement me = (SimpleModelElement) it.next();
				deleteSupertypeOfDirect(me, sub);
			}
			// Implicits on other side:
			madeImplicit = CollectionsDisjunktTester.sharedElements(
					sup.subtypes, sub.getAllSubtypes());
			it = madeImplicit.iterator();
			while (it.hasNext()) {
				// implicit ancestor of sub
				SimpleModelElement me = (SimpleModelElement) it.next();
				deleteSupertypeOfDirect(sup, me);
			}

			// Transaction is ready, modelSpace is in consistent state. Add Undo
			// point
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Adds a direct supertypeof relation (editor style). No check is done. This
	 * is the main step of all supertypeof relation mechanisms.
	 * 
	 * @param sup
	 * @param sub
	 */
	private void newSupertypeOfDirect(SimpleModelElement sup,
			SimpleModelElement sub) {
		sup.subtypes.add(sub);
		sub.supertype.add(sup);
		NotificationObject note = new NotificationObjectCreateSupertypeOf(sup,
				sub);
		/*
		 * NotificationObject note=new
		 * NotificationObject(ICoreNotificationObject
		 * .ACTION_CREATE_SUPERTYPEOF); note.addParameterElem(sup);
		 * note.addParameterElem(sub);
		 */
		sendNotification(note);
	}

	/**
	 * Adds a direct instanceof relation (editor style). No check is done. This
	 * is the main step of all instanceof relation mechanisms.
	 * 
	 * @param type
	 * @param inst
	 */
	private void newInstanceOfDirect(SimpleModelElement type,
			SimpleModelElement inst) {
		type.instances.add(inst);
		inst.type.add(type);
		NotificationObject note = new NotificationObjectCreateInstanceOf(type,
				inst);
		/*
		 * NotificationObject note=new
		 * NotificationObject(ICoreNotificationObject.ACTION_CREATE_INSTANCEOF);
		 * note.addParameterElem(type); note.addParameterElem(inst);
		 */
		sendNotification(note);
	}

	/**
	 * Creates a new containment, without notification of name change, if that
	 * happens. It is used, when creating new element.
	 * 
	 * @param cont
	 * @param elem
	 */
	private void firstContainment(IEntity cont, IEntity elem)
			throws VPMCoreException {
		sendNotification(new NotificationObjectMoreAtomicsToOneAtomStart());
		_newContainment(cont, elem);
		sendNotification(new NotificationObjectMoreAtomicsToOneAtomEnd());
	}

	private void addToNamespace(IModelElement cont, IModelElement elem) {
		SimpleModelElement e = (SimpleModelElement) elem;
		SimpleModelElement c = (SimpleModelElement) cont;
		String nameKey = e.getName();
		Long id = (Long) c.namespace.get(nameKey);
		if (id == null) {
			c.namespace.put(e.getName(), e.getLongID());
		} else {
			String newName = modelSpace.uniqueName(e.getName());
			log_warn_nameclash(e, c, newName);
			e.name = newName;
			NotificationObject note = new NotificationObjectSetName(e, newName,
					nameKey);
			sendNotification(note);
			addToNamespace(cont, elem);
		}
	}

	private void removeFromNamespace(IModelElement cont, IModelElement elem) {
		((SimpleModelElement) cont).namespace.remove(elem.getName());
	}

	/**
	 * Creates a new containment, without sending Atomic_step notification
	 * 
	 * @param cont
	 * @param elem
	 */
	private void _newContainment(IEntity cont, IEntity elem)
			throws VPMCoreException {
		SimpleEntity e = (SimpleEntity) elem;
		SimpleEntity c = (SimpleEntity) cont;
		if (e.parent != null)
			throw new VPMCoreException("core internal error");
		c.containment.put(e.getLongID(), e);
		e.parent = c;
		addToNamespace(cont, elem);
	}

	/**
	 * Deletes the instanceOf relation between the two model elements
	 * 
	 * @param type
	 *            the type element
	 * @param inst
	 *            the instance
	 * @throws VPMCoreNullParameterException
	 */
	public void deleteInstanceOf(IModelElement type, IModelElement inst)
			throws VPMCoreException {
		deleteInstanceOfMachine(type, inst);
	}

	/**
	 * Deletes the supertypeOf relation between the two elements
	 * 
	 * @param sup
	 *            the supertype
	 * @param sub
	 *            the subtype
	 * @throws VPMCoreNullParameterException
	 */
	public void deleteSupertypeOf(IModelElement sup, IModelElement sub)
			throws VPMCoreException {
		deleteSupertypeOfMachine(sup, sub);
	}

	public void deleteSupertypeOfEditor(IModelElement sup, IModelElement sub)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (sup == null || sub == null)
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			SimpleModelElement _sup = (SimpleModelElement) sup;
			SimpleModelElement _sub = (SimpleModelElement) sub;
			if (_sub.supertype.contains(_sup)) {
				deleteSupertypeOfDirect(_sup, _sub);
			} else {
				throw new VPMCoreException(
						EXCEPTION_DELETE_SUPERTYPE_EDITOR_NO_RELATION
								+ sup.getFullyQualifiedName() + ","
								+ sub.getFullyQualifiedName());
			}
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	public void deleteInstanceOfEditor(IModelElement type, IModelElement inst)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (inst == null || type == null)
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			SimpleModelElement _type = (SimpleModelElement) type;
			SimpleModelElement _inst = (SimpleModelElement) inst;
			if (_inst.type.contains(_type)) {
				deleteInstanceOfDirect(_type, _inst);
			} else {
				throw new VPMCoreException(
						EXCEPTION_DELETE_INSTANCEOF_NO_RELATION
								+ type.getFullyQualifiedName() + ","
								+ inst.getFullyQualifiedName());
			}
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	private boolean stupidMethod = false;

	public void deleteInstanceOfMachine(IModelElement type, IModelElement inst)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (inst == null || type == null)
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			SimpleModelElement _type = (SimpleModelElement) type;
			SimpleModelElement _inst = (SimpleModelElement) inst;
			boolean isDeleted = false;
			// if direct instanceof delete it
			if (_inst.type.contains(_type)) {
				deleteInstanceOfDirect(_type, _inst);
				isDeleted = true;
			}
			// if not direct delete all instance arrows that
			// implicate the ine to delete
			else {
				Iterator it = _inst.type.iterator();
				while (it.hasNext()) {
					SimpleModelElement me = (SimpleModelElement) it.next();
					if (me.getAllSupertypes().contains(_type)) {
						deleteInstanceOfDirect(me, _inst);
						isDeleted = true;
					}
				}
			}
			// Create all implicit instanceofs direct
			if (stupidMethod) {
				if (isDeleted) {
					Iterator iter = _type.supertype.iterator();
					while (iter.hasNext()) {
						SimpleModelElement supMeType = (SimpleModelElement) iter
								.next();
						if (!_inst.getAllTypes().contains(supMeType))
							newInstanceOfDirect(supMeType, _inst);
					}
				}
			}
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * delete supertype relation between sup and sub with machine semantics: no
	 * error if relation is not present no warn if it is an implicit relation
	 * delete the minimum number of direct relationships do not delete
	 * relationships (make new direct relationships) which were caused implicit
	 * by this relationship
	 * 
	 * @param sup
	 *            the supertype
	 * @param sub
	 *            the subtype
	 * @throws VPMCoreNullParameterException
	 * @throws VPMCoreException
	 */
	public void deleteSupertypeOfMachine(IModelElement sup, IModelElement sub)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (sup == null || sub == null)
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			SimpleModelElement _sup = (SimpleModelElement) sup;
			SimpleModelElement _sub = (SimpleModelElement) sub;
			ArrayList<IModelElement> deletedList = new ArrayList<IModelElement>();
			// Delete if direct arc exists
			if (_sub.supertype.contains(_sup)) {
				deleteSupertypeOfDirect(_sup, _sub);
				deletedList.add(_sup);
			}
			// Tricky delete
			else {
				// Find all arcs that makes it subtype of sup
				Iterator it = _sub.supertype.iterator();
				while (it.hasNext()) {
					SimpleModelElement me = (SimpleModelElement) it.next();
					if (me.getAllSupertypes().contains(_sup)) {
						deleteSupertypeOfDirect(me, _sub);
						deletedList.add(me);
					}
				}
			}
			if (stupidMethod) {
				// Add sub the super elements of sup (implicit arcs become
				// explicit)
				if (!deletedList.isEmpty()) {
					Iterator iter = _sup.supertype.iterator();
					while (iter.hasNext()) {
						SimpleModelElement supMeSup = (SimpleModelElement) iter
								.next();
						if (!_sub.getAllSupertypes().contains(supMeSup))
							newSupertypeOfDirect(supMeSup, _sub);
					}
				}
				// Add subSubtyped elements the old supertype (implicit arcs
				// become explicit)
				for (int i = 0; i < deletedList.size(); ++i) {
					SimpleModelElement supMe = (SimpleModelElement) deletedList
							.get(i);
					Iterator it = _sub.subtypes.iterator();
					while (it.hasNext()) {
						SimpleModelElement subMe = (SimpleModelElement) it
								.next();
						newSupertypeOfDirect(supMe, subMe);
					}
				}
			}
			if (deletedList.size() == 0) {
				log_warn(WARNING_DELETE_NOT_SUPERTYPE, _sup, _sub);
			} else {
				sendNotification(new NotificationObjectAtomicStepReady());
			}
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Delete the supertypeof relation between the two elements. The relation to
	 * be deleted must be direct
	 * 
	 * @param sup
	 *            supertype
	 * @param sub
	 *            subtype
	 * @throws VPMCoreNullParameterException
	 * @throws VPMCoreException
	 */
	private void deleteSupertypeOfDirect(SimpleModelElement sup,
			SimpleModelElement sub) throws VPMCoreNullParameterException,
			VPMCoreException {
		sup.subtypes.remove(sub);
		sub.supertype.remove(sup);
		NotificationObject note = new NotificationObjectDeleteSupertypeOf(sup,
				sub);
		sendNotification(note);
	}

	/**
	 * Delete the instanceof relation between the two elements. The relation to
	 * be deleted must be direct
	 * 
	 * @param type
	 * @param inst
	 * @throws VPMCoreNullParameterException
	 * @throws VPMCoreException
	 */
	private void deleteInstanceOfDirect(SimpleModelElement type,
			SimpleModelElement inst) throws VPMCoreNullParameterException,
			VPMCoreException {
		type.instances.remove(inst);
		inst.type.remove(type);
		NotificationObject note = new NotificationObjectDeleteInstanceOf(type,
				inst);
		sendNotification(note);
	}

	/**
	 * Sets the name of the element
	 * 
	 * @param e
	 *            the element
	 * @param n
	 *            the new name
	 * @throws VPMCoreException
	 */
	public void setName(IModelElement e, String n) throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (e == null || n == null) {
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			}
			// check name if valid
			checkElemName(n);
			_setName((SimpleModelElement) e, n);
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Sets the name of the element without sending ATOMIC_ACTION NOTIFICATION
	 * 
	 * @param e
	 *            the element
	 * @param n
	 *            the new name
	 */
	private void _setName(SimpleModelElement e, String n) {
		SimpleModelElement elem = (SimpleModelElement) e;
		String oldName = elem.getName();
		if (n.equals(oldName)) {
			// name not changing, nothing to do
			return;
		}
		SimpleModelElement par = (SimpleModelElement) elem.getNamespace();
		// check if elements has a namespace
		// all elements must have namespace, but this method may be called
		// in a non-consistent state
		if (par != null) {
			if (par.namespace.get(n) != null) {
				// if the container already has an element with the name to be
				// set
				// then generate a new name.
				log_warn(WARNING_NAME_CLASH_AT_NAME_CHANGE + n, e, par);
				n = modelSpace.uniqueName(n);
			}
			// We must not change name while in container, because the container
			// sorts contained
			// elements by name:
			removeFromNamespace(elem.getNamespace(), elem);
			elem.name = n;
			addToNamespace(elem.getNamespace(), elem);
		} else {
			elem.name = n;
		}
		NotificationObject note = new NotificationObjectSetName(elem, n,
				oldName);
		sendNotification(note);
	}

	/**
	 * Sets the value of the entity
	 * 
	 * @param e
	 *            the entity
	 * @param v
	 *            the new value
	 * @throws VPMCoreNullParameterException
	 */
	public void setValue(IEntity e, String v)
			throws VPMCoreNullParameterException {
		try {
			lock.writeLock().lock();
			if (e == null || v == null)
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			SimpleEntity ent = (SimpleEntity) e;
			String oldValue = ent.value;
			ent.value = v;
			NotificationObject note = new NotificationObjectSetValue(ent, v,
					oldValue);
			sendNotification(note);
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Removes the element from its current container and moves it to the
	 * specified new container
	 * 
	 * @param e
	 *            the element
	 * @param cont
	 *            new container
	 */
	public void moveEntityTo(IEntity e, IEntity cont) throws VPMCoreException {
		try {
			lock.writeLock().lock();
			// TODO mi van, ha pont abban a kontenerban van?
			if (e == null || cont == null) {
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			}
			SimpleEntity ent = (SimpleEntity) e;
			SimpleEntity par = ent.parent;
			if (ent.equals(cont))
				throw new VPMCoreException(
						"error in parameters of moveEntityTo: parent and children are equal");
			if (par == cont)
				return;
			par.containment.remove(ent.getLongID());
			ent.parent = null;
			removeFromNamespace(par, ent);
			_newContainment(cont, e);
			NotificationObject note = new NotificationObjectMoveTo(ent,
					(SimpleEntity) cont, par);
			sendNotification(note);
			sendNotification(new NotificationObjectAtomicStepReady());
			return;
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Send the given notification to all listeners, which have a category
	 * higher than the set notification level.
	 * 
	 * @param notification
	 */
	private void sendNotification(ICoreNotificationObject notification) {
		modelSpace.numberOfChanges++;
		if (nManager != null)
			nManager.sendNotification(notification);
	}

	private static final String WARNING_NAME_CLASH_AT_NAME_CHANGE = "name clash at name change. New name, element, container: ";

	private void log_warn(String s, SimpleModelElement sme,
			SimpleModelElement sme2) {
		String s1 = sme.getFullyQualifiedName();
		String s2 = sme2.getFullyQualifiedName();
		logger.warning(s + ":" + s1 + "," + s2);
	}

	private void log_warn_nameclash(SimpleModelElement element_with_old_name,
			SimpleModelElement container, String new_name) {
		String warn = "The creation / copy / renaming of element '"
				+ element_with_old_name.getFullyQualifiedName()
				+ "' caused name clash in container '"
				+ container.getFullyQualifiedName()
				+ "' and thus it was automatically renamed to '" + new_name
				+ "'";
		logger.warning(warn);
	}

	private void checkParameter(IModelElement me) throws VPMCoreException {
		if (me == null)
			throw new VPMCoreNullParameterException("");
		if (me.isDeleted())
			throw new VPMCoreException("parameter already deleted");
	}

	// TODO: check parameters
	public void setRelationFrom(IRelation rel, IModelElement from)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			checkParameter(rel);
			checkParameter(from);
			if (rel == from)
				throw new VPMCoreException("edge end must not be itself");
			if (checkEdgeCircle(rel, from)) {
				throw new VPMCoreException(
						"edge set from element error: would make infinite fully qualified name");
			}
			SimpleRelation srel = (SimpleRelation) rel;
			SimpleModelElement sfrom = (SimpleModelElement) from;
			NotificationObject no = new NotificationObjectSetRelationFrom(
					(SimpleRelation) rel, (SimpleModelElement) rel.getFrom(),
					(SimpleModelElement) from);
			setRelFrom(srel, sfrom);
			sendNotification(no);
			checkIsAggregationSemantics(srel);
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Check whether the isAggregation type relation's source and target are
	 * parent and child entities. If check fails then the state is repaired by
	 * moving the target entity below the parent
	 * 
	 * @param rel
	 * @throws VPMCoreException
	 */
	private void checkIsAggregationSemantics(SimpleRelation rel)
			throws VPMCoreException {
		for (IModelElement me : rel.getTypes()) {
			if (me instanceof SimpleRelation) {
				SimpleRelation tRel = (SimpleRelation) me;
				if (tRel.getIsAggregation()) {
					// The edge must be an aggregation:
					IModelElement to = rel.getTo(), from = rel.getFrom();
					if (!to.getNamespace().equals(from)) {
						if (to instanceof IEntity && from instanceof IEntity) {
							if (!to.equals(from)) {
								moveEntityTo((IEntity) to, (IEntity) from);
								log_warn("isAggregation=true set for "
										+ rel.getFullyQualifiedName()
										+ "'s type ("
										+ tRel.getFullyQualifiedName()
										+ "), containment should hold for ",
										(SimpleModelElement) to,
										(SimpleModelElement) from);
							} else {
								log_warn(
										"isAggregation=true set for "
												+ rel.getFullyQualifiedName()
												+ "'s type ("
												+ tRel.getFullyQualifiedName()
												+ "), containment should hold (but source and target are the same) for ",
										(SimpleModelElement) to,
										(SimpleModelElement) from);
							}
						}
					}
					return;
				}
			}
		}
	}

	public void setRelationTo(IRelation rel, IModelElement to)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (rel == to)
				throw new VPMCoreException("edge end must not be itself");
			SimpleRelation srel = (SimpleRelation) rel;
			SimpleModelElement sto = (SimpleModelElement) to;
			NotificationObject no = new NotificationObjectSetRelationTo(
					(SimpleRelation) rel, (SimpleModelElement) rel.getTo(),
					(SimpleModelElement) to);
			setRelTo(srel, sto);
			sendNotification(no);
			checkIsAggregationSemantics(srel);
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	public void setViewInfo(IModelElement me, String viewInfo)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			SimpleModelElement _me = (SimpleModelElement) me;
			sendNotification(new NotificationObjectSetViewInfo(me, me
					.getViewInfo(), viewInfo));
			_me.viewInfo = viewInfo;
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	public void deleteElement(IModelElement me, EDeleteSemantics semantics)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (me instanceof IEntity) {
				deleteEntity((IEntity) me, semantics);
			} else {
				deleteRelation((IRelation) me, semantics);
			}
		} finally {
			lock.writeLock().unlock();
		}
	}

	/**
	 * Check whether a recursive path from a relation to its source while the
	 * source is a relation generates an infinite path (a circle) or not. This
	 * must be checked when setting the source of a relation, because a circle
	 * would mean that the relations in the circle has no fully qualified name.
	 * 
	 * @param rel
	 *            the relation that must not be on the path (again)
	 * @param from
	 *            the element that the path is currently at (can be an entity or
	 *            a relation)
	 * @return true if the path contains a circle
	 */
	private boolean checkEdgeCircle(IRelation rel, IModelElement from) {
		try {
			while (!(from instanceof IEntity)) {
				if (from == rel)
					return true;
				if (from == null) {
					logger.fatal("Gaz van: " + rel.getFullyQualifiedName());
					break;
				}
				from = ((IRelation) from).getFrom();
			}
		} catch (NullPointerException e) {
			e.printStackTrace();
		}
		return false;
	}

	public void setRelationMultiplicity(IRelation rel,
			EMultiplicityKind multiplicity) throws VPMCoreException {
		try {
			lock.writeLock().lock();
			SimpleRelation _rel = (SimpleRelation) rel;
			_rel.multiplicity = multiplicity;
			NotificationObject note = new NotificationObjectSetRelationMultiplicity(
					_rel, multiplicity);
			sendNotification(note);
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	public void setRelationIsAggregation(IRelation rel, boolean isaggregation)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			SimpleRelation _rel = (SimpleRelation) rel;
			_rel.isAggregation = isaggregation;
			NotificationObject note = new NotificationObjectSetRelationIsAggregation(
					rel, isaggregation);
			sendNotification(note);
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	public void setRelationInverse(IRelation rel, IRelation inverse)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			SimpleRelation _rel = (SimpleRelation) rel;
			_rel.inverse = inverse;
			NotificationObject note = new NotificationObjectSetRelationInverse(
					rel, inverse);
			sendNotification(note);
			sendNotification(new NotificationObjectAtomicStepReady());
		} finally {
			lock.writeLock().unlock();
		}
	}

	public void setIsFinalType(IModelElement me, boolean isfinaltype)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (me == null)
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			SimpleModelElement sme = (SimpleModelElement) me;
			NotificationObject note = new NotificationObjectSetIsFinalType(me,
					isfinaltype, sme.isFinalType);
			sme.isFinalType = isfinaltype;
			sendNotification(note);
			sendNotification((new NotificationObjectAtomicStepReady()));
		} finally {
			lock.writeLock().unlock();
		}
	}

	public void setIsAnyFrom(IRelation me, boolean isAnyFrom)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (me == null)
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			SimpleRelation srel = (SimpleRelation) me;
			NotificationObject note = new NotificationObjectSetIsAny(me,
					isAnyFrom, true, srel.isAnyFrom);
			srel.isAnyFrom = isAnyFrom;
			sendNotification(note);
			sendNotification((new NotificationObjectAtomicStepReady()));
		} finally {
			lock.writeLock().unlock();
		}
	}

	public void setIsAnyTo(IRelation me, boolean isAnyTo)
			throws VPMCoreException {
		try {
			lock.writeLock().lock();
			if (me == null)
				throw new VPMCoreNullParameterException(parameterNullNotAllowed);
			SimpleRelation srel = (SimpleRelation) me;
			NotificationObject note = new NotificationObjectSetIsAny(me,
					isAnyTo, false, srel.isAnyTo);
			srel.isAnyTo = isAnyTo;
			sendNotification(note);
			sendNotification((new NotificationObjectAtomicStepReady()));
		} finally {
			lock.writeLock().unlock();
		}
	}
/*
	public void setRelationshipBetween(EConstraint relationType,
			IModelElement src, IModelElement tg) throws VPMCoreException {
		if (ConstraintTranslator.isOtherSide(relationType)) {
			relationType = ConstraintTranslator.getOtherSide(relationType);
			IModelElement _src = src;
			src = tg;
			tg = _src;
		}
		switch (relationType) {
		case INSTANCE:
			newInstanceOf(src, tg);
			break;
		case CHILD:
			moveEntityTo((IEntity) src, (IEntity) tg);
			break;
		case REL_FROM:
			setRelationFrom((IRelation) tg, src);
			break;
		default:
			logger.warning("create relationship not implemented: "
					+ relationType + ":" + src + ", " + tg);
			// throw new RuntimeException("not implemented: "+relationType);
		}
	}
*/
	/*
	private void deleteRelationshipBetween(EConstraint relationType,
			IModelElement src, IModelElement tg) throws VPMCoreException {
		if (ConstraintTranslator.isOtherSide(relationType)) {
			relationType = ConstraintTranslator.getOtherSide(relationType);
			IModelElement _src = src;
			src = tg;
			tg = _src;
		}
		switch (relationType) {
		case INSTANCE:
			deleteInstanceOf(src, tg);
			break;
		case CHILD:
			moveEntityTo((IEntity) src, (IEntity) tg);
			break;
		case REL_FROM:
			setRelationFrom((IRelation) tg, src);
			break;
		default:
			logger.warning("del relationship not implemented: " + relationType
					+ ":" + src + ", " + tg);
			// throw new RuntimeException("not implemented: "+relationType);
		}

	}
	*/

	public Set<? extends IModelElement> getElements() {
		try {
			lock.readLock().lock();
			TreeSet<IModelElement> ret = new TreeSet<IModelElement>();
			ret.addAll(modelSpace.entitiesSet);
			ret.addAll(modelSpace.relationsSet);
			return ret;
		} finally {
			lock.readLock().unlock();
		}
	}

	public void setProperty(IModelElement me, EModelElementProperty property,
			Object value) throws VPMCoreException {
		switch (property) {
		case Name:
			setName(me, (String) value);
			break;
		case Value:
			setValue((IEntity) me, (String) value);
			break;
		default:
			logger.warning("Set property not implemented:" + me + "("
					+ property + ")=" + value);
		}
	}

	public IModelElement copyModelElement(IModelElement source,
			IModelElement destinationContainer, boolean copyOutEdges)
			throws VPMCoreException {
		try {
			return ModelCopy.copyModel(modelSpace, modelSpace, source,
					destinationContainer, copyOutEdges);
		} catch (ModelCopyException e) {
			throw new VPMCoreException("error performing copy: "
					+ e.getMessage());
		}
	}

	public IModelElement copyModelElements(Set<IModelElement> sourceElements,
			IModelElement sourceRoot, IModelElement destinationContainer,
			boolean copyOutEdges) throws VPMCoreException {
		try {
			return ModelCopy.copyModels(sourceElements, modelSpace, modelSpace,
					sourceRoot, destinationContainer, copyOutEdges);
		} catch (ModelCopyException e) {
			throw new VPMCoreException("error performing copy: "
					+ e.getMessage());
		}
	}

	/**
	 * @return the lock
	 */
	public ReadWriteLock getLock() {
		return lock;
	}
}
