/*******************************************************************************
 * Copyright (c) 2004-2008 Andras Schmidt, Andras Balogh, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Andras Schmidt, Andras Balogh, Istvan Rath - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.core.simple;

import java.util.concurrent.locks.ReentrantLock;

import org.eclipse.viatra2.core.ICoreNotificationListener;
import org.eclipse.viatra2.core.IPause;
import org.eclipse.viatra2.core.ITransactionManager;
import org.eclipse.viatra2.core.notification.ICoreNotificationObject;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectSubTransactionAbort;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectSubTransactionBegin;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectSubTransactionEnd;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectTransactionAbort;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectTransactionBegin;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectTransactionEnd;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectUndoBegin;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectUndoEnd;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectUndoableTransactionBegin;
import org.eclipse.viatra2.errors.VPMRuntimeException;
import org.eclipse.viatra2.logger.Logger;

/**
 * This class implements a transaction manager for the Viatra core.
 * 
 * @author Andras Balogh, Istvan Rath
 * 
 */
public class SimpleTransactionManager implements ITransactionManager,
		ICoreNotificationListener {

	private int transactionCounter = 0;

	private ReentrantLock tLock = new ReentrantLock();

	protected SimpleUndoManager uManager;

	SimpleModelSpace modelSpace;

	public enum TransactionState {
		DEFAULT, IN_UNDOABLE_TRANSACTION, IN_TRANSACTION, PAUSED
	}

	protected TransactionState iState = TransactionState.DEFAULT;

	/*
	 * This method is provided temporarily for debug purposes. (CoreUtils)
	 */
	public SimpleUndoManager getUndoManager() {
		return uManager;
	}

	/**
	 * Initializes the transaction manager.
	 * 
	 * @param l
	 *            - the framework logger
	 * @param m
	 *            - the owning modelspace
	 * @throws VPMRuntimeException
	 */
	public void init(Logger l, SimpleModelSpace m) throws VPMRuntimeException {
		uManager = new SimpleUndoManager();
		modelSpace = m;

		uManager.init(l, m);
	}

	/**
	 * Begins a new transaction.
	 * 
	 */
	public String beginTransaction() {
		// FIXME what if a begin is issued without called end/rollback first?
		// modelSpace.lock.lock();
		tLock.lock();
		iState = TransactionState.IN_TRANSACTION;
		// if(modelSpace.lock.getHoldCount()==1)
		if (tLock.getHoldCount() == 1) {
			modelSpace.nManager
					.sendNotification(new NotificationObjectTransactionBegin());
		} else {
			modelSpace.nManager
					.sendNotification(new NotificationObjectSubTransactionBegin());
		}
		return "TRANS_" + (transactionCounter++);
	}

	public String beginTransaction(Object info) {
		// FIXME what if a begin is issued without called end/rollback first?
		// modelSpace.lock.lock();
		tLock.lock();
		iState = TransactionState.IN_TRANSACTION;
		// if(modelSpace.lock.getHoldCount()==1)
		if (tLock.getHoldCount() == 1) {
			modelSpace.nManager
					.sendNotification(new NotificationObjectTransactionBegin(
							info));
		} else {
			modelSpace.nManager
					.sendNotification(new NotificationObjectSubTransactionBegin(
							info));
		}
		return "TRANS_" + (transactionCounter++);
	}

	public String beginUndoableTransaction() {
		// modelSpace.lock.lock();
		tLock.lock();
		iState = TransactionState.IN_UNDOABLE_TRANSACTION;
		// if(modelSpace.lock.getHoldCount()==1)
		if (tLock.getHoldCount() == 1) {
			modelSpace.nManager
					.sendNotification(new NotificationObjectUndoableTransactionBegin());
		} else {
			modelSpace.nManager
					.sendNotification(new NotificationObjectSubTransactionBegin());
		}
		String id = "TRANS_" + (transactionCounter++);
		uManager.nextUndoBlock(id);
		return id;
	}

	public String beginUndoableTransaction(Object info) {
		// modelSpace.lock.lock();
		tLock.lock();
		iState = TransactionState.IN_UNDOABLE_TRANSACTION;
		// if(modelSpace.lock.getHoldCount()==1)
		if (tLock.getHoldCount() == 1) {
			modelSpace.nManager
					.sendNotification(new NotificationObjectUndoableTransactionBegin(
							info));
		} else {
			modelSpace.nManager
					.sendNotification(new NotificationObjectSubTransactionBegin(
							info));
		}
		String id = "TRANS_" + (transactionCounter++);
		uManager.nextUndoBlock(id);
		return id;
	}

	/**
	 * Commits the current transaction.
	 * 
	 */
	public void commitTransaction() {
		// int count=tLock.getHoldCount();
		if (tLock.getHoldCount() == 1) {
			iState = TransactionState.DEFAULT;
			// modelSpace.lock.unlock();
			tLock.unlock();
			modelSpace.nManager
					.sendNotification(new NotificationObjectTransactionEnd());
		} else {
			//iState = TransactionState.DEFAULT; <-- don't reach DEFAULT while only subtransactions are committed
			// modelSpace.lock.unlock();
			tLock.unlock();
			modelSpace.nManager
					.sendNotification(new NotificationObjectSubTransactionEnd());
		}
		// iState = TransactionState.DEFAULT;
		// modelSpace.lock.unlock();
	}

	public void commitCompositeTransaction()
	{
		if (tLock.getHoldCount() == 0)
			return; // not in transaction
		if (tLock.getHoldCount() == 1)
		{
			commitTransaction();
		}
		else
		{
			int i = tLock.getHoldCount();
			while (i>0)
			{
				commitTransaction();
				i--;
			}
		}
	}
	
	/**
	 * Aborts the current transaction.
	 * 
	 */
	public void abortTransaction() {
		if (tLock.getHoldCount() == 1) {
			iState = TransactionState.DEFAULT;
			
			
			tLock.unlock();
			modelSpace.nManager
					.sendNotification(new NotificationObjectTransactionAbort());
			
		} else {
			//iState = TransactionState.DEFAULT; <-- don't reach DEFAULT while only subtransactions are committed
			// modelSpace.lock.unlock();
			tLock.unlock();
			modelSpace.nManager
					.sendNotification(new NotificationObjectSubTransactionAbort());
		}
		
	}

	public boolean tryBeginTransaction() {
		// return modelSpace.lock.tryLock();
		return tLock.tryLock();
		// return true;
	}

	public IPause pauseTransaction() throws IllegalMonitorStateException {
		if (tLock.isHeldByCurrentThread()) {
			int c = tLock.getHoldCount();
			IPause ret = new Pause(c);
			for (int i = 0; i < c; ++i)
				commitTransaction();
			return ret;
		} else {
			throw new IllegalMonitorStateException(
					"owner of lock is not this thread");
		}
		// return null;
	}

	public void unPauseTransaction(IPause n) {
		Pause ret = (Pause) n;
		for (int i = 0; i < ret.n; ++i)
			beginTransaction();
	}

	public void undoTransaction(String aTransactionID) {
		modelSpace.nManager.sendNotification(new NotificationObjectUndoBegin());
		uManager.undo(aTransactionID);
		modelSpace.nManager.sendNotification(new NotificationObjectUndoEnd());
	}

	public void actionPerformed(ICoreNotificationObject notification) {
		if (iState == TransactionState.IN_UNDOABLE_TRANSACTION)
			uManager.actionPerformed(notification);
		// drop core notifications if not in an undoable transaction
	}

	public int getListenerCategory() {
		return uManager.getListenerCategory();
	}
}
