/*******************************************************************************
 * Copyright (c) 2004-2008 Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Istvan Rath - initial API and implementation
 *******************************************************************************/

 package org.eclipse.viatra2.editor.text.light;


import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.internal.ui.text.JavaPairMatcher;
import org.eclipse.jface.text.PaintManager;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.jface.text.source.MatchingCharacterPainter;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IFileEditorInput;
import org.eclipse.ui.dialogs.ListDialog;
import org.eclipse.ui.editors.text.TextEditor;
import org.eclipse.viatra2.editor.text.Activator;
import org.eclipse.viatra2.editor.text.light.vtcl.QuickLoadVTCLAction;
import org.eclipse.viatra2.editor.text.light.vtcl.QuickRunVTCLAction;
import org.eclipse.viatra2.editor.text.light.vtcl.VTCLConfiguration;
import org.eclipse.viatra2.editor.text.light.vtcl.VTCLMarkerManager;
import org.eclipse.viatra2.editor.text.markers.IVTCLErrorReporter;
import org.eclipse.viatra2.errors.info.ErrorInformation;
import org.eclipse.viatra2.framework.Framework;
import org.eclipse.viatra2.framework.FrameworkManager;
import org.eclipse.viatra2.framework.IFramework;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Machine;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Module;
import org.eclipse.viatra2.lpgparser.loader.VTCLParseController;
import org.eclipse.viatra2.lpgparser.loader.VTCLParserManager;

public class VTCLEditor extends TextEditor {

	
	private VTEColorProvider iColorProvider;
	
	protected QuickLoadVTCLAction iQuickParseAction;
	
	protected QuickRunVTCLAction iQuickRunAction;
	
	protected final static char[] BRACKETS = {'{', '}', '(', ')'};
	
	protected MatchingCharacterPainter fBracketPainter;
	private PaintManager fPaintManager;

	// -- Added by Daniel Varro
	private VTCLParseController fParseController;
	private VTCLMarkerManager fMarkerManager;

	@Override
	public void createPartControl(Composite parent) 
	{
		super.createPartControl(parent);
		fPaintManager = new PaintManager(getSourceViewer());
		startBracketHighlighting();	
		// -- Added by Daniel Varro		
		fParseController = null;   
		fMarkerManager = new VTCLMarkerManager();

	}
	
	

	public IVTCLErrorReporter getMarkerManager() {
		return fMarkerManager;
	}



	@Override
	public void doSave(IProgressMonitor progressMonitor) {
		super.doSave(progressMonitor);
		attachEditor();
		parseOnSave();
	}

	private IFramework attachedFramework;
	
	public IFramework getAttachedFramework()
	{
		return attachedFramework;
	}
	
	public void setAttachedFramework(IFramework fw)
	{
		attachedFramework = fw;
	}
	
	private Module module;
	
	public Module getModule()
	{
		return module;
	}
	
	@Override
	public String getPartName() {
		String t = super.getPartName();
		if (attachedFramework!=null)
		{
			t+= " ("+attachedFramework.getId()+")";
		}
//		else
//		{
//			t+= " (?)";
//		}
		return t;
	}

	private void startBracketHighlighting() 
	{
		if (fBracketPainter == null) 
		{
			ISourceViewer sourceViewer = getSourceViewer();
			fBracketPainter = new MatchingCharacterPainter(sourceViewer, new JavaPairMatcher(BRACKETS));
			fBracketPainter.setColor(new Color(null,0,0,128));
			fPaintManager.addPainter(fBracketPainter);
		}
	}
		
	private void stopBracketHighlighting() 
	{
		if (fBracketPainter != null) {
			fPaintManager.removePainter(fBracketPainter);
			fBracketPainter.deactivate(true);
			fBracketPainter.dispose();
			fBracketPainter = null;
		}
	} 
	
	public ISourceViewer getViewer() {
		return getSourceViewer();
	}
	
	public VTCLEditor() {
		super();
		iColorProvider = new VTEColorProvider();
		setSourceViewerConfiguration(new VTCLConfiguration(iColorProvider));
	}
	
	@Override
	protected void createActions() {
		super.createActions();
		iQuickParseAction = new QuickLoadVTCLAction();
		setAction("quickparse", iQuickParseAction);
		iQuickRunAction = new QuickRunVTCLAction();
		setAction("quickload", iQuickRunAction);
//		iLPGParseVTCLAction = new LPGParseVTCLAction();
//		setAction("lpgparse", iLPGParseVTCLAction);
//		initializeContextMenu();
	}
	
	public void updateActions()
	{
		//((VTCLActionbarContributor)getEditorSite().getActionBarContributor()).updateActions(this);
		//getEditorSite().getActionBars().getToolBarManager().update(true);
		//getEditorSite().getActionBars().updateActionBars();
	}	
		
	@Override
	public void setFocus() {
		super.setFocus();
		updateActions();
		Activator.getDefault().setCurrentVTCLEditor(this);
	}
	
	public void dispose() {
		iColorProvider.dispose();
		stopBracketHighlighting();
		Activator.getDefault().removeEditor(this);
		super.dispose();
	}
	
	public void attachEditor()
	{
		if (attachedFramework==null)
		{
			final String[] allfws = FrameworkManager.getInstance().getAllFrameWorks();
			if (allfws.length<1)
			{
				firePropertyChange(PROP_TITLE);
				return;
			}
			else if (allfws.length == 1)
			{
				attachedFramework = FrameworkManager.getInstance().getFramework(allfws[0]);	
				firePropertyChange(PROP_TITLE);
			}
			else
			{
				final ListDialog dialog = new ListDialog(this.getSite().getShell());
				dialog.setTitle("Select modelspace to attach");
				dialog.setAddCancelButton(true);
				dialog.setBlockOnOpen(true);
				dialog.setLabelProvider(new LabelProvider() {

					/* (non-Javadoc)
					 * @see org.eclipse.jface.viewers.LabelProvider#getText(java.lang.Object)
					 */
					@Override
					public String getText(Object element) {
						Framework framework = FrameworkManager.getInstance().getFramework((String)element);
						String filename = framework.getCurrentFilename();
						filename = filename.substring(filename.lastIndexOf('/')+1);
						return super.getText(element) + " (" + filename + ")";
					}
					
				});
				dialog.setContentProvider(new IStructuredContentProvider(){
					public Object[] getElements(Object inputElement){return allfws;}
					public void dispose(){}
					public void inputChanged(Viewer viewer, Object oldInput,Object newInput){}
				});
				dialog.setInput("blablabla");
				int result = dialog.open();
				if (result == ListDialog.OK)
				{
					Object[] o = dialog.getResult();
					if (o!=null && o.length == 1)
					{
						String id = o[0].toString();
						attachedFramework = FrameworkManager.getInstance().getFramework(id);	
						firePropertyChange(PROP_TITLE);
					}
				}
			}
		}
		else
		{
			firePropertyChange(PROP_TITLE);
		}
	}
	
	public void refreshTitle()
	{
		firePropertyChange(PROP_TITLE);
	}
	
	public void parseOnSave() {
//		attachEditor();
		
		if (attachedFramework!=null)
		{
			try {
				// -- Changed by Daniel Varro
				// Execute full parsing
				executeParsing(false);
				if (fParseController != null) {
					if (!fParseController.hasErrors())
					{
						// perform load
						if (fParseController.getModule()!=null)
						{
							// TODO: This code should be rewritten to accommodate the new MachineWrapper
							// check if machine with same name exists
							Machine newm = fParseController.getMachine();
							// TODO: This code should be rewritten to accommodate the new MachineWrapper
							Machine old = (Machine) attachedFramework.getMachineByFQN( newm.getFqn() );
							if (old!=null)
							{
								// TODO: This code should be rewritten to accommodate the new MachineWrapper
								// remove it
								attachedFramework.removeMachine(old.getFqn());
							}
							// load new one
							// TODO: This code should be rewritten to accommodate the new MachineWrapper
							attachedFramework.addMachine(newm.getFqn(), newm);
							this.module = fParseController.getModule();
						}
					}	
					else {
						if (fParseController.getMachine()!=null)
						{
							Machine old = (Machine) attachedFramework.getMachineByFQN( fParseController.getMachine().getFqn() );
							if (old!=null)
							{
								// TODO: This code should be rewritten to accommodate the new MachineWrapper
								// remove it
								attachedFramework.removeMachine(old.getFqn());
							}
						}
					}
					
				}
				
			} catch (Exception e) // TODO: This may be too generic
			{
				//e.printStackTrace();
				if (e.getMessage()!=null)	attachedFramework.getLogger().fatal(e.getMessage());
			}
		}
	}

	/**
	 * Calls the VTCL parser for the current open file as input.
	 * @param isParseOnly: if true then no model building is carried out
	 * @author Daniel Varro
	 */
	public void executeParsing(boolean isParseOnly) {
		try {
			IEditorInput editorInput= getEditorInput();
			IFile file= null;
			String fileName = null;

			VTCLEditor iEditor = Activator.getDefault().getCurrentVTCLEditor();
			if (iEditor == null) 
				return;

			if (editorInput instanceof IFileEditorInput) {
				// Get the file name and the file itself
				IFileEditorInput fileEditorInput= (IFileEditorInput) getEditorInput();
				file= fileEditorInput.getFile();
				fileName = file.getLocation().toOSString();
				
				// Reinitialize markers
				fMarkerManager.deleteMarkers(file, IMarker.PROBLEM);
				fMarkerManager.initialize(iEditor.getViewer(), file);

				// Initialize parsing
				VTCLParserManager parserManager = (VTCLParserManager) attachedFramework.getVTCLParserManager();
				fParseController = parserManager.lookupAndCreateParseController(fileName);
				// TODO: Check that parse controller is already initialized
				//fParseController.initialize(fileName, attachedFramework);
				// Execute parsing
				if (isParseOnly) {
					fParseController.parseOnly();
				}
				else {
					fParseController.parseAndBuild();
				}
				// Refresh marker annotations
				for (ErrorInformation error : fParseController.getErrors()) {
					fMarkerManager.reportError(error);
				}

			} 


		} catch (Exception e) {
			if (e.getMessage()!=null)	attachedFramework.getLogger().fatal("Parsing failed for input " + getEditorInput().getName());
		}

	}

}
