/*******************************************************************************
 * Copyright (c) 2004-2009 Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Istvan Rath - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.frameworkgui.views.console;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.Assert;
import org.eclipse.jface.action.Action;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.CTabFolder;
import org.eclipse.swt.custom.CTabItem;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.TraverseEvent;
import org.eclipse.swt.events.TraverseListener;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Item;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.ViewPart;
import org.eclipse.viatra2.framework.FrameworkManager;
import org.eclipse.viatra2.framework.IFramework;
import org.eclipse.viatra2.framework.IFrameworkListChangedListener;
import org.eclipse.viatra2.framework.properties.VPMProperties;
import org.eclipse.viatra2.frameworkgui.views.code.ColorFactory;
import org.eclipse.viatra2.frameworkgui.views.console.commands.CommandExecutor;
import org.eclipse.viatra2.frameworkgui.views.console.commands.IVIATRAConsoleCommandProvider;
import org.eclipse.viatra2.frameworkgui.views.console.commands.ParsedCommand;
import org.eclipse.viatra2.logger.Logger;

/**
 * GUI class to implements the VIATRA2 Console.
 * An interactive shell to the VIATRA engine.
 * @author istvan rath
 *
 */
public class ConsoleView extends ViewPart implements IFrameworkListChangedListener
{
	
	public static final String ID = "org.eclipse.viatra2.gui.console";
	
	/**
	 * True, if the output is buffered (i.e. should only be refreshed on
	 * transaction end).
	 */
	public boolean isBuffering = true;
	
	/**
	 * Internal GUI back-end class that implements a "model" for a given page
	 * corresponding to a model space. 
	 * @author istvan
	 *
	 */
	protected class ConsolePage implements Logger
	{
		/**
		 * Text used to display contents of the log.
		 */
		protected Text defaultDisplay;
	
		/**
		 * We accumulate log output here.
		 */
		protected StringBuffer defaultBuffer;
				
		/**
		 * Reference to the framework we're listening to.
		 */
		protected String frameworkId;
		
		/**
		 * Points to the tabItem in the tabFolder we're displayed in.
		 */
		protected Item externalTabItem;

		/**
		 * Text field where commands are entered.
		 */
		private Text commandField;
		
		
		public ConsolePage(String id)
		{
			frameworkId = id;
			defaultBuffer = new StringBuffer();
			
			// register as code output plugin
			IFramework fw =  FrameworkManager.getInstance().getFramework(frameworkId);
			fw.addLoggerListener(this);
			
			// register as transaction listener
			//fw.getTopmodel().getNotificationManager().addAllListener(this);
		}
		
		public void disposePage()
		{
			// deregister as logger listener
			IFramework fw =  FrameworkManager.getInstance().getFramework(frameworkId);
			if (fw!=null) {
				fw.removeLoggerListener(this);
			}
			// deregister as transaction listener
			//fw.getTopmodel().getNotificationManager().removeAllListener(this);
			// dispose item
			externalTabItem.dispose();
		}
		
		private int currentBackwardIndex = 0;
		
		private int currentContentAssistIndex = 0;
		
		public Control createPageControl(Composite parent)
		{
			parent.setLayout(new FillLayout());
			Composite c = new Composite(parent, SWT.NONE);
			c.setLayout(new SimpleBorderLayout());
			
			
			defaultDisplay = new Text(c, SWT.MULTI | SWT.V_SCROLL | SWT.H_SCROLL | SWT.READ_ONLY);
			defaultDisplay.setLayoutData(SimpleBorderLayout.CENTER);
			
			if (System.getProperty("file.separator").equals("/"))
			{
				FontData fd = new FontData("Monospace", 10, SWT.NONE);
				defaultDisplay.setFont(new Font(Display.getDefault(), fd));
			}
			else
			{
				FontData fd = new FontData("Courier New", 10, SWT.NONE);
				defaultDisplay.setFont(new Font(Display.getDefault(), fd));
			}
			
			defaultDisplay.setBackground(ColorFactory.getColor(255,255,255));
			defaultDisplay.setForeground(ColorFactory.getColor(0, 0, 0));
			
	

			commandField = new Text(c, SWT.BORDER|SWT.SINGLE);
			commandField.setLayoutData(SimpleBorderLayout.SOUTH);
			commandField.setText("");
			commandField.addTraverseListener(new TraverseListener() {
				
				public void keyTraversed(TraverseEvent e) {
					// disable traversal on tab, to allow for tab-triggered content assist
					if (e.keyCode == SWT.TAB) {
						e.doit = false;
					}
					
				}
			});

			commandField.addKeyListener(new KeyListener() {
				
				public void keyReleased(KeyEvent e) {
					IFramework fw =  FrameworkManager.getInstance().getFramework(frameworkId);
					if (e.keyCode == SWT.CR) {
						// print first execution feedback
						ParsedCommand cmd = CommandExecutor.getParsedCommand(commandField.getText());
						if (cmd.isValid) {
							defaultBuffer.append("> "+cmd.cmdName+"("+cmd.params.toString()+")\n");
							if (CommandExecutor.executeCommand(fw,cmd)) {
								// add to command history
								CommandExecutor.storeCommand(fw, commandField.getText());
								// clear the current contents of the command entry field
								commandField.setText("");
								// reset the backward index
								currentBackwardIndex = 0;
							}
							else {
								fw.getLogger().error("Could not find VIATRA Console command by name: "+cmd.cmdName);
							}		
						}
						else {
							fw.getLogger().error("Parsing error for command: " + commandField.getText());
						}
					}
					else if (e.keyCode == SWT.ARROW_UP) {
						// fetch the previous command in the history if available
						currentBackwardIndex++;
						String oldCmd = CommandExecutor.getCommandAtBackwardIndex(fw, currentBackwardIndex);
						if (oldCmd!=null) {
							commandField.setText(oldCmd);
							commandField.selectAll();
						}
					}
					else if (e.keyCode == SWT.ARROW_DOWN) {
						// fetch the previous command in the history if available
						currentBackwardIndex--;
						String oldCmd = CommandExecutor.getCommandAtBackwardIndex(fw, currentBackwardIndex);
						if (oldCmd!=null) {
							commandField.setText(oldCmd);
							commandField.selectAll();
						}
						else {
							commandField.setText("");
						}
					}
					else if (e.keyCode == SWT.TAB) {
						// provide very simple content-assist :)
						// consider only non-selected part of text as a prefix
						String prefix = commandField.getText();
						if (commandField.getSelectionCount()>0) {
							prefix = commandField.getText().substring(0, commandField.getText().indexOf(commandField.getSelectionText()));
						}
						List<IVIATRAConsoleCommandProvider> matches = CommandExecutor.getCommandByNamePrefix(fw, prefix);
						if (matches.size()>0) {
							currentContentAssistIndex = Math.max(0, currentContentAssistIndex);
							currentContentAssistIndex = Math.min(matches.size()-1, currentContentAssistIndex);
							commandField.setText(matches.get(currentContentAssistIndex).getCommandSignature());
							commandField.setSelection(prefix.length(), commandField.getText().length());
							if (matches.size() == currentContentAssistIndex+1) {
								currentContentAssistIndex=0;
							}
							else {
								currentContentAssistIndex++;
							}
						}
					}
				}
				
				public void keyPressed(KeyEvent e) {
				}
			});
			
			return c;
		}
		
		
		
		

		public void refreshContent()
		{
			Display.getDefault().syncExec(new Runnable()
			{
				public void run() 
				{
					defaultDisplay.setText(defaultBuffer.toString());
					// autoscroll to end 
					defaultDisplay.setSelection(defaultBuffer.length()-1);
				}	
			});
		}
		
		public void clear()
		{
			defaultBuffer = new StringBuffer();
			refreshContent();
		}
/*				
		public void actionPerformed(ICoreNotificationObject notification) 
		{
			if (!isBuffering) return;
			String type = notification.getActionType();
			if (type.equals(ICoreNotificationObject.TA_TRANSACTION_END))
			{
				refreshContent();
			}
		}
*/
		public int getListenerCategory() { return 0; }

		public void debug(String s) {
			if (debugs) {
			defaultBuffer.append("[debug] "+s+"\n");
			refreshContent();
			}
		}

		public void error(String s) {
			if (errors) {
			defaultBuffer.append("[error] "+s+"\n");
			refreshContent();
			}
		}

		public void fatal(String s) {
			defaultBuffer.append("[fatal] "+s+"\n");
			refreshContent();
		}

		public void info(String s) {
			if (infos) {
			defaultBuffer.append("[info] "+s+"\n");
			refreshContent();
			}
		}

		public void init(VPMProperties p) {
			// empty
			
		}

		public void message(int level, String msg, Throwable cause) {
			message(level,msg);
			printStackTrace(cause);
		}

		public void message(int level, String msg) {
			defaultBuffer.append("[message ("+level+")] "+msg+"\n");
			refreshContent();
		}

		public void printStackTrace(Throwable t) {
			StringWriter sw = new StringWriter();
			t.printStackTrace(new PrintWriter(sw));
			defaultBuffer.append("[exception] stack trace follows for: "+
					t.getClass().getCanonicalName()+": "+t.getMessage()+"\n");
			defaultBuffer.append(sw.getBuffer());
			refreshContent();
		}

		public void setLevel(int l) {
			
		}

		public void warning(String s) {
			if (warnings) {
			defaultBuffer.append("[warning] "+s+"\n");
			refreshContent();
			}
		}

	}
	
	
	protected CTabFolder tabFolder;

	/**
	 * frameworkId ==> ConsolePage map
	 */
	private Map<String, ConsolePage> pages = new HashMap<String, ConsolePage>();
		
	
	private CTabFolder getTabFolder()
	{
		return tabFolder;
	}
	
	protected int getActivePage() {
		CTabFolder tabFolder = getTabFolder();
		if (tabFolder != null && !tabFolder.isDisposed()) {
			return tabFolder.getSelectionIndex();
		}
		return -1;
	}
	
	protected int getPageCount() {
		CTabFolder folder = getTabFolder();
		// May not have been created yet, or may have been disposed.
		if (folder != null && !folder.isDisposed()) {
			return folder.getItemCount();
		}
		return 0;
	}
	
	protected void setActivePage(int pageIndex) {
		Assert.isTrue(pageIndex >= 0 && pageIndex < getPageCount());
		getTabFolder().setSelection(pageIndex);
		pageChange(pageIndex);
	}
	
	private CTabItem getItem(int pageIndex) {
		return getTabFolder().getItem(pageIndex);
	}
	
	protected void setPageText(int pageIndex, String text) {
		getItem(pageIndex).setText(text);
	}
	
	protected String getPageText(int pageIndex) {
		return getItem(pageIndex).getText();
	}
	
	@Override
	public void createPartControl(Composite parent) 
	{
		parent.setLayout(new FillLayout());
		tabFolder = new CTabFolder(parent, SWT.TOP | SWT.FLAT);
		tabFolder.setSimple(false);
		tabFolder.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				int newPageIndex = tabFolder.indexOf((CTabItem) e.item);
				pageChange(newPageIndex);
			}
		});
		
		createPages();
		if (getPageCount() > 0 && getActivePage() == -1)
			setActivePage(0);
		
		// register frameworklistchanged listener
		FrameworkManager.getInstance().addFrameworkListChangedListener(this);
		
		// add toolbar actions
		addToolbarActions();
	}

	private void addToolbarActions()
	{
		// add save action
//		getViewSite().getActionBars().getToolBarManager().add(new SaveAllBuffersAction(this));
//		getViewSite().getActionBars().getToolBarManager().add(new SaveOutputAction(this));
		
		Action action = new Action("Clear log") 
		{
			public void run() 
			{
				int outerindex = getActivePage();
				ConsolePage p = ((ConsolePage)getItem( outerindex ).getData());
				p.clear();
			}
		};
		action.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages().getImageDescriptor(ISharedImages.IMG_TOOL_DELETE));
		action.setDisabledImageDescriptor(PlatformUI.getWorkbench().getSharedImages().getImageDescriptor(ISharedImages.IMG_TOOL_DELETE_DISABLED));
		//action.setHoverImageDescriptor(PlatformUI.getWorkbench().getSharedImages().getImageDescriptor(ISharedImages.IMG_TOOL_DELETE_HOVER));
		action.setToolTipText("Clear log");
		getViewSite().getActionBars().getToolBarManager().add(action);
	
		// add toggle actions
		getViewSite().getActionBars().getToolBarManager().add(new ToggleAction(this, "Info", "infos"));
		getViewSite().getActionBars().getToolBarManager().add(new ToggleAction(this, "Warning", "warnings"));
		getViewSite().getActionBars().getToolBarManager().add(new ToggleAction(this, "Error", "errors"));
		getViewSite().getActionBars().getToolBarManager().add(new ToggleAction(this, "Debug", "debugs"));
	}
	
	public boolean warnings = true;
	public boolean errors = true;
	public boolean infos = true;
	public boolean debugs = false;
	
	public ConsolePage getCurrentOuterPage()
	{
		int outerindex = getActivePage();
		ConsolePage p = ((ConsolePage)getItem( outerindex ).getData());
		return p;
	}
	
	protected void createPages()
	{
		for (String id : FrameworkManager.getInstance().getAllFrameWorks())
		{
			createNewPage(id);
		}
	}
	
	private void createNewPage(String _id)
	{
		final String id = _id;
		
//		Display.getDefault().syncExec(new Runnable(){

//			public void run() {
				ConsolePage p = new ConsolePage(id);
				pages.put(id, p);
				int index = getPageCount();
				Control control = p.createPageControl(getTabFolder());
				CTabItem item = new CTabItem(getTabFolder(), SWT.NONE, index);
				item.setControl(control);
				item.setData(p);
				p.externalTabItem = item;
				setPageText(index, id);
//			}
			
//		});
	}
	
	private void removePage(String _id)
	{
		final String id = _id;
		Display.getDefault().syncExec(new Runnable(){

			public void run() 
			{
				ConsolePage p = pages.get(id);
				if (p!=null)
				{
					p.disposePage();
					pages.remove(id);
				}
			}
			
		});
	}
	
	protected void pageChange(int newPageIndex)
	{
		// currently empty
	}
	
	
	@Override
	public void setFocus() {
		setFocus(getActivePage());
	}
	
	private void setFocus(int pageIndex)
	{
		// currently empty
	}

	public void frameworkAdded(final IFramework f) 
	{
		Display.getDefault().syncExec(new Runnable(){
		
			public void run() {
				// add a new page
				createNewPage(f.getId());
				// set it as active
				setActivePage(getPageCount()-1);
			}
		});	
	}

	public void frameworkRemoved(final IFramework f) 
	{
		Display.getDefault().syncExec(new Runnable(){
			
			public void run() {
				// remove page
				removePage(f.getId());
				
			}
		});
	}
	
	
	@Override
	public void dispose() {
		FrameworkManager.getInstance().removeFrameworkListChangedListener(this);
		for (String id :  pages.keySet()) {
			removePage(id);
		}
		super.dispose();
		
	}
	
}
