/*******************************************************************************
 * Copyright (c) 2008-2010, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *   Istvan Rath - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.frameworkgui.views.console.commands.xform;

import java.util.Map;
import java.util.TreeMap;

import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.PlatformUI;
import org.eclipse.viatra2.errors.VPMRuntimeControlledException;
import org.eclipse.viatra2.errors.VPMRuntimeException;
import org.eclipse.viatra2.framework.FrameworkException;
import org.eclipse.viatra2.framework.IFramework;
import org.eclipse.viatra2.frameworkgui.FrameworkGUIPlugin;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Machine;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Rule;
import org.eclipse.viatra2.interpreters.IProgressReport;

/**
 * Helper class which contains utility functions that provide
 * a (hopefully temporary) "wrapper" around the core VIATRA2 engine API to:
 * - execute transformations
 * - invoke GTASM rules
 * - retrieve graph pattern matching sets
 * 
 * @author istvan rath
 *
 */
public class TransformationExecutionHelper {

	
	
	// phases: 
	// 1. parameter map assembly, datatype conversion
	// 2. set up IRunnableWithProgress
	// 2a. if necessary, print out transaction ID for undoable transactions
	// 3. execute through FrameworkGUIPlugin
	
	/**
	 * Takes a parameter string in the key=value;key2=value2; format and provides
	 * a key->value, key2->value structured map in return.
	 */
	public static Map<String, Object> parseParameters(String paramStr) {
		TreeMap<String, Object> ret = new TreeMap<String, Object>();
		for (String entry : paramStr.split(";")) {
			if (entry.length()>0 && entry.indexOf("=")>-1) {
				String key = entry.substring(0, entry.indexOf("=")); 
				String value = entry.substring(entry.indexOf("=")+1);
				// datatype conversion 
				ret.put(key, convertToPassableType(value));
			}
		}
		return ret;
	}
	
	public static boolean checkParameterCompatibility(IFramework fw, Object m, Map<String, Object> params) {
		String[] paramNames;
		try {
			paramNames = fw.getEntrypointParameters(m);
		} catch (FrameworkException e) {
			return false;
		}
		boolean ret = true;
		if (paramNames!=null && params!=null) {
			for (String param : paramNames) {
				ret &= (params.get(param)!=null);
			}
		}
		return ret;
	}
	
	public static Object convertToPassableType(String val) {
		Object o = val;
		try { o = new Double(Double.parseDouble(val)); } catch (NumberFormatException e1) { }
		try { o = new Integer(Integer.parseInt(val)); } catch (NumberFormatException e1) {}
		//try { o = new Boolean(Boolean.parseBoolean(val)); } catch () // this is not needed for the current core
		// no lookup performed here!
		return o;
	}
	
	
	private static IRunnableWithProgress createRunnable(final IFramework fw, final Object entryPoint, final Map<String, Object> params, final boolean undoable) {
		return new IRunnableWithProgress()
		{
			public void run(final org.eclipse.core.runtime.IProgressMonitor monitor) throws java.lang.reflect.InvocationTargetException ,InterruptedException 
			{
				// running phase
				long start = 0;
				String transID = "";
				try
				{
					monitor.beginTask("VIATRA2 Transformation executing", 1000000); // guess
					start = System.currentTimeMillis();
					if (undoable) {
						transID = fw.getTopmodel().getTransactionManager().beginUndoableTransaction(Boolean.TRUE);
					}
					else {
						transID = fw.getTopmodel().getTransactionManager().beginTransaction(Boolean.TRUE);
					}
					monitor.worked(2);
					fw.runEntrypoint(entryPoint, params, new IProgressReport(){
						public void progress(int indicator) {
							monitor.worked(indicator);
						}
					}); 
				}
				catch (final VPMRuntimeControlledException rex)
				{
					Display.getDefault().syncExec(new Runnable(){
					public void run() {
						MessageDialog.openInformation(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "VIATRA2 R3", rex.getMessage() + " (see the Error Log view for details)");							
						}	
					});
				}				
				catch (final VPMRuntimeException rex)
				{
					Display.getDefault().syncExec(new Runnable(){
					public void run() {
						MessageDialog.openInformation(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "VIATRA2 R3", rex.getMessage() + " (see the Error Log view for details)");							
						}	
					});
				}
				finally
				{
					fw.getTopmodel().getTransactionManager().commitTransaction();
					long stop = System.currentTimeMillis();
					monitor.done();
					fw.getLogger().info("Transformation engine ran for "+(stop-start)+" milliseconds.");
					if (undoable) {
						fw.getLogger().info("Transformation transaction ID: "+transID);
					}
				}	
			}
		};
	}
	
	
	public static void runRule(final IFramework fw, final Rule r, final Map<String, Object> params, final boolean undoable) {
		FrameworkGUIPlugin.runAsLongTransaction(createRunnable(fw, r, params, undoable), "VIATRA2 Transformation: "+r.getFqn(), fw);
	}
	
	
	public static void runMachine(final IFramework fw, final Machine m, final Map<String, Object> params, final boolean undoable) {
		FrameworkGUIPlugin.runAsLongTransaction(createRunnable(fw, m, params, undoable), "VIATRA2 Transformation: "+m.getFqn(), fw);
	}
}
