/*******************************************************************************
 * Copyright (c) 2004-2008 Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Istvan Rath - initial API and implementation
 *******************************************************************************/

 package org.eclipse.viatra2.frameworkgui.wizards.jet;

import java.io.ByteArrayInputStream;
import java.io.InputStream;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.emf.codegen.jet.JETEmitter;
import org.eclipse.emf.codegen.jet.JETException;
import org.eclipse.viatra2.frameworkgui.wizards.model.FileTemplateExtension;

/**
 * This class encapsulates access to the JET engine.
 * 
 * @author Istvan Rath based on work by Remko Popma
 */
public class JETGateway {

	private FileTemplateExtension template = null;

	public JETGateway(FileTemplateExtension t) {
		template = t;
	}

	/**
	 * Invokes the JET template specified in the <code>Config</code> with the
	 * model specified in the <code>Config</code> and returns the generated text
	 * as a String.
	 * <p>
	 * This implementation uses a <code>JETEmitter</code> to translate the
	 * template to a Java implementation class. The translated class is created
	 * in a hidden project called <code>.JETEmitters</code>.
	 * <p>
	 * In order to be able to compile the translated template implementation
	 * class, the classes used by the model specified in the <code>Config</code>
	 * must be available in the classpath. For this reason, this method sets the
	 * first runtime library of the plugin specified in the <code>Config</code>
	 * as a classpath variable to the <code>.JETEmitters</code> project.
	 * 
	 * @param monitor
	 *            the progress monitor to use. May be <code>null</code>.
	 * @return the source code text generated by the JET template
	 * @throws CoreException
	 * @throws JETException
	 */
	public String generate(IProgressMonitor monitor) throws CoreException,
			JETException {
		monitor = createIfNull(monitor);

		String pluginId = template.contributorID;
		String base = Platform.getBundle(pluginId).getEntry("/").toString();
		String relativeUri = template.templateFile;

		JETEmitter emitter = new JETEmitter(base + relativeUri, getClass()
				.getClassLoader());

		IProgressMonitor sub = new SubProgressMonitor(monitor, 1);
		try {
			//String result = emitter.generate(sub, template.params.toArray());
			String result = emitter.generate(sub, new Object[]{template.getParameterMap()});
			monitor.worked(1);
			return result;	
		} catch (Exception e)
		{
			e.printStackTrace();
		}
		return "";
	}

	/**
	 * Saves the specified contents to a location specified by the
	 * <code>Config</code> settings. The location of the file to save is found
	 * by finding or creating the container (folder) for the <code>Config</code>
	 * 's package in the <code>Config</code>'s target folder. The name of the
	 * file to save is the <code>Config</code>'s target file.
	 * 
	 * @param monitor
	 *            the progress monitor to use. May be <code>null</code>.
	 * @param contents
	 *            the byte contents of the file to save
	 * @throws CoreException
	 */
	public void save(IFile targetFile, IProgressMonitor monitor, byte[] contents)
			throws CoreException {
		monitor = createIfNull(monitor);

		InputStream newContents = new ByteArrayInputStream(contents);
		if (targetFile.exists()) {
			targetFile.setContents(newContents, true, true,
					new SubProgressMonitor(monitor, 1));
		} else {
			targetFile.create(newContents, true, new SubProgressMonitor(
					monitor, 1));
		}
	}

	/**
	 * Returns a non-null progress monitor.
	 * 
	 * @param monitor
	 *            an existing progress monitor
	 * @return a new <code>NullProgressMonitor</code> if the specified monitor
	 *         is <code>null</code>, or the existing monitor otherwise
	 */
	private IProgressMonitor createIfNull(IProgressMonitor monitor) {
		if (monitor == null) {
			return new NullProgressMonitor();
		}
		return monitor;
	}

}
