/*******************************************************************************
 * Copyright (c) 2004-2008 Peter Pasztor, Akos Horvath, Gergely Varro, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Peter Pasztor, Akos Horvath, Gergely Varro, Istvan Rath - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.gtasm.interpreter.impl.rules;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import java.util.Vector;

import org.eclipse.emf.common.util.EList;
import org.eclipse.viatra2.core.IEntity;
import org.eclipse.viatra2.core.IModelElement;
import org.eclipse.viatra2.gtasm.interpreter.exception.ASMInterpreterErrorStrings;
import org.eclipse.viatra2.gtasm.interpreter.exception.ViatraTransformationException;
import org.eclipse.viatra2.gtasm.interpreter.executionEnvironment.IExecutionEnvironment;
import org.eclipse.viatra2.gtasm.interpreter.impl.executionEnvironment.BlockRuleExecutionEnvironment;
import org.eclipse.viatra2.gtasm.interpreter.impl.executionEnvironment.ExecutionEnvironment;
import org.eclipse.viatra2.gtasm.interpreter.impl.machine.ASMInterpreterException;
//import org.eclipse.viatra2.gtasm.interpreter.impl.threads.ThreadPool;
import org.eclipse.viatra2.gtasm.interpreter.term.rules.TermEvaluator;
import org.eclipse.viatra2.gtasm.patternmatcher.ExecutionMode;
import org.eclipse.viatra2.gtasm.patternmatcher.IMatching;
import org.eclipse.viatra2.gtasm.patternmatcher.ParameterMode;
import org.eclipse.viatra2.gtasm.patternmatcher.PatternCallSignature;
import org.eclipse.viatra2.gtasm.patternmatcher.PatternMatcherParameters;
import org.eclipse.viatra2.gtasm.patternmatcher.Scope;
import org.eclipse.viatra2.gtasm.patternmatcher.patterns.IGTRuleMatcher;
import org.eclipse.viatra2.gtasm.patternmatcher.patterns.IPatternMatcher;
import org.eclipse.viatra2.gtasm.patternmatcher.patterns.PatternMatcherProvider;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.compoundRules.BlockRule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.compoundRules.ChooseRule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.compoundRules.CollectionIteratorRule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.compoundRules.ForallRule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.compoundRules.LetRule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.core.RuntimeAnnotation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.core.RuntimeAnnotationElement;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.SymbolicRuleParameter;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Variable;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.VariableDefinition;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.enums.DirectionKind;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.enums.ValueKind;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.simpleRules.ASMRuleInvocation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.simpleRules.GTRuleInvocation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.GTPatternCall;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.Term;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.VariableReference;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.ContainmentConstraint;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTPattern;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTRule;
import org.eclipse.viatra2.interpreters.IProgressReport;



/** Interprets the block rules of the GTASM language 
 * @author Peter Pasztor, Akos Horvath and Istvan Rath
 *
 */
public class BlockRuleInterpreter extends RuleInterpreter {

	private static BlockRuleInterpreter _instance = new BlockRuleInterpreter();

	private BlockRuleInterpreter()
	{
		;
	}
	
	public static BlockRuleInterpreter getInstance() { 
		return _instance; 
	} 

	@Override
	public Boolean interpretRule(IExecutionEnvironment executionEnvironment, ASMRuleInvocation ruleToBeInterpreted, IProgressReport pr) 
		throws ViatraTransformationException
	{
		
		BlockRule blockRule=(BlockRule)ruleToBeInterpreted;
		if(blockRule instanceof LetRule)
		{
			LetRule letBlockRule = (LetRule)ruleToBeInterpreted;
			return interpretLetRule(executionEnvironment,letBlockRule,pr);
		}
		else if(blockRule instanceof ForallRule)
		{
			ForallRule forallBlockRule = (ForallRule) blockRule;
			if (forallBlockRule.getGtrule()==null)
			{
				if(forallBlockRule.getCondition() instanceof GTPatternCall)
				{
//FORALL with GTPATTERN
					GTPatternCall patternCall = (GTPatternCall)forallBlockRule.getCondition();
					boolean isParallel = false;
					int threadPoolSize = 0;
					for (Object _ra : patternCall.getCalledPattern().getRuntimeAnnotations())
					{
						RuntimeAnnotation ra = (RuntimeAnnotation) _ra;
						if (ra.getAnnotationName().equalsIgnoreCase("@parallelforall"))
						{
							isParallel = true;
							for(Object _p : ra.getElements())
							{
								RuntimeAnnotationElement p = (RuntimeAnnotationElement) _p;
								if (p.getKey().equalsIgnoreCase("threadpoolsize"))
								{
									threadPoolSize = Integer.parseInt(p.getValue());
								}
							}
							break;
						}
					}
//					if (isParallel)
//						return interpretForallRulewithGTPattern_Parallel(executionEnvironment, forallBlockRule, pr, threadPoolSize);
//					else
						return interpretForallRulewithGTPattern(executionEnvironment, forallBlockRule,pr);
				}
				else 
				{
//FORALL with ASMFUNCTION
					return interpretForallRulewithASMFunction(executionEnvironment, forallBlockRule,pr);
				}
			}
			else
			{
//FORALL with GTRULE
				return interpretForallRulewithGTRule(executionEnvironment, forallBlockRule,pr);
			}
		}
		else if(blockRule instanceof ChooseRule)
		{
			ChooseRule chooseBlockRule = ((ChooseRule)blockRule);

			if (chooseBlockRule.getGtrule()==null) {
				if(chooseBlockRule.getCondition() instanceof GTPatternCall)
				{
//CHOOSE with GTPATTERN			
					return interpretChooseRulewithGTPattern(executionEnvironment, chooseBlockRule,pr);	
				}
				else
				{
//CHOOSE with ASMFUNCTION				
				return interpretChooseRulewithASMFunction(executionEnvironment, chooseBlockRule,pr);
				}
			}
			else
			{
//CHOOSE with GTRULE
				return interpretChooseRulewithGTRule(executionEnvironment, chooseBlockRule,pr);
			}
		}
		return Boolean.FALSE;
	}
	
	
	private Boolean interpretLetRule(IExecutionEnvironment executionEnvironment, LetRule letBlockRule, IProgressReport pr) 
		throws ViatraTransformationException
	{
		BlockRuleExecutionEnvironment letRuleExecutionEnvironment =
			new BlockRuleExecutionEnvironment(executionEnvironment.getFramework(),letBlockRule.getBody());
		Hashtable<Variable, Object> childExecutionEnvironmentVariables=new Hashtable<Variable, Object>();
		childExecutionEnvironmentVariables.putAll(executionEnvironment.getVariableValues());
		
		//evaluate the variable definitions' values.
		for (Object variableDefinition : letBlockRule.getDefinitions()) {
            // This long line puts the variable's values into the execution environment. If the parser gave a null value,
            // a ValueKind.UNDEF_LITERAL value is substituted.
			VariableDefinition varDef  = (VariableDefinition)variableDefinition;
			childExecutionEnvironmentVariables.put(varDef.getVariable(),(varDef.getValue())==null?
					ValueKind.UNDEF_LITERAL:TermEvaluator.getInstance().evaluate(executionEnvironment, varDef.getValue()));
		}
		
		letRuleExecutionEnvironment.onBegin(childExecutionEnvironmentVariables);
		Boolean result= RuleInterpreter.getInstance().interpretRule(letRuleExecutionEnvironment,letBlockRule.getBody(),pr);
		childExecutionEnvironmentVariables=(Hashtable<Variable, Object>)letRuleExecutionEnvironment.onTerminate();

		for (Variable variable : executionEnvironment.getVariableValues().keySet()) {
			try {
				executionEnvironment.setVariableValue(variable, childExecutionEnvironmentVariables.get(variable));
			} catch (ViatraTransformationException e) {
				String[] context = {variable.getName(),e.getMessage()};
				throw new ASMInterpreterException(
						ASMInterpreterErrorStrings.SET_VAR_VALUE_AFTER_NATIVE_CALL
						,context
						,letBlockRule);
			}
		}
		return result;
	}
	
	private Boolean interpretForallRulewithGTRule(IExecutionEnvironment executionEnvironment, ForallRule forallBlockRule, IProgressReport pr) 
		throws ViatraTransformationException{
		HashMap<Variable, Scope> parameterParameterScopes = new HashMap<Variable, Scope>();
		GTRule gtRule = forallBlockRule.getGtrule().getRule();
		Collection<IMatching> lHSmatches;
		//Needs a new Exec.env. for the ***DO*** part of the rule
		BlockRuleExecutionEnvironment forallRuleExecutionEnvironment = getNewExecutionEnvironment(executionEnvironment, forallBlockRule);

		//New GTEXEC. env. for the action part of the gtRule 
		IExecutionEnvironment gtEnvironment= new BlockRuleExecutionEnvironment(executionEnvironment.getFramework());
		
		// get scopes for the gtRule
		parameterParameterScopes = getScopesFromContainmentConstraints(forallRuleExecutionEnvironment
				, forallBlockRule);
	
		// get matcher parameters
		PatternMatcherParameters mp=getGTRuleMatchParameters(forallRuleExecutionEnvironment
				,forallBlockRule
				, parameterParameterScopes
				, gtRule
				, forallBlockRule.getGtrule());
		
			// Get patternMatcher
		IGTRuleMatcher gtMatcher;
		
			//evaluate the matches of the LHS for the gtRule
		try{
				gtMatcher = PatternMatcherProvider.getInstance().getGtRuleMatcher(gtEnvironment,gtRule,mp);
				lHSmatches =gtMatcher.initMatchAll(mp.getCallSignature(), mp.getInputMapping(), mp.getQuantificationOrder());
			} catch (ViatraTransformationException e1) {
				throw e1.addNewStackElement(forallBlockRule);
			}
				
		if(lHSmatches != null && lHSmatches.size() != 0)
		{
			// reorder if neccessary
			lHSmatches = reorder(gtRule.getPrecondition().getCalledPattern(), lHSmatches);

		// the RHS\LHS and action part of the rule are evaluated by the matchAll method	
			for(IMatching lHSMatch: lHSmatches)
			{		//get matches
					Object[] match;
					try {
						match = gtMatcher.matchAll(lHSMatch, mp.getCallSignature(),mp.getInputMapping());
					} catch (ViatraTransformationException e1) {
						throw e1.addNewStackElement(forallBlockRule);
					}
					
					//Bind the variables for the 'DO' part of the rule
					evaulateDoPartofGTRuleInvocation(gtRule
							,match
							,forallBlockRule
							,forallRuleExecutionEnvironment
							,forallBlockRule.getGtrule()
							,pr);
			}// end of for cycle for the matches of the GTRULE
		}
	/*************************************************/
		// write the changed values back
		forallRuleExecutionEnvironment.onTerminate();
		for(Variable var :executionEnvironment.getVariableValues().keySet())
			executionEnvironment.setVariableValue(var, forallRuleExecutionEnvironment.getVariableValue(var));
	/*************************************************/	
	return Boolean.TRUE;
	}
	
	private Boolean interpretForallRulewithASMFunction(IExecutionEnvironment executionEnvironment, ForallRule forallBlockRule, IProgressReport pr)
		throws ViatraTransformationException{
		// This will list all the values a symbolic variable might hold.
		HashMap<Variable, Vector<Object>> valueList = new HashMap<Variable, Vector<Object>>();
		Map<Variable, Object> variables = new Hashtable<Variable, Object>(executionEnvironment.getVariableValues());

		for (Object variable : forallBlockRule.getLocalVariables()) {
			valueList.put(((Variable) variable), new Vector<Object>());
			variables.put((Variable)variable, ValueKind.UNDEF_LITERAL);
		}
		
		executionEnvironment.fetchVariableVariations(valueList,forallBlockRule.getCondition());

		// We have a list of possible values for the variables.
		// Now we must get each configuration (just as a Cartesian multiplication
		// would, but being a lot more efficient), and evaluate them using brute force.

		Object[] keys = valueList.keySet().toArray();

		// Needs a new Exec.env. the new Variables will be added later.
		
		BlockRuleExecutionEnvironment forallRuleExecutionEnvironment = 
			new BlockRuleExecutionEnvironment(executionEnvironment.getFramework()
				,forallBlockRule.getBody());

		forallRuleExecutionEnvironment.onBegin(variables);
		// We need a divider vector to be able to iterate through all the variations.
		Vector<Integer> divVect = new Vector<Integer>();
		divVect.add(new Integer(valueList.get(keys[0]).size()));
		Integer maxCount = new Integer(valueList.get(keys[0]).size());
			
		for (int iter = 1; iter < valueList.size(); iter++) {
			maxCount*=valueList.get(keys[iter]).size();
			divVect.add(valueList.get(keys[iter]).size());
			// Don't bother, if the function doesn't have any values. In such cases, this speeds up a lot.
			if(valueList.get(keys[iter]).size()==0) return Boolean.FALSE;
		}
		// To speed up things a bit.
		//Integer maxCount = divVect.lastElement();
		Integer size = divVect.size();
		//Evaluate all configurations
		
		for (Integer iter_config = 0; iter_config < maxCount; iter_config++) {
			Integer config_no = new Integer(iter_config);
		
			// bind the variables
			for (int iter_var = 0; iter_var < size; iter_var++) { 
				//Integer num=config_no-config_no%(iter_var==0?1:divVect.get(iter_var-1))/divVect.get(iter_var);
				Variable variable = (Variable) (valueList.keySet().toArray()[iter_var]);
				try {
					forallRuleExecutionEnvironment.setVariableValue(
							variable,
							valueList.get(valueList.keySet().toArray()[iter_var]).get(config_no	% divVect.get(iter_var)));
				} catch (ViatraTransformationException e) {
					String[] context = {variable.getName(),e.getMessage()};
					throw new ASMInterpreterException(ASMInterpreterErrorStrings.RULE_INIT_VAR_ASMFUNCTION
							,context
							,forallBlockRule);
				}
						config_no = (config_no - config_no % divVect.get(iter_var))/divVect.get(iter_var);
			}
			// Now for the last variable
			//evaluate it in a sandbox-like execution environment.
			
			if (Boolean.TRUE.equals(
					TermEvaluator.getInstance().evaluate(forallRuleExecutionEnvironment,forallBlockRule.getCondition()))) 
			{
				// The condition evaluated to true --> this is a good configuration to interpret
				// the actual rule
				//TODO: if it fails, what to do?
				RuleInterpreter.getInstance().interpretRule(forallRuleExecutionEnvironment,forallBlockRule.getBody(),pr);
			}
		}
		// all runs are done now
		// write the changed values back
		forallRuleExecutionEnvironment.onTerminate();
		for(Variable var :executionEnvironment.getVariableValues().keySet())
			executionEnvironment.setVariableValue(var, forallRuleExecutionEnvironment.getVariableValue(var));
		
		return Boolean.TRUE;
	}
/*	
	private Boolean interpretForallRulewithGTPattern_Parallel(IExecutionEnvironment executionEnvironment, final ForallRule forallBlockRule, final IProgressReport pr, final int threadPoolSize) 
		throws ViatraTransformationException
	{	
		int i=0;
		GTPatternCall  forallGTPatternCall = (GTPatternCall)forallBlockRule.getCondition();
		PatternMatcherParameters mp;
		Collection<IMatching> matches;
		
		// make new execution environment
		final BlockRuleExecutionEnvironment forallRuleExecutionEnvironment = getNewExecutionEnvironment(executionEnvironment, forallBlockRule);
		
		//Get scopes from containmentConstraints
		HashMap<Variable, Scope> parameterParameterScopes =getScopesFromContainmentConstraints(forallRuleExecutionEnvironment,forallBlockRule);
		
		//Get patternMatcher
		IPatternMatcher patternMatcher;
		try
		{
			patternMatcher = PatternMatcherProvider.getInstance().getPatternMatcher(forallRuleExecutionEnvironment,
						forallGTPatternCall.getCalledPattern());
			
			// get matcher parameters
			mp=getMatchParameters(forallRuleExecutionEnvironment,forallBlockRule, parameterParameterScopes);
			
			// get matches
			matches =patternMatcher.matchAll(mp.getInputMapping(), mp.getCallSignature(),mp.getQuantificationOrder() );
			
		} catch (ViatraTransformationException e1) {
			throw e1.addNewStackElement(forallBlockRule);
		}
		
		// create threadpool for execution
		ThreadPool pool = new ThreadPool(threadPoolSize);
		
		// reorder if neccessary
		matches = reorder(forallGTPatternCall.getCalledPattern(), matches);
		
		// Assign the parameters to the new exec.environment
		for (IMatching match : matches) {
			
			// make a copy of the execution environment
			final IExecutionEnvironment _forallRuleExecutionEnvironment = new ExecutionEnvironment(forallRuleExecutionEnvironment);
			
			
			// bind the variables
			// Update variables values from the matching
			
			for (i=0;i< forallGTPatternCall.getActualParameters().size();i++) {
				// OUTPUT checking
				if(mp.getCallSignature()[i].getParameterMode().equals(ParameterMode.OUTPUT)) 
				{ Variable variable = ((VariableReference)forallGTPatternCall.getActualParameters().get(i)).getVariable();
					try {
						_forallRuleExecutionEnvironment.setVariableValue(
								variable,
								match.lookup(i));
					} catch (ViatraTransformationException e) {
						String[] context = {variable.getName(),e.getMessage()};
						throw new ASMInterpreterException(ASMInterpreterErrorStrings.RULE_INIT_VAR_PARALLEL_FORALL
								,context
								,forallBlockRule);
					}
				}
			}
			pool.assign(new Runnable(){
				public void run() {
					try {
						RuleInterpreter.getInstance().interpretRule(_forallRuleExecutionEnvironment,forallBlockRule.getBody(),null);
					} catch (ViatraTransformationException e) {
						e.printStackTrace();
					}		
				}
			});
			//TODO: what if fail???
		}
		pool.complete();

		// write the changed values back -- TODO NO INFORMATION PROPAGATION SUPPORT FOR NOW
		//forallRuleExecutionEnvironment.onTerminate();
		//for(Variable var :org.eclipse.viatra2.gtasm.interpreter.executionEnvironment.getVariableValues().keySet())
		//	org.eclipse.viatra2.gtasm.interpreter.executionEnvironment.setVariableValue(var, forallRuleExecutionEnvironment.getVariableValue(var));
		return Boolean.TRUE;
	}
*/	
	
	
	private Boolean interpretForallRulewithGTPattern(IExecutionEnvironment executionEnvironment, ForallRule forallBlockRule, IProgressReport pr)
		throws ViatraTransformationException {
		
		int i=0;
				GTPatternCall  forallGTPatternCall = (GTPatternCall)forallBlockRule.getCondition();
				PatternMatcherParameters mp;
				Collection<IMatching> matches;
				
				// make new execution environment
				BlockRuleExecutionEnvironment forallRuleExecutionEnvironment = getNewExecutionEnvironment(executionEnvironment, forallBlockRule);
				
				//Get scopes from containmentConstraints
				HashMap<Variable, Scope> parameterParameterScopes =getScopesFromContainmentConstraints(forallRuleExecutionEnvironment,forallBlockRule);
				
				//Get patternMatcher
				IPatternMatcher patternMatcher;
				try{
				patternMatcher = PatternMatcherProvider.getInstance().getPatternMatcher(forallRuleExecutionEnvironment,
							forallGTPatternCall.getCalledPattern());
				
				// get matcher parameters
				mp=getMatchParameters(forallRuleExecutionEnvironment,forallBlockRule, parameterParameterScopes);
				
				// get matches
				matches =patternMatcher.matchAll(mp.getInputMapping(), mp.getCallSignature(),mp.getQuantificationOrder() );
				
				} catch (ViatraTransformationException e1) {
					throw e1.addNewStackElement(forallBlockRule);
				}

				// reorder if neccessary
				matches = reorder(forallGTPatternCall.getCalledPattern(), matches);

				// Assign the parameters to the new exec.environment
				for (IMatching match : matches) {
					// bind the variables
					// Update variables values from the matching
					
					for (i=0;i< forallGTPatternCall.getActualParameters().size();i++) {
						// OUTPUT checking
						if(mp.getCallSignature()[i].getParameterMode().equals(ParameterMode.OUTPUT)) 
						{ Variable variable = ((VariableReference)forallGTPatternCall.getActualParameters().get(i)).getVariable();
							try {
								forallRuleExecutionEnvironment.setVariableValue(
										variable,
										match.lookup(i));
							} catch (ViatraTransformationException e) {
								String[] context = {variable.getName(),e.getMessage()};
								throw new ASMInterpreterException(ASMInterpreterErrorStrings.RULE_INIT_VAR_FORALL
										,context
										,forallBlockRule );
							}
						}
					}
					RuleInterpreter.getInstance().interpretRule(forallRuleExecutionEnvironment,forallBlockRule.getBody(),pr);
					//TODO: what if fail???
				}
				/*************************************************/
				// write the changed values back
				forallRuleExecutionEnvironment.onTerminate();
				for(Variable var :executionEnvironment.getVariableValues().keySet())
					executionEnvironment.setVariableValue(var, forallRuleExecutionEnvironment.getVariableValue(var));
				/*************************************************/
			return Boolean.TRUE;
	}


	private Boolean interpretChooseRulewithGTPattern(
			IExecutionEnvironment executionEnvironment,
			ChooseRule chooseBlockRule, IProgressReport pr) throws ViatraTransformationException {
		
				int i= 0;
				GTPatternCall chooseGTPatternCall = (GTPatternCall) chooseBlockRule.getCondition();
				PatternMatcherParameters mp; 
				IMatching match;
				
				HashMap<Variable, Scope> parameterParameterScopes = new HashMap<Variable, Scope>();
		
				// Needs a new Exec.env. the new Variables will be added also.
				BlockRuleExecutionEnvironment chooseRuleExecutionEnvironment = 
					getNewExecutionEnvironment(executionEnvironment, chooseBlockRule);
		
				// get scopes
				parameterParameterScopes = getScopesFromContainmentConstraints(chooseRuleExecutionEnvironment, chooseBlockRule);
		
				// is true pseudorandom required?
				GTPattern gtPattern = chooseGTPatternCall.getCalledPattern();
				boolean isRandom = false;
				for (Object annot : gtPattern.getRuntimeAnnotations()) {
					if ("@random".equals(((RuntimeAnnotation) annot).getAnnotationName().toLowerCase())) {
						isRandom = true; break;
					}
				}
				
				// Get patternMatcher
				IPatternMatcher patternMatcher;
				try{
					patternMatcher = PatternMatcherProvider.getInstance().getPatternMatcher(chooseRuleExecutionEnvironment, gtPattern);
					// get matcher parameters
					mp = getMatchParameters(chooseRuleExecutionEnvironment, chooseBlockRule,
							parameterParameterScopes);
			
					// get matches
					if (!isRandom) match = patternMatcher.match(mp.getInputMapping(), mp.getCallSignature()); 
					else match = patternMatcher.matchRandomly(mp.getInputMapping(), mp.getCallSignature()); 
				} catch (ViatraTransformationException e1) {
					throw e1.addNewStackElement(chooseBlockRule);
				}
				
				// no match found
				if (match == null)
					return Boolean.FALSE;
		
				// Update variables values from the matching
				for (i = 0; i < chooseGTPatternCall.getActualParameters().size(); i++) {
					if (mp.getCallSignature()[i].getParameterMode().equals(
							ParameterMode.OUTPUT)) {
						Variable variable = ((VariableReference) chooseGTPatternCall.getActualParameters().get(i)).getVariable();
						try {
							chooseRuleExecutionEnvironment.setVariableValue(
									variable , match.lookup(i));
						} catch (ViatraTransformationException e) {
							String[] context = {variable.getName(), e.getMessage()};
							throw new ASMInterpreterException(
									ASMInterpreterErrorStrings.RULE_INIT_VAR_CHOOSE
									,context
									,chooseBlockRule);
						}
					}
				}
				Boolean result = RuleInterpreter.getInstance().interpretRule(chooseRuleExecutionEnvironment
								,chooseBlockRule.getBody(),pr);
				/** ********************************************** */
				// write the changed values back
				chooseRuleExecutionEnvironment.onTerminate();
				for (Variable var : executionEnvironment.getVariableValues().keySet())
					executionEnvironment.setVariableValue(var,
							chooseRuleExecutionEnvironment.getVariableValue(var));
				/** ********************************************** */
				
				return result;
	}
	
	private Boolean interpretChooseRulewithASMFunction(IExecutionEnvironment executionEnvironment, ChooseRule chooseBlockRule, IProgressReport pr)
		throws ViatraTransformationException{

		// This will list all the values a symbolic variable might hold.
		HashMap<Variable, Vector<Object>> valueList = new HashMap<Variable, Vector<Object>>();
		Map<Variable, Object> variables = new Hashtable<Variable, Object>(executionEnvironment.getVariableValues());
		for (Object variable : chooseBlockRule.getLocalVariables()) {
			valueList.put(((Variable) variable), new Vector<Object>());
			variables.put((Variable)variable, ValueKind.UNDEF_LITERAL);
		}
		
		executionEnvironment.fetchVariableVariations(valueList,chooseBlockRule.getCondition());

		// We have a list of possible values for the variables.
		// Now we must get each configuration (just as a Cartesian multiplication
		// would, but being a lot more efficient), and evaluate them using brute force.

		Object[] keys = valueList.keySet().toArray();

		// Needs a new Exec.env. the new Variables will be added later.
		
		BlockRuleExecutionEnvironment chooseRuleExecutionEnvironment = 
			new BlockRuleExecutionEnvironment(executionEnvironment.getFramework(),
				chooseBlockRule.getBody());
		
		chooseRuleExecutionEnvironment.onBegin(variables);
		// We need a divider vector to be able to iterate through all the variations.
		Vector<Integer> divVect = new Vector<Integer>();
		divVect.add(new Integer(valueList.get(keys[0]).size()));
		Integer maxCount = new Integer(valueList.get(keys[0]).size());
		
		for (int iter = 1; iter < valueList.size(); iter++) {
			//divVect.add(valueList.get(keys[iter]).size()
				//	* divVect.get(iter - 1));
			maxCount*=valueList.get(keys[iter]).size();
			divVect.add(valueList.get(keys[iter]).size());
					

			// Don't bother, if the function doesn't have any values. In such cases, this speeds up a lot.
			if(valueList.get(keys[iter]).size()==0) return Boolean.FALSE;
		}
		// To speed up things a bit.
		//Integer maxCount = divVect.lastElement();
		Integer size = divVect.size();
		//Evaluate all configurations
		
		for (Integer iter_config = 0; iter_config < maxCount; iter_config++) {
			Integer config_no = new Integer(iter_config);
		
			// bind the variables
			for (int iter_var = 0; iter_var < size; iter_var++) { 

				//Integer num=config_no-config_no%(iter_var==0?1:divVect.get(iter_var-1))/divVect.get(iter_var);
				Variable variable = (Variable) (valueList.keySet().toArray()[iter_var]);
				try {
					chooseRuleExecutionEnvironment.setVariableValue(
							variable,
							valueList.get(valueList.keySet().toArray()[iter_var]).get(config_no	% divVect.get(iter_var)));
				} catch (ViatraTransformationException e) {
					String[] context = {variable.getName(),e.getMessage()};
					throw new ASMInterpreterException(ASMInterpreterErrorStrings.RULE_INIT_VAR_ASMFUNCTION
							,context
							,chooseBlockRule);
				}
						config_no = (config_no - config_no % divVect.get(iter_var))/divVect.get(iter_var);
			}
			//evaluate it in a sandbox-like execution environment.

			if (Boolean.TRUE.equals(TermEvaluator.getInstance().evaluate(
							chooseRuleExecutionEnvironment,
							chooseBlockRule.getCondition()))) {
				// The condition evaluated to true --> this is a good configuration to interpret
				// the actual rule
				//TODO: if it fails, what to do?
				return RuleInterpreter.getInstance()
						.interpretRule(chooseRuleExecutionEnvironment,
								chooseBlockRule.getBody(),pr);
				
				/* ***********************************************************
				 * The only difference to ForAllRule!!!!!!!!!!!!!!!!!!!!!!!!!
				 *
				 * THIS RETURNS ON THE FIRST MATCH!	
				/* ***********************************************************/
			}
			// write the changed values back
			chooseRuleExecutionEnvironment.onTerminate();
			for(Variable var :executionEnvironment.getVariableValues().keySet())
				executionEnvironment.setVariableValue(var, chooseRuleExecutionEnvironment.getVariableValue(var));
		}
		return Boolean.FALSE;
	}		
	
	private Boolean interpretChooseRulewithGTRule(IExecutionEnvironment executionEnvironment, ChooseRule chooseBlockRule, IProgressReport pr) 
		throws ViatraTransformationException{
		
		//CHOOSE with GTRULE
		HashMap<Variable, Scope> parameterParameterScopes = new HashMap<Variable, Scope>();
		GTRule gtRule = (chooseBlockRule.getGtrule()).getRule();
		//Collection<IMatching> lHSmatches;
		// Needs a new Exec.env. for the ***DO*** part of the rule
		BlockRuleExecutionEnvironment chooseRuleExecutionEnvironment = 
			getNewExecutionEnvironment(executionEnvironment, chooseBlockRule);

		// New GTEXEC. env. for the action part of the gtRule 
		IExecutionEnvironment gtEnvironment= new BlockRuleExecutionEnvironment(executionEnvironment.getFramework());
		
		// get scopes for the gtRule
		parameterParameterScopes = getScopesFromContainmentConstraints(chooseRuleExecutionEnvironment 
				, chooseBlockRule);
	
		// get matcher parameters
		PatternMatcherParameters mp=getGTRuleMatchParameters(chooseRuleExecutionEnvironment 
				,chooseBlockRule
				, parameterParameterScopes
				, gtRule
				, chooseBlockRule.getGtrule());
		
		// Get patternMatcher
		IGTRuleMatcher gtMatcher;
		
		// get matches
		Object[] match;
		try {
			gtMatcher = PatternMatcherProvider.getInstance().getGtRuleMatcher(gtEnvironment,gtRule,mp);
			match = gtMatcher.match(mp.getCallSignature(), mp.getInputMapping());
			} catch (ViatraTransformationException e1) {
				throw e1.addNewStackElement(chooseBlockRule);
			}
		//Bind the variable for the Do part of the rule
		if(match != null)
		{
			Boolean result= evaulateDoPartofGTRuleInvocation(gtRule
					, match
					, chooseBlockRule
					, chooseRuleExecutionEnvironment
					,chooseBlockRule.getGtrule(),pr);
		//*************************************************/
			// writes the changed values back
			chooseRuleExecutionEnvironment.onTerminate();
			for(Variable var :executionEnvironment.getVariableValues().keySet())
				executionEnvironment.setVariableValue(var, chooseRuleExecutionEnvironment.getVariableValue(var));
		//*************************************************/	
			return result;
		}
		else // there is no match found, choose rule fails
			return Boolean.FALSE;
	}

	/** Returns the transformed scope constraints of a GT pattern or GT rule invocation 
	 * @param org.eclipse.viatra2.gtasm.interpreter.executionEnvironment The execution environment on which the elements are evaluated
	 * @param containmentConstraints The constraints of the invocations
	 * @return The variable and the corresponding Scope in a HashMap
	 * @throws ViatraTransformationException
	 */
	private HashMap<Variable,Scope> getScopesFromContainmentConstraints(IExecutionEnvironment executionEnvironment
			,CollectionIteratorRule rule) throws ViatraTransformationException
	{
		EList<ContainmentConstraint> containmentConstraints = rule.getContainmentConstraints();
		HashMap<Variable,Scope> parameterScopes = new HashMap<Variable,Scope>();
		for (Object cC: containmentConstraints) {
			ContainmentConstraint containmentConstraint = (ContainmentConstraint) cC;
			if(containmentConstraint.getParent()!=null)
			{
				Object constraintParent = TermEvaluator.getInstance().evaluate(executionEnvironment, containmentConstraint.getParent());
						if(constraintParent instanceof IEntity)
						{
							parameterScopes .put((containmentConstraint).getVariable(),
								new Scope( containmentConstraint.getMode().getValue(),
										(IModelElement)(TermEvaluator.getInstance().evaluate(executionEnvironment, containmentConstraint.getParent() ))));
						}
						else
						{
							// Scope parent!= ENTITY, the element the scope references does not exist
							String[] context = {containmentConstraint.getVariable().getName()};
							ViatraTransformationException e = new ASMInterpreterException(ASMInterpreterErrorStrings.SCOPE_NOT_AN_ENTITY
									,context
									,containmentConstraint.getVariable());
							throw e.addNewStackElement(rule);
						}
			}
			else
				parameterScopes .put(containmentConstraint.getVariable(), new Scope());
		}
		return parameterScopes;
	}
	
	/** Evaluates the do part of the GT Rule if it exists
	 * @param gtRule The GTRule invoked
	 * @param match The matching produced by the application of the GT rule
	 * @param blockRule the block rule representing the DO part
	 * @param blockRuleExecutionEnvironment the corresponding execution environment on which we the TERMS evaluated
	 * @param ruleInvocation the actual gtRule invocation
	 * @return true if it is successfully evaluated (or no DO part) else false
	 * @throws ViatraTransformationException
	 */
	private boolean evaulateDoPartofGTRuleInvocation(GTRule gtRule
			, Object[] match
			, BlockRule blockRule
			, IExecutionEnvironment blockRuleExecutionEnvironment
			, GTRuleInvocation ruleInvocation
			, IProgressReport pr)
	throws ViatraTransformationException
	{

	// the block rule does not have a do part
	if(blockRule.getBody() == null)
		return Boolean.TRUE;
	
	GTRuleInvocation gtRuleInvocation = ((CollectionIteratorRule) blockRule).getGtrule();
	
		for (int i=0;i<gtRule.getSymParameters().size();i++) {
			//Only OUT and INOUT parameters have to be updated
			if(!(gtRule.getSymParameters().get(i)).getDirection().equals(DirectionKind.IN_LITERAL) ) 
			{
				if(match[i].equals(ValueKind.UNDEF_LITERAL))
					//have to check that it is not a quantified variable
					{
					Object ActualTerm = ruleInvocation.getActualParameters().get(i);
					if(!(ActualTerm instanceof VariableReference 
							// the CollectionIterator rule's actual parameter
							&& ((CollectionIteratorRule)blockRule).getLocalVariables().contains(((VariableReference)ActualTerm).getVariable())))
						{
						String[] context = {(gtRule.getSymParameters().get(i)).getName(), gtRule.getName()};
						throw new ASMInterpreterException(ASMInterpreterErrorStrings.GTOUTPUT_IN_INOUT_NOT_BOUND
								,context
								,blockRule);
						}
					}
//				else	
//					throw new ASMInterpreterException(((SymbolicRuleParameter)gtRule.getSymParameters().get(i)).getName()+ASMInterpreterErrorStrings.GTOUTPUT_IN_INOUT_NOT_BOUND+gtRule.getName(),blockRule);
//				//have to check that an input parameter is not changed during the execution of the GTRule
//				if(!match[i]
//						.equals( 
//								blockRuleExecutionEnvironment.getTermEvaluator().evaluate(blockRuleExecutionEnvironment
//										,(Term)((CollectionIteratorRule)blockRule).getGtrule().getActualParameters().get(i))) )
//				{
//					ViatraTransformationException e = new ASMInterpreterException(ASMInterpreterErrorStrings.GTIN_OUTDIFFERENCE,(GTASMElement)chooseBlockRule.getGtrule().getActualParameters().get(i));
//					e.addNewStackElement(blockRule);
//					throw e;
//				}
			}
//			else
			//INOUT or OUT parameter has to be changed
//			if(!((SymbolicRuleParameter)gtRule.getSymParameters().get(i)).getDirection().equals(DirectionKind.IN_LITERAL))
//			{
			Variable variable = ((VariableReference)(gtRuleInvocation).getActualParameters().get(i)).getVariable();
			try {
				if(gtRuleInvocation.getActualParameters().get(i) instanceof VariableReference)
					blockRuleExecutionEnvironment.
						setVariableValue(
							variable,
							match[i]);
				} catch (ViatraTransformationException e) {
					String[] context = {variable.getName(),e.getMessage()};
					throw new ASMInterpreterException(
							blockRule instanceof ForallRule?ASMInterpreterErrorStrings.RULE_INIT_VAR_FORALL:ASMInterpreterErrorStrings.RULE_INIT_VAR_CHOOSE
							,context
							,blockRule);
				}
//			}
		}
	return RuleInterpreter.getInstance().interpretRule(blockRuleExecutionEnvironment,blockRule.getBody(),pr);
	}

	
	
	/** Returns the appropriate parameters for the GTRUleMatcher invocation.
	 * @param .executionEnvironment The environment in which the terms will be evaluated
	 * @param ruleToBeInterpreted the Block rule containing the GT rule invocation
	 * @param parameterScopes Scope parameters for the quantified variables
	 * @param gtRule the invoked GT rule
	 * @param invocation the GT rule invocation rule
	 * @return 
	 * @throws ViatraTransformationException
	 */
	private PatternMatcherParameters getGTRuleMatchParameters(IExecutionEnvironment executionEnvironment
			, CollectionIteratorRule ruleToBeInterpreted
			, Map<Variable,Scope> parameterScopes
			, GTRule gtRule
			, GTRuleInvocation invocation)
	throws ViatraTransformationException
	{

		Object[] patternParams = new Object[invocation.getActualParameters().size()];
		Integer[] quantificationOrder=new Integer[invocation.getActualParameters().size()];
		PatternCallSignature[] patternCallSignatures=new PatternCallSignature[invocation.getActualParameters().size()];						
//		quantification order in case of forall rule
		int outerVariableCounterForQuantificationOrder=(ruleToBeInterpreted).getLocalVariables().size();
		int localVariableCounterForQuantificationOrder=0;
		
		
		for (int i = 0 ; i< invocation.getActualParameters().size(); i++) {
			Object ActualTerm = invocation.getActualParameters().get(i);
			SymbolicRuleParameter symParam = gtRule.getSymParameters().get(i); 
			PatternCallSignature pcs=new PatternCallSignature();
			// the value of the actual parameter
			patternParams[i] = TermEvaluator.getInstance().evaluate(executionEnvironment , (Term)ActualTerm);
		
			/************* Execution mode handling*********************/
			if( ActualTerm instanceof VariableReference 
				// the CollectionIterator rule's actual parameter
				&& (ruleToBeInterpreted).getLocalVariables().contains(((VariableReference)ActualTerm).getVariable()))
				{
					if( ruleToBeInterpreted instanceof ForallRule)
						pcs.setExecutionMode(ExecutionMode.MULTIPLE_RESULTS);
					else
						pcs.setExecutionMode(ExecutionMode.SINGLE_RESULT);
					//	pcs.setParameterMode(ParameterMode.OUTPUT);	//the variable is quantified by the BlockRule
					// checks that the quantified variable is a parameter of the LHS, if not exception	
					boolean isQuantifiedVariableinLHS = false;	
					for(int j = 0; j< gtRule.getPrecondition().getActualParameters().size(); j++){
						if(((VariableReference)gtRule.getPrecondition().getActualParameters().get(j)).getVariable().equals(symParam.getVariable()))
							isQuantifiedVariableinLHS = true;
						}
					if(!isQuantifiedVariableinLHS)	
						{
						String[] context = {((VariableReference)ActualTerm).getVariable().getName(),gtRule.getName()};
						ViatraTransformationException e = new ASMInterpreterException(
								ASMInterpreterErrorStrings.GTINPUT_QUANTIFIED_NOTIN_LHS
								,context
								,gtRule);
						throw e.addNewStackElement(ruleToBeInterpreted);
						}
				}
			else
				{
				pcs.setExecutionMode(ExecutionMode.SINGLE_RESULT);
				}
			/************* Input / Output parameter handling*********************/
			if(symParam.getDirection().equals(DirectionKind.IN_LITERAL))	//INPUT
				{	if((patternParams[i]==null 
							|| ValueKind.UNDEF_LITERAL.equals(patternParams[i])))
						{
							String[] context = {symParam.getVariable().getName(),gtRule.getName()};
							ViatraTransformationException e  = new ASMInterpreterException(ASMInterpreterErrorStrings.GTINPUT_IN
									,context
									,gtRule);
							throw e.addNewStackElement(ruleToBeInterpreted);
						}
					else
							pcs.setParameterMode(ParameterMode.INPUT);
				}
			else
			if(symParam.getDirection().equals(DirectionKind.OUT_LITERAL))		//OUTPUT		
				{	//if((patternParams[i]==null 
					//		|| ValueKind.UNDEF_LITERAL.equals(patternParams[i])))
						pcs.setParameterMode(ParameterMode.OUTPUT);
						patternParams[i] = ValueKind.UNDEF_LITERAL;
						//else
				//		{
						//if(pcs.getParameterMode() == null) // Its is a quantificated variable if paramterMode != null 
					//		{	
						//log.warning(symParam.getName()+" is an OUTPUT parameter but has a value at the invocation of the ");
								//GTASMException e = new ASMInterpreterException(symParam.getVariable().getName()+ASMInterpreterErrorStrings.GTINPUT_OUT+gtRule.getName()+" gt rule.",gtRule);
								//throw e.addNewStackElement(ruleToBeInterpreted);
					//		}
				//		}
				}
			else
			//if(symParam.getDirection().equals(DirectionKind.INOUT_LITERAL))		//INOUT
					{
				// in and inout parameter has to have an input value! 
				if((patternParams[i]==null 
						|| ValueKind.UNDEF_LITERAL.equals(patternParams[i])))
						{
						String[] context = {symParam.getVariable().getName(),gtRule.getName()};
						ViatraTransformationException e  = new ASMInterpreterException(ASMInterpreterErrorStrings.GTINPUT_IN
								,context
								,gtRule);
						throw e.addNewStackElement(ruleToBeInterpreted);
						}
				else
						pcs.setParameterMode(ParameterMode.INPUT);
					}
			/************* Scope of the parameters handling*********************/
			if(ActualTerm instanceof VariableReference) // variable type element can be in the scope order
			{
				if(parameterScopes.containsKey(((VariableReference)ActualTerm).getVariable()))
					pcs.setParameterScope(parameterScopes.get(((VariableReference)ActualTerm).getVariable()));
				else
					pcs.setParameterScope(new Scope());
			}
			else	//node varibaleReference type input parameter
			{	
				pcs.setParameterScope(new Scope());
			}
			/************* Quantification order in case of forall rule*********************/
			if(ruleToBeInterpreted instanceof ForallRule)
				if(ActualTerm instanceof VariableReference)
					{
						if(((ruleToBeInterpreted).getLocalVariables().indexOf(((VariableReference)ActualTerm).getVariable()))==-1)
						{	
							// an error occured, there is a variable which is quanitified by the BlockRule but not used in the GT rule
							if(invocation.getActualParameters().size() == outerVariableCounterForQuantificationOrder)
								getNoneUsedVariable(ruleToBeInterpreted);

							quantificationOrder[outerVariableCounterForQuantificationOrder]=i;
							outerVariableCounterForQuantificationOrder++;
						}
						else
						{
						quantificationOrder[((ruleToBeInterpreted).getLocalVariables().indexOf(((VariableReference)ActualTerm).getVariable()))]= i;
						localVariableCounterForQuantificationOrder++;
						}
					}
				else
					{
					//error a variable is not quantified varibale
					if(invocation.getActualParameters().size() == outerVariableCounterForQuantificationOrder)
						getNoneUsedVariable(ruleToBeInterpreted);
					
					quantificationOrder[outerVariableCounterForQuantificationOrder]=i;
					outerVariableCounterForQuantificationOrder++;
					}	
		// the signature is assigned to the array	
		patternCallSignatures[i]=pcs;	
		}
	return new PatternMatcherParameters(patternCallSignatures,patternParams,quantificationOrder);
	}
	
	/** Evaluates which parameter is quantified but not used in the (GT rule or pattern) invocation
	 * @param ruleToBeInterpreted the Block rule which invokes the Pattern or GT rule
	 * @throws ViatraTransformationException
	 */
	private void getNoneUsedVariable(CollectionIteratorRule ruleToBeInterpreted)
	throws ViatraTransformationException{
	//Called from a gtPattern or a gtRule
	String name = (ruleToBeInterpreted).getCondition() instanceof GTPatternCall? 
				((GTPatternCall)(ruleToBeInterpreted).getCondition()).getCalledPattern().getName()+ " pattern"
				:(ruleToBeInterpreted).getGtrule().getRule().getName() + " GT Rule";
		
	EList<Term> actualParamater = (ruleToBeInterpreted).getCondition() instanceof GTPatternCall? 
				((GTPatternCall)(ruleToBeInterpreted).getCondition()).getActualParameters()
				:(ruleToBeInterpreted).getGtrule().getActualParameters();
		
	for(int j=0; j< ruleToBeInterpreted.getLocalVariables().size(); j++)
		{// search for the errorfull variable
		Variable locVar = ruleToBeInterpreted.getLocalVariables().get(j);
		boolean usedInRule = false;
		for(int k = 0; k < actualParamater.size(); k++)
			{//if it is used in the gtRule then the usedInRule is set to true 
			if(actualParamater.get(k) instanceof VariableReference
					&& ((VariableReference)actualParamater.get(k)).getVariable().equals(locVar))
				usedInRule = true;
			}
		if(!usedInRule) //we have the errorful variable
			{
			String[] context = {locVar.getName(),name };
			throw new ASMInterpreterException(ASMInterpreterErrorStrings.BLOCKRULE_PARAMS_NOT_USED 
					,context
					,ruleToBeInterpreted);
			}
		}
	
		String[] context = {name};
		throw new ASMInterpreterException(ASMInterpreterErrorStrings.FATAL_ERROR
				,context
				,ruleToBeInterpreted);
		
	}
	
	/** Returns the appropriate parameters for the PatternMatcher invocation
	 * @param org.eclipse.viatra2.gtasm.interpreter.executionEnvironment the execution environment of the invocation context
	 * @param ruleToBeInterpreted the choose/forall rule that is invoked
	 * @param parameterScopes the scope of the parameters
	 * @return the parameters of the rule 
	 * @throws ViatraTransformationException
	 */
	private PatternMatcherParameters getMatchParameters(IExecutionEnvironment executionEnvironment
			, CollectionIteratorRule ruleToBeInterpreted
			, Map<Variable,Scope> parameterScopes)
	throws ViatraTransformationException
	{
		CollectionIteratorRule iteratorBlockRule = (ruleToBeInterpreted);
		int paramSize = ((GTPatternCall)iteratorBlockRule.getCondition()).getActualParameters().size();
		//a quantified variable is not used
		if(paramSize < iteratorBlockRule.getLocalVariables().size())
			getNoneUsedVariable(ruleToBeInterpreted);
		
		Object[] patternParams = new Object[paramSize];
		Integer[] quantificationOrder=new Integer[paramSize];
		PatternCallSignature[] patternCallSignatures=new PatternCallSignature[paramSize];						
		int outerVariableCounterForQuantificationOrder=iteratorBlockRule.getLocalVariables().size();
		int localVariableCounterForQuantificationOrder=0;
			
	//for (Object term : ((GTPatternCall)((CollectionIteratorRule)ruleToBeInterpreted).getCondition()).getActualParameters()) {
		for (int i = 0 ; i< paramSize; i++) {
			Object ActualTerm = ((GTPatternCall)iteratorBlockRule.getCondition()).getActualParameters().get(i);
			
			PatternCallSignature pcs=new PatternCallSignature();
			patternParams[i] = TermEvaluator.getInstance().evaluate(executionEnvironment , (Term)ActualTerm);

			/************* Execution mode handling*********************/
			if((Term)ActualTerm instanceof VariableReference 
					&& ruleToBeInterpreted instanceof ForallRule
					&& iteratorBlockRule.getLocalVariables().contains(((VariableReference)ActualTerm).getVariable()))
				pcs.setExecutionMode(ExecutionMode.MULTIPLE_RESULTS);
			else
				pcs.setExecutionMode(ExecutionMode.SINGLE_RESULT);
				
			/************* Input / Output parameter handling*********************/
			if(patternParams[i]==null||ValueKind.UNDEF_LITERAL.equals(patternParams[i]))
				{
				if((Term)ActualTerm instanceof VariableReference
						&& iteratorBlockRule.getLocalVariables().contains(((VariableReference)ActualTerm).getVariable()))
						pcs.setParameterMode(ParameterMode.OUTPUT);
				else
					{
					String[] context = {""+i};
					throw new ASMInterpreterException(ASMInterpreterErrorStrings.BLOCKRULE_PARAM_NOT_QUANTIFIED
							,context
							,iteratorBlockRule);
					}
				}
			else
				pcs.setParameterMode(ParameterMode.INPUT);
			
			/************* Scope of the parameters handling*********************/
			if(ActualTerm instanceof VariableReference)
			{	Variable variable = ((VariableReference)ActualTerm).getVariable();
				if(parameterScopes.containsKey(variable))
					pcs.setParameterScope(parameterScopes.get(variable));
				else
					pcs.setParameterScope(new Scope());
			}
			else
			{
				pcs.setParameterScope(new Scope());
			}
				
			patternCallSignatures[i]=pcs;
			/************* Quantification order in case of forall rule*********************/
			if(ruleToBeInterpreted instanceof ForallRule)
				if(ActualTerm instanceof VariableReference)
				{ Variable variable = ((VariableReference)ActualTerm).getVariable();
					//is it a variable which is quantified
					if(iteratorBlockRule.getLocalVariables().indexOf(variable)==-1)
					{	
						//an error occured, there is a variable which is quanitificated by the BlockRule but not used in the GT pattern. Throws ViatraTransformationException
						if(paramSize == outerVariableCounterForQuantificationOrder)
							getNoneUsedVariable(ruleToBeInterpreted);
						
						quantificationOrder[outerVariableCounterForQuantificationOrder]=i;
						outerVariableCounterForQuantificationOrder++;
					}
					else
					{
					quantificationOrder[iteratorBlockRule .getLocalVariables().indexOf(variable)]= i;
					localVariableCounterForQuantificationOrder++;
					}
				}
			else
				{
				quantificationOrder[outerVariableCounterForQuantificationOrder]=i;
				outerVariableCounterForQuantificationOrder++;
				}	
		}// end of for cycle
		return new PatternMatcherParameters(patternCallSignatures,patternParams,quantificationOrder);
	}
	
	/** Creates a new execution environment for a CollectionIterator rule
	 * @param executionEnvironment the parent execution environment
	 * @param ruleToBeInterpreted the CollectionIterator rule
	 * @return new execution environment, where parent = executionEnviroment
	 */
	private BlockRuleExecutionEnvironment getNewExecutionEnvironment(IExecutionEnvironment executionEnvironment,CollectionIteratorRule ruleToBeInterpreted)
	{
		// Needs a new Exec.env. for the new Variables.
		BlockRuleExecutionEnvironment forallRuleExecutionEnvironment = 
			new BlockRuleExecutionEnvironment(executionEnvironment.getFramework(),ruleToBeInterpreted.getBody());
		Map<Variable, Object> variables = new Hashtable<Variable, Object>(executionEnvironment.getVariableValues());
		for (Object variable : ruleToBeInterpreted.getLocalVariables()) {
			variables.put((Variable)variable, ValueKind.UNDEF_LITERAL);
		}
		forallRuleExecutionEnvironment.onBegin(variables);
		return forallRuleExecutionEnvironment;
	}
	
	/**
	 * Reorders the matches of the pattern according to the ordering settings, if specified. 
	 * 
	 * @param gtPattern the pattern that was matched
	 * @param matches the already retrieved matches of the pattern
	 * @return the reordered matches, or the original collection if no ordering was specified
	 */
	@SuppressWarnings("unchecked")
	private Collection<IMatching> reorder(GTPattern gtPattern, Collection<IMatching> matches) {
		// is ordering required?
		Integer orderColoumn = null;
		boolean orderDesc = false;
		String orderType = "string";
		for (Object annot : gtPattern.getRuntimeAnnotations()) {
			RuntimeAnnotation annot2 = (RuntimeAnnotation) annot;
			if ("@orderby".equals(annot2.getAnnotationName().toLowerCase())) {
				for (Object oRel: annot2.getElements()) {
					RuntimeAnnotationElement rel = (RuntimeAnnotationElement)oRel;
					if ("col".equals(rel.getKey().toLowerCase())) {
						try {
							orderColoumn = Integer.parseInt(rel.getValue());
						} catch (NumberFormatException ex) {
							//throw new ASMInterpreterException("", , forallGTPatternCall);
						}
					} 
					else if ("order".equals(rel.getKey().toLowerCase()) && "desc".equals(rel.getValue().toLowerCase())) 
						orderDesc = true;
					else if ("type".equals(rel.getKey().toLowerCase()))
						orderType = rel.getValue().toLowerCase(); 
				}
				break;
			}
		}
		// if yes, do the ordering
		if (orderColoumn !=null) 
		{
			final boolean desc = orderDesc;
			final Map<IMatching, Comparable> keys = new HashMap<IMatching, Comparable>();
			for (IMatching match : matches) {
				Object col = match.lookup(orderColoumn);
				Comparable key = null;
				if (col==null) key = null;
				else if (col instanceof IEntity) {
					if ("name".equals(orderType)) key = ((IModelElement)col).getName();
					else if ("name-nocase".equals(orderType)) key = ((IModelElement)col).getName().toLowerCase();
					else if ("fqn".equals(orderType)) key = ((IModelElement)col).getFullyQualifiedName();
					else if ("fqn-nocase".equals(orderType)) key = ((IModelElement)col).getFullyQualifiedName().toLowerCase();
					else {
						String value = ((IEntity)col).getValue();
						if ("double".equals(orderType)) key = Double.parseDouble(value);
						else if ("integer".equals(orderType)) key = Integer.parseInt(value);
						else if ("nocase".equals(orderType)) key = value.toLowerCase();
						else if ("value-nocase".equals(orderType)) key = value.toLowerCase();
						else key = value;
					}
				} else if (col instanceof Comparable)
					key = (Comparable) col;
				else key = col.toString();
				keys.put(match, key);
			}
			ArrayList<IMatching> matchList = new ArrayList<IMatching>(matches);
			Collections.sort(matchList, new Comparator<IMatching>() {
				public int compare(IMatching o1, IMatching o2) {
					int compareResult = keys.get(o1).compareTo(keys.get(o2));
					return desc ? -compareResult : compareResult;
				}
			});
			matches = matchList;
		}
		return matches;
	}
	
	
}
