/*******************************************************************************
 * Copyright (c) 2004-2008 Peter Pasztor, Akos Horvath, Gergely Varro, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Peter Pasztor, Akos Horvath, Gergely Varro, Istvan Rath - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.gtasm.interpreter.term.rules;

import org.eclipse.viatra2.core.IModelElement;
import org.eclipse.viatra2.gtasm.interpreter.exception.ViatraTransformationException;
import org.eclipse.viatra2.gtasm.interpreter.executionEnvironment.IExecutionEnvironment;
import org.eclipse.viatra2.gtasm.interpreter.term.internal.TermInterpreterErrorString;
import org.eclipse.viatra2.gtasm.interpreter.term.internal.TermInterpreterException;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.enums.MultiplicityKind;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.enums.ValueKind;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.Term;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.builtInFunctions.ConversionOperation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.builtInFunctions.ToBoolean;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.builtInFunctions.ToDouble;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.builtInFunctions.ToInt;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.builtInFunctions.ToMultiplicity;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.builtInFunctions.ToString;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.modelmanagement.queryFunctions.impl.MultiplicityImpl;


/** Holds all Conversion operations 
 * @author Peter Pasztor, Akos Horvath
 *
 */
public class ConversionOperationEvaluator extends TermEvaluator {
	private static ConversionOperationEvaluator _instance = new ConversionOperationEvaluator();

	private ConversionOperationEvaluator() {
		;
	}

	public static ConversionOperationEvaluator getInstance() {
		return _instance;
	}

	@Override
	public Object evaluate(IExecutionEnvironment executionEnvironment,
			Term termToBeEvaluated) throws ViatraTransformationException{
		ConversionOperation conversionOperation = ((ConversionOperation) termToBeEvaluated);

		if (conversionOperation.getActualParameters().size() != 1)
			{
			String[] context = {conversionOperation.getName()};
			throw new TermInterpreterException(TermInterpreterErrorString.PARAM_NUM_CONVERSION
					,context
					,conversionOperation);
			}
		// First we have to evaluate the parameter, and then perform the
		// operation.

		Term term = (conversionOperation.getActualParameters().get(0));
		Object op1 = TermEvaluator.getInstance().evaluate(
				executionEnvironment,
				term);

		if (conversionOperation instanceof ToBoolean) {
			Object result = object2Boolean(op1, conversionOperation); // if null, it's undef!
			return (result == null) ? (ValueKind.UNDEF_LITERAL) : (result);
		} else if (conversionOperation instanceof ToString)
			return op1.toString();
		else if (conversionOperation instanceof ToInt) {
			if (op1 instanceof Integer)
				return op1;
			else if (op1 instanceof Double)
				return ((Double) op1).intValue();
			else if (op1 instanceof String)
				try {
					return Integer.valueOf((String) op1);
				} catch (NumberFormatException e) {
					String[] context = {(String) op1,"Integer",conversionOperation.getName()};
					throw new TermInterpreterException(
							TermInterpreterErrorString.TERM_CAN_NOT_BE_CONVERTED
							,context
							,termToBeEvaluated);
					//return ValueKind.UNDEF_LITERAL;
				}
			else if (op1 instanceof Boolean)
				return (Boolean) op1 ? (Integer) 1 : (Integer) 0;
			else if (ValueKind.UNDEF_LITERAL.equals(op1))
				return ValueKind.UNDEF_LITERAL;
			else
				{
				String[] context = {convertToJavaType(op1),conversionOperation.getName()};
				throw new TermInterpreterException(
						TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_ONE_PARAM
						,context
						,termToBeEvaluated);
				}

		} else if (conversionOperation instanceof ToDouble) {
			if (op1 instanceof Double)
				return op1;
			else if (op1 instanceof Integer)
				return new Double(((Integer) op1).doubleValue());
			else if (op1 instanceof String)
				try {
					return new Double((String) op1);
				} catch (NumberFormatException e) {
					String[] context = {(String) op1,"Double",conversionOperation.getName()};
					throw new TermInterpreterException(
							TermInterpreterErrorString.TERM_CAN_NOT_BE_CONVERTED
							,context
							,termToBeEvaluated);
					//return ValueKind.UNDEF_LITERAL;
				}
			else if (op1 instanceof Boolean)
				return (Boolean) op1 ? (Double) 1.0 : (Double) 0.0;
			else if (ValueKind.UNDEF_LITERAL.equals(op1))
				return ValueKind.UNDEF_LITERAL;
			else
				{
				String[] context = {convertToJavaType(op1),conversionOperation.getName()};
				throw new TermInterpreterException(
						TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_ONE_PARAM
						,context
						,termToBeEvaluated);
				}
		} else if (conversionOperation instanceof ToMultiplicity) {
			if(op1 instanceof MultiplicityImpl)
				return op1;
			else if(op1 instanceof String)
				{
				String value = (String) op1;
				
				if(value.compareToIgnoreCase("one-to-one") == 0)
					return MultiplicityKind.ONE_TO_ONE_LITERAL;
				
				if(value.compareToIgnoreCase("one-to-many") == 0)
					return MultiplicityKind.ONE_TO_MANY_LITERAL;
				
				if(value.compareToIgnoreCase("many-to-one") == 0)
					return MultiplicityKind.MANY_TO_ONE_LITERAL;
					
				if(value.compareToIgnoreCase("many-to-many") == 0)
					return MultiplicityKind.MANY_TO_MANY_LITERAL;
				
				String[] context = {value,"Multiplicity",conversionOperation.getName()};
				throw new TermInterpreterException(
						TermInterpreterErrorString.TERM_CAN_NOT_BE_CONVERTED
						,context
						,termToBeEvaluated);
				
				}
			else if (ValueKind.UNDEF_LITERAL.equals(op1))
				return ValueKind.UNDEF_LITERAL;
			else
				{
				String[] context = {convertToJavaType(op1),conversionOperation.getName()};
				throw new TermInterpreterException(
						TermInterpreterErrorString.TERM_NOT_MULTIPLICITY_COMPATIBLE
						,context
						,termToBeEvaluated);
				}
		}
		// the control gets here, if a non-implemented Term was evaluated.
		else
			{
			String[] context = {conversionOperation.getName()};
			throw new TermInterpreterException(
					TermInterpreterErrorString.UNIMP_CONVERSION
					,context
					,termToBeEvaluated);
			}
	}

	protected Boolean object2Boolean(Object obj,ConversionOperation toBoolean)
			throws TermInterpreterException  {
		if (obj instanceof String) {
			return "true".equals(obj) ? (Boolean.TRUE) : ("false"
					.equals(obj) ? Boolean.FALSE : null);
		} else if (obj instanceof Double) {
			return (((Double) obj).equals(0.0)) ? Boolean.FALSE : Boolean.TRUE;
		} else if (obj instanceof Integer) {
			return (((Integer) obj).equals(0)) ? Boolean.FALSE : Boolean.TRUE;
		} else if (obj instanceof Boolean) {
			return (Boolean) obj;
		} else if (obj.equals(ValueKind.UNDEF_LITERAL)) {
			return Boolean.FALSE;

		} else if (obj instanceof IModelElement) {
			return Boolean.TRUE;
		}
		String[] context = {toBoolean.getName(),convertToJavaType(obj)};
		throw new TermInterpreterException(
				TermInterpreterErrorString.TERM_NOT_BOOL_COMPATIBLE
				, context
				, toBoolean);
	}
}
