/*******************************************************************************
 * Copyright (c) 2004-2008 Akos Horvath, Gergely Varro and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Akos Horvath, Gergely Varro - initial API and implementation
 *******************************************************************************/

 package org.eclipse.viatra2.gtasm.patternmatcher.impl.gtmatcher.internal;

import java.util.ArrayList;
import java.util.Collection;

import org.eclipse.viatra2.gtasm.interpreter.exception.ViatraTransformationException;
import org.eclipse.viatra2.gtasm.interpreter.executionEnvironment.IExecutionEnvironment;
import org.eclipse.viatra2.gtasm.interpreter.impl.rules.RuleInterpreter;
import org.eclipse.viatra2.gtasm.patternmatcher.IMatching;
import org.eclipse.viatra2.gtasm.patternmatcher.ParameterMode;
import org.eclipse.viatra2.gtasm.patternmatcher.PatternCallSignature;
import org.eclipse.viatra2.gtasm.patternmatcher.PatternMatcherParameters;
import org.eclipse.viatra2.gtasm.patternmatcher.Scope;
import org.eclipse.viatra2.gtasm.patternmatcher.exceptions.PatternMatcherCompileTimeException;
import org.eclipse.viatra2.gtasm.patternmatcher.exceptions.PatternMatcherRuntimeException;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.gtmatcher.exceptions.GTErrorStrings;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.gtmatcher.exceptions.GTOperationException;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.gtmatcher.exceptions.GTRuleBuildingException;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.gtmatcher.exceptions.GraphTransformationException;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.gtmatcher.internal.operation.IUpdatePlanOperation;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.patternmatcher.core.PatternBuilder;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.patternmatcher.internal.MatchingFrame;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.patternmatcher.internal.operation.ISearchPlanOperation;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.patternmatcher.term.BasicTermHandler;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.patternmatcher.term.ITermHandler;
import org.eclipse.viatra2.gtasm.patternmatcher.patterns.IPatternMatcher;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.core.RuntimeAnnotation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.SymbolicRuleParameter;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Variable;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.enums.ValueKind;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.GTPatternCall;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.VariableReference;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTRule;
import org.eclipse.viatra2.core.IModelElement;
import org.eclipse.viatra2.core.IModelManager;

 
/**
 *
 * The class interprets all traditionally defined GT rule containing
 * LHS, RHS and action part (or a subset of these elements).
 * @author Akos Horvath
 */
public class TraditionalGTRuleMatcher extends GTRuleMatcher {
	IPatternMatcher leftHandSidePatternMatcher;
 	RightHandSideMatcher rightHandSideMatcher;
	boolean isRandom;
	
	Collection<GTElementMapping> elementMappings;
	Collection<IUpdatePlanOperation> nonchangeableGTOperations;
	Collection<ISearchPlanOperation> checkSet;
	
	public TraditionalGTRuleMatcher( GTRule gt
			, IExecutionEnvironment enviroment
			, ITermHandler handler
			, PatternMatcherParameters mp) throws ViatraTransformationException {
		super(enviroment.getFramework().getLogger()
				, enviroment.getFramework().getTopmodel().getModelManager()
				, gt);
		elementMappings = new ArrayList<GTElementMapping>();
		this.initGTRule(enviroment);
		this.construct(manager,mp.getCallSignature());
	}

	/** 
	 * Returns a Pattern Matcher to work with, in case of alternative pattern matchers this is the Method to override
	 * @return Pattern matcher for the
	 * @throws GTASMException
	 */
	protected IPatternMatcher getPatternMatcherforPreCondition() throws ViatraTransformationException{
		PatternBuilder pb = new PatternBuilder(this.logger, manager, handler);
		try {
			return pb.construct(gtRule.getPrecondition().getCalledPattern());
		} catch (PatternMatcherCompileTimeException e) {
			//logger.error(e.getMessage());
			throw e.addNewStackElement(gtRule);
		}

	}
	
	/**
	 * Determines whether true pseudorandom matching is required for ChooseRules.
	 * @return true if random is required
	 */
	protected boolean isLeftHandSideRandom() {
		// is true pseudorandom required?
		for (Object annot : gtRule.getPrecondition().getCalledPattern().getRuntimeAnnotations()) {
			if ("@random".equals(((RuntimeAnnotation) annot).getAnnotationName().toLowerCase())) {
				return true;
			}
		}	
		return false;
	}
	
	/**
	 * Creates the GTrule matcher by constructing the search graph of the LHS and RHS (optional) part, and the operations from RHS\LHS
	 * @param manager model manager of the VIATRA model space
	 * @throws GTASMException
	 */
	private void construct(IModelManager manager, PatternCallSignature[] symSignatures) throws ViatraTransformationException{
		
		if(gtRule.getPostcondition() != null && gtRule.getPrecondition() == null)
			logger.error("There is only PostCondition defined for the GTRule "+ gtRule.getName());
		
		if (gtRule.getPostcondition() != null) {
			try {
				leftHandSidePatternMatcher = getPatternMatcherforPreCondition();
				isRandom = isLeftHandSideRandom();
				rightHandSideMatcher =
					new RightHandSideMatcher(gtRule.getPostcondition().getCalledPattern(), logger, manager, handler); 
			} catch (PatternMatcherCompileTimeException e) {
				logger.error(e.getMessage());
				throw e.addNewStackElement(gtRule);
			}
			//maps the parameters between the RHS and the LHS
			elementMappings = new ArrayList<GTElementMapping>();
			int actualIndex = 0;
			for(Object obj: gtRule.getPrecondition().getActualParameters()){
				VariableReference varReference = (VariableReference) obj;
				Variable actualVariable = varReference.getVariable();
				if(getActualParameterIndexof(gtRule.getPostcondition(),actualVariable) != -1)
				{
					GTElementMapping map = new GTElementMapping(actualIndex,getActualParameterIndexof(gtRule.getPostcondition(),actualVariable));
					elementMappings.add(map);
				}
				actualIndex++;
			}

			PatternCallSignature[] rhsSignatures = new PatternCallSignature[gtRule.getPostcondition().getActualParameters().size()];
			//Input parameters from the symmetric parameters
			if(gtRule.getPostcondition().getActualParameters().size() > 0)
				for(int i = 0; i < symSignatures.length; i++){
					SymbolicRuleParameter symParam = (SymbolicRuleParameter)gtRule.getSymParameters().get(i);
					if(symParam.getVariable().getReferences().size() > 0) 
						if(getActualParameterIndexofSymbolicParameter(gtRule.getPostcondition(), symParam) != -1)
						{
							if(symSignatures[i].getParameterMode().compareTo(ParameterMode.INPUT) == 0)
							{
								PatternCallSignature sig = new PatternCallSignature();
								sig.setParameterMode(ParameterMode.INPUT);
								rhsSignatures[getActualParameterIndexofSymbolicParameter(gtRule.getPostcondition(), symParam)] = sig;
							}
							else
								rhsSignatures[getActualParameterIndexofSymbolicParameter(gtRule.getPostcondition(), symParam)] = symSignatures[i].clone();
						}
				}

//			//Element mapping
			for(GTElementMapping map: elementMappings)
				if(rhsSignatures[map.getRhsInputOrderIndex()] == null)
				{
					PatternCallSignature sig = new PatternCallSignature();
					sig.setParameterMode(ParameterMode.INPUT);
					rhsSignatures[map.getRhsInputOrderIndex()] = sig;
				}
				else // it is a GT Rule parameter which can be output type but as it is represented in both of the patterns
					// to the RHS it is an input parameter
				{
					rhsSignatures[map.getRhsInputOrderIndex()].setParameterMode(ParameterMode.INPUT);
				}
			//output parameter of the RHS, which is used in the action part only
			for(int j=0; j < gtRule.getPostcondition().getActualParameters().size(); j++)
			{
				if(rhsSignatures[j] == null)
				{rhsSignatures[j] = new PatternCallSignature();
				rhsSignatures[j].setParameterMode(ParameterMode.OUTPUT);
				}
			}
			try {
				GTOperationContext _temp = rightHandSideMatcher.generateGTOperation(elementMappings,rhsSignatures);
				nonchangeableGTOperations = _temp.getFinalOperations();
				checkSet = _temp.getCheckSet();

			} catch (GTRuleBuildingException e) {
				throw e.addNewStackElement(gtRule);
			}
		}
		else //only a precondition is available
		{
			leftHandSidePatternMatcher = getPatternMatcherforPreCondition();
			elementMappings = null;
			nonchangeableGTOperations = null;
		}
	}
	
	
	public void initGTRule(IExecutionEnvironment actualEnvironment) throws ViatraTransformationException{
		//sets the actual context of the gt rule invocation 
		this.gtEnvironment = actualEnvironment;
		this.handler = new BasicTermHandler(actualEnvironment); //handler;
		
		try {
			// adds the local variables & sym parameters to the execution environment -> later only the values need to be changed
			for(Object o:gtRule.getLocalVariables()){
				gtEnvironment.addVariable((Variable) o, ValueKind.UNDEF_LITERAL);
			}
		} catch (ViatraTransformationException e) {
			throw e.addNewStackElement(gtRule);
			// throw new GTASMException(e.getLocalizedMessage() + GTErrorStrings.INIT+gtRule.getName(), gtRule);
		}
	}
	
	/**
	 * Executes the GT operations created from the RHS/LHS and the optional DO part of the rule
	 * @param symSignatures the signature of the rule
	 * @param symInputMapping mapping of the input parameters
	 * @param lhsFrame the signature, input mapping and the quantification order for he LHS
	 * @return Returns the values of the rule's parameters after executing the operations constructed form the rule
	 * @throws GTASMException
	 */
	protected Object[] executeGTOperations(PatternCallSignature[] symSignatures
			, Object[] symInputMapping
			, IMatching lhsFrame) throws ViatraTransformationException{
		
		GTOperationContext actualGTContext = null;
		GTPatternCall preConditionCall = gtRule.getPrecondition();
		GTPatternCall postConditionCall = gtRule.getPostcondition();
		// *****RHS of the GTRule
		logger.debug("[debug] Starting GT rule: "+ gtRule.getName());
		
		if (lhsFrame != null && postConditionCall != null) {
			PatternCallSignature[] rhsSignatures = new PatternCallSignature[postConditionCall.getActualParameters().size()];
			Object[] rhsInputMapping = new Object[postConditionCall.getActualParameters().size()];
			
			//Input parameters from the symmetric parameters
			if(postConditionCall.getActualParameters().size() > 0)
			for(int i = 0; i < symSignatures.length; i++){
				SymbolicRuleParameter symParam = (SymbolicRuleParameter)gtRule.getSymParameters().get(i);
				if(symParam.getVariable().getReferences().size() > 0) 
					if(getActualParameterIndexofSymbolicParameter(postConditionCall, symParam) != -1)
						{
						rhsInputMapping[getActualParameterIndexofSymbolicParameter(postConditionCall, symParam)] = symInputMapping[i];
						if(symSignatures[i].getParameterMode().compareTo(ParameterMode.INPUT) == 0)
							{
							PatternCallSignature sig = new PatternCallSignature();
							sig.setParameterMode(ParameterMode.INPUT);
							sig.setParameterScope(new Scope());
							rhsSignatures[getActualParameterIndexofSymbolicParameter(postConditionCall, symParam)] = sig;
							}
						else//input parameter
							rhsSignatures[getActualParameterIndexofSymbolicParameter(postConditionCall, symParam)] = symSignatures[i];
						}
				}
			
			//Element mapping
			for(GTElementMapping map: elementMappings)
				if(rhsSignatures[map.getRhsInputOrderIndex()] == null)
					{rhsInputMapping[map.getRhsInputOrderIndex()]= lhsFrame.lookup(map.getLhsInputOrderIndex()); 
					PatternCallSignature sig = new PatternCallSignature();
					sig.setParameterMode(ParameterMode.INPUT);
					sig.setParameterScope(new Scope());
					rhsSignatures[map.getRhsInputOrderIndex()] = sig;
					}
				else // it is a GT Rule parameter which can be any in any parametermode but as it is represented in both LHS and RHS
					 // to the RHS it is an input parameter
					{
					rhsInputMapping[map.getRhsInputOrderIndex()]= lhsFrame.lookup(map.getLhsInputOrderIndex()); 
					rhsSignatures[map.getRhsInputOrderIndex()].setParameterMode(ParameterMode.INPUT);
					}
			// it is an input parameter for the RHS form the GT Rule head
			for(int j=0; j < postConditionCall.getActualParameters().size(); j++)
				{
			 	 if(rhsSignatures[j] == null)
			 	 	{rhsSignatures[j] = new PatternCallSignature();
					 rhsSignatures[j].setParameterMode(ParameterMode.OUTPUT);
					 rhsSignatures[j].setParameterScope(new Scope());
					 rhsInputMapping[j] = ValueKind.UNDEF_LITERAL;
			 	 	}
				}
			try {
				MatchingFrame frame = rightHandSideMatcher.generateMatchingFrameofInputContext(this.elementMappings, rhsInputMapping,rhsSignatures);
				actualGTContext = new GTOperationContext(frame,nonchangeableGTOperations,checkSet);
			} catch (GTRuleBuildingException e) {
				throw e.addNewStackElement(gtRule);
			}
		}
		
		// Running the operations from the difference of the RHS and the LHS
		if (lhsFrame != null && actualGTContext != null) {
			logger.debug("GT Opeations of the GtRule: "+this.gtRule.getName());
		
			//* manipulation operations */
			for (IUpdatePlanOperation op :actualGTContext.getFinalOperations()) {
				try {
					if (op.update(actualGTContext.getReturnFrame())) {
						logger.debug(op.toString(actualGTContext.getReturnFrame()));
					} else {
						String[] context ={op.toString(actualGTContext.getReturnFrame()), gtRule.getName()};
						GraphTransformationException e = 
							new GraphTransformationException(GTErrorStrings.GTRULE_RHS_FAILED_MANIPULATION
									,context
									,op.getErrorfulElement(actualGTContext.getReturnFrame())); 
						e.addNewStackElement(postConditionCall.getCalledPattern());
						throw e.addNewStackElement(gtRule);
					}
				} catch (GTOperationException e) {
					e.addNewStackElement(postConditionCall.getCalledPattern());
					throw e.addNewStackElement(gtRule);
				}
			}
			//* check operations */
			for(ISearchPlanOperation op: actualGTContext.getCheckSet()){
				try
				{
					op.preprocess(actualGTContext.getReturnFrame());
					if(!op.execute(actualGTContext.getReturnFrame())){
						String[] context = {op.toString(actualGTContext.getReturnFrame()), gtRule.getName()};
						GraphTransformationException ex = 
							new GraphTransformationException(GTErrorStrings.GTRULE_RHS_FAILED_CHECK
									,context
									,op.getErrorfulElement(actualGTContext.getReturnFrame()));
						ex.addNewStackElement(postConditionCall.getCalledPattern());
						throw ex.addNewStackElement(gtRule);  
					}
				} catch(PatternMatcherRuntimeException e) {
					e.addNewStackElement(postConditionCall.getCalledPattern());
					throw e.addNewStackElement(gtRule);
				}
			}
		}

		// * ACTION part of the GTRULE */
		//there is an action part in the gtrule
		if(gtRule.getAction() != null)
		{	// fill up the input parameter for the action part
			try
			{	//The rule has only PreCondition
				if(postConditionCall == null)
						{	//symbolic parameter its not in the LHS
							for(int i = 0; i< symInputMapping.length; i++)
							{SymbolicRuleParameter symParam = (SymbolicRuleParameter) gtRule.getSymParameters().get(i);
							// its a variable which isn't in the LHS 
							if(getActualParameterIndexofSymbolicParameter(preConditionCall, symParam) == -1)
								{//this variable can only be founded in the action part
							 	if(symParam.getVariable().getReferences().size() > 0)
									 this.gtEnvironment.setVariableValue(symParam.getVariable(),symInputMapping[i]);
							 	else
							 	 	 logger.warning("SymbolicRuleParameter "+symParam.getName()+" in gtRule " + gtRule.getName()+ "is never used");
								}
							}
							//Local variables: have to check where are they initialized at the first time
							for(Object oVar: gtRule.getLocalVariables())
							{	Variable actualVariable = (Variable) oVar;
								//The variable is in the LHS Pattern
								if(getActualParameterIndexof(preConditionCall,actualVariable) != -1)
									{
									Object actualValue = lhsFrame.lookup(getActualParameterIndexof(preConditionCall,actualVariable));
									gtEnvironment.setVariableValue(actualVariable, returnValidValue(actualValue));
									}
							}
						}
				//both Pre and Post Condition
				else
						{
						//symbolic parameter
						for(int i = 0; i< symInputMapping.length; i++)
						{SymbolicRuleParameter symParam = (SymbolicRuleParameter) gtRule.getSymParameters().get(i);
						// it is a variable which isn't in the RHS or the LHS 
						if(getActualParameterIndexofSymbolicParameter(postConditionCall, symParam) == -1
								&& getActualParameterIndexofSymbolicParameter(preConditionCall, symParam) == -1)
							//this variable can only be found in the action part
						 	if(symParam.getVariable().getReferences().size() > 0)
								 this.gtEnvironment.setVariableValue(symParam.getVariable(),symInputMapping[i]);
						 	else
						 	 	 logger.warning("SymbolicRuleParameter "+symParam.getName()+" in gtRule " + gtRule.getName()+ "is never used");
						}
						//Local variables, have to check where are they initialized at the first time
						for(Object oVar: gtRule.getLocalVariables())
						{	Variable actualVariable = (Variable) oVar;
							//The variable is in the RHS
							if(getActualParameterIndexof(postConditionCall,actualVariable) != -1)
									{
									Object actualValue = actualGTContext.getReturnFrame().getValue(getActualParameterIndexof(postConditionCall,actualVariable));
									gtEnvironment.setVariableValue(actualVariable, returnValidValue(actualValue));
									}
							else
							if(getActualParameterIndexof(preConditionCall,actualVariable) != -1)
									{
									Object actualValue = lhsFrame.lookup(getActualParameterIndexof(preConditionCall,actualVariable));
									gtEnvironment.setVariableValue(actualVariable, returnValidValue(actualValue));
									}
							}
						}
			}
			catch (ViatraTransformationException e) {
				// Variable setting failed
				throw e.addNewStackElement(gtRule);
				// throw new GTASMException(e.getMessage()+GTErrorStrings.INIT_ACTION+gtRule.getName(), gtRule);
			}
			//GTAction Execution 
			try {
				RuleInterpreter.getInstance().interpretRule(gtEnvironment, gtRule.getAction(),null); 
				return getOutPutSymbolicParatemerRules();
			} catch (ViatraTransformationException e) {
				// GTAction Execution Failed
				throw e.addNewStackElement(gtRule);
			}
		}
		else //******** Only LHS and RHS!
		{
			for(int i = 0; i< symInputMapping.length; i++)
			{SymbolicRuleParameter symParam = (SymbolicRuleParameter) gtRule.getSymParameters().get(i);
			// its a variable which isn't in the RHS or the LHS 
			if(getActualParameterIndexofSymbolicParameter(postConditionCall, symParam) == -1
					&& getActualParameterIndexofSymbolicParameter(preConditionCall, symParam) == -1
					)
				//this variable can only be founded in the action part
			 	 if(symParam.getVariable().getReferences().size() > 0)
					 gtEnvironment.setVariableValue(symParam.getVariable(),symInputMapping[i] );
			 	 else
			 	 	logger.warning("SymbolicRuleParameter "+symParam.getName()+" in gtRule " + gtRule.getName()+ "is never used");
			}
			if(postConditionCall != null && preConditionCall != null)
			{
				for(Object oVar: gtRule.getLocalVariables())
				{	Variable actualVariable = (Variable) oVar;
					//The variable is in the RHS
					if(getActualParameterIndexof(postConditionCall,actualVariable) != -1)
						{
						Object actualValue = actualGTContext.getReturnFrame().getValue(getActualParameterIndexof(postConditionCall,actualVariable));
						gtEnvironment.setVariableValue(actualVariable, returnValidValue(actualValue));
						}
					else
					if(getActualParameterIndexof(preConditionCall,actualVariable) != -1)
						{
						Object actualValue = lhsFrame.lookup(getActualParameterIndexof(preConditionCall,actualVariable));
						gtEnvironment.setVariableValue(actualVariable, returnValidValue(actualValue));
						}
				}
			}
			return getOutPutSymbolicParatemerRules();
		}
	}
	
	/** Returns a valid equivalent value (for the Viatra interpreter) of the input value. For example instead of a deleted element or null pointer it returns an UNDEF literal
	 * @param o input parameter o
	 * @return The 
	 */
	private Object returnValidValue(Object obj){
		
		if(obj == null)
			return ValueKind.UNDEF_LITERAL;
		
		if(obj instanceof IModelElement)
			{
			IModelElement element = (IModelElement)obj;
			if(element.isDeleted())
				return ValueKind.UNDEF_LITERAL;
			}
		
		return obj;
	}
	
	
	public Object[] match(PatternCallSignature[] symSignatures, Object[] symInputMapping)
		throws ViatraTransformationException{
	
	IMatching lhsFrame;
	PatternMatcherParameters lhsParamters;
	lhsParamters = initLHSPattern(symSignatures, symInputMapping, null);	
	lhsFrame = isRandom? leftHandSidePatternMatcher.matchRandomly(lhsParamters.getInputMapping(), lhsParamters.getCallSignature()) 
			: leftHandSidePatternMatcher.match(lhsParamters.getInputMapping(), lhsParamters.getCallSignature());
	
	if(lhsFrame != null)
		return executeGTOperations( symSignatures
			,symInputMapping
			,lhsFrame);
	else
		return null;
	}
	
	public Collection<IMatching> initMatchAll(PatternCallSignature[] symSignatures, Object[] symInputMapping, Integer[] quantificationOrder)
			throws ViatraTransformationException {
		PatternMatcherParameters lhsParamters;
		lhsParamters = initLHSPattern(symSignatures, symInputMapping,quantificationOrder);	
		return leftHandSidePatternMatcher.matchAll(lhsParamters.getInputMapping(), lhsParamters.getCallSignature(),lhsParamters.getQuantificationOrder());	
	}

	public Object[] matchAll(IMatching LHSMatching ,PatternCallSignature[] symSignatures, Object[] symInputMapping)
		throws ViatraTransformationException {
		IMatching lhsFrame = (IMatching) LHSMatching;
		return executeGTOperations( symSignatures
				,symInputMapping
				,lhsFrame);
	}
	
}
