/*******************************************************************************
 * Copyright (c) 2004-2008 Gabor Bergmann and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Gabor Bergmann - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.gtasm.patternmatcher.incremental.adapters;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import java.util.Map.Entry;

import org.eclipse.emf.common.util.EList;
import org.eclipse.viatra2.framework.IFramework;
import org.eclipse.viatra2.gtasm.interpreter.exception.ViatraTransformationException;
import org.eclipse.viatra2.gtasm.interpreter.executionEnvironment.ASMFunctionContent;
import org.eclipse.viatra2.gtasm.interpreter.executionEnvironment.IExecutionEnvironment;
import org.eclipse.viatra2.gtasm.interpreter.term.rules.TermEvaluator;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.boundary.AbstractEvaluator;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.boundary.PredicateEvaluatorNode;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.matcher.IPatternMatcherContext;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.matcher.ReteEngine;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.network.ReteContainer;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.tuple.FlatTuple;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.tuple.Tuple;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.ASMFunction;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Variable;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.Term;


/**
 * @author Gabor Bergmann
 * 
 *         Permits the traversal of update notifications if a given Term (with
 *         variables mapped to given positions in the Tuple) equals a
 *         right-hand-side (also mapped to a given position). If right-hand-side
 *         is omitted, the Term should evaluate to true.
 * 
 *         The Term is reevaluated on the Tuple each time an element affected by
 *         the term experiences a move, name or value change. Furthermore, it is
 *         also reevaluated if any ASMfunctions called at the previous
 *         evaluation are changed at the positions that were used.
 * 
 *         Uses unwrapped tuples. In distributed environments, TermEvaluatorNodes 
 *         are always be built on the head container, because they need access to Viatra.
 * 
 */
public class GTASMTermEvaluatorNode extends PredicateEvaluatorNode {
	Map<String, Integer> variableIndices;
	
	/**
	 * @param term
	 *            the GTASM Term that should be evaluated on all inserted
	 *            tuples.
	 * @param rhsIndex
	 *            the index of the element in the Tuple that should equals the
	 *            result of the term; if null, the right-hand-side will be the
	 *            Boolean true.
	 * @param variableIndices
	 *            maps variable names to values.
	 */
	public GTASMTermEvaluatorNode(ReteEngine<?> engine, ReteContainer container, Term term,
			Integer rhsIndex, Map<String, Integer> variableIndices,
			Map<String, String> variableEquivalence,
			int[] affectedIndices, int tupleWidth) 
	{
		super(engine, container, rhsIndex, affectedIndices, tupleWidth,
				new GTASMTermEvaluator(engine.getContext(), term, variableIndices, variableEquivalence));
		this.variableIndices = variableIndices;
	}
	
	@Override
	protected String prettyPrintTuple(Tuple ps) {
		StringBuilder sb = new StringBuilder();
		sb.append("{ ");
		for (Entry<String, Integer> ent : variableIndices.entrySet()) {
			sb.append(" '" + ent.getKey() + "'=" + ps.get(ent.getValue()).toString());
		}
		sb.append("}");
		return sb.toString();
	}

}

class GTASMTermEvaluator extends AbstractEvaluator {

	protected Term term;
	protected IPatternMatcherContext<?> context;
	protected Map<String, Integer> variableIndices;
	protected Map<String, String> variableEquivalence;

	public GTASMTermEvaluator(IPatternMatcherContext<?> context,
			Term term, Map<String, Integer> variableIndices,
			Map<String, String> variableEquivalence) {
		super();
		this.context = context;
		this.term = term;
		this.variableEquivalence = variableEquivalence;
		this.variableIndices = variableIndices;
	}

	@Override
	public Object doEvaluate(Tuple tuple) throws ViatraTransformationException{
		TermCheckExecutionEnvironment env = new TermCheckExecutionEnvironment(this, tuple);
		return TermEvaluator.getInstance().evaluate(env, term);
	}
	
	public int findVariableIndexByName(String name) {
		String equivalentName = variableEquivalence.get(name);
		Integer integer = variableIndices.get(equivalentName);
		return integer;
	}

	public IPatternMatcherContext<?> getContext() {
		return context;
	}
	

	class TermCheckExecutionEnvironment implements IExecutionEnvironment {
		IFramework framework;
		GTASMTermEvaluator evaluator;
		Tuple ps;
		Set<Tuple> traces;

		public TermCheckExecutionEnvironment(GTASMTermEvaluator evaluator, Tuple ps) {
			this.framework = (((VPMContext<?>)evaluator.getContext()).getFramework()); // FIXME
			this.evaluator = evaluator;
			this.ps = ps;

			this.traces = new HashSet<Tuple>();
		}

		public Object getValueOfASMFunction(ASMFunction asmFunction,
				EList<Object> location) {
			Object[] traceElements = new Object[location.size() + 1];
			int k = 0;
			traceElements[k++] = asmFunction;
			for (Object o : location)
				traceElements[k++] = o;
			Tuple trace = new FlatTuple(traceElements);
			traces.add(trace);

			HashMap<EList<Object>, Object> af = 
				ASMFunctionContent.getInstance().get(asmFunction);
			if (af == null)
				return null;
			return af.get(location);
		};

		public Object getVariableValue(Variable variable) {
			String varName = variable.getName();
			int index = evaluator.findVariableIndexByName(varName);
			return ps.get(index);
		}

		/**
		 * Each ASMFunction invocation is recorded as a trace Tuple. Each trace
		 * tuple contains the ASMFunction as the first element, followed by the
		 * parameter sequence ("location").
		 * 
		 * @return the traces
		 */
		public Set<Tuple> getTraces() {
			return traces;
		}
		public void addVariable(Variable key, Object value) throws ViatraTransformationException {
			throw new UnsupportedOperationException();
		}

		public void fetchVariableVariations(HashMap<Variable, Vector<Object>> possibleVariableValues, Term termToBeEvaluated) {
			throw new UnsupportedOperationException();
		}

		public IFramework getFramework() {
			return framework;
		}

		public Map<Variable, Object> getVariableValues() {
			throw new UnsupportedOperationException();
		}

		public void setVariableValue(Variable variable, Object value) throws ViatraTransformationException {
			throw new UnsupportedOperationException();
		}

		public void updateASMFunction(ASMFunction asmFunction, EList<Object> location, Object value) throws ViatraTransformationException {
			throw new UnsupportedOperationException();
		}

	}
}
