/*******************************************************************************
 * Copyright (c) 2004-2008 Gabor Bergmann and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Gabor Bergmann - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.misc;

import java.util.Collection;
import java.util.LinkedHashSet;

import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.network.Direction;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.network.Network;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.network.ReteContainer;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.single.SingleInputNode;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.tuple.Clearable;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.tuple.Tuple;


/**
 * A monitoring object that connects to the rete network as a receiver to
 * reflect changes since an arbitrary state acknowledged by the client.
 * 
 * If a new matching is found, it appears in the matchFoundEvents collection,
 * and disappears when that particular matching cannot be found anymore. If the
 * event of finding a match has been processed by the client, it can be removed
 * manually. In this case, when a previously found matching is lost, the Tuple
 * will appear in the matchLostEvents collection, and disappear upon finding the
 * same matching again. "Matching lost" events can also be acknowledged by
 * removing a Tuple from the collection. If the matching is found once again, it
 * will return to matchFoundEvents.
 * 
 * Does NOT propagate updates!
 * 
 * @author Gabor Bergmann
 * 
 */
public class DeltaMonitor extends SingleInputNode implements Clearable {

	/**
	 * matches that are newly found
	 */
	public Collection<Tuple> matchFoundEvents;
	/**
	 * matches that are newly lost
	 */
	public Collection<Tuple> matchLostEvents;

	/**
	 * @param reteContainer
	 */
	public DeltaMonitor(ReteContainer reteContainer) {
		super(reteContainer);
		matchFoundEvents = new LinkedHashSet<Tuple>();
		matchLostEvents = new LinkedHashSet<Tuple>();
		reteContainer.registerClearable(this);
	}

	/**
	 * Override this method to provide a lightweight, stateless filter on the tuples
	 * @param tuple the occurrence that is to be filtered
	 * @return true if this tuple should be monitored, false if ignored
	 */
	public boolean statelessFilter(Tuple tuple) {
		return true;		
	}
	
	/**
	 * Build a delta monitor into the head container of the network.
	 * 
	 * @param network
	 */
	public DeltaMonitor(Network network) {
		this(network.getHeadContainer());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.network.Receiver#update(org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.network.Direction, org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.tuple.Tuple)
	 */
	public void update(Direction direction, Tuple updateElement) {
		if (statelessFilter(updateElement)) {
			if (direction == Direction.INSERT) {
				if (!matchLostEvents.remove(updateElement)) // either had before but
															// lost
					matchFoundEvents.add(updateElement); // or brand-new
			} else // revoke
			{
				if (!matchFoundEvents.remove(updateElement)) // either never found
																// in the first
																// place
					matchLostEvents.add(updateElement); // or newly lost
			}
		}
	}

	public void clear() {
		matchFoundEvents.clear();
		matchLostEvents.clear();
	}

	public void pullInto(Collection<Tuple> collector) {
		// IGNORE
		throw new UnsupportedOperationException(
				"DeltaMonitor contents cannot be pulled");
	}

}
