/*******************************************************************************
 * Copyright (c) 2004-2008 Gabor Bergmann and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Gabor Bergmann - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.gtasm.patternmatcher.incremental.simple;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.construction.RetePatternBuildException;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.matcher.IPatternMatcherContext.EdgeInterpretation;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.util.Options;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.util.UnionFind;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Variable;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.enums.ContainmentMode;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.enums.ValueKind;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.Constant;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.GTPatternCall;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.Term;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.VariableReference;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.ContainmentConstraint;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTPattern;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTPatternBody;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.NonInjectivityConstraint;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.PatternVariable;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.PatternVariableAssignment;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.modelmanagement.queryFunctions.ElementReference;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.Entity;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.ModelElement;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.Relation;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.Relationship;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.SupertypeOf;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.TypeOf;



class PatternGraph {

	protected GTPattern gtPattern;
	protected GTPatternBody body;
	protected SimpleReteBuilder<?, ?> builder;
	protected Entity graphEntity;

	/**
	 * pNodes are pattern variables, indexed by their names.
	 * 
	 * VariableAssignments are already taken care of by variableEquivalence,
	 * pNodes are equivalence groups.
	 */
	protected Map<String, PatternNode> pNodeByName = new HashMap<String, PatternNode>();
	//protected Map<String, PatternEdge> pEdgeByName = new HashMap<String, PatternEdge>();
	/**
	 * pEdges are the pattern constraints / hyperedges
	 */
	protected Collection<PatternEdge> pEdges = new LinkedList<PatternEdge>();
	protected Map<PatternNode, Collection<PatternEdge>> edgeList = new HashMap<PatternNode, Collection<PatternEdge>>();
	// protected Set<PatternNode> boundNodes = new HashSet<PatternNode>();

	protected int nextVirtualNodeID = 0;

	/**
	 * contains the equivalence classes induced by variableassignments the
	 * factor set corresponds to the set of pNodes pNodes are named after
	 * equivalence class representatives
	 */
	protected UnionFind<String> variableEquivalence = new UnionFind<String>();

	/**
	 * Valid in shareable patterns.
	 * 
	 * For each pNode, this structure contains the set of pNodes against which
	 * it is explicitly tested for inequality during the injectivity check phase. 
	 * 
	 * For virtual nodes, it should always contain null, reflecting that
	 * virtual nodes are permitted to be equal with anything.
	 * 
	 * For regular nodes, it symmetrically gathers explicity inequality assertions.
	 */	
	protected Map<PatternNode, Set<PatternNode>> explicitInequality = new HashMap<PatternNode, Set<PatternNode>>();
	
	/**
	 * For each pNode, this structure contains the set of pNodes against which
	 * it is NOT tested for inequality during the injectivity check phase. 
	 *  
	 * For injective (distinct) patterns, it overrides the default injectivity check. 
	 * For shareable patterns, it overrides explicit inequality checking.  
	 *  
	 * For virtual nodes, it should always contain null, reflecting that
	 * virtual nodes are permitted to be equal with anything.
	 * 
	 * For regular nodes, it always contains identity. 
	 * Furthermore, parameters of a positive pattern call of an injective pattern 
	 * 		MAY be included in each others permittableEquality sets to avoid double-checking, 
	 * 		but only under conditions that if it is safe.
	 */
	protected Map<PatternNode, Set<PatternNode>> permittableEquality = new HashMap<PatternNode, Set<PatternNode>>();


	/**
	 * contains the values of constant nodes
	 */
	protected Map<PatternNode, Object> constantValues = new HashMap<PatternNode, Object>();

	/**
	 * @param gtPattern
	 * @param body
	 */
	public PatternGraph(GTPattern gtPattern, GTPatternBody body,
			SimpleReteBuilder<?, ?> builder)
			throws RetePatternBuildException {
		super();
		this.gtPattern = gtPattern;
		this.body = body;
		this.builder = builder;
		graphEntity = body.getPatternGraph();

		preProcessAssignments();
		preProcessPatternNodes();
		gatherInequalityAssertions();
		gatherContainmentConstraints();
		gatherRelationsAndContainments();
		gatherRelationships();
		gatherPositivePatternCalls();
		gatherCheckExpressions();
		expressTypeChecks();
		gatherNegativePatternCalls(); // should go last to have node.isTouched
										// ready
		touchQuantifiedVariables();
		connectPEdges(); 
		countTypeRestrictions();
	}










	public PatternNode getPNode(String name) {
		return pNodeByName.get(variableEquivalence.find(name));
	}

	public PatternNode getPNode(Variable variable) {
		return getPNode(variable.getName());
	}

	public PatternNode getPNode(ModelElement element) {
		return getPNode(element.getName());
	}

	public Collection<PatternEdge> getEdgeList(PatternNode pNode) {
		return edgeList.get(pNode);
	}

	/**
	 * @return the constantValues
	 */
	public Map<PatternNode, Object> getConstantValues() {
		return constantValues;
	}

	/**
	 * @return the graphEntity
	 */
	public Entity getGraphEntity() {
		return graphEntity;
	}

	protected void preProcessAssignments() {
		for (Object o : body.getVariableAssignments()) {
			PatternVariableAssignment assignment = (PatternVariableAssignment) o;
			String lName = assignment.getLeftValue().getVariable().getName();
			String rName = assignment.getRightValue().getVariable().getName();
			variableEquivalence.unite(lName, rName);
		}
	}

	protected void preProcessPatternNodes()
			throws RetePatternBuildException {
		for (Object o : gtPattern.getSymParameters())
			preProcessPatternNodeVariables((PatternVariable) o);
		for (Object o : body.getLocalVariables())
			preProcessPatternNodeVariables((PatternVariable) o);
		preProcessPatternNodeEntities(graphEntity.getComponents());
	}

	protected void preProcessPatternNodeVariables(PatternVariable variable) {
		/*
		 * Collection elements = variable.getElementInPattern(); boolean
		 * isRelation = (!elements.isEmpty() && elements.iterator().next()
		 * instanceof Relation);
		 */

		String name = variable.getName();
		if (variableEquivalence.isRoot(name)) // only if this is the
												// representative of an
												// equivalence class
		{
			PatternNode pNode = new PatternNode(variable/* , !isRelation */);
			pNodeByName.put(name, pNode);

			edgeList.put(pNode, new LinkedList<PatternEdge>());

			HashSet<PatternNode> permitted = new HashSet<PatternNode>();
			permitted.add(pNode);
			permittableEquality.put(pNode, permitted);
			explicitInequality.put(pNode, new HashSet<PatternNode>());
		}
	}

	@SuppressWarnings("unchecked")
	protected void preProcessPatternNodeEntities(Collection entities)
			throws RetePatternBuildException {
		for (Object o : entities) {
			Entity entity = (Entity) o;
			String name = entity.getName();
			getPNode(name).setSource(entity);
			// new PatternEdge(this, entity);
			preProcessPatternNodeEntities(entity.getComponents());
		}
	}

	public PatternNode newConstant(Constant constant) {
		Object value;
		if (constant.getKind().getValue() == ValueKind.MODELELEMENT) {
			value = builder.context.resolveConstant(constant.getValue());
		} else
			value = constant.getValue();

		return newConstantValue(value);
	}

	public PatternNode newConstant(ElementReference reference) {
		String fqn = ((Constant) reference.getArgument()).getValue();
		return newConstant(fqn);
	}

	public PatternNode newConstant(String fqn) {
		Object value = builder.context.resolveConstant(fqn);
		return newConstantValue(value);
	}

	public PatternNode substituteTerm(Term term)
			throws RetePatternBuildException {
		PatternNode pNode = newVirtualNode();
		pEdges.add(new PatternEdge(this, term, pNode));
		return pNode;
	}

	public void instantiationIndirect(PatternNode type, PatternNode instance) {
		PatternNode virtual = newVirtualNode();
		pEdges.add(new PatternEdge(this, type, virtual,
				HyperEdgeClass.GENERALIZATION_TRANSITIVE));
		pEdges.add(new PatternEdge(this, virtual, instance,
				HyperEdgeClass.INSTANTIATION_DIRECT));
	}

	PatternNode newConstantValue(Object value) {
		PatternNode pNode = newVirtualNode();
		constantValues.put(pNode, value);
		return pNode;
	}

	PatternNode newVirtualNode() {
		int newId = nextVirtualNodeID++;

		PatternNode pNode = new PatternNode(newId);
		pNodeByName.put(pNode.name, pNode);
		edgeList.put(pNode, new LinkedList<PatternEdge>());
		// permittedEqualities = null

		return pNode;
	}

	void touchQuantifiedVariables() {
		for (Object o : gtPattern.getSymParameters())
			getPNode((PatternVariable) o).isTouched = true;
		for (PatternEdge edge : pEdges)
			edge.touchQuantifiedNodes();
	}
	//PRE: quantified nodes have been touched
	void connectPEdges() {
		for (PatternEdge edge : pEdges)
			edge.connectNodes();
	} 
	
	protected void gatherInequalityAssertions() {
		for(Object o: body.getNonInjectivityConstraints()) {
			NonInjectivityConstraint nic = (NonInjectivityConstraint)o;
			PatternNode left = getPNode(nic.getLeftValue().getVariable());
			PatternNode right = getPNode(nic.getRightValue().getVariable());
			explicitInequality.get(left).add(right);
			explicitInequality.get(right).add(left);
		}
	}	
	
	protected void gatherRelationsAndContainments() {
		// using queue to protect against concurrentModificationException for created virtual nodes +
		// Relations can be added back after being gathered
		TreeSet<PatternNode> nodeQueue = new TreeSet<PatternNode>(pNodeByName.values());
		for (Object rel : body.getDanglingRelations()) // dangling relations cannot be collected otherwise
			gatherRelation((Relation) rel, nodeQueue); 

		while (!nodeQueue.isEmpty()) {
			PatternNode pNode = nodeQueue.first();
			nodeQueue.remove(pNode);
			if (pNode.source != null) {

				for (Object o : pNode.source.getRelationsFrom()) {
					Relation rel = (Relation) o;
					gatherRelation(rel, nodeQueue);
				}
				for (Object o : pNode.source.getRelationsTo()) {
					Relation rel = (Relation) o;
					gatherRelation(rel, nodeQueue);
				}
				if (pNode.source instanceof Entity) {
					Entity entity = (Entity) pNode.source;
					for (Object o : entity.getComponents()) {
						pEdges.add(new PatternEdge(this, pNode, getPNode((Entity) o),
								HyperEdgeClass.CONTAINMENT_DIRECT));
					}
				}

			}
		}
	}
	// idempotent
	protected void gatherRelation(Relation rel, Set<PatternNode> nodeQueue) {
		PatternNode pNode = getPNode(rel);
		if (pNode.source == null) {
			pNode.setSource(rel);
			// new PatternEdge(this, rel);
			if (!nodeQueue.contains(pNode))
				nodeQueue.add(pNode);
		}
	}
	
	protected void gatherRelationships()
			throws RetePatternBuildException {
		// against concurrentModificationException for created virtual nodes
		Collection<PatternNode> nodes = new ArrayList<PatternNode>(pNodeByName.values());
		Collection<Relationship> relationships = new LinkedHashSet<Relationship>(body.getDanglingRelationships());

		for (PatternNode pNode : nodes) {
			if (pNode.source != null) {
				boolean typed = false;
				
				for (Object o : pNode.source.getSuperRelationships()) {
					Relationship rship = (Relationship) o;
					if (rship instanceof TypeOf
							&& rship.getSupplier()!=null
							&& rship.getSupplier().getRealElement() != null
							&& !rship.getSupplier().getRealElement().equals("")) 
					{
						typed = true;
						//String supplierName = rship.getSupplier().getName(); // LOL what a coincidence
						String supplierName = rship.getSupplier().getRealElement();
						if (pNode.source instanceof Relation) {							
							Relation relation = (Relation) pNode.source;
							if (builder.context.edgeInterpretation() == EdgeInterpretation.TERNARY) {
								Object typeObject = builder.context.retrieveTernaryEdgeType(supplierName);
								pNode.assertedTypes.add(typeObject);
								if (typeObject != null) 
								{
									Object srcType = builder.context.ternaryEdgeSourceType(typeObject);
									Object trgType = builder.context.ternaryEdgeTargetType(typeObject);
									implyTypes(srcType, getPNode(relation.getFromStr()).impliedTypes);
									implyTypes(trgType, getPNode(relation.getToStr()).impliedTypes);
								}
							} else {
								Object typeObject = builder.context.retrieveBinaryEdgeType(supplierName);
								pNode.assertedTypes.add(typeObject);
								if (typeObject != null) 
								{
									Object srcType = builder.context.binaryEdgeSourceType(typeObject);
									Object trgType = builder.context.binaryEdgeTargetType(typeObject);
									implyTypes(srcType, getPNode(relation.getFromStr()).impliedTypes);
									implyTypes(trgType, getPNode(relation.getToStr()).impliedTypes);
								}
							}
						} else { // pNode.source instanceof Entity
							Object typeObject = builder.context.retrieveUnaryType(supplierName);
							pNode.assertedTypes.add(typeObject);
						}
							
					}
					else relationships.add(rship);
				}
				if (!typed) // manually add entity(A), relation(B, C, D) edges
				{
					pNode.assertedTypes.add(null);
				}

				for (Object o : pNode.source.getSubRelationships()) relationships.add((Relationship) o);
			}
		}
		for (Relationship rship : relationships) gatherRelationship(rship);
	}
	
	// not idempotent, unique invocation required
	protected void gatherRelationship(Relationship rship) throws RetePatternBuildException {
		String clientName = rship.getClientStr();
		PatternNode clientNode = getPNode(clientName);
		if (clientNode == null) clientNode = newConstant(clientName);

		String supplierName = rship.getSupplierStr();
		PatternNode supplierNode = getPNode(supplierName);
		if (supplierNode == null) {
			//if (rship instanceof TypeOf) 
			supplierNode = newConstant(supplierName);
		}
	
		HyperEdgeClass edgeClass;
		if (rship instanceof TypeOf) edgeClass = HyperEdgeClass.INSTANTIATION_TRANSITIVE;
		else if (rship instanceof SupertypeOf) edgeClass = HyperEdgeClass.GENERALIZATION_TRANSITIVE;
		else {
			String[] args = {rship.eClass().getName(), supplierName, clientName, gtPattern.getFqn()};
			String msg = "Unsupported VPM Relationship type: {1} encountered between {2} and {3} in pattern {4}";
			throw new RetePatternBuildException(msg, args, gtPattern);
		}
		
		pEdges.add(new PatternEdge(this, supplierNode, clientNode, edgeClass));
	}
	
	/**
	 * Implies that relation type typeObj and all supertypes are implicitly checked for the owner pNode of impliedTypes
	 * @param typeObj
	 * @param impliedTypes
	 */
	private void implyTypes(Object typeObj, Set<Object> impliedTypes) {
		if (Options.calcImpliedTypes && typeObj!=null && impliedTypes.add(typeObj)) {
			for (Object superType: builder.context.enumerateDirectSupertypes(typeObj)) {
				implyTypes(superType, impliedTypes);
			}
		}		
	}
	
//	/**
//	 * Implies that relation type typeObj and all supertypes are implicitly checked for the owner pNode of impliedTypes
//	 * @param typeObj
//	 * @param impliedTypes
//	 */
//	private void implyRelationTypes(Object typeObj, Set<Object> impliedTypes) {
//		if (impliedTypes.add(typeObj)) {
//			for (Object superType: builder.context.enumerateDirectRelationSupertypes(typeObj)) {
//				implyRelationTypes(superType, impliedTypes);
//			}
//		}
//	}
//	/**
//	 * Implies that reference type typeObj and all supertypes are implicitly checked for the owner pNode of impliedTypes
//	 * @param typeObj
//	 * @param impliedTypes
//	 */
//	private void implyReferenceTypes(Object typeObj, Set<Object> impliedTypes) {
//		if (impliedTypes.add(typeObj)) {
//			for (Object superType: builder.context.enumerateDirectReferenceSupertypes(typeObj)) {
//				implyReferenceTypes(superType, impliedTypes);
//			}
//		}
//	}



	protected void expressTypeChecks() {
		for (PatternNode pNode : pNodeByName.values()) {
			for (Object assertedType: pNode.assertedTypes) {
				if (pNode.source instanceof Relation) 
					pEdges.add(new PatternEdge(this, (Relation) pNode.source, assertedType));
				// entity optimization by implied types
				else if (assertedType==null && pNode.impliedTypes.isEmpty() || 
						assertedType!=null && !pNode.impliedTypes.contains(assertedType))
					pEdges.add(new PatternEdge(this, (Entity) pNode.source, assertedType));
				else {
					// Object o = 2*15; // No need to assert type, already implied!
				}
			}
		}
	}
	protected void gatherContainmentConstraints()
			throws RetePatternBuildException {
		for (Object o : body.getContainmentConstraints()) {
			ContainmentConstraint cc = (ContainmentConstraint) o;
			PatternNode child = getPNode(cc.getVariable());
			boolean transitive = cc.getMode().equals(
					ContainmentMode.BELOW_LITERAL);
			HyperEdgeClass relationshipClass = transitive ? HyperEdgeClass.CONTAINMENT_TRANSITIVE
					: HyperEdgeClass.CONTAINMENT_DIRECT;

			if (cc.getParent() instanceof VariableReference) {
				VariableReference parentVar = (VariableReference) cc
						.getParent();
				PatternNode parent = getPNode(parentVar.getVariable());

				pEdges.add(new PatternEdge(this, parent, child, relationshipClass));
			} else if (cc.getParent() instanceof Constant) {
				PatternNode parent = newConstant((Constant) cc.getParent());

				pEdges.add(new PatternEdge(this, parent, child, relationshipClass));

			} else if (cc.getParent() instanceof ElementReference) {
				PatternNode parent = newConstant((ElementReference) cc
						.getParent());

				pEdges.add(new PatternEdge(this, parent, child, relationshipClass));

			} else {
				PatternNode parent = substituteTerm(cc.getParent());

				pEdges.add(new PatternEdge(this, parent, child, relationshipClass));
				// // ???
				// hu.bme.mit.viatra.gtasmmodel.gtasm.metamodel.modelmanagement
				// .queryFunctions.impl.ElementReferenceImpl
				//				
				// builder.logger.warning(
				// "ContainmentConstraint parent must be a variable or constant or element reference instead of "
				// + cc.getParent().toString() );
				// //throw new RetePatternBuildException();
			}
		}

	}

	protected void gatherPositivePatternCalls()
			throws RetePatternBuildException {
		for (Object o : body.getCalledPatterns()) {
			GTPatternCall call = (GTPatternCall) o;
			pEdges.add(new PatternEdge(this, call, false));
		}
	}

	protected void gatherNegativePatternCalls()
			throws RetePatternBuildException {
		for (Object o : body.getNegativePatterns()) {
			GTPatternCall call = (GTPatternCall) o;
			pEdges.add(new PatternEdge(this, call, true));
		}
	}

	protected void gatherCheckExpressions()
			throws RetePatternBuildException {
		for (Object o : body.getCheckExpressions()) {
			Term term = (Term) o;
			pEdges.add(new PatternEdge(this, term, null/* equals true */));
		}

	}
	
	public void countTypeRestrictions() {
		for (PatternEdge e: pEdges)
			for (int index : e.conveysTypeInformationIndices()) {
				PatternNode node = e.nodes[index];
				node.lingeringTypeInfoCounter++;
				for (PatternEdge f: edgeList.get(node))
					f.totalLingeringTypeInfoCounter++;
			}

	}

	// protected void addEdge(PatternEdge edge)
	// {
	// if (edge.relMe!=null)
	// {
	// String name = edge.relMe.getName();
	// PatternNode pNode = pNodeByName.get(name);
	// if (pNode != null)
	// {
	// //pNodeByName.remove(name);
	// edgeList.remove(pNode);
	// pEdgeByName.put(name, edge);
	// }
	// }
	// }

}
