/*******************************************************************************
 * Copyright (c) 2004-2008 Peter Pasztor, Akos Horvath, Gergely Varro, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Peter Pasztor, Akos Horvath, Gergely Varro, Istvan Rath - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.gtasm.patternmatcher.patterns;


import org.eclipse.emf.common.util.EList;
import org.eclipse.viatra2.gtasm.interpreter.exception.ASMInterpreterErrorStrings;
import org.eclipse.viatra2.gtasm.interpreter.exception.GTASMException;
import org.eclipse.viatra2.gtasm.interpreter.exception.ViatraTransformationException;
import org.eclipse.viatra2.gtasm.interpreter.executionEnvironment.IExecutionEnvironment;
import org.eclipse.viatra2.gtasm.patternmatcher.PatternMatcherParameters;
import org.eclipse.viatra2.gtasm.patternmatcher.exceptions.PatternMatcherCompileTimeException;
import org.eclipse.viatra2.gtasm.patternmatcher.exceptions.PatternMatcherRuntimeException;
import org.eclipse.viatra2.gtasm.patternmatcher.extension.ViatraGTASMPatternMatcherPlugin;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.core.RuntimeAnnotation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTPattern;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTRule;


public class PatternMatcherProvider {

	private static PatternMatcherProvider _instance = new PatternMatcherProvider();
	private static String patternMathcerIDToUse = ViatraGTASMPatternMatcherPlugin.DEFAULT_PM_FACTORY_ID;
	private static String gtRuleMathcerIDToUse = ViatraGTASMPatternMatcherPlugin.DEFAULT_GT_FACTORY_ID;

	public static PatternMatcherProvider getInstance() {
		return _instance;
	}

	public static void setPatternMatcherIDToUse(String id) {
		patternMathcerIDToUse = id;
	}

	public static String getPatternMatcherIDToUse() {
		return patternMathcerIDToUse;
	}

	public static String getGTRuleMathcerIDToUse() {
		return gtRuleMathcerIDToUse;
	}

	public static void setGTRuleMathcerIDToUse(String ruleMathcerIDToUse) {
		gtRuleMathcerIDToUse = ruleMathcerIDToUse;
	}

	/**
	 * Returns a pattern matcher to work with
	 * 
	 * @param org.eclipse.viatra2.gtasm.interpreter.executionEnvironment
	 *            the execution environment of the invocation context
	 * @param gtPattern
	 *            the invoked pattern
	 * @return a patternMatcher for the pattern
	 * @throws PatternMatcherCompileTimeException
	 */
	public IPatternMatcher getPatternMatcher(
			IExecutionEnvironment executionEnvironment, GTPattern gtPattern)
			throws ViatraTransformationException {
		EList rtAnnotations = gtPattern.getRuntimeAnnotations();
		if (rtAnnotations != null && rtAnnotations.size() > 0) {
			for (Object obj : rtAnnotations) {
				RuntimeAnnotation an = null;
				if (obj instanceof RuntimeAnnotation) 
					an = (RuntimeAnnotation) obj;
				else {
					String[] context = {obj.toString()};
					PatternMatcherRuntimeException e = new PatternMatcherRuntimeException(
							ASMInterpreterErrorStrings.ANNOTATION_ILLEGAL, context, gtPattern);
					throw e;
				}
				if ("@incremental".equals(an.getAnnotationName().toLowerCase())){
//					boolean parallel = false;
//					for (Object o2 : an.getElements()) {
//						if (o2 instanceof RuntimeAnnotationElement) {
//							RuntimeAnnotationElement rae = (RuntimeAnnotationElement) o2;
//							if ("parallel".equals(rae.getKey()))
//							{
//								String lcValue = rae.getValue().toLowerCase();
//								parallel = "true".equals(lcValue) || "parallel".equals(lcValue) || "1".equals(lcValue); 
//								break;
//							}
//						}
//					}					
					return ViatraGTASMPatternMatcherPlugin
							.getDefault()
							.getPatternMatcherFactory(
									ViatraGTASMPatternMatcherPlugin.INCREMENTAL_PM_FACTORY_ID)
							.getPatternMatcher(executionEnvironment, gtPattern);					
				}
				else if ("@localsearch".equals(an.getAnnotationName().toLowerCase()))
					return ViatraGTASMPatternMatcherPlugin
							.getDefault()
							.getPatternMatcherFactory(
									ViatraGTASMPatternMatcherPlugin.LOCAL_PM_FACTORY_ID)
							.getPatternMatcher(executionEnvironment, gtPattern);
			}
		}
		return ViatraGTASMPatternMatcherPlugin.getDefault()
				.getPatternMatcherFactory(patternMathcerIDToUse)
				.getPatternMatcher(executionEnvironment, gtPattern);
	}

	/**
	 * Returns a GT rule matcher to work with
	 * 
	 * @param gtEnviroment
	 *            The execution environment on which the TERMs are evaluated
	 * @param gtRule
	 *            The invoked GT rule
	 * @param mp
	 *            The invocation context of the GT rule
	 * @return a gt rule matcher for the rule
	 * @throws GTASMException
	 */
	public IGTRuleMatcher getGtRuleMatcher(IExecutionEnvironment gtEnviroment,
			GTRule gtRule, PatternMatcherParameters mp)
			throws ViatraTransformationException {

		// GT rule level annotations
		// EList rtAnnotations = gtRule.getRuntimeAnnotations();
		// LHS level annotation
		EList rtAnnotations = gtRule.getPrecondition().getCalledPattern()
				.getRuntimeAnnotations();
		if (rtAnnotations != null && rtAnnotations.size() > 0) {
			// TODO: what to do with GT rule level annotations?
			for (Object obj : rtAnnotations) {
				RuntimeAnnotation an = null;
				if (obj instanceof RuntimeAnnotation)
					an = (RuntimeAnnotation) obj;
				else {
					String[] context = {obj.toString()};
					PatternMatcherRuntimeException e = new PatternMatcherRuntimeException(
							ASMInterpreterErrorStrings.ANNOTATION_ILLEGAL, context, gtRule);
					throw e;
				}
				if ("@incremental".equals(an.getAnnotationName().toLowerCase()))
					return ViatraGTASMPatternMatcherPlugin
							.getDefault()
							.getGTMatcherFactory(
									ViatraGTASMPatternMatcherPlugin.ALTERNATIVE_PATTERN_MATCHER_FOR_GT_FACTORY_ID)
							.getGTMatcher(gtEnviroment, gtRule, mp);
				else if ("@localsearch".equals(an.getAnnotationName()
						.toLowerCase()))
					return ViatraGTASMPatternMatcherPlugin
							.getDefault()
							.getGTMatcherFactory(
									ViatraGTASMPatternMatcherPlugin.LOCAL_GT_FACTORY_ID)
							.getGTMatcher(gtEnviroment, gtRule, mp);
			}
		}

		// //Annotation about LHS pattern matching
		// EList ptCallAnnotations =
		// gtRule.getPrecondition().getRuntimeAnnotations();
		// EList ptAnnotations =
		// gtRule.getPrecondition().getCalledPattern().getRuntimeAnnotations();
		//		
		// if(ptCallAnnotations != null && ptCallAnnotations.size()> 0)
		// {
		// for(Object obj: ptCallAnnotations)
		// {
		// RuntimeAnnotation an = null;
		// if(obj instanceof RuntimeAnnotation)
		// an =(RuntimeAnnotation)obj;
		// else
		// throw new
		// GTASMException(ASMInterpreterErrorStrings.ANNOTATION_ILLEGAL,gtRule);
		// if ( "@incremental".equals( an.getAnnotationName().toLowerCase() ))
		// return
		// ViatraGTASMPatternMatcherPlugin.getDefault().getGTMatcherFactory(ViatraGTASMPatternMatcherPlugin.ALTERNATIVE_PATTERN_MATCHER_FOR_GT_FACTORY_ID).getGTMatcher(gtEnviroment,
		// gtRule, mp);
		// else if ("@localsearch".equals( an.getAnnotationName().toLowerCase()
		// ) )
		// return
		// ViatraGTASMPatternMatcherPlugin.getDefault().getGTMatcherFactory(ViatraGTASMPatternMatcherPlugin.LOCAL_GT_FACTORY_ID).getGTMatcher(gtEnviroment,
		// gtRule, mp);
		// }
		// }
		//		
		// if(ptAnnotations != null && ptAnnotations.size()> 0)
		// {
		// for(Object obj: ptAnnotations )
		// {
		// RuntimeAnnotation an = null;
		// if(obj instanceof RuntimeAnnotation)
		// an =(RuntimeAnnotation)obj;
		// else
		// throw new
		// GTASMException(ASMInterpreterErrorStrings.ANNOTATION_ILLEGAL,gtRule);
		// if ( "@incremental".equals( an.getAnnotationName().toLowerCase() ))
		// return
		// ViatraGTASMPatternMatcherPlugin.getDefault().getGTMatcherFactory(ViatraGTASMPatternMatcherPlugin.ALTERNATIVE_PATTERN_MATCHER_FOR_GT_FACTORY_ID).getGTMatcher(gtEnviroment,
		// gtRule, mp);
		// else if ("@localsearch".equals( an.getAnnotationName().toLowerCase()
		// ) )
		// return
		// ViatraGTASMPatternMatcherPlugin.getDefault().getGTMatcherFactory(ViatraGTASMPatternMatcherPlugin.LOCAL_GT_FACTORY_ID).getGTMatcher(gtEnviroment,
		// gtRule, mp);
		// }
		// }

		return ViatraGTASMPatternMatcherPlugin.getDefault().getGTMatcherFactory(
				gtRuleMathcerIDToUse).getGTMatcher(gtEnviroment, gtRule, mp);
	}
	
	/**
	 *  Sets the pattern matcher provider to the default pattern and GT matcher
	 */
	public static void resetPatternMatcher() {
		setPatternMatcherIDToUse(ViatraGTASMPatternMatcherPlugin.DEFAULT_PM_FACTORY_ID);
		setGTRuleMathcerIDToUse(ViatraGTASMPatternMatcherPlugin.DEFAULT_GT_FACTORY_ID);
	}
}
