/*******************************************************************************
 * Copyright (c) 2004-2008 Andras Okros, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Andras Okros - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.gtasm.trigger.engine;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.PlatformUI;
import org.eclipse.viatra2.core.ICoreNotificationListener;
import org.eclipse.viatra2.core.notification.ICoreNotificationObject;
import org.eclipse.viatra2.core.simple.notification.NotificationObjectTransactionEnd;
import org.eclipse.viatra2.framework.IFramework;
import org.eclipse.viatra2.framework.IMachineSetChangedListener;
import org.eclipse.viatra2.gtasm.interpreter.exception.ViatraTransformationException;
import org.eclipse.viatra2.gtasm.interpreter.executionEnvironment.ASMFunctionContent;
import org.eclipse.viatra2.gtasm.interpreter.executionEnvironment.IExecutionEnvironment;
import org.eclipse.viatra2.gtasm.interpreter.impl.rules.RuleInterpreter;
import org.eclipse.viatra2.gtasm.interpreter.term.rules.TermEvaluator;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.EngineManager;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.matcher.ReteEngine;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.matcher.RetePatternMatcher;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.misc.DeltaMonitor;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.tuple.Tuple;
import org.eclipse.viatra2.gtasm.trigger.model.Trigger;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.core.RuntimeAnnotation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.core.RuntimeAnnotationElement;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.ASMFunction;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.InitialValue;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Machine;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.SymbolicRuleParameter;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Variable;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.enums.ValueKind;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.simpleRules.ASMRuleInvocation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.Term;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.VariableReference;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTPattern;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTRule;
import org.eclipse.viatra2.logger.Logger;


/**
 * Trigger Execution Engine; handles the proper execution of the trigger's action-sequences. 
 * @author Andr�s �kr�s
 */
public class TriggerExecutionEngine implements ICoreNotificationListener, IMachineSetChangedListener
{

	protected IFramework fw;
	protected ReteEngine<GTPattern> re = null;
	protected HashSet<Trigger> triggers;
	protected HashSet<ITriggerSetChangedListener> trigger_set_listeners;
	protected boolean trigger_running_flag;
	protected boolean serialExecutionMode; // if false > parallel mode
	
	public TriggerExecutionEngine(IFramework afw)
	{
		fw = afw;
		//getReteEngine();
		triggers = new HashSet<Trigger>();
		trigger_set_listeners = new HashSet<ITriggerSetChangedListener>();
		fw.getTopmodel().getNotificationManager().addAllListener(this);
		fw.addMachineSetListener(this);
		trigger_running_flag = false;
		serialExecutionMode = true; 
	}

	private ReteEngine<GTPattern> getReteEngine() {
		if (re==null) re = EngineManager.getInstance().getReteEngine(fw);
		return re;
	}
 		
	public void initTrigger(Machine current)
	{
		EList<GTRule> gtrules;
		boolean anyImport=false;
		gtrules = current.getGtRuleDefinitions();
		for (Object obj: gtrules)
		{
			GTRule gtrule = (GTRule) obj;
			EList<RuntimeAnnotation> runtimeAnnotations = gtrule.getRuntimeAnnotations();
			for (Object obj1: runtimeAnnotations) 
			{
				RuntimeAnnotation ran = (RuntimeAnnotation) obj1;
				if (ran.getAnnotationName().equals("@Trigger")) {
					Trigger input = new Trigger(gtrule.getName());
					anyImport=true;
					input.setMachine(current);
					input.setGtrule(gtrule);
					boolean isRunnable = false;
					EList<RuntimeAnnotationElement> annotations = ran.getElements();
					for (Object obj2: annotations) {
						RuntimeAnnotationElement ranElement = (RuntimeAnnotationElement) obj2;
						String key = ranElement.getKey();
						String value = ranElement.getValue();
						if (key.equals("priority")) 
						{
							try {
								int p = Integer.parseInt(value); 
								input.setPriority(p); // default priority: 0 
							}
							catch (NumberFormatException nfe) {
								fw.getLogger().info("Trigger engine: the " + input.getName() + " trigger's priority annotation is not a valid integer number.");
							}
						}
						if (key.equals("mode")) 
						{
							if (value.equals("always")) input.setMode("always"); // default
							if (value.equals("once")) input.setMode("once");
						}
						if (key.equals("sensitivity")) 
						{
							if (value.equals("rise")) input.setSensitivity("rise"); // default
							if (value.equals("fall")) input.setSensitivity("fall");
						}
						if (key.equals("startup")) 
						{
							if (value.equals("active")) input.setStartup("active"); // default
							if (value.equals("passive")) input.setStartup("passive");
						}
						if (key.equals("execution")) 
						{
							if (value.equals("forall")) input.setExecution("forall"); // default
							if (value.equals("iterate")) input.setExecution("iterate");
						}
						if (key.equals("autoStart")) 
						{
							if (value.equals("true")) isRunnable=true; 
						}
					}
					
					addTrigger(input);
					
					if (isRunnable) startTrigger(input);
				}
			}
		}
		if (anyImport) actionPerformed(new NotificationObjectTransactionEnd());
	}

	
	public void initTriggerRete (Trigger input)
	{
		//Setup a new RetePatternMatcher for the current trigger
		RetePatternMatcher rpm = null;
		try 
		{ 
			rpm=getReteEngine().accessMatcher(input.getGtrule().getPrecondition().getCalledPattern());
		} 
		catch (Exception e) { e.printStackTrace(); }
		//Setup a DeltaMontior Receiver for the pattern matcher
		DeltaMonitor dm = new DeltaMonitor(re.getReteNet());
		if (input.getStartup().equals("active"))
		{
			rpm.connect(dm, true);
		}
		else
		{
			rpm.connect(dm, false);
		}
		input.setReceiver(dm);
		input.setRetepatternmatcher(rpm);
		input.setReteEngineSet(true);
	}
	
	public void initTriggerASMFunctions(Trigger input)
	{
		IExecutionEnvironment ee = new TriggerExecutionEnvironment(fw);
		for (Object o : input.getMachine().getAsmFunctionDefinitions())
		{
			ASMFunction fun =(ASMFunction)o;
			HashMap<EList<Object>,Object> map= new HashMap<EList<Object>, Object>();
			for (Object o2 : fun.getInitialValues())
			{
				InitialValue initialValue =(InitialValue)o2;
				try {
					EList<Object> locationsList= new BasicEList<Object>();
					for (Object o3 : initialValue.getLocations())
					{
						Term locationsTerm=(Term)o3;
						locationsList.add(TermEvaluator.getInstance().evaluate(ee, locationsTerm));
					}
					map.put(locationsList,TermEvaluator.getInstance().evaluate(ee,initialValue.getValue()));
				} catch (ViatraTransformationException e) {	e.printStackTrace(); } 			
			}
			if (!ASMFunctionContent.getInstance().containsKey(fun))	ASMFunctionContent.getInstance().put(fun, map);
		}
		input.setASMFunctionsSet(true);
	}

	public void startTrigger(Trigger input)
	{
		if (!input.isReteEngineSet())
			initTriggerRete(input);
		input.setRunning(true);
		for (ITriggerSetChangedListener tsl: trigger_set_listeners) {
			tsl.triggerRefreshed(input);
			}
	}
	
	public void startTriggerWithCommit(Trigger input)
	{
		startTrigger(input);
		actionPerformed(new NotificationObjectTransactionEnd());
	}
	
	public void stopTrigger(Trigger input)
	{
		input.setRunning(false);
		for (ITriggerSetChangedListener tsl: trigger_set_listeners) {
			tsl.triggerRefreshed(input);
			}
	}
	
	public void addTrigger(Trigger input)
	{
		// Remove the previous version
		for (Trigger trigger : triggers) {
			if (trigger.getName().equals(input.getName())){
				deleteTrigger(trigger);
				break;
			}
		}
		
		triggers.add(input);
		for (ITriggerSetChangedListener tsl: trigger_set_listeners) {
			tsl.triggerAdded(input);
			}
		fw.getLogger().info("Trigger engine: the " + input.getName() + " trigger has been imported.");
	}
	
	
	public void deleteTrigger(Trigger input)
	{
		triggers.remove(input);
		for (ITriggerSetChangedListener tsl: trigger_set_listeners) {
			tsl.triggerRemoved(input);
			}
		fw.getLogger().info("Trigger engine: the " + input.getName() + " trigger has been removed.");
	}
	
	
	//Run the trigger's action sequence with the correct variable settings.
	public void runTriggerAction(IFramework fw, final Trigger input, Tuple ps)
	{
		IExecutionEnvironment ee = new TriggerExecutionEnvironment(fw);
		HashMap <Object,Integer> params = input.getRetepatternmatcher().getPosMapping();
		Set<Object> variables = params.keySet();
		

		try {
			// sym parameters
			for (SymbolicRuleParameter param: input.getGtrule().getSymParameters()) {
				ee.addVariable(param.getVariable(), ValueKind.UNDEF_LITERAL);
			}			
			// the actual parameters of the precondition are used
			EList<Term> actpars = input.getGtrule().getPrecondition().getActualParameters();
			int pos=0;
			for (Object _o : actpars)
			{
				Object value = ps.get(pos++);
				VariableReference ref = (VariableReference) _o;
				Variable var = ref.getVariable(); // "asm" variable
				if (!ee.getVariableValues().keySet().contains(var)) // only add if not added yet
					ee.addVariable(var, value);
				else
					ee.setVariableValue(var, value);
			}
			
			ASMRuleInvocation asmri = input.getGtrule().getAction();
			RuleInterpreter.initInterpreters(fw.getLogger(), fw.getCodeOutput());
			
			//trigger_running_flag = true;
			fw.getTopmodel().getTransactionManager().beginTransaction();
			try {
				RuleInterpreter.getInstance().interpretRule(ee, asmri,null);
			} finally {
				fw.getTopmodel().getTransactionManager().commitTransaction();
			}
			
			//trigger_running_flag = false;
		}
		catch (final ViatraTransformationException e) {
			fw.getLogger().message(Logger.ERROR, e.getMessage(), e);
			Display.getDefault().syncExec(new Runnable(){
				public void run() {
					MessageDialog.openInformation(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), 
						"VIATRA2 R3", 
						"Error executing trigger "+ input.getName() +": " + 
							e.getMessage() + " (see the Error Log view for details)");							
					}	
				});
		}
		
	}

	
	// Runs the trigger with the correct execution semantics and sensitivity settings
	public void runTrigger(Trigger triggertorun)
	{
		// This is default execution semantics; the forall semantics > executes all of the possible matchings 
		if (triggertorun.getExecution().equals("forall"))
		{
			if (triggertorun.getSensitivity().equals("rise"))
			{
				while (triggertorun.getReceiver().matchFoundEvents.iterator().hasNext())
				{
					Tuple ps = triggertorun.getReceiver().matchFoundEvents.iterator().next();
					triggertorun.getReceiver().matchFoundEvents.remove(ps);
					runTriggerAction(fw, triggertorun, ps);
				}
			}
			if (triggertorun.getSensitivity().equals("fall"))
			{
				while (triggertorun.getReceiver().matchLostEvents.iterator().hasNext())
				{
					Tuple ps = triggertorun.getReceiver().matchLostEvents.iterator().next();
					triggertorun.getReceiver().matchLostEvents.remove(ps);
					runTriggerAction(fw, triggertorun, ps);
				}
			}
		} else if (triggertorun.getExecution().equals("iterate"))
		// the iterate semantics > only executes one matching of the trigger
		{
			if (triggertorun.getSensitivity().equals("rise"))
			{
				if (triggertorun.getReceiver().matchFoundEvents.iterator().hasNext())
				{
					Tuple ps = triggertorun.getReceiver().matchFoundEvents.iterator().next();
					triggertorun.getReceiver().matchFoundEvents.remove(ps);
					runTriggerAction(fw, triggertorun, ps);
				}
			}
			if (triggertorun.getSensitivity().equals("fall"))
			{
				if (triggertorun.getReceiver().matchLostEvents.iterator().hasNext())
				{
					Tuple ps = triggertorun.getReceiver().matchLostEvents.iterator().next();
					triggertorun.getReceiver().matchLostEvents.remove(ps);
					runTriggerAction(fw, triggertorun, ps);
				}
			}
		}
	}
	
	
	public void execute_serial()
	{
		// This trigger will be the one to run in this "procession-round"
		Trigger triggertorun = null;
		
		// Find the max priority trigger to run first
		// Note: the action sequence of that trigger will make a new TA_TRANSACTION_END (which starts the trigger execution engine)
		for (Trigger tr :triggers)
		{
			if (tr.isRunning()) 
			{
				if (!tr.getReceiver().matchFoundEvents.isEmpty() && tr.getSensitivity().equals("rise"))
				{
					if (triggertorun==null)	{ triggertorun = tr; }
					else if (tr.getPriority()>triggertorun.getPriority())
					{ triggertorun = tr; }
				}
				if (!tr.getReceiver().matchLostEvents.isEmpty() && tr.getSensitivity().equals("fall"))
				{
					if (triggertorun==null)	{ triggertorun = tr; }
					else if (tr.getPriority()>triggertorun.getPriority())
					{ triggertorun = tr; }
				}
			}
		}
						 
		if (triggertorun != null) 
		{
			// Set up the ASMFunctions before the first run of the trigger.
			if (!triggertorun.isASMFunctionsSet()) {
				initTriggerASMFunctions(triggertorun);
			}
			
			runTrigger(triggertorun);
			
			// Drop the trigger if it had a "once" mode
			if (triggertorun.getMode().equals("once"))
			{
				// Drop the trigger from the engine.
				deleteTrigger(triggertorun);
			}					
		}
	}
	
	
	public void execute_parallel()
	{
		// All of the triggers on the list will be executed
		HashSet<Trigger> triggertorunlist;
		triggertorunlist = new HashSet<Trigger>();
		for (Trigger tr :triggers)
		{
			if (tr.isRunning()) 
			{
				if (!tr.getReceiver().matchFoundEvents.isEmpty() && tr.getSensitivity().equals("rise"))
				{
					triggertorunlist.add(tr);
				}
				if (!tr.getReceiver().matchLostEvents.isEmpty() && tr.getSensitivity().equals("fall"))
				{
					triggertorunlist.add(tr);
				}
			}
		}
		
		for (Trigger triggertorun :triggertorunlist)
		{
			// Set up the ASMFunctions before the first run of the trigger.
			if (!triggertorun.isASMFunctionsSet()) {
				initTriggerASMFunctions(triggertorun);
			}
			
			runTrigger(triggertorun);
			
			// Drop the trigger if it had a "once" mode
			if (triggertorun.getMode().equals("once"))
			{
				// Drop the trigger from the engine.
				deleteTrigger(triggertorun);
			}					
		}
	}
	
	
	//Listeners for tracking transactions.
	public void actionPerformed(ICoreNotificationObject notification) {
		
		if ( notification.getActionType().equals(ICoreNotificationObject.TA_TRANSACTION_BEGIN) || 
				notification.getActionType().equals(ICoreNotificationObject.TA_UNDOABLE_TRANSACTION_BEGIN) )
		{
			if (trigger_running_flag==false) {
				for (Trigger tr :triggers) 
				{
					if (tr.isRunning())
						tr.getReceiver().clear();
				}
				fw.getLogger().info("Trigger engine: Receiver cleared.");
			}
		}
		
		if ( notification.getActionType().equals(ICoreNotificationObject.TA_TRANSACTION_END) ||
				notification.getActionType().equals(ICoreNotificationObject.TA_UNDO_END) )
			{
				fw.getLogger().info("Trigger engine: A transactional operation ended in the system.");
				
				if(!triggers.isEmpty()) {
					getReteEngine().settle();
					trigger_running_flag = true;
					if (serialExecutionMode) { 
						execute_serial();
					} else { 
						execute_parallel();
					}		
					trigger_running_flag = false;
				}
			}
	}

	
	public int getListenerCategory() {
		return 0;
	}

	
	public HashSet<Trigger> getTriggers() {
		return triggers;
	}
	
	
	public void addTriggerSetListener(ITriggerSetChangedListener l) 
	{
		if (!trigger_set_listeners.contains(l)) trigger_set_listeners.add(l);
	}

	
	public void removeTriggerSetListener(ITriggerSetChangedListener l) 
	{
		trigger_set_listeners.remove(l);
	}


	public boolean isSerialExecutionMode() {
		return serialExecutionMode;
	}


	public void setSerialExecutionMode(boolean serialExecutionMode) {
		this.serialExecutionMode = serialExecutionMode;
	}


	public boolean isTrigger_running_flag() {
		return trigger_running_flag;
	}


	public void machineAdded(Object objMachine) {
		Machine machine = (Machine) objMachine;
		fw.getLogger().info("Trigger engine: The " + machine.getName() + " machine has been added to the system.");
		initTrigger(machine);		
	}


	public void machineRemoved(Object objMachine) {
		Machine machine = (Machine) objMachine;
		fw.getLogger().info("Trigger engine: The " + machine.getName() + " machine has been removed from the system.");
	}


}
