/*******************************************************************************
 * Copyright (c) 2009-2010, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *   Istvan Rath - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.imports.uml2.galileo.modelexporter;

import org.eclipse.viatra2.core.IEntity;
import org.eclipse.viatra2.core.IModelElement;
import org.eclipse.viatra2.core.IRelation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.util.EcoreUtil;

public class UmlModelExportModel {

	private Map<IEntity, EPackage> packageMap;

	public Map<IEntity, EPackage> getPackageMap() {
		if (packageMap==null) packageMap = new HashMap<IEntity, EPackage>();
		return packageMap;
	}

	private EPackage getMappedPackage(IEntity ent)
	{
		for (IModelElement me : ent.getTypes())
		{
			if (me instanceof IEntity && getPackageMap().containsKey(((IEntity)me).getParent()))
				return getPackageMap().get(((IEntity)me).getParent());
		}
		return null;
	}

	private EClassifier getEClassifier(IEntity ent)
	{
		EPackage pack = getMappedPackage(ent);
		if (pack==null) return null;
		for (IModelElement me : ent.getTypes())
		{
			if (me instanceof IEntity)
			{
				EClassifier ecl = pack.getEClassifier(me.getName());
				if (ecl!=null) return ecl;
			}
		}
		return null;
	}

	private IRelation getRelationMetaForName(IEntity ent, String name)
	{

		for (IModelElement meta : ent.getAllTypes())
		{
			if (meta.getRelationFromByName(name)!=null)
				return meta.getRelationFromByName(name);
		}


		return null;
	}
	private List<IEntity> listReferencedEntities(IEntity ent, String name)
	{
		List<IEntity> result = new ArrayList<IEntity>();

		IRelation refrel = getRelationMetaForName(ent,name);

		if (refrel!=null)
		{
			for (IRelation rr : ent.getRelationsFrom())
			{
				if (rr.getTypes().contains(refrel) && rr.getTo() instanceof IEntity)
					result.add((IEntity)rr.getTo());
			}
		}
		return result;
	}
	public EObject exportModel(IEntity ent)
	{
		elementMap = new HashMap<IEntity, Object>();
		references = new ArrayList<ReferenceValue>();
		try{		
			Object o = mapEntity(ent,null);
		if (o instanceof EObject)
		{
			resolveReferences();
			return (EObject)o;
		}
		} catch (Throwable t)
		{
			t.printStackTrace();
		}

		return null;
	}
	private class ReferenceValue
	{
		private EObject object;
		private EReference feature;
		private IEntity value;

		ReferenceValue(EObject object, EReference feature, IEntity value)
		{
			this.object=object;
			this.feature=feature;
			this.value=value;
		}

		public EReference getFeature() {
			return feature;
		}

		public void setFeature(EReference feature) {
			this.feature = feature;
		}

		public EObject getObject() {
			return object;
		}

		public void setObject(EObject object) {
			this.object = object;
		}

		public IEntity getValue() {
			return value;
		}

		public void setValue(IEntity value) {
			this.value = value;
		}


	}
	List<ReferenceValue> references;
	Map<IEntity,Object> elementMap;
	private void resolveReferences()
	{
		for (ReferenceValue ref : references)
		{
			Object o = elementMap.get(ref.getValue());
			if (o!=null)
			{
				setFeatureValue(ref.getObject(), ref.getFeature(), o);
			} else java.lang.System.out.println("!reference unresolved: "+ref.getValue().getFullyQualifiedName());
		}
	}
	private Object mapEntity(IEntity ent, EClassifier typeHint)
	{
		Object outObject = null;

		EClassifier cf = getEClassifier(ent);
		if (cf==null)
		{
			cf = typeHint;
			if (cf==null) return null;
		}
		
		if (cf instanceof EClass)
		{
			EClass cl = (EClass)cf;
			outObject = EcoreUtil.create(cl);
			elementMap.put(ent, outObject);
			for (Object oattr : cl.getEAllAttributes())
			{
				EAttribute attr = (EAttribute)oattr;
				List<IEntity> rets = listReferencedEntities(ent,attr.getName());
				for (IEntity e : rets)
				{
					Object o = mapEntity(e,attr.getEType());
					if (o!=null)
					{
						setFeatureValue((EObject)outObject,attr,o);
					}
				}
			}
			for (Object oref : cl.getEAllReferences())
			{
				EReference ref = (EReference)oref;
				List<IEntity> rets = listReferencedEntities(ent,ref.getName());
				if (ref.isContainment())
				{
					for (IEntity e : rets)
					{
						Object o = mapEntity(e,null);
						if (o!=null)
						{
							setFeatureValue((EObject)outObject,ref,o);
						}
					}
				}
				else
				{
					for (IEntity e : rets)
					{
						references.add(new ReferenceValue((EObject)outObject,ref,e));
					}
				}
			}
		}
		else
		{
			if (ent.getValue()!=null)
				try{
				outObject = EcoreUtil.createFromString((EDataType)cf, ent.getValue());
				}
			catch (Throwable t) {}
		}
		return outObject;
	}

	private void setFeatureValue(EObject obj, EStructuralFeature feat, Object value)
	{
		if (value==null) return;
		try
		{
		if (feat.isMany())
		{
			((List)obj.eGet(feat)).add(value);
		}
		else
			obj.eSet(feat, value);
		}
		catch (Throwable t)
		{
			java.lang.System.out.println(obj.toString()+" "+feat.getName()+"   "+value.toString());
			t.printStackTrace();
		}
	}


}
