/**
* Copyright (c) 2007 OptXware Research and Development LLC.
* All rights reserved.   This program and the accompanying materials
* are made available under the terms of the Eclipse Public License v1.0
* which accompanies this distribution, and is available at
* http://www.eclipse.org/legal/epl-v10.html
*
* Contributors:
*   OptXware - Initial API and implementation
*
* Created: 
* 		2007.09.28	Daniel Varro
*/

package org.eclipse.viatra2.lpgparser.loader;

import java.io.InputStream;
import java.util.Iterator;
import java.util.List;

import org.eclipse.viatra2.errors.VPMRuntimeException;
import org.eclipse.viatra2.errors.info.ErrorInformation;
import org.eclipse.viatra2.errors.info.ErrorInformation.ErrorSeverity;
import org.eclipse.viatra2.framework.IFramework;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Machine;
import org.eclipse.viatra2.loaders.Loader;
import org.eclipse.viatra2.logger.Logger;

/**
 * Summary: 
 * 	This class manages how to parse and load a VTCL transformation to the 
 *  VIATRA framework.  
 * 
 * Description: 
 * 	LPGVTCLLoader initiates the parsing of a VTCL file with respect to an open 
 *  VIATRA framework. In case of successful parsing (i.e when no error messages
 *  are retrieved by the parser or the validator), it returns the EMF model of
 *  the transformation.    
 *   
 * @author Daniel Varro
 *
 */
public class LPGVTCLLoader implements Loader {

	/**
	 * This method of the interface {@link Loader} is not currently in use
	 */
	public void process(InputStream f, IFramework fw)	throws VPMRuntimeException 
	{
		// return null;
	}
	
	/**
	 * This method of the interface {@link Loader} is not currently in use
	 */
	protected void loadNativeFunctions(IFramework framework) {
		
	}

	/**
	 * 	This method initiates the parsing of a VTCL file with respect to an open 
	 *  VIATRA framework. In case of successful parsing (i.e when no error messages
	 *  are retrieved by the parser or the validator), it returns the EMF model of
	 *  the transformation. If there are errors a null value is returned as result    
	 * 
	 */
	public void processFile(String fileName, IFramework framework) throws VPMRuntimeException 
	{
		// Lookup parseController in framework and create one if none is present
		VTCLParserManager parserManager = (VTCLParserManager) framework.getVTCLParserManager();
		VTCLParseController parseController = parserManager.lookupAndCreateParseController(fileName);
		//VTCLParseController parseController = new VTCLParseController();
    	Logger log = framework.getLogger();

    	try {
    		// TODO: Check that parserController is already initialized
			//parseController.initialize(fileName, framework);
			
			// Call parser
			//ParserResult result = parseController.parseAndBuild();
			parseController.parseAndBuild();
			
			// Report if there are no errors
			if (parseController.getErrors().size() == 0) {
				String msg = "Parsing completed successfully. Transformation " +
				parseController.getMachine().getName() +
				" can now be executed in the VIATRA Frameworks view";
				log.info(msg);
				// Add the GTASM Machine to the framework
				//Machine m = (Machine) result.module.getMachine().get(0);
				Machine m = parseController.getMachine();
				// TODO: This code should be rewritten to accommodate the new MachineWrapper				
				framework.addMachine(m.getFqn(), m);
			}
			else {
				boolean hasErrors = false;
				// First check if there are errors
				Iterator<ErrorInformation> it1 = parseController.getErrors().iterator();
				while (!hasErrors && it1.hasNext()) {
					ErrorInformation error = it1.next();
					if (error.getErrorSeverity().compareTo(ErrorSeverity.ERROR) == 0) {
						hasErrors = true;
					}
				}
				
				if (!hasErrors) {
					String msg = "Parsing completed, but there are some warnings in the Error Log. Transformation " +
								parseController.getMachine().getName() +
								" can now be executed in the VIATRA Frameworks view";
					log.info(msg);
					// Add the GTASM Machine to the framework
					//Machine m = (Machine) result.module.getMachine().get(0);
					Machine m = parseController.getMachine();
					framework.addMachine(m.getFqn(), m);
				}
				// Report errors and warnings
				List<ErrorInformation> errorList = parseController.getErrors();
				for (int i = 0; i < errorList.size(); i++) {
					ErrorInformation errorInfo = errorList.get(i);
					String errorMsg = errorInfo.getMessageWithLocation();
					ErrorSeverity errorSeverity = errorInfo.getErrorSeverity();
					switch (errorSeverity) {
					case ERROR:
						log.error(errorMsg);
						break;
					case WARNING:
						log.warning(errorMsg);
						break;
					default:
						break;
					}
				}
				
			}
		} catch (Exception e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
		}
//
//		String[] args = new String[]{"-d", fileName};
//        Option option;
//        VTCLLexer vtcl_lexer;
//        VTCLParser vtcl_parser;
//        Module vtclModule;
//
//        ASTNode ast;
//
//        try
//        {
//            option = new Option(args);
//            option.readInputChars();
//            vtcl_lexer = new VTCLLexer(option.getInputChars(), option.getFileName()); // Create the lexer
//            vtcl_lexer.setTab(4);
//            vtcl_parser = new VTCLParser(vtcl_lexer);	// Create the parser
//            vtcl_lexer.lexer(vtcl_parser); // Lex the stream to produce the token stream
////            if (option.dumpTokens())
////            {
////                System.out.println("\n****Output Tokens: \n");
////                vtcl_parser.dumpTokens();
////            }
//            ast = vtcl_parser.parser(100); // Parse the token stream to produce an AST
//        	Logger log = framework.getLogger();
//            if (ast == null) {
////            	System.out.println("****Failure");
////            	System.out.println("Number of parser errors: " + vtcl_parser.getAllErrorInfos().size());
//            	for (int i = 0; i < vtcl_parser.getAllErrorInfos().size(); i++) {
//            		ErrorInformation errorInfo = vtcl_parser.getAllErrorInfos().get(i);
//            		String errorMsg = errorInfo.getMessageWithLocation();
//            		ErrorSeverity errorSeverity = errorInfo.getErrorSeverity();
//            		switch (errorSeverity) {
//					case ERROR:
//						log.error(errorMsg);
//						break;
//					case WARNING:
//						log.warning(errorMsg);
//						break;
//					default:
//						break;
//					}
//            	}
//            }
//            else
//            {
////            	System.out.println("*** Parsing completed ***");
//            	// Initializing a sample model space
//            	IModelSpace modelSpace = framework.getTopmodel();
//            	// Retrieve native functions from framework
//            	ASMNativeFunction[] asmNativeFunctions = framework.getNativeFunctionManager().getAllNativeFunctions();
//
//            	// Starting the model builder
//            	VTCLModelBuilder modelBuilder = new VTCLModelBuilder(modelSpace, asmNativeFunctions);
//            	// Setting the file name for the model builder
//            	modelBuilder.setFileName(fileName);
//            	vtclModule = (Module) ast.accept(modelBuilder);
//            	
//            	// Reporting parser errors
////            	System.out.println("Number of parser errors: " + vtcl_parser.getAllErrorInfos().size());
//            	for (int i = 0; i < vtcl_parser.getAllErrorInfos().size(); i++) {
//            		ErrorInformation errorInfo = vtcl_parser.getAllErrorInfos().get(i);
//            		String errorMsg = errorInfo.getMessageWithLocation();
////            	    System.out.println(errorMsg);
//            		ErrorSeverity errorSeverity = errorInfo.getErrorSeverity();
//            		switch (errorSeverity) {
//					case ERROR:
//						log.error(errorMsg);
//						break;
//					case WARNING:
//						log.warning(errorMsg);
//						break;
//					default:
//						break;
//					}
//            	}
//            	
//            	// Reporting model resolution and validation errors
////            	System.out.println("*** Starting model builder ***");
//            	VTCLModelResolver modelResolver = modelBuilder.getModelResolver();
////            	System.out.println("Number of validation errors: " + 
////            		modelResolver.getAllReferenceErrorInfos().size());
//            	// Reporting model validation errors 
//            	for (int i = 0; i < modelResolver.getAllReferenceErrorInfos().size(); i++) {
//            		ErrorInformation errorInfo = modelResolver.getAllReferenceErrorInfos().get(i);
//            		String errorMsg = errorInfo.getMessageWithLocation();
////            	    System.out.println(errorMsg);
//            		ErrorSeverity errorSeverity = errorInfo.getErrorSeverity();
//            		switch (errorSeverity) {
//					case ERROR:
//						log.error(errorMsg);
//						break;
//					case WARNING:
//						log.warning(errorMsg);
//						break;
//					default:
//						break;
//					}
//            	}
//            	if (modelResolver.getAllReferenceErrorInfos().size() == 0 && 
//            		vtcl_parser.getAllErrorInfos().size() == 0) {
//            		String msg = "Parsing completed successfully. Transformation " +
//            					((Machine) vtclModule.getMachine().get(0)).getName() +
//            					" can now be executed in the VIATRA Frameworks view";
//            		log.info(msg);
//            		//return vtclModule.getMachine().get(0);
//            		Machine m = (Machine)vtclModule.getMachine().get(0);
//            		framework.addMachine(m.getFqn(), m);
//            	}
//            	// Checks if we have only warnings
//            	else {
//                	boolean hasErrors = false;
//                	// First check if there are parser errors
//                	List<ErrorInformation> parserErrors = vtcl_parser.getAllErrorInfos();
//                	Iterator<ErrorInformation> it1 = parserErrors.iterator();
//                	while (!hasErrors && it1.hasNext()) {
//						ErrorInformation error = it1.next();
//						if (error.getErrorSeverity().compareTo(ErrorSeverity.ERROR) == 0) {
//							hasErrors = true;
//						}
//					}
//                	
//                	// Then check if there are validation errors
//                	List<ErrorInformation> errors = modelResolver.getAllReferenceErrorInfos();
//                	Iterator<ErrorInformation> it = errors.iterator();
//                	while (!hasErrors && it.hasNext()) {
//						ErrorInformation error = it.next();
//						if (error.getErrorSeverity().compareTo(ErrorSeverity.ERROR) == 0) {
//							hasErrors = true;
//						}
//					}
//                	if (!hasErrors) {
//                		String msg = "Parsing completed successfully. Transformation " +
//                					((Machine) vtclModule.getMachine().get(0)).getName() +
//                					" can now be executed in the VIATRA Frameworks view";
//                		log.info(msg);
//                		//return vtclModule.getMachine().get(0);
//                		Machine m = (Machine)vtclModule.getMachine().get(0);
//                		framework.addMachine(m.getFqn(), m);
//                	}
//                	else {
//                		//return null;
//                	}
//            		
//            	}
//            }
//        }
//        catch (Exception e)
//        {
//            System.err.println(e.getMessage());
//            e.printStackTrace();
//        }
//        //return null;
	}

}
