/**
* Copyright (c) 2007 OptXware Research and Development LLC.
* All rights reserved.   This program and the accompanying materials
* are made available under the terms of the Eclipse Public License v1.0
* which accompanies this distribution, and is available at
* http://www.eclipse.org/legal/epl-v10.html
*
* Contributors:
*   Daniel Varro - Initial API and implementation
*/

package org.eclipse.viatra2.lpgparser.loader;

import java.io.IOException;
import java.util.List;

import lpg.lpgjavaruntime.DiagnoseParser;
import lpg.lpgjavaruntime.ParseTable;

import org.eclipse.viatra2.core.IModelSpace;
import org.eclipse.viatra2.errors.VPMRuntimeException;
import org.eclipse.viatra2.errors.info.ErrorInformation;
import org.eclipse.viatra2.errors.info.ErrorInformation.ErrorSeverity;
import org.eclipse.viatra2.framework.IFramework;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Machine;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Module;
import org.eclipse.viatra2.lpgparser.Option;
import org.eclipse.viatra2.lpgparser.VTCLLexer;
import org.eclipse.viatra2.lpgparser.VTCLParser;
import org.eclipse.viatra2.lpgparser.VTCLParserprs;
import org.eclipse.viatra2.lpgparser.ast.ASTNode;
import org.eclipse.viatra2.lpgparser.modelbuilder.VTCLModelBuilder;
import org.eclipse.viatra2.lpgparser.modelbuilder.VTCLModelResolver;
import org.eclipse.viatra2.lpgparser.typechecker.VTCLTypeChecker;
import org.eclipse.viatra2.natives.ASMNativeFunction;

/**
 * This class is responsible for calling the parser and the model builder for 
 * a given file. It stores the AST and, in case of successful build process, 
 * it also stores the GTASM machine.  
 * @author Daniel Varro
 *
 */
public class VTCLParseController  {

	protected ASTNode fAst;
	protected VTCLParser fParser;
	protected VTCLModelResolver fModelResolver;
	protected VTCLModelBuilder fModelBuilder;
	protected VTCLTypeChecker fTypeChecker;
	protected VTCLLexer fLexer;
//	protected IASTNodeLocator fNodeLocator;

	protected Module fVtclModule;
	protected String fFileName;
	protected IFramework fFramework;
	
	protected List<ErrorInformation> errors;

    protected char fKeywords[][];
    protected boolean fIsKeyword[];

    public VTCLModelResolver getModelResolver() {
    	return fModelResolver;
    }

    public VTCLModelBuilder getModelBuilder() {
    	return fModelBuilder;
    }
    
    public VTCLTypeChecker getTypeChecker() {
    	return fTypeChecker;
    }
    
	public String getFileName() {
		return fFileName;
	}
	
//	public IFramework getFramework() {
//		return fFramework;
//	}
	
	public VTCLParser getParser() {
		return fParser;
	}

	public VTCLLexer getLexer() {
		return fLexer;
	}

	public Module getModule() {
		return fVtclModule;
	}
	
//	public IASTNodeLocator getNodeLocator() {
//		return fNodeLocator;
//	}
	
	public Machine getMachine() {
		if (fVtclModule!=null)
		{
			if (fVtclModule.getMachine()!=null)
			{
				return fVtclModule.getMachine().get(0);
			}
		}
		return null;
	}
	
	public ASTNode getCurrentAst() {
		return fAst;
	}

	public boolean hasErrors() {
		if (!errors.isEmpty()) {
			for (ErrorInformation error : errors) {
				if (error.getErrorSeverity().compareTo(ErrorSeverity.ERROR) == 0) {
					return true;
				}
			}
			return false;
		}
		else return false;
	}
	
	public List<ErrorInformation> getErrors() {
		return errors;
	}



	public void initialize(String filename, IFramework framework) {
		fFileName = filename;
		fFramework = framework;
		if (errors != null) {
			errors.clear();
		}
		//if (fParser == null || fLexer == null) {
		createLexerAndParser();
		//}
		if (fModelBuilder == null) {
			createModelBuilder();
		}
		// Clearing error lists
    	fParser.clearParseErrorList();
    	fModelResolver.clearAllReferenceErrors();
    	fTypeChecker.clearAllTypeErrors();
	}
	
//	public void terminate() {
//		fFileName = null;
//		fFramework = null;
//		fLexer = null;
//		fParser = null;
//		errors.clear();
//		errors.null;
//	}

	private void createModelBuilder() {
        if (fFramework != null) {
        	IModelSpace modelSpace = fFramework.getTopmodel();
        	// Retrieve native functions from framework
        	ASMNativeFunction[] asmNativeFunctions = fFramework.getNativeFunctionManager().getAllNativeFunctions();
        	// Creating model builder
        	fModelBuilder = new VTCLModelBuilder(modelSpace, asmNativeFunctions);
        	// Setting the file name for the model builder
        	fModelBuilder.setFileName(fFileName);
        	// Setting the model resolver
        	fModelResolver = fModelBuilder.getModelResolver();
        	fTypeChecker = fModelBuilder.getTypeChecker();
        }
	}
	
	private void createLexerAndParser() {
		String[] args = new String[]{"-d", fFileName};
		Option option;
		try {
            option = new Option(args);
            option.readInputChars();
            fLexer = new VTCLLexer(option.getInputChars(), option.getFileName()); // Create the lexer
            fLexer.setTab(4);
            fParser = new VTCLParser(fLexer);
			//fParser = new VTCLParser(fLexer.getLexStream());
            // Setting message handler (available from LPG 2.0)
			//IMessageHandler handler = new ErrorMessageCreator(fParser);
            //fParser.setMessageHandler(handler);
//            fNodeLocator = new AstLocator();
		} catch (IOException e) {
			throw new Error(e);
		}
	}

//	public ParserResult parseAndBuild(IProgressMonitor monitor) throws VPMRuntimeException
	public void parseAndBuild() throws VPMRuntimeException
	{
//		MyMonitor my_monitor= new MyMonitor(monitor);		
    	parse();
    	if (fAst != null && !hasErrors())
    	{
    		build();
            // Returns with ParserResult
            //return new ParserResult(errors, fVtclModule);
        }
		//return null;
	}

	/**
	 * build() should be never called without a previous successful call to parse()
	 */
	protected void build() {
		if (fAst != null)
		{
			// Clear reference errors
        	fModelResolver.clearAllReferenceErrors();
        	// Clear all type errors
        	fTypeChecker.clearAllTypeErrors();
			// Initiate model building
			Module vtclModule = (Module) fAst.accept(fModelBuilder);

			// Reporting model resolution and validation errors
			errors.addAll(fModelResolver.getAllReferenceErrorInfos());
			// Update the VTCL module if not null
			fVtclModule = vtclModule;
			if (!hasErrors()) {
				// Initiate type checking if all references are resolved
				fTypeChecker.doTypeChecking();
				// Reporting type errors
				errors.addAll(fTypeChecker.getAllTypeErrorInfos());
			}
			else {
				// Clear all elements scheduled for type checking in case of errors
				fTypeChecker.initializeTypeChecker();
			}
		}	
	}

	protected void parse() {
        try
        {
        	// Initialization (available from LPG 2.0)
//        	fLexer.initialize(fFileName);
//        	fParser.resetTokenStream();        	
        	// Clear all errors
        	fParser.clearParseErrorList();
    		if (errors != null) {
    			errors.clear();
    		}
        	
        	// Lex the stream to produce the token stream
            fLexer.lexer(fParser); 
            
            // Parse the token stream to produce an AST
//            ASTNode ast = fParser.parser(-1); // TODO: Needed for LPG 2.0 
            ASTNode ast = fParser.parser(100);
            
            // Check if there are parser errors 
            int errorIndex = fParser.getTokens().size() - 1;
            if (fParser.getFirstRealToken(errorIndex) != errorIndex) {
            	ParseTable prsTable = new VTCLParserprs();
    			//ParseTable prsTable = fParser.getParseTable();
    			DiagnoseParser diagnoseParser = new DiagnoseParser(fParser, prsTable );
    			diagnoseParser.diagnose();
            }

            // If the AST is valid, then save it
            if (ast != null) {
            	fAst = ast;
            }
            errors = fParser.getAllErrorInfos();
            // Initialize keywords on first run
        	if (fKeywords == null)
        	    initKeywords();
            // Return ParserResult (with existing VTCL Module) 
        }
        catch (Exception e)
        {
            System.err.println(e.getMessage());
            e.printStackTrace();
        }
		
	}
	
	public void parseOnly() throws VPMRuntimeException 
//	public ParserResult parseOnly(IProgressMonitor monitor) throws VPMRuntimeException 
	{
//		MyMonitor my_monitor= new MyMonitor(monitor);
    	parse();
        //return new ParserResult(errors, fVtclModule);
    	// Return no VTCLModule in case of parseOnly
    	//return new ParserResult(errors, null);
	}

    private void initKeywords() {
    	String tokenKindNames[]= fParser.getParseStream().orderedTerminalSymbols();

    	this.fIsKeyword= new boolean[tokenKindNames.length];
    	this.fKeywords= new char[tokenKindNames.length][];

    	int[] keywordKinds= fLexer.getKeywordKinds();

    	for(int i= 1; i < keywordKinds.length; i++) {
    	    int index= fParser.getParseStream().mapKind(keywordKinds[i]);
    	    fIsKeyword[index]= true;
    	    fKeywords[index]= fParser.getParseStream().orderedTerminalSymbols()[index].toCharArray();
    	}
        }

//    public void exportAST(String fileName) throws IOException {
//    	System.out.println(fileName);
//		File f = new File(fileName);
//		// f.createNewFile();
//		FileOutputStream stream = new FileOutputStream(f);
//		exportAST(stream);
//    }
//    
//    private void exportAST(FileOutputStream stream) {
//		PrintWriter out = new PrintWriter(stream);
//		out.println("<?xml version=\"1.0\" ?>");
//		out.println("<AST>");
//		if (fAst != null) {
//			visitAstNode(out, fAst);
//		}
//		out.println("</AST>");
//		out.flush();
//		out.close();
//		
//	}
//    
//    private void visitAstNode (PrintWriter out, ASTNode node) {
//    	out.println("<" + node.getClass().getSimpleName() + ">");
//    	out.print("<leftToken>");
//    	out.print(node.getLeftIToken().toString());
//    	out.println("</leftToken>");
//    	out.print("<rightToken>");
//    	out.print(node.getRightIToken().toString());
//    	out.println("</rightToken>");
//    	for (Iterator iter = node.getChildren().iterator(); iter.hasNext();) {
//			ASTNode child = (ASTNode) iter.next();
//			visitAstNode(out, child);
//		}
//    	out.println("</" + node.getClass().getSimpleName() + ">");
//    }

}
