/**
 * Copyright (c) 2007 - 2009OptXware Research and Development LLC.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *   Daniel Varro - Initial API and implementation
 */

package org.eclipse.viatra2.lpgparser.modelbuilder;

import java.util.ArrayList;
import java.util.EmptyStackException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.eclipse.viatra2.core.IEntity;
import org.eclipse.viatra2.core.IModelElement;
import org.eclipse.viatra2.core.IModelManager;
import org.eclipse.viatra2.core.IModelSpace;
import org.eclipse.viatra2.core.IRelation;
import org.eclipse.viatra2.core.simple.UniqueNameProvider;
import org.eclipse.viatra2.errors.info.ErrorInformation;
import org.eclipse.viatra2.errors.info.Location;
import org.eclipse.viatra2.errors.info.ErrorInformation.ErrorKind;
import org.eclipse.viatra2.errors.info.ErrorInformation.ErrorSeverity;
import org.eclipse.viatra2.framework.IFramework;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.compoundRules.BlockRule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.compoundRules.CompoundRule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.compoundRules.NestedRule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.core.AnnotatedElement;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.core.Annotation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.core.GTASMElement;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.ASMFunction;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.ImportDeclaration;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.InitialValue;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Machine;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Module;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Rule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.TypeConstant;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Variable;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.enums.ValueKind;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.simpleRules.ASMRuleInvocation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.simpleRules.CallRule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.simpleRules.ConditionalRuleIf;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.simpleRules.ConditionalRuleTry;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.simpleRules.GTRuleInvocation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.simpleRules.RuleUpdateASMFunction;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.ASMFunctionInvocation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.Constant;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.GTPatternCall;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.Term;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.VariableReference;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTPattern;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTPatternBody;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTRule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GtFactory;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.PatternVariable;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.PatternVariableAssignment;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.PatternVariableConstraint;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.modelmanagement.manipulationRules.creation.ElementCreateRule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.modelmanagement.manipulationRules.creation.EntityCreateRule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.modelmanagement.manipulationRules.creation.RelationCreateRule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.modelmanagement.queryFunctions.ElementReference;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.EditmodelFactory;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.Entity;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.ModelElement;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.Relation;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.Relationship;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.SupertypeOf;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.TypeOf;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.VPMElement;
import org.eclipse.viatra2.lpgparser.loader.VTCLMessages;
import org.eclipse.viatra2.lpgparser.typechecker.TypeResolver;
import org.eclipse.viatra2.lpgparser.typechecker.VTCLTypeNameConstants;
import org.eclipse.viatra2.natives.ASMNativeFunction;

/**
 * This class is responsible for resolving references when parsing a VTCL file.
 * Some of the methods are also accessed by TransformationContent.java when
 * reloading transformations to model spaces
 * @author Daniel Varro
 */
public class VTCLModelResolver {

	private static final String PARSER_MARKER = "org.eclipse.viatra2.loaders.vtclparsermarker";

	public VTCLModelResolver() {
		modelSpace = null;
		modelManager = null;
		asmNativeFunctions = null;
		typeResolver = new TypeResolver();
		uniqueNameProvider = null;
	}

	public VTCLModelResolver(IModelSpace aModelSpace,
			ASMNativeFunction[] allAsmNativeFunctions) {
		modelSpace = aModelSpace;
		modelManager = modelSpace.getModelManager();
		asmNativeFunctions = allAsmNativeFunctions;
		typeResolver = new TypeResolver();
		uniqueNameProvider = new UniqueNameProvider();
	}

	private ASMNativeFunction[] asmNativeFunctions;
	private IModelSpace modelSpace;
	private IModelManager modelManager;
	private TypeResolver typeResolver;
	// private TypeAssignments typeAssignments;
	private UniqueNameProvider uniqueNameProvider;

	public IModelSpace getModelSpace() {
		return modelSpace;
	}

	public IModelManager getModelManager() {
		return modelManager;
	}

	public ASMNativeFunction[] getNativeFunctions() {
		return asmNativeFunctions;
	}

	private Machine storedMachine;

	/**
	 * Returns the machine that is stored inside the model resolver for
	 * resolution.
	 * @return the machine to be resolved.
	 */
	public Machine getMachine() {
		return storedMachine;
	}

	/**
	 * Sets the {@link Machine} whose elements have to be resolved for the model
	 * resolver.
	 * @param machine
	 *            the machine to be resolved.
	 */
	public void setMachine(Machine machine) {
		this.storedMachine = machine;
	}

	/**
	 * Clears all buffers in the model resolver. It is the callers
	 * responsibility to call this method before the model elements are
	 * resolved.
	 */
	public void clearBuffers() {
		unresolvedPatternCalls.clear();
		unresolvedAsmFunctionInvocations.clear();
		unresolvedAsmFunctionUpdates.clear();
		unresolvedAsmRuleCalls.clear();
		unresolvedGTRuleInvoc.clear();
	}
	
	/**
	 * Resolves the elements of the previously loaded machine. It is the callers
	 * responsibility to call both {@link VTCLModelResolver#clearBuffers()} and
	 * {@link VTCLModelResolver#setMachine(Machine)} before calling this method.
	 */
	public void resolveModel() {
		resolvePatternCalls(storedMachine);
		resolveAsmFunInvocations(storedMachine);
		resolveAsmFunUpdates(storedMachine);
		resolveAsmRulesInvocations(storedMachine);
		resolveGTRuleInvocations(storedMachine);
		validateDistinctVariables(storedMachine);
		validateNoDuplicateNames(storedMachine);
	}

	private List<ErrorInformation> referenceResolutionErrorInfos = new ArrayList<ErrorInformation>();

	public List<ErrorInformation> getAllReferenceErrorInfos() {
		return referenceResolutionErrorInfos;
	}

	public void clearAllReferenceErrors() {
		referenceResolutionErrorInfos.clear();
	}

	private Map<GTPatternBody, List<VPMElement>> orphanBodyElements = new HashMap<GTPatternBody, List<VPMElement>>();

	public List<VPMElement> getOrphanBodyElements(GTPatternBody body) {
		if (orphanBodyElements.get(body) == null) {
			List<VPMElement> list = new ArrayList<VPMElement>();
			orphanBodyElements.put(body, list);
			return list;
		} else
			return orphanBodyElements.get(body);
	}

	private Stack<List<VariableReference>> unresolvedVariableRefsStack = new Stack<List<VariableReference>>();

	// public List<VariableReference> getTopUnresolvedVariableRefs() {
	// if (unresolvedVariableRefsStack.peek() == null) {
	// return new ArrayList<VariableReference>();
	// }
	// else
	// return unresolvedVariableRefsStack.peek();
	// }
	public Stack<List<VariableReference>> getUnresolvedVariableRefsStack() {
		return unresolvedVariableRefsStack;
	}

	public void addVariableReferenceToStack(VariableReference varRef) {
		// added by Istvan
		try {
			unresolvedVariableRefsStack.peek();
		} catch (EmptyStackException e) {
			unresolvedVariableRefsStack
					.push(new ArrayList<VariableReference>());
		}
		// end added section
		unresolvedVariableRefsStack.peek().add(varRef);
	}

	private List<GTPatternCall> unresolvedPatternCalls = new ArrayList<GTPatternCall>();

	public List<GTPatternCall> getUnresolvedPatternCalls() {
		return unresolvedPatternCalls;
	}

	private List<RuleUpdateASMFunction> unresolvedAsmFunctionUpdates = new ArrayList<RuleUpdateASMFunction>();

	public List<RuleUpdateASMFunction> getUnresolvedAsmFunctionUpdates() {
		return unresolvedAsmFunctionUpdates;
	}

	private List<ASMFunctionInvocation> unresolvedAsmFunctionInvocations = new ArrayList<ASMFunctionInvocation>();

	public List<ASMFunctionInvocation> getUnresolvedAsmFunctionInvocations() {
		return unresolvedAsmFunctionInvocations;
	}

	private List<CallRule> unresolvedAsmRuleCalls = new ArrayList<CallRule>();

	public List<CallRule> getUnresolvedAsmRuleCalls() {
		return unresolvedAsmRuleCalls;
	}

	private List<GTRuleInvocation> unresolvedGTRuleInvoc = new ArrayList<GTRuleInvocation>();

	public List<GTRuleInvocation> getUnresolvedGTRuleInvoc() {
		return unresolvedGTRuleInvoc;
	}

	private Map<GTPatternBody, List<VariableReference>> unresolvedPatternVariableRefs = new HashMap<GTPatternBody, List<VariableReference>>();

	public List<VariableReference> getUnresolvedPatternVariableRefs(
			GTPatternBody body) {
		if (unresolvedPatternVariableRefs.get(body) == null) {
			List<VariableReference> list = new ArrayList<VariableReference>();
			unresolvedPatternVariableRefs.put(body, list);
			return list;
		} else
			return unresolvedPatternVariableRefs.get(body);
	}

	private Map<GTPatternBody, Map<PatternVariable, ModelElement>> bodyMapElement2Var = new HashMap<GTPatternBody, Map<PatternVariable, ModelElement>>();

	public Map<PatternVariable, ModelElement> getBodyMapElement4Var(
			GTPatternBody body) {
		if (bodyMapElement2Var.get(body) == null) {
			HashMap<PatternVariable, ModelElement> map = new HashMap<PatternVariable, ModelElement>();
			bodyMapElement2Var.put(body, map);
			return map;
		} else
			return bodyMapElement2Var.get(body);
	}

	private Map<GTPatternBody, Map<ModelElement, PatternVariable>> bodyMapVar2Element = new HashMap<GTPatternBody, Map<ModelElement, PatternVariable>>();

	public Map<ModelElement, PatternVariable> getBodyMapVar4Element(
			GTPatternBody body) {
		if (bodyMapVar2Element.get(body) == null) {
			HashMap<ModelElement, PatternVariable> map = new HashMap<ModelElement, PatternVariable>();
			bodyMapVar2Element.put(body, map);
			return map;
		} else
			return bodyMapVar2Element.get(body);
	}

	private Map<ASMRuleInvocation, GTRule> actionRuleMap = new HashMap<ASMRuleInvocation, GTRule>();

	public Map<ASMRuleInvocation, GTRule> getActionRuleMap() {
		return actionRuleMap;
	}

	public void resolveNamespaceImport(ImportDeclaration declaration) {
		if (modelManager.getElementByName(declaration.getImportValue()) == null) {
			String[] context = new String[] { declaration.getImportValue() };
			addReferenceResolutionError(declaration.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_IMPORT_UNRESOLVABLE,
					ErrorSeverity.ERROR, context);
		}
	}

	/**
	 * Tries to resolve a VariableReference to one of the local Variable of the
	 * GTRule. If no such Variable is found, then a new Variable is created.
	 * @param varRef
	 *            : a VariableReference
	 * @param rule
	 *            : a GTRule
	 * @return varDef : a Variable
	 */
	public Variable resolveLocalVariableInGTRule(VariableReference varRef,
			GTRule rule) {
		Variable varDef = lookupVariableDefByName(varRef.getName(), rule);
		if (varDef != null) {
			varRef.setVariable(varDef);
		}
		// If a corresponding Variable definition is not found
		else {
			varDef = createPatternVariableFromReference(varRef, rule
					.getLocalVariables());
			Annotation annotation = lookupAnnotation(
					IAnnotationKeys.NODE_INFORMATION, varRef.getAnnotations());
			if (annotation != null) {
				varDef.getAnnotations().add(annotation);
			}
		}
		return varDef;
	}

	/**
	 * Tries to resolve a VariableReference to one of the local PatternVariable
	 * of the Pattern. If no such PatternVariable is found, then a new
	 * PatternVariable is created.
	 * @param varRef
	 *            : a {@link VariableReference}
	 * @param body
	 *            : the container pattern body ({@link GTPatternBody})
	 * @return varDef : a PatternVariable
	 */
	public PatternVariable resolveLocalVariableInPattern(
			VariableReference varRef, GTPatternBody body) {
		PatternVariable varDef = lookupVariableDefByName(varRef.getName(), body);
		if (varDef != null) {
			varRef.setVariable(varDef);
		}
		// If a corresponding PatternVariable definition is not found
		else {
			varDef = createPatternVariableFromReference(varRef, body
					.getLocalVariables());
			Annotation annotation = lookupAnnotation(
					IAnnotationKeys.NODE_INFORMATION, varRef.getAnnotations());
			if (annotation != null) {
				varDef.getAnnotations().add(annotation);
			}
		}
		return varDef;
	}

	/**
	 * Tries to resolve unresolved VariableReferences to existing
	 * PatternVariables of the GTPatternBody.
	 * @param body
	 *            : a GTPatternBody
	 */
	public void resolvePatternVariableRefs(GTPatternBody body) {
		List<VariableReference> list = getUnresolvedPatternVariableRefs(body);
		for (VariableReference varRef : list) {
			PatternVariable varDef = lookupVariableDefByName(varRef.getName(),
					body);
			if (varDef != null) {
				varRef.setVariable(varDef);
			}
			// Report error if a corresponding PatternVariable cannot be found
			else {
				String[] context = new String[] { varRef.getName() };
				addReferenceResolutionError(
						varRef.getAnnotations(),
						VTCLMessages.VTCLValidationErrorCodes_VARIABLE_REF_UNRESOLVABLE,
						ErrorSeverity.ERROR, context);
			}

		}
	}

	@SuppressWarnings("unchecked")
	public PatternVariable createPatternVariableFromReference(
			VariableReference varRef, List list) {
		PatternVariable varDef = GtFactory.eINSTANCE.createPatternVariable();
		varRef.setVariable(varDef);
		// Generate a unique variable name for the new PatternVariable
		if (varRef.getName().charAt(0) == '_') {
			varDef.setName(uniqueNameProvider.uniqueName("Var"));
		} else {
			varDef.setName(varRef.getName());
		}
		list.add(varDef);
		return varDef;
	}

	public void resolveModelElementReferences(GTPatternBody body) {
		List<VPMElement> orphanElements = getOrphanBodyElements(body);
		for (Object object : orphanElements) {
			if (object instanceof Relation) {
				Relation relation = (Relation) object;
				resolveRelationReferences(relation, body);
			} else if (object instanceof SupertypeOf) {
				SupertypeOf relship = (SupertypeOf) object;
				resolveRelationshipReferences(relship, body);
			} else if (object instanceof TypeOf) {
				TypeOf relship = (TypeOf) object;
				resolveRelationshipReferences(relship, body);
			}

		}
	}

	/**
	 * Resolves the source and target of a {@link Relation} if possible by using
	 * the model elements defined in the {@link GTPatternBody}. If no
	 * corresponding model element is found then we check if the corresponding
	 * VariableReference exists as a pattern parameter or in a pattern
	 * composition, thus a relation will be partially resolved in such a case.
	 * @param relation
	 *            : {@link Relation} to be resolved
	 * @param body
	 *            : {@link GTPatternBody}
	 */
	private void resolveRelationReferences(Relation relation, GTPatternBody body) {
		// Seek for the source model element
		PatternVariable srcVar = lookupVariableDefByName(relation.getFromStr(),
				body);
		if (srcVar != null && bodyMapElement2Var.get(body).get(srcVar) != null) {
			// Lookup the variable in the currently processed Body
			ModelElement srcElem = bodyMapElement2Var.get(body).get(srcVar);
			// Setting the source
			relation.setFrom(srcElem);
			// Setting the PatternVariable of the VariableReference
			VariableReference srcRef = relation.getVariableReferences().get(1);
			srcRef.setVariable(srcVar);
		}
		// If there is a PatternVariable, but no corresponding model element ->
		// warning
		else if (srcVar != null
				&& bodyMapElement2Var.get(body).get(srcVar) == null) {
			// If the source cannot be resolved locally, then add the relation
			// to the ElementWrappers of the body
			body.getDanglingRelations().add(relation);
			addReferenceResolutionError(
					lookupVariableRefByName(relation.getFromStr(), relation)
							.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_SOURCE_LOCALLY_UNRESOLVABLE,
					ErrorSeverity.WARNING, new String[] {}); // empty context
			// Add the relation to the elementWrappers buffer to preserve it in
			// the EMF model
		}
		// If no corresponding PatternVariable exists -> error
		else {
			// addReferenceResolutionError( relation.getAnnotations(),
			addReferenceResolutionError(lookupVariableRefByName(
					relation.getFromStr(), relation).getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_SOURCE_UNRESOLVABLE,
					ErrorSeverity.ERROR, new String[] {}); // empty context
		}

		// Seek for the target model element
		PatternVariable trgVar = lookupVariableDefByName(relation.getToStr(),
				body);
		if (trgVar != null && bodyMapElement2Var.get(body).get(trgVar) != null) {
			// Assume that the last element is the current element processed
			ModelElement trgElem = bodyMapElement2Var.get(body).get(trgVar);
			// Setting the target
			relation.setTo(trgElem);
			// Setting the PatternVariable of the VariableReference
			VariableReference trgRef = relation.getVariableReferences().get(2);
			trgRef.setVariable(trgVar);
		}
		// If there is a PatternVariable, but no corresponding model element ->
		// warning
		else if (trgVar != null
				&& bodyMapElement2Var.get(body).get(trgVar) == null) {
			addReferenceResolutionError(
					lookupVariableRefByName(relation.getToStr(), relation)
							.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_TARGET_LOCALLY_UNRESOLVABLE,
					ErrorSeverity.WARNING, new String[] {}); // empty context
		}
		// If no corresponding PatternVariable exists -> error
		else {
			// addReferenceResolutionError( relation.getAnnotations(),
			addReferenceResolutionError(lookupVariableRefByName(
					relation.getToStr(), relation).getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_TARGET_UNRESOLVABLE,
					ErrorSeverity.ERROR, new String[] {}); // empty context
		}

	}

	private void resolveRelationshipReferences(Relationship relship,
			GTPatternBody body) {
		// Seek for the source model element
		PatternVariable clientVar = lookupVariableDefByName(relship
				.getClientStr(), body);
		if (clientVar != null
				&& bodyMapElement2Var.get(body).get(clientVar) != null) {
			// Lookup the variable
			ModelElement clientElem = bodyMapElement2Var.get(body).get(
					clientVar);
			// Setting the source
			relship.setClient(clientElem);
			// Setting the PatternVariable of the VariableReference
			VariableReference clientRef = relship.getVariableReferences()
					.get(1);
			clientRef.setVariable(clientVar);
		}
		// If there is a PatternVariable, but no corresponding model element
		else if (clientVar != null
				&& bodyMapElement2Var.get(body).get(clientVar) == null) {
			// If the client cannot be resolved locally, then add the
			// relationship
			// as to the list of dangling relationship in order to get
			// serialized
			body.getDanglingRelationships().add(relship);
			addReferenceResolutionError(
					lookupVariableRefByName(relship.getClientStr(), relship)
							.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_CLIENT_LOCALLY_UNRESOLVABLE,
					ErrorSeverity.WARNING, new String[] {}); // empty context
		}
		// If no corresponding PatternVariable exists, this is an error
		else {
			// addReferenceResolutionError( relship.getAnnotations(),
			addReferenceResolutionError(lookupVariableRefByName(
					relship.getClientStr(), relship).getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_CLIENT_UNRESOLVABLE,
					ErrorSeverity.ERROR, new String[] {}); // empty context
		}

		// Seek for the source model element
		PatternVariable supplierVar = lookupVariableDefByName(relship
				.getSupplierStr(), body);
		if (supplierVar != null
				&& bodyMapElement2Var.get(body).get(supplierVar) != null) {
			// Lookup the variable
			ModelElement supplierElem = bodyMapElement2Var.get(body).get(
					supplierVar);
			// Setting the target
			relship.setSupplier(supplierElem);
			// Setting the PatternVariable of the VariableReference
			VariableReference supplierRef = relship.getVariableReferences()
					.get(0);
			supplierRef.setVariable(supplierVar);
		}
		// If there is a PatternVariable, but no corresponding model element
		else if (supplierVar != null
				&& bodyMapElement2Var.get(body).get(supplierVar) == null) {
			addReferenceResolutionError(
					lookupVariableRefByName(relship.getSupplierStr(), relship)
							.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_SUPPLIER_LOCALLY_UNRESOLVABLE,
					ErrorSeverity.WARNING, new String[] {}); // empty context
		}
		// If no corresponding PatternVariable exists, this is an error
		else {
			addReferenceResolutionError(
					lookupVariableRefByName(relship.getSupplierStr(), relship)
							.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_SUPPLIER_UNRESOLVABLE,
					ErrorSeverity.ERROR, new String[] {}); // empty context
		}
	}

	/**
	 * Creates an ErrorInformation
	 * @param annotations
	 *            : location information of the error
	 * @param errorString
	 *            : textual representation of the error (VTCLMessages should be
	 *            used)
	 * @param errorSeverity
	 *            : severity of the error
	 * @param context
	 *            : list of contextual information
	 */
	public void addReferenceResolutionError(List<?> annotations,
			String errorString, ErrorSeverity errorSeverity, String[] context) {
		// Report reference resolution error
		Annotation locInfo = lookupAnnotation(IAnnotationKeys.NODE_INFORMATION,
				annotations);
		// Handle empty annotations
		Location location = new Location((locInfo != null) ? locInfo.getValue()
				: "0,0,0,0");
		ErrorInformation errorInfo = new ErrorInformation(errorString,
				PARSER_MARKER, ErrorKind.VALIDATION_ERROR, location,
				errorSeverity);
		// Bind contextual parameters of error strings
		errorInfo.bind(context);
		referenceResolutionErrorInfos.add(errorInfo);
	}

	public void storeVarDefForModelElement(GTPatternBody body,
			PatternVariable varDef, ModelElement me) {
		if (getBodyMapElement4Var(body).get(varDef) == null) {
			getBodyMapElement4Var(body).put(varDef, me);
		}
		// Add an error if a variable is defined multiple times
		else {
			String[] context = new String[] { varDef.getName() };
			addReferenceResolutionError(
					varDef.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_DUPLICATE_PATTERNVAR_DEFINITION,
					ErrorSeverity.ERROR, context);
		}
		// Add the inverse reference between model element and its variable
		getBodyMapVar4Element(body).put(me, varDef);
	}

	public GTPattern reportGtPatternCallResolutionError(GTPatternCall call,
			GTPattern patternDef, IFramework fw) {
		String nameStr = call.getFqn();
		int callArity = call.getActualParameters().size();
		// If not found locally, initiate lookup in other machines
		if (patternDef == null) {
			int lastDotIndex = nameStr.lastIndexOf(".");
			// If ruleName is a fully qualified name which can be separated to
			// machine+rule
			if (lastDotIndex != -1) {
				String machineStr = nameStr.substring(0, lastDotIndex);
				// TODO: This code should be rewritten to accommodate the new
				// MachineWrapper
				Machine extMachine = (Machine) fw.getMachineByFQN(machineStr);
				if (extMachine != null) {
					patternDef = lookupGTPattern(nameStr, callArity, extMachine);
				}
			}
		}
		// Set the GTPatternCall correspondingly
		if (patternDef != null) {
			if (patternDef.getSymParameters().size() == call
					.getActualParameters().size()) {
				call.setCalledPattern(patternDef);
			}
			// Report if the number of actual parameters do not match with the
			// number of formal parameters
			else {
				String arity = String.valueOf(patternDef.getSymParameters()
						.size());
				String[] context = new String[] { patternDef.getFqn() + "/"
						+ arity };
				addReferenceResolutionError(
						call.getAnnotations(),
						VTCLMessages.VTCLValidationErrorCodes_PATTERN_PARAMETER_MISMATCH,
						ErrorSeverity.ERROR, context);
			}
		}
		// Otherwise report an error that pattern call cannot be resolved
		else {
			String arity = String.valueOf(call.getActualParameters().size());
			String[] context = new String[] { call.getFqn() + "/" + arity };
			addReferenceResolutionError(
					call.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_PATTERN_CALL_UNRESOLVABLE,
					ErrorSeverity.ERROR, context);
		}
		return patternDef;
	}

	public void resolvePatternCalls(Machine machine) {
		List<GTPatternCall> list = getUnresolvedPatternCalls();
		for (GTPatternCall call : list) {
			String nameStr = call.getFqn();
			int arity = call.getActualParameters().size();
			// Seek for GTPattern definition locally
			GTPattern patternDef = lookupGTPattern(nameStr, arity, machine);
			reportGtPatternCallResolutionError(call, patternDef, modelSpace
					.getFramework());
		}
	}

	public GTRule reportGtRuleResolutionError(GTRuleInvocation gtInvoc,
			GTRule gtRuleDef, IFramework fw) {
		String nameStr = gtInvoc.getFqn();
		int callArity = gtInvoc.getActualParameters().size();
		// If not found locally, initiate lookup in other machines
		if (gtRuleDef == null) {
			int lastDotIndex = nameStr.lastIndexOf(".");
			// If ruleName is a fully qualified name which can be separated to
			// machine+rule
			if (lastDotIndex != -1) {
				String machineStr = nameStr.substring(0, lastDotIndex);
				// TODO: This code should be rewritten to accommodate the new
				// MachineWrapper
				Machine extMachine = (Machine) fw.getMachineByFQN(machineStr);
				if (extMachine != null) {
					gtRuleDef = lookupGTRule(nameStr, callArity, extMachine);
				}
			}
		}
		if (gtRuleDef != null) {
			// Check if parameters are matching
			if (gtRuleDef.getSymParameters().size() == gtInvoc
					.getActualParameters().size()) {
				gtInvoc.setRule(gtRuleDef);
			} else {
				String arity = String.valueOf(gtRuleDef.getSymParameters()
						.size());
				String[] context = new String[] { gtRuleDef.getName() + "/"
						+ arity };
				addReferenceResolutionError(
						gtInvoc.getAnnotations(),
						VTCLMessages.VTCLValidationErrorCodes_GTRULE_PARAMETER_MISMATCH,
						ErrorSeverity.ERROR, context);
			}
		}
		// Otherwise report an error
		else {
			String arity = String.valueOf(gtInvoc.getActualParameters().size());
			String[] context = new String[] { gtInvoc.getName() + "/" + arity };
			addReferenceResolutionError(
					gtInvoc.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_GTRULE_CALL_UNRESOLVABLE,
					ErrorSeverity.ERROR, context);
		}
		return gtRuleDef;
	}

	public void resolveGTRuleInvocations(Machine machine) {
		List<GTRuleInvocation> list = getUnresolvedGTRuleInvoc();
		for (GTRuleInvocation gtInvoc : list) {
			String nameStr = gtInvoc.getFqn();
			int arity = gtInvoc.getActualParameters().size();
			// Seek for GTPattern definition locally
			GTRule gtRuleDef = lookupGTRule(nameStr, arity, machine);
			reportGtRuleResolutionError(gtInvoc, gtRuleDef, modelSpace
					.getFramework());
		}
	}

	public ASMFunction reportFunInvocResolutionError(
			ASMFunctionInvocation funInvoc, ASMFunction asmFunDef, IFramework fw) {
		String nameStr = funInvoc.getFqn();
		int callArity = funInvoc.getActualParameters().size();
		// If not found locally, initiate lookup in other machines
		if (asmFunDef == null) {
			int lastDotIndex = nameStr.lastIndexOf(".");
			// If ruleName is a fully qualified name we separate the machine
			// string
			if (lastDotIndex != -1) {
				String machineStr = nameStr.substring(0, lastDotIndex);
				Machine extMachine = (Machine) fw.getMachineByFQN(machineStr);
				// TODO: This code should be rewritten to accommodate the new
				// MachineWrapper
				if (extMachine != null) {
					asmFunDef = lookupAsmFunction(nameStr, callArity,
							extMachine, true);
				}
			}
		}
		// If found, set the ASMFunctionInvocation correspondingly
		if (asmFunDef != null) {
			// Check the number of parameters
			if (asmFunDef.getArity() == funInvoc.getActualParameters().size()) {
				funInvoc.setCalledFunction(asmFunDef);
			}
			// Check that the number of actual parameters match with formal
			// definition
			else {
				String arity = String.valueOf(asmFunDef.getArity());
				String[] context = new String[] { asmFunDef.getName() + "/"
						+ arity };
				addReferenceResolutionError(
						funInvoc.getAnnotations(),
						VTCLMessages.VTCLValidationErrorCodes_ASMFUN_PARAMETER_MISMATCH,
						ErrorSeverity.ERROR, context);
			}
		}
		// Otherwise report an error
		else {
			String arity = String
					.valueOf(funInvoc.getActualParameters().size());
			String[] context = new String[] { funInvoc.getName() + "/" + arity };
			addReferenceResolutionError(
					funInvoc.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_ASMFUN_INVOC_UNRESOLVABLE,
					ErrorSeverity.ERROR, context);
		}
		return asmFunDef;
	}

	public void resolveAsmFunInvocations(Machine machine) {
		List<ASMFunctionInvocation> list = getUnresolvedAsmFunctionInvocations();
		for (ASMFunctionInvocation funInvoc : list) {
			String nameStr = funInvoc.getFqn();
			int arity = funInvoc.getActualParameters().size();
			// Lookup with conditions for right values (i.e. RHS of an
			// assignment)
			ASMFunction asmFunDef = lookupAsmFunction(nameStr, arity, machine,
					true);
			reportFunInvocResolutionError(funInvoc, asmFunDef, modelSpace
					.getFramework());
		}
	}

	public ASMFunction reportAsmFunUpdateResolutionError(
			RuleUpdateASMFunction updateRule, ASMFunction asmFunDef,
			IFramework fw) {
		String nameStr = updateRule.getFqn();
		int callArity = updateRule.getLocations().size();
		// If not found locally, initiate lookup in other machines
		if (asmFunDef == null) {
			int lastDotIndex = nameStr.lastIndexOf(".");
			// If ruleName is a fully qualified name we separate the machine
			// string
			if (lastDotIndex != -1) {
				String machineStr = nameStr.substring(0, lastDotIndex);
				// TODO: This code should be rewritten to accommodate the new
				// MachineWrapper
				Machine extMachine = (Machine) fw.getMachineByFQN(machineStr);
				if (extMachine != null) {
					asmFunDef = lookupAsmFunction(nameStr, callArity,
							extMachine, false);
				}
			}
		}
		// If found, set the ASMFunctionInvocation correspondingly
		if (asmFunDef != null) {
			if (asmFunDef.getArity() == updateRule.getLocations().size()) {
				updateRule.setFunction(asmFunDef);
			} else {
				String arity = String.valueOf(asmFunDef.getArity());
				String[] context = new String[] { asmFunDef.getName() + "/"
						+ arity };
				addReferenceResolutionError(
						updateRule.getAnnotations(),
						VTCLMessages.VTCLValidationErrorCodes_ASMFUN_PARAMETER_MISMATCH,
						ErrorSeverity.ERROR, context);
			}
		}
		// Otherwise report an error
		else {
			String arity = String.valueOf(updateRule.getLocations().size());
			String[] context = new String[] { updateRule.getFqn() + "/" + arity };
			addReferenceResolutionError(
					updateRule.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_ASMFUN_UPDATE_UNRESOLVABLE,
					ErrorSeverity.ERROR, context);
		}
		return asmFunDef;
	}

	public void resolveAsmFunUpdates(Machine machine) {
		List<RuleUpdateASMFunction> list = getUnresolvedAsmFunctionUpdates();
		for (RuleUpdateASMFunction updateRule : list) {
			String nameStr = updateRule.getFqn();
			int arity = updateRule.getLocations().size();
			// Lookup with conditions for right values (i.e. RHS of an
			// assignment)
			ASMFunction asmFunDef = lookupAsmFunction(nameStr, arity, machine,
					false);
			reportAsmFunUpdateResolutionError(updateRule, asmFunDef, modelSpace
					.getFramework());
		}
	}

	public Rule reportAsmRuleInvocResolutionError(CallRule ruleInvoc,
			Rule ruleDef, IFramework fw) {
		String ruleName = ruleInvoc.getFqn();
		int callArity = ruleInvoc.getActualParameters().size();
		// If not found locally, initiate lookup in other machines
		if (ruleDef == null) {
			int lastDotIndex = ruleName.lastIndexOf(".");
			// If ruleName is a fully qualified name which can be separated to
			// machine+rule
			if (lastDotIndex != -1) {
				String machineStr = ruleName.substring(0, lastDotIndex);
				// TODO: This code should be rewritten to accommodate the new
				// MachineWrapper
				Machine extMachine = (Machine) fw.getMachineByFQN(machineStr);
				if (extMachine != null) {
					ruleDef = lookupAsmRule(ruleName, callArity, extMachine);
				}
			}
		}
		// If found, set the ASMFunctionInvocation correspondingly
		if (ruleDef != null) {
			if (ruleDef.getSymParameters().size() == ruleInvoc
					.getActualParameters().size()) {
				ruleInvoc.setRule(ruleDef);
			} else {
				String arity = String
						.valueOf(ruleDef.getSymParameters().size());
				String[] context = new String[] { ruleDef.getName() + "/"
						+ arity };
				addReferenceResolutionError(
						ruleInvoc.getAnnotations(),
						VTCLMessages.VTCLValidationErrorCodes_ASMRULE_PARAMETER_MISMATCH,
						ErrorSeverity.ERROR, context);
			}
		}
		// Otherwise report an error
		else {
			String arity = String.valueOf(ruleInvoc.getActualParameters()
					.size());
			String[] context = new String[] { ruleName + "/" + arity };
			addReferenceResolutionError(
					ruleInvoc.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_ASMRULE_INVOC_UNRESOLVABLE,
					ErrorSeverity.ERROR, context);
		}
		return ruleDef;
	}

	public void resolveAsmRulesInvocations(Machine machine) {
		List<CallRule> list = getUnresolvedAsmRuleCalls();
		for (CallRule ruleInvoc : list) {
			String ruleName = ruleInvoc.getFqn();
			int arity = ruleInvoc.getActualParameters().size();
			// Lookup with conditions for Left values (i.e. LHS of an
			// assignment)
			Rule ruleDef = lookupAsmRule(ruleName, arity, machine);
			reportAsmRuleInvocResolutionError(ruleInvoc, ruleDef, modelSpace
					.getFramework());
		}
	}

	public void resolveVariableRefsInRule(GTASMElement defScope) {
		List<VariableReference> list = unresolvedVariableRefsStack.peek();
		for (VariableReference varRef : list) {
			Variable varDef = null;
			// If resolution is initiated from an ASMRuleInvocation
			if (defScope instanceof ASMRuleInvocation) {
				ASMRuleInvocation ruleInvoc = (ASMRuleInvocation) defScope;
				// Lookup VariableDefinition in the ASM Rule
				varDef = traverseCallHierarchy(varRef.getName(), ruleInvoc);
			}
			// If resolution is initiated from a GTPatternBody
			else if (defScope instanceof GTPatternBody) {
				GTPatternBody body = (GTPatternBody) defScope;
				varDef = lookupVariableDefByName(varRef.getName(), body);
			}
			{

			}
			// If found, set the variable reference appropriately
			if (varDef != null) {
				varRef.setVariable(varDef);
			}
			// Otherwise report an error
			else {
				String[] context = new String[] { varRef.getName() };
				addReferenceResolutionError(
						varRef.getAnnotations(),
						VTCLMessages.VTCLValidationErrorCodes_VARIABLE_REF_UNRESOLVABLE,
						ErrorSeverity.ERROR, context);
			}
		}
	}

	/*
	 * public ModelElement resolveElementWrapper(String typeStr, GTPatternBody
	 * body) { boolean found = false; ModelElement modelElement = null; Iterator
	 * iter = body.getElementWrappers().iterator(); while (!found &&
	 * iter.hasNext()) { modelElement = (ModelElement) iter.next(); } return
	 * null; }
	 */

	// public void typeChecking(Machine machine) {
	// // Processing each GTPattern defined in the machine
	// for (Iterator iter = machine.getGtPatternDefinitions().iterator();
	// iter.hasNext();) {
	// GTPattern pattern = (GTPattern) iter.next();
	// for (Iterator iter2 = pattern.getPatternBodies().iterator();
	// iter2.hasNext();) {
	// GTPatternBody body = (GTPatternBody) iter2.next();

	// }

	// }
	// }

	public void resolveEntityTypesInBody(GTPatternBody body) {
		Entity top = body.getPatternGraph();
		for (Iterator<?> iter = top.getComponents().iterator(); iter.hasNext();) {
			Entity entity = (Entity) iter.next();
			List<String> fqnList = lookupEntityType(getMachine(), entity
					.getTypeStr());
			resolveTypeNames(entity, fqnList, body);
		}
	}

	// protected Machine lookupMachine(GTPatternBody body) {
	// GTPattern pattern= body.getHeader();
	// Machine machine = pattern.getNamespace();
	// if (machine != null) {
	// return machine;
	// }
	// // If pattern is part of a GTRule
	// else if (pattern.eContainer() instanceof GTRule) {
	// GTRule gtRule = (GTRule) pattern.eContainer();
	// return gtRule.getNamespace();
	// }
	// // If pattern is defined as a negative pattern of another pattern body
	// else if (pattern.eContainer() instanceof GTPatternBody) {
	// GTPatternBody parentBody = (GTPatternBody) pattern.eContainer();
	// return lookupMachine(parentBody);
	// }
	// return null;
	// }

	public void resolveRelationTypesInBody(GTPatternBody body) {
		// Resolve the types of dangling relations
		for (Iterator<?> iter = body.getDanglingRelations().iterator(); iter
				.hasNext();) {
			Relation relation = (Relation) iter.next();
			// Resolving type strings
			List<String> fqnList = lookupRelationType(getMachine(), relation
					.getTypeStr());
			resolveTypeNames(relation, fqnList, body);
			// Type checking for the end of relations
			checkRelationEndTypes(relation);
		}

		// Resolve types of relations leading out from an entity
		Entity top = body.getPatternGraph();
		for (Iterator<?> iter = top.getComponents().iterator(); iter.hasNext();) {
			Entity entity = (Entity) iter.next();
			checkRelationsFrom(body, entity);
		}
	}

	private void checkRelationsFrom(GTPatternBody body, ModelElement element) {
		for (Iterator<?> iter2 = element.getRelationsFrom().iterator(); iter2
				.hasNext();) {
			Relation relation = (Relation) iter2.next();
			// Resolving type names
			List<String> fqnList2 = lookupRelationType(getMachine(), relation
					.getTypeStr());
			resolveTypeNames(relation, fqnList2, body);
			// Type checking for the end of relations
			checkRelationEndTypes(relation);
			checkRelationsFrom(body, relation);
		}
	}

	private class TypeInfo {
		boolean isValid;
		IModelElement type;

		public TypeInfo(boolean isValid, IModelElement type) {
			this.isValid = isValid;
			this.type = type;
		}
	}

	/**
	 * Calculates the type of a model element while fulfilling a type equality
	 * constraint.
	 * @param element
	 * @param expectedType
	 * @return
	 */
	private TypeInfo calculateVariableType(ModelElement element,
			IModelElement expectedType) {
		boolean found = false;
		Iterator<?> it = element.getType().iterator();
		IModelElement actualType = null;
		while (!found && it.hasNext()) {
			ModelElement typeSrc = (ModelElement) it.next();
			// Seeking for the actual VPM representation of the type
			actualType = modelManager
					.getElementByName(typeSrc.getRealElement());
			// If the source of the relation type is a supertype of the expected
			// vpmTypeOfSrcElementInPattern
			// -> proper type found
			if (actualType != null) {
				if ((expectedType == actualType || expectedType
						.isSupertypeOf(actualType))) {
					found = true;
				}
			}
		}
		return new TypeInfo(found, actualType);
	}

	/**
	 * Checking that both ends of a relation are appropriately typed The source
	 * (target) end of the type of a relation should be a supertype or exactly
	 * the same type as the type of the source model element. Formally, if
	 * type() denote the type of a model element and src() and trg() denote its
	 * source and target end, then src(type(r)) >= type(src(r)) and trg(type(r))
	 * >= type(trg(r))
	 * @param relation
	 */
	private void checkRelationEndTypes(Relation relation) {
		if (!relation.getType().isEmpty()) {
			// TODO: Type checking for multiple types
			Relation typeRel = (Relation) relation.getType().get(0);
			IRelation vpmTypeOfRelInPattern = modelManager
					.getRelationByName(typeRel.getRealElement());
			// If there is a VPM relation corresponding to the type string
			if (vpmTypeOfRelInPattern != null) {
				// Type checking the source model element
				if (relation.getFrom() != null) {
					// Accessing the source element of the relation
					ModelElement source = relation.getFrom();
					if (!source.getType().isEmpty()) {
						TypeInfo info = calculateVariableType(source,
								vpmTypeOfRelInPattern.getFrom());
						if (!info.isValid) {
							IModelElement vpmTypeOfSrcElementInPattern = info.type;
							String expectedTypeStr = vpmTypeOfRelInPattern
									.getFrom().getFullyQualifiedName();
							if (vpmTypeOfSrcElementInPattern != null) {
								String actualTypeStr = vpmTypeOfSrcElementInPattern
										.getFullyQualifiedName();
								String[] context = new String[] {
										actualTypeStr, expectedTypeStr };
								addReferenceResolutionError(
										relation.getAnnotations(),
										VTCLMessages.VTCLValidationErrorCodes_RELATION_SRC_TYPE_ERROR,
										ErrorSeverity.ERROR, context);
							} else {
								addReferenceResolutionError(
										relation.getAnnotations(),
										VTCLMessages.VTCLValidationErrorCodes_SOURCE_UNRESOLVABLE,
										ErrorSeverity.WARNING, new String[0]);
							}
						}
					}
				}
				// Type checking the target model element
				if (relation.getTo() != null) {
					ModelElement target = relation.getTo();
					if (!target.getType().isEmpty()) {
						TypeInfo info = calculateVariableType(target,
								vpmTypeOfRelInPattern.getTo());
						if (!info.isValid) {
							IModelElement vpmTypeOfTrgElementInPattern = info.type;
							String expectedTypeStr = vpmTypeOfRelInPattern
									.getTo().getFullyQualifiedName();
							if (vpmTypeOfTrgElementInPattern != null) {
								String actualTypeStr = vpmTypeOfTrgElementInPattern
										.getFullyQualifiedName();
								String[] context = new String[] {
										actualTypeStr, expectedTypeStr };
								addReferenceResolutionError(
										relation.getAnnotations(),
										VTCLMessages.VTCLValidationErrorCodes_RELATION_TRG_TYPE_ERROR,
										ErrorSeverity.ERROR, context);
							} else {
								addReferenceResolutionError(
										relation.getAnnotations(),
										VTCLMessages.VTCLValidationErrorCodes_TARGET_UNRESOLVABLE,
										ErrorSeverity.WARNING, new String[0]);
							}

						}

					}
				}
			}

		}
	}

	/**
	 * Resolves the types used in an {@link EntityCreateRule} or a
	 * {@link RelationCreateRule}.
	 * @param rule
	 *            : The {@link ElementCreateRule} to be resolved
	 * @param typeStr
	 *            : The string representation of the type of the element to be
	 *            created
	 */
	public void resolveTypesInASMRules(ElementCreateRule rule, String typeStr) {
		List<String> fqnList = null;
		if (rule instanceof EntityCreateRule) {
			fqnList = lookupEntityType(getMachine(), typeStr);
		} else if (rule instanceof RelationCreateRule) {
			fqnList = lookupRelationType(getMachine(), typeStr);
		}
		if (fqnList != null) {
			// If the type can be resolved uniquely: change the type in
			// ElementReference
			if (fqnList.size() == 1) {
				ElementReference elemRef = (ElementReference) rule.getType();
				Constant typeCons = (Constant) elemRef.getArgument();
				typeCons.setValue(fqnList.get(0));
			}
			// If there are multiple resolutions
			if (fqnList.size() > 1) {
				// Collect all possible names
				String[] context = new String[fqnList.size()];
				for (int i = 0; i < fqnList.size(); i++) {
					context[i] = fqnList.get(i);
				}

				addReferenceResolutionError(
						rule.getAnnotations(),
						VTCLMessages.VTCLValidationErrorCodes_AMBIGUOUS_TYPE_RESOLUTION,
						ErrorSeverity.ERROR, context);
			}
			// If the type cannot be resolved
			else if (fqnList.size() == 0) {
				if (rule instanceof EntityCreateRule) {
					addReferenceResolutionError(
							rule.getAnnotations(),
							VTCLMessages.VTCLValidationErrorCodes_ENTITY_TYPE_UNRESOLVABLE,
							ErrorSeverity.ERROR, new String[] {}); // empty
					// context

				} else if (rule instanceof RelationCreateRule) {
					addReferenceResolutionError(
							rule.getAnnotations(),
							VTCLMessages.VTCLValidationErrorCodes_RELATION_TYPE_UNRESOLVABLE,
							ErrorSeverity.ERROR, new String[] {}); // empty
					// context
				}
			}
		}

	}

	private void resolveTypeNames(AnnotatedElement element, List<String> fqnList) {
		resolveTypeNames(element, fqnList, null);
	}

	private void resolveTypeNames(AnnotatedElement element,
			List<String> fqnList, GTPatternBody body) {
		// If the type can be resolved
		if (fqnList.size() == 1) {
			if (element instanceof Entity) {
				Entity entity = (Entity) element;
				resolveEntity(entity, fqnList.get(0), body);
			} else if (element instanceof Relation) {
				Relation relation = (Relation) element;
				resolveRelation(relation, fqnList.get(0), body);
			} else if (element instanceof Variable) {
				Variable variable = (Variable) element;
				resolveAsmVariableType(variable, fqnList.get(0));
			} else if (element instanceof ASMFunction) {
				ASMFunction asmFun = (ASMFunction) element;
				resolveAsmFunctionType(asmFun, fqnList.get(0));
			} else if (element instanceof TypeConstant) {
				TypeConstant typeConst = (TypeConstant) element;
				resolveTypeConstantType(typeConst, fqnList.get(0));
			}
		}
		// If there are multiple resolutions
		else if (fqnList.size() > 1) {
			// Collect all possible names
			String[] context = new String[fqnList.size()];
			for (int i = 0; i < fqnList.size(); i++) {
				context[i] = fqnList.get(i);
			}
			addReferenceResolutionError(
					element.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_AMBIGUOUS_TYPE_RESOLUTION,
					ErrorSeverity.ERROR, context);
		}
		// If the type cannot be resolved
		else if (fqnList.size() == 0) {
			if (element instanceof Entity) {
				Entity entity = (Entity) element;
				addReferenceResolutionError(
						entity.getAnnotations(),
						VTCLMessages.VTCLValidationErrorCodes_ENTITY_TYPE_UNRESOLVABLE,
						ErrorSeverity.ERROR, new String[] {}); // empty context

			} else if (element instanceof Relation) {
				Relation relation = (Relation) element;
				addReferenceResolutionError(
						relation.getAnnotations(),
						VTCLMessages.VTCLValidationErrorCodes_RELATION_TYPE_UNRESOLVABLE,
						ErrorSeverity.ERROR, new String[] {}); // empty context
			} else if (element instanceof Variable) {
				Variable variable = (Variable) element;
				String[] context = new String[] { variable.getName() };
				addReferenceResolutionError(
						variable.getVariableType().getAnnotations(),
						VTCLMessages.VTCLValidationErrorCodes_VARIABLE_TYPE_UNRESOLVABLE,
						ErrorSeverity.ERROR, context);
			} else if (element instanceof ASMFunction) {
				ASMFunction asmFun = (ASMFunction) element;
				String[] context = new String[] { asmFun.getName() };
				addReferenceResolutionError(
						asmFun.getReturnType().getAnnotations(),
						VTCLMessages.VTCLValidationErrorCodes_ASMFUN_TYPE_UNRESOLVABLE,
						ErrorSeverity.ERROR, context);
			}
		}
	}

	private void resolveTypeConstantType(TypeConstant typeConst, String fqn) {
		if (!fqn.equals("entity") && !fqn.equals("vpm.entity")
				&& !fqn.equals("ENTITY") && !fqn.equals("relation")
				&& !fqn.equals("vpm.entity.relation")
				&& !fqn.equals("RELATION")) {
			if (typeConst != null) {
				typeConst.setFqn(fqn);
			}
		} else {
			typeConst.setFqn(typeConst.getName());
		}
	}

	private void resolveAsmFunctionType(ASMFunction asmFun, String fqn) {
		if (!fqn.equals("entity") && !fqn.equals("vpm.entity")
				&& !fqn.equals("relation")
				&& !fqn.equals("vpm.entity.relation")) {
			TypeConstant typeConst = asmFun.getReturnType();
			if (typeConst != null) {
				typeConst.setFqn(fqn);
			}
		}

	}

	private void resolveAsmVariableType(Variable variable, String fqn) {
		if (!fqn.equals("entity") && !fqn.equals("vpm.entity")
				&& !fqn.equals("relation")
				&& !fqn.equals("vpm.entity.relation")) {
			TypeConstant typeConst = variable.getVariableType();
			if (typeConst != null) {
				typeConst.setFqn(fqn);
			}
		}
	}

	private void resolveEntity(Entity entity, String fqn, GTPatternBody body) {
		if (!fqn.contentEquals(VTCLTypeNameConstants.TOP)
				&& !fqn.contentEquals(VTCLTypeNameConstants.ENTITY)
				&& !fqn.equals("entity") && !fqn.equals("vpm.entity")) {
			Entity typeEntity = (Entity) resolveElementInElementWrapper(fqn,
					body, true);
			if (typeEntity != null) {
				// Adding the type
				entity.getType().add(typeEntity);
				// Adding a TypeOf element required by the pattern matcher
				TypeOf relship = EditmodelFactory.eINSTANCE.createTypeOf();
				// Setting the type
				relship.setSupplier(typeEntity);
				relship.setSupplierStr(typeEntity.getName());
				// Setting the instance
				relship.setClient(entity);
				relship.setClientStr(entity.getName());
			}
		}
	}

	private void resolveRelation(Relation relation, String fqn,
			GTPatternBody body) {
		if (!fqn.contentEquals(VTCLTypeNameConstants.TOP)
				&& !fqn.contentEquals(VTCLTypeNameConstants.RELATION)
				&& !fqn.equals("relation")
				&& !fqn.equals("vpm.entity.relation")) {
			Relation typeRelation = (Relation) resolveElementInElementWrapper(
					fqn, body, false);
			if (typeRelation != null) {
				// Adding the type
				relation.getType().add(typeRelation);
				// Adding a TypeOf element required by the pattern matcher
				TypeOf relship = EditmodelFactory.eINSTANCE.createTypeOf();
				// Setting the type
				relship.setSupplier(typeRelation);
				relship.setSupplierStr(typeRelation.getName());
				// Setting the instance
				relship.setClient(relation);
				relship.setClientStr(relation.getName());
			}
		}
	}

	private ModelElement resolveElementInElementWrapper(String fqn,
			GTPatternBody body, boolean isEntity) {
		ModelElement element = lookupElementWrapperInBody(fqn, body
				.getElementWrappers());
		if (element == null) {
			IModelElement type = modelManager.getElementByName(fqn);
			if (type != null && type.isEntity()) {
				Entity typeEntity = EditmodelFactory.eINSTANCE.createEntity();
				typeEntity.setRealElement(fqn);
				typeEntity.setName(fqn);
				typeEntity.setTypeStr(fqn);
				body.getElementWrappers().add(typeEntity);
				return typeEntity;
			} else if (type != null && type.isRelation()) {
				Relation typeRelation = EditmodelFactory.eINSTANCE
						.createRelation();
				typeRelation.setRealElement(fqn);
				typeRelation.setName(fqn);
				typeRelation.setTypeStr(fqn);
				body.getElementWrappers().add(typeRelation);
				return typeRelation;
			}
		}
		return element;
	}

	// -------------------------- Lookup methods ----------------------------

	/**
	 * Returns a local {@link PatternVariable} definition in a
	 * {@link GTPatternBody} (if any) having a name corresponding to varStr. If
	 * multiple variables have the same name in GTRule, this error is ignored
	 * here.
	 * @param varStr
	 *            : String, the name of the variable we are seeking
	 * @param body
	 *            : {@link GTPatternBody} a graph pattern
	 * @return varDef : {@link Variable}, which is found in body
	 */
	public PatternVariable lookupVariableDefByName(String varStr,
			GTPatternBody body) {
		// Skip lookup if the variable reference starts with '_' character
		if (varStr.charAt(0) == '_') return null;
		boolean found = false;
		PatternVariable varDef = null;
		// Seek for pattern variable in the symbolic parameters of the pattern
		// header
		Iterator<?> iter1 = body.getHeader().getSymParameters().iterator();
		while (!found && iter1.hasNext()) {
			varDef = (PatternVariable) iter1.next();
			// Check if the names are corresponding
			if (varStr.equals(varDef.getName())) {
				found = true;
			}
		}

		// Seek for pattern variable in the local variables
		Iterator<?> iter2 = body.getLocalVariables().iterator();
		while (!found && iter2.hasNext()) {
			varDef = (PatternVariable) iter2.next();
			// Check if the names are corresponding
			if (varStr.equals(varDef.getName())) {
				found = true;
			}
		}
		if (found) {
			return varDef;
		} else {
			return null;
		}
	}

	/**
	 * Returns a local {@link Variable} definition in a {@link GTRule} (if any)
	 * having a name corresponding to varStr. If multiple variables have the
	 * same name in GTRule, this error is ignored here.
	 * @param varStr
	 *            : String, the name of the variable we are seeking
	 * @param rule
	 *            : {@link GTRule} a graph transformation rule
	 * @return varDef : the first {@link Variable} with proper name in GTRule
	 *         (if found)
	 */
	public Variable lookupVariableDefByName(String varStr, GTRule rule) {
		boolean found = false;
		Variable varDef = null;

		// Seek for pattern variable in the local variables
		Iterator<?> iter2 = rule.getLocalVariables().iterator();
		while (!found && iter2.hasNext()) {
			varDef = (Variable) iter2.next();
			// Check if the names are corresponding
			if (varStr.equals(varDef.getName())) {
				found = true;
			}
		}
		if (found) {
			return varDef;
		} else {
			return null;
		}
	}

	/**
	 * Returns a {@link VariableReference} corresponding to a {@link VPMElement}
	 * (if any) having a name corresponding to varStr. If multiple variables
	 * have the same name in GTRule, this error is ignored here.
	 * @param varStr
	 *            : String, the name of the variable we are seeking
	 * @param localElement
	 *            : a local {@link VPMElement}
	 * @return varDef : the first {@link VariableReference} with corresponding
	 *         name in {@link VPMElement} (if found)
	 */
	public VariableReference lookupVariableRefByName(String varStr,
			VPMElement localElement) {
		boolean found = false;
		VariableReference varRef = null;
		// Seek for pattern variable in the symbolic parameters of the pattern
		// header
		Iterator<?> iter1 = localElement.getVariableReferences().iterator();
		while (!found && iter1.hasNext()) {
			varRef = (VariableReference) iter1.next();
			// Check if the names are corresponding
			if (varStr.equals(varRef.getName())) {
				found = true;
			}
		}
		if (found) {
			return varRef;
		} else {
			return null;
		}
	}

	/**
	 * Traverses the call hierarchy of ASMRuleInvocations in order to resolve
	 * the VariableReference with an name varName
	 * @param varName
	 *            : the name of the variable reference to be resolved
	 * @param ruleInvoc
	 *            : the current ASMRuleInvocation
	 * @return The resolved variable (definition)
	 */

	public Variable traverseCallHierarchy(String varName,
			ASMRuleInvocation ruleInvoc) {
		Variable var = null;
		// Try to find Variable locally in the lists of localVariables
		if (ruleInvoc instanceof BlockRule) {
			BlockRule rule = (BlockRule) ruleInvoc;
			var = lookupVariableDefInRule(varName, rule.getLocalVariables());
		}

		// If the variable is not found locally, traverse recursively up in the
		// call hierarchy
		if (var == null && ruleInvoc.getCaller() != null) {
			return traverseCallHierarchy(varName, ruleInvoc.getCaller());
		}
		// If it is already the body of an ASMRule, lookup in rule parameter
		// list
		else if (var == null && ruleInvoc.getAsmRule() != null) {
			Rule asmRule = ruleInvoc.getAsmRule();
			var = lookupVariableDefInRule(varName, asmRule.getLocalVariables());
		}
		// An action part of the GT rule is reached (no link to the GTRule
		// itself)
		else if (var == null && ruleInvoc.getAsmRule() == null) {
			GTRule gtRule = actionRuleMap.get(ruleInvoc);
			if (gtRule != null) {
				var = lookupVariableDefInRule(varName, gtRule
						.getLocalVariables());
			}
		}
		return var;
	}

	/**
	 * Returns a local {@link Variable} definition in an ASM Rule (if any)
	 * having a name corresponding to varName. If multiple variables have the
	 * same name in Rule, this error is ignored here.
	 * @param varName
	 *            : String, the name of the variable we are seeking
	 * @param list
	 *            : List of {@link Variable} definitions
	 * @return varDef : the first {@link Variable} in the list (if found)
	 */
	protected Variable lookupVariableDefInRule(String varName, List<?> list) {
		boolean found = false;
		Variable varDef = null;
		// Seek for pattern definition in the machine
		Iterator<?> iter1 = list.iterator();
		while (!found && iter1.hasNext()) {
			varDef = (Variable) iter1.next();
			// Check if the names are corresponding
			if (varName.equals(varDef.getName())) {
				found = true;
			}
		}
		if (found) {
			return varDef;
		} else {
			return null;
		}
	}

	/**
	 * Returns a {@link GTPattern} defined in a {@link Machine} (if any) having
	 * a name corresponding to patternName.
	 * @param patternName
	 *            : String, the name of the {@link GTPattern} we are seeking
	 * @param arity
	 * @param machine
	 *            : {@link Machine} we are seeking in
	 * @return patternDef : the first {@link GTPattern} (if found)
	 */
	public GTPattern lookupGTPattern(String patternName, int arity,
			Machine machine) {
		boolean found = false;
		GTPattern patternDef = null;
		// Seek for pattern definition in the machine
		Iterator<?> iter1 = machine.getGtPatternDefinitions().iterator();
		while (!found && iter1.hasNext()) {
			patternDef = (GTPattern) iter1.next();
			// Check if the names are corresponding
			if ((patternName.equals(patternDef.getFqn()) || patternName
					.equals(patternDef.getName()))
					&& arity == patternDef.getSymParameters().size()) {
				found = true;
			}
		}
		if (found) {
			return patternDef;
		} else {
			return null;
		}
	}

	/**
	 * Returns a {@link GTRule} defined in a {@link Machine} (if any) having a
	 * name corresponding to ruleName.
	 * @param ruleName
	 *            : String, the name of the {@link GTRule} we are seeking
	 * @param arity
	 * @param machine
	 *            : {@link Machine} we are seeking in
	 * @return gtRuleDef : the first {@link GTRule} in machine (if found)
	 */
	public GTRule lookupGTRule(String ruleName, int arity, Machine machine) {
		boolean found = false;
		GTRule gtRuleDef = null;
		// Seek for pattern definition in the machine
		Iterator<?> iter1 = machine.getGtRuleDefinitions().iterator();
		while (!found && iter1.hasNext()) {
			gtRuleDef = (GTRule) iter1.next();
			// Check if the names are corresponding
			if ((ruleName.equals(gtRuleDef.getFqn()) || ruleName
					.equals(gtRuleDef.getName()))
					&& arity == gtRuleDef.getSymParameters().size()) {
				found = true;
			}
		}
		if (found) {
			return gtRuleDef;
		} else {
			return null;
		}
	}

	/**
	 * Returns an ASM {@link Rule} defined in a {@link Machine} (if any) having
	 * a name corresponding to ruleName.
	 * @param ruleName
	 *            : String, the name of the ASM {@link Rule} we are seeking
	 * @param arity
	 * @param machine
	 *            : {@link Machine} we are seeking in
	 * @return ruleDef : the first {@link Rule} in machine (if found)
	 */
	public Rule lookupAsmRule(String ruleName, int arity, Machine machine) {
		boolean found = false;
		Rule ruleDef = null;
		// Seek for pattern definition in the machine
		Iterator<?> iter1 = machine.getAsmRuleDefinitions().iterator();
		while (!found && iter1.hasNext()) {
			ruleDef = (Rule) iter1.next();
			// Check if the names are corresponding
			if ((ruleName.equals(ruleDef.getFqn()) || ruleName.equals(ruleDef
					.getName()))
					&& arity == ruleDef.getSymParameters().size()) {
				found = true;
			}
		}
		if (found) {
			return ruleDef;
		} else {
			return null;
		}
	}

	/**
	 * Returns an {@link ASMFunction} defined in a {@link Machine} (if any)
	 * having a name corresponding to the String asmFunName.
	 * @param asmFunName
	 *            : String, the name of the {@link ASMFunction} we are seeking
	 * @param arity
	 * @param machine
	 *            : {@link Machine} we are seeking in
	 * @param isRValue
	 *            : denotes if the ASMFunction is used as a right value in a
	 *            term
	 * @return asmFunDef : the first {@link ASMFunction} in machine (if found)
	 */
	public ASMFunction lookupAsmFunction(String asmFunName, int arity,
			Machine machine, boolean isRValue) {
		boolean found = false;
		ASMFunction asmFunDef = null;
		// Seek for ASMFunction definition in the machine
		Iterator<?> iter1 = machine.getAsmFunctionDefinitions().iterator();
		while (!found && iter1.hasNext()) {
			asmFunDef = (ASMFunction) iter1.next();
			if ((asmFunName.equals(asmFunDef.getFqn()) || asmFunName
					.equals(asmFunDef.getName()))
					&& arity == asmFunDef.getArity()) {
				found = true;
			}
		}
		// // Check ASMFunction names of Native Functions if (isRValue == true)
		// if (!found && isRValue) {
		// for (int i = 0; (!found && i < getNativeFunctions().length); i++) {
		// ASMNativeFunction nativeFunction = getNativeFunctions()[i];
		// if ( asmFunName.equals(nativeFunction.getName()) ) {
		// found = true;
		// }
		// }
		// }
		if (found) {
			return asmFunDef;
		} else {
			return null;
		}
	}

	/**
	 * Returns true if there is a native ASM function having a name
	 * corresponding to the String asmFunName. Native ASM functions can only be
	 * used as a right value
	 * @param asmFunName
	 *            : String, the name of the ASMNativeFunction we are seeking
	 * @return true, if there is a native ASM function with the corresponding
	 *         name
	 */
	public boolean lookupNativeAsmFunction(String asmFunName) {
		boolean found = false;
		// Check ASMFunction names of Native Functions if (isRValue == true)
		if (!found) {
			for (int i = 0; (!found && i < getNativeFunctions().length); i++) {
				ASMNativeFunction nativeFunction = getNativeFunctions()[i];
				if (asmFunName.equals(nativeFunction.getName())) {
					found = true;
				}
			}
		}
		if (found) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Returns a {@link ModelElement} listed in elementWrappers of a pattern
	 * body (if any) having a name corresponding to the String fqn.
	 * @param fqn
	 *            : Fully qualified name of the element being sought
	 * @param elementWrappers
	 *            : List of {@link ModelElement}s (Entities and Relations)
	 * @return {@link ModelElement}
	 */
	@SuppressWarnings("unchecked")
	public ModelElement lookupElementWrapperInBody(String fqn,
			List elementWrappers) {
		boolean found = false;
		ModelElement element = null;
		Iterator iter1 = elementWrappers.iterator();
		// Check if a model element with the corresponding FQN already exists in
		// elementWrappers
		while (!found && iter1.hasNext()) {
			element = (ModelElement) iter1.next();
			if (fqn.equals(element.getName())) {
				found = true;
			}
		}
		if (found) {
			return element;
		} else {
			return null;
		}
	}

	/**
	 * Returns the annotation of a ModelElement corresponding to a specific
	 * annotation key
	 * @param key
	 *            : {@link IAnnotationKeys}, typically the location
	 * @param annotations
	 *            : the list of annotation of the model element
	 * @return annotation : the first {@link Annotation} stored at the specific
	 *         key
	 */
	public Annotation lookupAnnotation(String key, List<?> annotations) {
		boolean found = false;
		Annotation annotation = null;
		// Seek for pattern variable in the symbolic parameters of the pattern
		// header
		Iterator<?> iter1 = annotations.iterator();
		while (!found && iter1.hasNext()) {
			annotation = (Annotation) iter1.next();
			// Check if the names are corresponding
			if (annotation.getKey().equals(key)) {
				found = true;
			}
		}
		return annotation;

	}

	/**
	 * Searches for a VPM equivalent of the "type" parameter in the model space.
	 * Type is tried to be resolved first as an absolute FQN and then as a
	 * relative FQN (prefixed by the import declarations of the module)
	 * @param machine
	 *            : {@link Machine}
	 * @param type
	 *            : String representation of the type to be resolved
	 * @return List<String> : the fully qualified name of the VPM elements found
	 */
	public List<String> lookupEntityType(Machine machine, String type) {
		List<String> nameList = new ArrayList<String>();
		// If the type name is "entity", handle differently
		if (type.equals("entity") || type.equals("vpm.entity")
				|| type.equals("ENTITY")) {
			// nameList.add(type);
			nameList.add("ENTITY");
			return nameList;
		}
		// Accessing the container module
		Module module = machine.getModule();
		// It is irrelevant if the type of the entity is a local name or FQN
		// Try to resolve type name as an fqn
		IEntity e = modelManager.getEntityByName(type);
		if (e != null) {
			nameList.add(e.getFullyQualifiedName());
			return nameList;
		}

		// check in imports
		for (ImportDeclaration id : module.getImport()) {
			String fqn = fqn(id.getImportValue(), type);
			e = modelManager.getEntityByName(fqn);
			if (e != null) {
				nameList.add(e.getFullyQualifiedName());
			}
		}
		return nameList;
	}

	/**
	 * Searches for a VPM equivalent of the "type" parameter in the model space.
	 * Type is tried to be resolved first as an absolute FQN and then as a
	 * relative FQN (prefixed by the import declarations of the module)
	 * @param machine
	 *            : {@link Machine}
	 * @param type
	 *            : String representation of the type to be resolved
	 * @return List<String> : the fully qualified name of the VPM elements found
	 */
	public List<String> lookupRelationType(Machine machine, String type) {
		List<String> nameList = new ArrayList<String>();
		// If the type name is "relation", handle differently
		if (type.equals("relation") || type.equals("vpm.entity.relation")
				|| type.equals("RELATION")) {
			// nameList.add(type);
			nameList.add("RELATION");
			return nameList;
		}
		// Accessing the container module
		Module module = (Module) machine.eContainer();
		// Try to resolve type name as an fqn
		IRelation r = modelManager.getRelationByName(type);
		if (r != null) {
			nameList.add(r.getFullyQualifiedName());
		}

		// check it also in imports, and report
		for (ImportDeclaration id : module.getImport()) {
			String fqn = fqn(id.getImportValue(), type);
			r = modelManager.getRelationByName(fqn);
			if (r != null) {
				nameList.add(r.getFullyQualifiedName());
			}
		}
		return nameList;
	}

	/**
	 * Returns the fully qualified name concatenated from <b>ns</b> and <b>n</b>
	 * @param ns
	 *            : The namespace string
	 * @param n
	 *            : The local name
	 * @return the string concatenated as <b>ns+"."+n</b>
	 */
	private String fqn(String ns, String n) {
		if (ns == null) return n;
		if (ns.length() == 0) return n;
		// if (n.indexOf('.')>=0) return n;
		return ns + "." + n;
	}

	// public Machine lookupMachine(ASMRuleInvocation ruleInvoc) {
	// // If the variable is not found locally, traverse recursively up in the
	// call hierarchy
	// if (ruleInvoc.getCaller() != null) {
	// return lookupMachine(ruleInvoc.getCaller());
	// }
	// // If it is already the body of an ASMRule, lookup in rule parameter list
	// else if (ruleInvoc.getAsmRule() != null) {
	// Rule asmRule = ruleInvoc.getAsmRule();
	// return asmRule.getNamespace();
	// }
	// // An action part of the GT rule is reached (no link to the GTRule
	// itself)
	// else if (ruleInvoc.getAsmRule() == null) {
	// GTRule gtRule = actionRuleMap.get(ruleInvoc);
	// if (gtRule != null) {
	// return gtRule.getNamespace();
	// }
	// }
	// return null;
	// }

	// ------------- Additional validations ---------------------

	/**
	 * Runs validations to reveal multiply defined variables in machine
	 * @param machine
	 *            : the {@link Machine} being validated
	 */
	public void validateDistinctVariables(Machine machine) {
		for (Iterator<?> iter = machine.getAsmRuleDefinitions().iterator(); iter
				.hasNext();) {
			Rule asmRule = (Rule) iter.next();
			validateDistinctVariablesAsmRule(asmRule);
		}

		for (Iterator<?> iter = machine.getGtRuleDefinitions().iterator(); iter
				.hasNext();) {
			GTRule gtRule = (GTRule) iter.next();
			validateDistinctVariablesGTRule(gtRule);
			validatePostcondition(gtRule);
		}

		for (Iterator<?> iter = machine.getGtPatternDefinitions().iterator(); iter
				.hasNext();) {
			GTPattern pattern = (GTPattern) iter.next();
			validateDistinctVariablesPattern(pattern);

		}
	}

	/**
	 * Validates if the postcondition of gtRule does not contain OR patterns,
	 * check expressions or negative patterns.
	 * @param gtRule
	 *            : {@link GTRule} to be checked
	 */
	private void validatePostcondition(GTRule gtRule) {
		GTPatternCall rhsCall = gtRule.getPostcondition();
		if (rhsCall != null) {
			GTPattern rhs = rhsCall.getCalledPattern();
			if (rhs != null) {
				if (rhs.getPatternBodies().size() > 1) {
					String[] context = new String[] { rhs.getName(),
							gtRule.getName() };
					addReferenceResolutionError(
							rhs.getAnnotations(),
							VTCLMessages.VTCLValidationErrorCodes_POSTCONDITION_OR_PATTERN_NOT_ALLOWED,
							ErrorSeverity.ERROR, context);
				}
				for (Iterator<?> it = rhs.getPatternBodies().iterator(); it
						.hasNext();) {
					GTPatternBody body = (GTPatternBody) it.next();
					for (Iterator<?> it2 = body.getCheckExpressions()
							.iterator(); it2.hasNext();) {
						Term term = (Term) it2.next();
						String[] context = new String[] { rhs.getName(),
								gtRule.getName() };
						addReferenceResolutionError(
								term.getAnnotations(),
								VTCLMessages.VTCLValidationErrorCodes_POSTCONDITION_IGNORE_CHECK,
								ErrorSeverity.WARNING, context);
					}
					for (Iterator<?> it3 = body.getNegativePatterns()
							.iterator(); it3.hasNext();) {
						GTPatternCall gtCall = (GTPatternCall) it3.next();
						String[] context = new String[] { rhs.getName(),
								gtRule.getName() };
						addReferenceResolutionError(
								gtCall.getAnnotations(),
								VTCLMessages.VTCLValidationErrorCodes_POSTCONDITION_IGNORE_NAC,
								ErrorSeverity.WARNING, context);

					}

				}
			}

		}

	}

	/**
	 * All variables of a GTRule should be named differently. This is checked
	 * for locally defined variables, and validations are initiated for each
	 * locally defined subpattern and the action part
	 * @param gtRule
	 *            : GTRule to be checked for distinctness of variable names
	 */

	public void validateDistinctVariablesGTRule(GTRule gtRule) {
		Set<String> variableNames = new HashSet<String>();
		// Check rule parameters for distinct variables
		for (Iterator<?> iter = gtRule.getLocalVariables().iterator(); iter
				.hasNext();) {
			Variable varDef = (Variable) iter.next();
			checkNameExistence(
					varDef,
					variableNames,
					VTCLMessages.VTCLValidationErrorCodes_DUPLICATE_GTVAR_DEFINITION);
			checkVariableIsInUse(varDef,
					VTCLMessages.VTCLValidationErrorCodes_UNUSED_VARIABLE);
		}
		// Check the action part
		if (gtRule.getAction() != null) {
			checkSubruleInvocations(gtRule.getAction(), variableNames);
		}

		// Check distinct variables in subpatterns
		for (Iterator<?> iter = gtRule.getLocalPatternDefinition().iterator(); iter
				.hasNext();) {
			GTPattern pattern = (GTPattern) iter.next();
			validateDistinctVariablesPattern(pattern);
		}
	}

	/**
	 * All variables of a GTPattern should be named differently. This is checked
	 * for locally defined variables in each pattern body, and validations are
	 * initiated for each locally defined subpattern
	 * @param pattern
	 *            : GTPattern to be checked for distinctness of variable names
	 */
	public void validateDistinctVariablesPattern(GTPattern pattern) {
		Set<String> variableNames = new HashSet<String>();
		// Check rule parameters for distinct variables
		for (Iterator<?> iter = pattern.getSymParameters().iterator(); iter
				.hasNext();) {
			PatternVariable varDef = (PatternVariable) iter.next();
			checkNameExistence(
					varDef,
					variableNames,
					VTCLMessages.VTCLValidationErrorCodes_DUPLICATE_PATTERNVAR_DEFINITION);
			checkVariableIsInUse(varDef,
					VTCLMessages.VTCLValidationErrorCodes_UNUSED_VARIABLE);
		}
		Set<String> localVarNames = new HashSet<String>();
		for (Iterator<?> iter = pattern.getPatternBodies().iterator(); iter
				.hasNext();) {
			// Create a copy of variable names
			localVarNames.clear();
			for (String nameStr : variableNames) {
				localVarNames.add(nameStr);
			}
			GTPatternBody body = (GTPatternBody) iter.next();
			// Validate the distinctness of local pattern variables
			for (Iterator<?> iter2 = body.getLocalVariables().iterator(); iter2
					.hasNext();) {
				PatternVariable bodyVarDef = (PatternVariable) iter2.next();
				checkNameExistence(
						bodyVarDef,
						localVarNames,
						VTCLMessages.VTCLValidationErrorCodes_DUPLICATE_PATTERNVAR_DEFINITION);
				checkVariableIsInUse(bodyVarDef,
						VTCLMessages.VTCLValidationErrorCodes_UNUSED_VARIABLE);
			}
			// Validate locally defined subpatterns
			for (Iterator<?> iter3 = body.getLocalPatternDefinition()
					.iterator(); iter3.hasNext();) {
				GTPattern subPattern = (GTPattern) iter3.next();
				validateDistinctVariablesPattern(subPattern);
			}
		}
	}

	/**
	 * All variables of an ASM rule should be named differently. This is checked
	 * for local variables of the rule, and recursively, for all subrule
	 * invocations within the ASM rule. Critical cases are those with local
	 * variable definitions (LetRule, ChooseRule, and ForallRule) and those that
	 * call subrule(s)
	 * @param asmRule
	 *            : an ASM {@link Rule}
	 */
	public void validateDistinctVariablesAsmRule(Rule asmRule) {
		Set<String> varNames = new HashSet<String>();
		// Check rule parameters for distinct variables
		for (Iterator<?> iter = asmRule.getLocalVariables().iterator(); iter
				.hasNext();) {
			Variable varDef = (Variable) iter.next();
			checkNameExistence(
					varDef,
					varNames,
					VTCLMessages.VTCLValidationErrorCodes_DUPLICATE_ASMVAR_DEFINITION);
			checkVariableIsInUse(varDef,
					VTCLMessages.VTCLValidationErrorCodes_UNUSED_VARIABLE);
		}
		// Check distinct variables in rule body
		checkSubruleInvocations(asmRule.getBody(), varNames);
	}

	/**
	 * Helper function for validateDistinctVariablesAsmRule. It recursively
	 * traverses all the subrules within the ASM rule
	 * @param ruleInvoc
	 *            : The currently processed {@link ASMRuleInvocation}
	 * @param varNames
	 *            : The set of variable names obtained up to now
	 */
	protected void checkSubruleInvocations(ASMRuleInvocation ruleInvoc,
			Set<String> varNames) {
		// Checking variables defined locally -> Let rule, Choose rule, Forall
		// rule
		if (ruleInvoc instanceof BlockRule) {
			BlockRule rule = (BlockRule) ruleInvoc;
			for (Iterator<?> iter = rule.getLocalVariables().iterator(); iter
					.hasNext();) {
				Variable varDef = (Variable) iter.next();
				checkNameExistence(
						varDef,
						varNames,
						VTCLMessages.VTCLValidationErrorCodes_DUPLICATE_ASMVAR_DEFINITION);
				checkVariableIsInUse(varDef,
						VTCLMessages.VTCLValidationErrorCodes_UNUSED_VARIABLE);
			}
		}
		// // Checking that ASM rule is called with proper number of parameters
		// if (ruleInvoc instanceof CallRule) {
		// CallRule callRule = (CallRule) ruleInvoc;
		// Rule ruleDef = callRule.getAsmRule();
		// if (ruleDef != null) {
		// if (callRule.getActualParameters().size() !=
		// ruleDef.getSymParameters().size()) {
		// modelResolver.addReferenceResolutionError(
		// callRule.getAnnotations(),
		// VTCLMessages.VTCLValidationErrorCodes_DUPLICATE_ASMFUN_NAME,
		// ErrorSeverity.ERROR);
		// }
		// }

		// }

		// Recursive call for ConditionalRuleIf to process bodies
		if (ruleInvoc instanceof ConditionalRuleIf) {
			ConditionalRuleIf rule = (ConditionalRuleIf) ruleInvoc;
			ASMRuleInvocation subRule1 = rule.getRuleTrue();
			if (subRule1 != null) {
				Set<String> varNamesIf = new HashSet<String>(varNames);
				checkSubruleInvocations(subRule1, varNamesIf);
			}
			ASMRuleInvocation subRule2 = rule.getRuleFalse();
			if (subRule2 != null) {
				Set<String> varNamesIf = new HashSet<String>(varNames);
				checkSubruleInvocations(subRule2, varNamesIf);
			}
		}
		// Recursive call for ConditionalRuleTry to process bodies
		else if (ruleInvoc instanceof ConditionalRuleTry) {
			ConditionalRuleTry rule = (ConditionalRuleTry) ruleInvoc;
			ASMRuleInvocation subRule1 = rule.getRuleToTry();
			if (subRule1 != null) {
				Set<String> varNamesTry = new HashSet<String>(varNames);
				checkSubruleInvocations(subRule1, varNamesTry);
			}
			ASMRuleInvocation subRule2 = rule.getRuleElse();
			if (subRule2 != null) {
				Set<String> varNamesTry = new HashSet<String>(varNames);
				checkSubruleInvocations(subRule2, varNamesTry);
			}
		}
		// Recursive call for CompoundRules to process body
		else if (ruleInvoc instanceof CompoundRule) {
			CompoundRule rule = (CompoundRule) ruleInvoc;
			ASMRuleInvocation subRule = rule.getBody();
			if (subRule != null) {
				checkSubruleInvocations(subRule, varNames);
			}
		}
		// Recursive call for NestedRules to process subrules
		else if (ruleInvoc instanceof NestedRule) {
			NestedRule rule = (NestedRule) ruleInvoc;
			// Iterate over all subrules
			for (Iterator<?> iter = rule.getSubrules().iterator(); iter
					.hasNext();) {
				ASMRuleInvocation subRule = (ASMRuleInvocation) iter.next();
				// Store varNames as oldVarNames
				Set<String> oldVarNames = new HashSet<String>(varNames);
				// Check subrules
				checkSubruleInvocations(subRule, varNames);
				// Restore varNames for next iteration
				varNames = oldVarNames;
			}
		}
	}

	/**
	 * Helper function to check if the name of a GTASMElement is not distinct,
	 * i.e. already contained in names. Name existence is not checked when the
	 * name of the model element starts with an _ In such a case, an error
	 * message is reported otherwise, the name of the GTASMElement is added to
	 * names.
	 * @param element
	 *            : The current GTASMElement to be checked for existence
	 * @param names
	 *            : The set of names obtained up to now
	 * @param errorCode
	 *            : The type of the error message that should be reported
	 */
	protected void checkNameExistence(GTASMElement element, Set<String> names,
			String errorCode) {
		if (!names.contains(element.getName())) {
			if (element.getName().charAt(0) != '_') {
				names.add(element.getName());
			}
		} else {
			String[] context = new String[] { element.getName() };
			addReferenceResolutionError(element.getAnnotations(), errorCode,
					ErrorSeverity.ERROR, context);
		}
	}

	/**
	 * Helper function to check if the signature of a GTASMElement (rule,
	 * pattern, gtrule or ASM function) is not distinct, i.e. already contained
	 * in names. In such a case, an error message is reported otherwise, the
	 * signature of the GTASMElement is added to names.
	 * @param element
	 *            : The GTASM element currently processed
	 * @param signature
	 *            : The signature of the current GTASMElement to be checked for
	 *            existence
	 * @param names
	 *            : The set of signatures obtained up to now
	 * @param errorCode
	 *            : The type of the error message that should be reported
	 */
	protected void checkSignatureExistence(GTASMElement element,
			String signature, Set<String> names, String errorCode) {
		if (!names.contains(signature)) {
			names.add(signature);
		} else {
			String[] context = new String[] { signature };
			addReferenceResolutionError(element.getAnnotations(), errorCode,
					ErrorSeverity.ERROR, context);
		}
	}

	/**
	 * Duplicates (name clashes) between ASM functions, GT rules, patterns and
	 * ASM rules defined within the same Machine.
	 * @param machine
	 *            : {@link Machine} to be validated
	 */

	public void validateNoDuplicateNames(Machine machine) {
		Set<String> asmFunNames = new HashSet<String>();
		for (ASMFunction asmFun : machine.getAsmFunctionDefinitions()) {
			String signature = asmFun.getName() + "/"
					+ String.valueOf(asmFun.getArity());
			checkSignatureExistence(asmFun, signature, asmFunNames,
					VTCLMessages.VTCLValidationErrorCodes_DUPLICATE_ASMFUN_NAME);
		}

		Set<String> asmRuleNames = new HashSet<String>();
		for (Rule asmRule : machine.getAsmRuleDefinitions()) {
			String signature = asmRule.getName() + "/"
					+ String.valueOf(asmRule.getSymParameters().size());
			checkSignatureExistence(
					asmRule,
					signature,
					asmRuleNames,
					VTCLMessages.VTCLValidationErrorCodes_DUPLICATE_ASMRULE_NAME);
		}

		Set<String> patternNames = new HashSet<String>();
		for (GTPattern pattern : machine.getGtPatternDefinitions()) {
			String signature = pattern.getName() + "/"
					+ String.valueOf(pattern.getSymParameters().size());
			checkSignatureExistence(
					pattern,
					signature,
					patternNames,
					VTCLMessages.VTCLValidationErrorCodes_DUPLICATE_PATTERN_NAME);
		}

		Set<String> gtRuleNames = new HashSet<String>();
		for (GTRule gtRule : machine.getGtRuleDefinitions()) {
			String signature = gtRule.getName() + "/"
					+ String.valueOf(gtRule.getSymParameters().size());
			checkSignatureExistence(gtRule, signature, gtRuleNames,
					VTCLMessages.VTCLValidationErrorCodes_DUPLICATE_GTRULE_NAME);
			Set<String> containedPatternNames = new HashSet<String>();
			//Check for local pattern name uniquity as well
			for (GTPattern pattern : gtRule.getLocalPatternDefinition()) {
				String pSignature = pattern.getName() + "/"
						+ String.valueOf(pattern.getSymParameters().size());
				checkSignatureExistence(
						pattern,
						pSignature,
						containedPatternNames,
						VTCLMessages.VTCLValidationErrorCodes_DUPLICATE_PATTERN_NAME);
			}
		}
	}

	/**
	 * The initial parameter definitions of an {@link ASMFunction} should always
	 * correspond to its Arity
	 * @param asmFun
	 *            : {@link ASMFunction} to be validated
	 */
	public void validateAsmFunctionInitialValues(ASMFunction asmFun) {
		for (Iterator<?> initIter = asmFun.getInitialValues().iterator(); initIter
				.hasNext();) {
			InitialValue initValue = (InitialValue) initIter.next();
			if (initValue.getLocations().size() != asmFun.getArity()) {
				addReferenceResolutionError(
						initValue.getAnnotations(),
						VTCLMessages.VTCLValidationErrorCodes_ASMFUN_ARITY_INITVALUE_MISMATCH,
						ErrorSeverity.ERROR, new String[] {});
			}
		}
	}

	/**
	 * Checks if a model element constant meConst is existing in the model
	 * space. Issues an error if this is not the case:
	 * @param meConst
	 *            : model element constant to be checked
	 * @return The fully qualified name of the resolved model element, or "" if
	 *         no such element exists
	 */
	public String resolveModelElementConstant(Constant meConst) {
		// Get the value of the model element constant (FQN of model element)
		String name = meConst.getValue();
		// Lookup model element
		IModelElement element = modelManager.getElementByName(name);
		// Add type information to the constant
		if (element != null) {
			// Type checking
			String typeStr = typeResolver.lookupType(element);
			meConst.setType(typeStr);
			// ValueKind kind = typeResolver.getValueKind(typeStr);
			meConst.setKind(ValueKind.MODELELEMENT_LITERAL);
		}
		// Issue an error message if model element is non-existing
		else {
			String[] context = new String[] { name };
			addReferenceResolutionError(
					meConst.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_NONEXISTING_MODELELEM_CONSTANT,
					ErrorSeverity.ERROR, context);
			// Add type information
			meConst.setType(VTCLTypeNameConstants.TOP);
			meConst.setKind(ValueKind.UNDEF_LITERAL);
			return "";
		}
		// Fully qualified name of model element is retrieved
		return element.getFullyQualifiedName();

	}

	/**
	 * Checks if a model element constant meConst is existing in the model
	 * space. Issues an error if this is not the case:
	 * @param element
	 *            : annotated element to be checked
	 * @param typeStr
	 *            : the assumed type string
	 */
	public void resolveAsmType(AnnotatedElement element, String typeStr) {
		List<String> fqnList1 = lookupEntityType(getMachine(), typeStr);
		List<String> fqnList2 = lookupRelationType(getMachine(), typeStr);
		fqnList1.addAll(fqnList2);
		resolveTypeNames(element, fqnList1);
	}

	/**
	 * Checks if a variable definition is actually referred to by some variable
	 * references, and issues a warning if it is not the case.
	 * @param varDef
	 * @param errorCode
	 */
	protected void checkVariableIsInUse(Variable varDef, String errorCode) {
		if (varDef.getReferences().isEmpty()) {
			String[] context = new String[] { varDef.getName() };
			addReferenceResolutionError(varDef.getAnnotations(), errorCode,
					ErrorSeverity.WARNING, context);
		}
	}

	/**
	 * Reports an error message when a match counter is used in a negative
	 * pattern call.
	 * @param patternCall
	 * @param element
	 */
	public void reportInvalidMatchCounter(GTPatternCall patternCall,
			String element) {
		String[] context = new String[] { patternCall.getName(), element };
		addReferenceResolutionError(patternCall.getAnnotations(),
				VTCLMessages.VTCLValidationErrorCodes_INVALID_MATCH_COUNTER,
				ErrorSeverity.ERROR, context);
	}

	/**
	 * Issues a warning if a formal parameter of a graph pattern does not appear
	 * in the body (as an entity, a relation or in a pattern variable
	 * assignment)
	 * @param body
	 */
	public void reportUnusedPatternParameter(GTPatternBody body) {
		for (Iterator<?> iter = body.getHeader().getSymParameters().iterator(); iter
				.hasNext();) {
			PatternVariable varDef = (PatternVariable) iter.next();
			boolean found = false;
			Iterator<?> itEnt = body.getPatternGraph().getComponents()
					.iterator();
			// Check all entities inside the top-level pattern graph
			while (!found && itEnt.hasNext()) {
				Entity ent = (Entity) itEnt.next();
				if (ent.getName().equals(varDef.getName())) {
					found = true;
				}
				// Check all relations leading out of the entity
				Iterator<?> itRels = ent.getRelationsFrom().iterator();
				while (!found && itRels.hasNext()) {
					Relation rel = (Relation) itRels.next();
					if (rel.getName().equals(varDef.getName())) {
						found = true;
					}
				}
			}
			// Check all (positive) pattern calls
			for (Iterator<?> itPattCall = body.getCalledPatterns().iterator(); !found
					&& itPattCall.hasNext();) {
				GTPatternCall pattCall = (GTPatternCall) itPattCall.next();
				// Check actual parameters
				for (Iterator<?> itParams = pattCall.getActualParameters()
						.iterator(); !found && itParams.hasNext();) {
					Term param = (Term) itParams.next();
					// If actual parameter is a VariableReference
					if (param instanceof VariableReference) {
						if (param.getName().equals(varDef.getName())) {
							found = true;
						}
					}

				}
			}
			// Check all dangling relations
			for (Iterator<?> itDangRels = body.getDanglingRelations()
					.iterator(); !found && itDangRels.hasNext();) {
				Relation rel = (Relation) itDangRels.next();
				if (rel.getName().equals(varDef.getName())) {
					found = true;
				}
			}

			// Check all pattern variable assignments
			for (Iterator<?> itPVA = body.getVariableAssignments().iterator(); !found
					&& itPVA.hasNext();) {
				PatternVariableAssignment pva = (PatternVariableAssignment) itPVA
						.next();
				if (pva.getLeftValue().getName().equals(varDef.getName())) {
					found = true;
				}
				if (pva.getRightValue().getName().equals(varDef.getName())) {
					found = true;
				}
			}
			// // Check containment constraints
			// for (Iterator<?> itContCons =
			// body.getContainmentConstraints().iterator(); !found &&
			// itContCons.hasNext();) {
			// ContainmentConstraint contConstr = (ContainmentConstraint)
			// itContCons.next();
			// Term parentTerm = (Term) contConstr.getParent();
			// if (parentTerm instanceof VariableReference) {
			// if (parentTerm.getName().equals(varDef.getName())) {
			// found = true;
			// }
			// }
			// }
			// Issue a warning if no reference is found
			if (!found) {
				String[] context = new String[] { varDef.getName() };
				addReferenceResolutionError(
						varDef.getAnnotations(),
						VTCLMessages.VTCLValidationErrorCodes_PATTPARAM_NOT_LOCALLY_DEFINED,
						ErrorSeverity.WARNING, context);
			}
		}

	}

	/**
	 * Check that the left and right variables are different
	 * @param varAssign
	 */
	public void validatePatternVariableConstraint(
			PatternVariableConstraint varAssign) {
		if (varAssign.getLeftValue().getName().equals(
				varAssign.getRightValue().getName())) {
			addReferenceResolutionError(
					varAssign.getAnnotations(),
					VTCLMessages.VTCLValidationErrorCodes_PATTVVARIABLECONSTRAINT_AMBIGOUS,
					ErrorSeverity.ERROR, new String[] {}); // Empty context
		}
	}

}
