/*******************************************************************************
 * Copyright (c) 2009 Zoltan Ujhelyi and Daniel Varro.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Zoltan Ujhelyi - initial API and implementation
 *******************************************************************************/
package org.eclipse.viatra2.visualisation.modelspace;

import java.util.Collection;
import java.util.List;
import java.util.Stack;

import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.ViewerFilter;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PlatformUI;
import org.eclipse.viatra2.core.ICoreNotificationListener;
import org.eclipse.viatra2.core.IModelElement;
import org.eclipse.viatra2.core.IModelManager;
import org.eclipse.viatra2.core.notification.ICoreNotificationObject;
import org.eclipse.viatra2.framework.FrameworkManager;
import org.eclipse.viatra2.framework.IFramework;
import org.eclipse.viatra2.framework.IFrameworkListChangedListener;
import org.eclipse.viatra2.treeeditor.ViatraTreeEditor;
import org.eclipse.viatra2.visualisation.IVisualisationDescriptor;
import org.eclipse.viatra2.visualisation.ViatraGraphViewer;
import org.eclipse.viatra2.visualisation.layouts.IViatraLayoutAlgorithm;
import org.eclipse.viatra2.visualisation.modelspace.actions.groups.FilterGroup;
import org.eclipse.viatra2.visualisation.modelspace.actions.groups.ModelSpaceProviderGroup;
import org.eclipse.viatra2.visualisation.modelspace.datasource.AbstractModelSpaceProvider;
import org.eclipse.viatra2.visualisation.modelspace.datasource.filter.PostFilter;
import org.eclipse.viatra2.visualisation.modelspace.datasource.filter.PreFilter;
import org.eclipse.zest.core.widgets.ZestStyles;
import org.eclipse.zest.layouts.LayoutStyles;
import org.eclipse.zest.layouts.algorithms.RadialLayoutAlgorithm;

/**
 * Describes the contributions of the visualisation descriptor
 * @author Zoltan Ujhelyi
 */
public class ModelSpaceVisualisationDescriptor implements
		IVisualisationDescriptor {

	// View elements
	ModelSpaceGraphViewer gv;
	PreFilter preFilter = new PreFilter();
	PostFilter postFilter = new PostFilter();
	// Viatra source elements
	IModelManager modelManager;
	IFramework framework;
	ViatraTreeEditor editor;
	// Actions and action groups
	ModelSpaceProviderGroup providerGroup = new ModelSpaceProviderGroup();
	// Old layout filters
	ViewerFilter[] filters = new ViewerFilter[0];
	// Event Listeners
	private IFrameworkListChangedListener frameworkListener = new IFrameworkListChangedListener() {

		public void frameworkRemoved(IFramework f) {
			if (f.equals(framework)) {
				setFramework(null);
				setModelManager(null);
				clearGraphModel();
			}
		}

		public void frameworkAdded(IFramework f) {
		}
	};

	/**
	 * Stored previous selections to avoid infinite selection changes
	 * LocalSelection is only used to represent the remote selection, while the
	 * remoteSelection stores the values from local selection.
	 */
	private ISelection localSelection, remoteSelection;
	private ISelectionChangedListener selectionListener = new ISelectionChangedListener() {

		public void selectionChanged(SelectionChangedEvent event) {
			// XXX this code is a bit hacky
			if (framework != null) {
				ISelection selection = event.getSelection();
				if (event.getSource().equals(gv)) {
					// graph viewer selection control
					if (localSelection != null
							&& localSelection.equals(selection)) return;
					ViatraTreeEditor editor = getEditorPart(framework);
					remoteSelection = selection;
					editor.getTreeViewer().setSelection(selection);
				} else {
					// remote selection changed
					if (remoteSelection != null
							&& remoteSelection.equals(selection)) return;
					gv.setSelection(selection);
					localSelection = selection;
				}
			}
		}
	};

	private ICoreNotificationListener modelSpaceListener = new ICoreNotificationListener() {

		Stack<ModelspaceTransactionChanges> transactions = new Stack<ModelspaceTransactionChanges>();
		ModelspaceTransactionChanges actual = new ModelspaceTransactionChanges(
				preFilter);

		/**
		 * Decides whether a selected string is a member of a set of strings
		 * @param item
		 * @param set
		 * @return true, if item is present in set
		 */
		private boolean isOneOf(String item, String[] set) {
			for (String setElement : set) {
				if (item.contentEquals(setElement)) return true;
			}
			return false;
		}

		private void updateGraph() {
			actual.handleLooseEdges();
			Collection<IModelElement> newElements = actual.getNewElements();
			Collection<IModelElement> deletedElements = actual
					.getDeletedElements();
			if (newElements.isEmpty() && deletedElements.isEmpty()) {
				return;
			}
			if (newElements.isEmpty()) {
				refreshGraph();
				return;
			}
			for (IModelElement element : newElements) {
				if (!preFilter.isFiltered(element)) {
					redrawGraph(); // if a new element is added the graph layout
					// shall be updated as well
					return;
				}
			}
		}

		public int getListenerCategory() {
			// TODO what is listener category
			return 0;
		}

		private ICoreNotificationObject notificationObject;

		public void actionPerformed(ICoreNotificationObject notification) {
			this.notificationObject = notification;
			Display.getDefault().syncExec(new Runnable() {

				public void run() {
					String action = notificationObject.getActionType();

					if (isOneOf(action, new String[] {
							ICoreNotificationObject.TA_TRANSACTION_END,
							ICoreNotificationObject.TA_UNDO_END,
							ICoreNotificationObject.TA_SUBTRANSACTION_END })) {
						updateGraph();
						transactions.pop();
					} else if (isOneOf(
							action,
							new String[] {
									ICoreNotificationObject.TA_TRANSACTION_BEGIN,
									ICoreNotificationObject.TA_UNDO_BEGIN,
									ICoreNotificationObject.TA_SUBTRANSACTION_BEGIN,
									ICoreNotificationObject.TA_UNDOABLE_TRANSACTION_BEGIN })) {
						actual = new ModelspaceTransactionChanges(preFilter);
						transactions.push(actual);
					} else if (isOneOf(action, new String[] {
							ICoreNotificationObject.ACTION_CREATE_ENTITY,
							ICoreNotificationObject.ACTION_CREATE_RELATION,
							// this is intentionally commented out - it
							// introduces
							// new, unwanted metamodel nodes in the graph
							// ICoreNotificationObject.ACTION_CREATE_INSTANCEOF,
							// ICoreNotificationObject.ACTION_CREATE_SUPERTYPEOF,
							ICoreNotificationObject.ACTION_SET_ISANY,
							ICoreNotificationObject.ACTION_SET_NAME,
							ICoreNotificationObject.ACTION_SET_RELATION_FROM,
							ICoreNotificationObject.ACTION_SET_RELATION_TO })) {
						actual.addNewElements(notificationObject
								.getNotifiedObjects());
					} else if (isOneOf(action, new String[] {
							ICoreNotificationObject.ACTION_DELETE_ENTITY,
							ICoreNotificationObject.ACTION_DELETE_RELATION,
							ICoreNotificationObject.ACTION_DELETE_INSTANCEOF,
							ICoreNotificationObject.ACTION_DELETE_SUPERTYPEOF })) {
						actual.addRemovedElements(notificationObject
								.getNotifiedObjects());
					}
				}

			});

		}
	};

	/**
	 * Initializes the descriptor.
	 */
	public ModelSpaceVisualisationDescriptor() {
		FrameworkManager.getInstance().addFrameworkListChangedListener(
				frameworkListener);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seeorg.eclipse.viatra2.visualisation.IVisualisationDescriptor#
	 * addMenuContributions(org.eclipse.jface.action.IMenuManager)
	 */
	public void addMenuContributions(IMenuManager manager) {
		providerGroup.fillContextMenu(manager);
		FilterGroup filters = new FilterGroup();
		filters.setFilter(postFilter);
		filters.fillContextMenu(manager);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seeorg.eclipse.viatra2.visualisation.IVisualisationDescriptor#
	 * addToolbarContributions(org.eclipse.jface.action.IToolBarManager)
	 */
	public void addToolbarContributions(IToolBarManager manager) {

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.viatra2.visualisation.IVisualisationDescriptor#getViewer()
	 */
	public ViatraGraphViewer getViewer() {
		return gv;
	}

	private ViatraGraphViewer initializeViewer(Composite parent) {
		gv = new ModelSpaceGraphViewer(parent, SWT.NONE);
		gv.setConnectionStyle(ZestStyles.CONNECTIONS_DIRECTED);
		gv.setLayoutAlgorithm(new RadialLayoutAlgorithm(
				LayoutStyles.NO_LAYOUT_NODE_RESIZING
						| ZestStyles.IGNORE_INVISIBLE_LAYOUT), true);
		gv.setContentProvider(providerGroup.getDefaultProvider());
		gv.addFilter(postFilter);
		gv.setDescriptor(this);
		gv.addSelectionChangedListener(selectionListener);
		// TODO this is required for selection synchronizing - not sure why
		gv.setUseHashlookup(false);
		return gv;
	}

	public void clearGraphModel() {
		preFilter = new PreFilter();
		for (AbstractModelSpaceProvider provider : providerGroup.getProviders()) {
			provider.setFilter(preFilter);
		}
		gv.setInput(new Object());
	}

	public void removeSelection() {
		ISelection selection = gv.getSelection();
		if (!selection.isEmpty()) {
			excludeItems(((StructuredSelection) selection).toList());
			// refresh();
		}
	}

	/**
	 * Sets the filter to allow the selected model element and it's contained
	 * elements to appear in the graph.
	 * @param modelElement
	 *            a model element
	 */
	public void includeItem(IModelElement modelElement) {
		preFilter.includeSubtree(modelElement);
	}

	/**
	 * Sets the filter to allow the selected model elements (and only those
	 * elements) to appear in the graph.
	 * @param elements
	 *            the list of model elements
	 */
	public void includeItems(List<IModelElement> elements) {
		for (IModelElement element : elements)
			preFilter.includeElement(element);
	}

	/**
	 * Sets the filter to disallow the selected model element and it's contained
	 * elements to appear in the graph.
	 * @param modelElement
	 *            a model element
	 */
	public void excludeItem(IModelElement modelElement) {
		preFilter.excludeSubtree(modelElement);
	}

	/**
	 * Sets the filter to disallow the selected model elements (and only those
	 * elements) to appear in the graph.
	 * @param elements
	 *            the list of model elements
	 */
	public void excludeItems(List<IModelElement> elements) {
		for (IModelElement element : elements)
			preFilter.excludeElement(element);
	}

	/**
	 * @return the modelManager
	 */
	public IModelManager getModelManager() {
		return modelManager;
	}

	/**
	 * @return the framework
	 */
	public IFramework getFramework() {
		return framework;
	}

	/**
	 * Sets the Viatra framework associated with the graph viewer
	 * @param framework
	 *            the framework to set
	 */
	public void setFramework(IFramework framework) {
		if (this.framework != null) {
			this.framework.getTopmodel().getNotificationManager()
					.removeAllListener(modelSpaceListener);
		}
		this.framework = framework;
		if (editor != null) {
			editor.getTreeViewer().removeSelectionChangedListener(
					selectionListener);
		}
		if (framework != null) {
			framework.getTopmodel().getNotificationManager().addAllListener(
					modelSpaceListener);
		}
		clearGraphModel();
		editor = getEditorPart(framework);
		assert editor != null : "An editor should be open now";
		editor.getTreeViewer().addSelectionChangedListener(selectionListener);
	}

	/**
	 * Adds a model manager component to the visualisation descriptor. The Model
	 * Manager contains the elements to visualised.
	 * @param manager
	 */
	public void setModelManager(IModelManager manager) {
		this.modelManager = manager;
		for (AbstractModelSpaceProvider provider : providerGroup.getProviders()) {
			provider.setModelManager(modelManager);
		}
		gv.setInput(new Object());
	}

	/**
	 * Searches for the open ViatraTreeEditor for a framework object.
	 * @param framework
	 *            the framework to search for
	 * @return the found editor, or null if none is found
	 */
	public ViatraTreeEditor getEditorPart(IFramework framework) {
		if (editor == null) {
			IWorkbenchPage page = PlatformUI.getWorkbench()
					.getActiveWorkbenchWindow().getActivePage();

			for (IEditorReference ref : page.getEditorReferences()) {
				IEditorPart editorRef = ref.getEditor(true);
				if (editorRef instanceof ViatraTreeEditor) {
					ViatraTreeEditor viatraEditor = (ViatraTreeEditor) editorRef;
					if (viatraEditor.getFramework().equals(framework))
						return viatraEditor;
				}
			}
		}
		return editor;
	}

	public void dispose() {
		FrameworkManager.getInstance().removeFrameworkListChangedListener(
				frameworkListener);
		if (framework != null) {
			framework.getTopmodel().getNotificationManager().removeAllListener(
					modelSpaceListener);
		}
	}

	public Object getAdapter(Class adapter) {
		if (adapter == org.eclipse.ui.views.properties.IPropertySheetPage.class
				&& framework != null) {
			ViatraTreeEditor editor = getEditorPart(framework);
			if (editor != null) {
				return editor.getAdapter(adapter);
			}
		}
		return null;
	}

	/**
	 * Redraws the corresponding graph elements, including a reapplication of
	 * the layout, and reapplies the selected components. To avoid reapplying
	 * the layout see {@link ModelSpaceVisualisationDescriptor#refreshGraph}.
	 */
	public void redrawGraph() {
		gv.setInput(new Object());
		// gv.applyLayout();
		// refreshGraph();
	}

	/**
	 * Redraws the graph elements without reapplying the layout. For reapplying
	 * the layout see {@link ModelSpaceVisualisationDescriptor#redrawGraph()}.
	 */
	public void refreshGraph() {
		gv.refresh();
		if (framework != null && modelManager != null && localSelection != null)
			gv.setSelection(localSelection);
	}

	public void createFormContributions(Composite parent) {
		initializeViewer(parent);

	}

	public void setLayoutAlgorithm(IViatraLayoutAlgorithm algorithm) {
		for (ViewerFilter filter : filters) {
			gv.removeFilter(filter);
		}
		gv.setLayoutAlgorithm(algorithm.getAlgorithm(), true);
		filters = algorithm.getFilters();
		for (ViewerFilter filter : filters) {
			gv.addFilter(filter);
		}
		gv.applyLayout();
	}
}
