/*******************************************************************************
 * Copyright (c) 2009 Zoltan Ujhelyi and Daniel Varro.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Zoltan Ujhelyi - initial API and implementation
 *******************************************************************************/
package org.eclipse.viatra2.visualisation.modelspace.actions.groups;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.TreeSet;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.ui.actions.ActionGroup;
import org.eclipse.viatra2.visualisation.layouts.IViatraLayoutAlgorithm;
import org.eclipse.viatra2.visualisation.modelspace.actions.SelectLayoutAction;
import org.eclipse.zest.layouts.LayoutAlgorithm;

/**
 * @author Zoltan Ujhelyi
 */
public class LayoutGroup extends ActionGroup {

	private static final String ExtensionID = "org.eclipse.viatra2.visualisation.layout";
	ArrayList<LayoutAlgorithm> algorithms;
	IViatraLayoutAlgorithm defaultAlgorithm;

	private class LayoutCategory {
		String ID;
		String name = "";
		String order = "";
		ArrayList<SelectLayoutAction> layouts = new ArrayList<SelectLayoutAction>();

		public LayoutCategory(String ID) {
			this.ID = ID;
			name = "";
			order = "";
		}
		
		public void setName(String name) {
			if (name != null) {
				this.name = name;
			}
		}
		
		public void setOrder(String order) {
			if (order != null) {
				this.order = order;
			}
		}
	}

	private class CategoryComparator implements Comparator<LayoutCategory> {

		public int compare(LayoutCategory o1, LayoutCategory o2) {
			String order1 = o1.order;
			String order2 = o2.order;
			int order = order1.compareTo(order2);
			if (order == 0) {
				order = o1.name.compareTo(o2.name);
			}
			if (order == 0) {
				order = o1.ID.compareTo(o2.ID);
			}
			return order;
		}

	}
	
	private Hashtable<String, LayoutCategory> categories = new Hashtable<String, LayoutCategory>();
	
	private LayoutCategory getCategory(String ID) {
		if (categories.containsKey(ID)) {
			return categories.get(ID);
		} else {
			LayoutCategory category = new LayoutCategory(ID);
			categories.put(ID, category);
			return category;
		}
	}

	public IViatraLayoutAlgorithm getDefaultLayout() {
		return defaultAlgorithm;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.ui.actions.ActionGroup#fillContextMenu(org.eclipse.jface.
	 * action.IMenuManager)
	 */
	@Override
	public void fillContextMenu(IMenuManager manager) {
		algorithms = new ArrayList<LayoutAlgorithm>();
		IMenuManager menu = new MenuManager("Layouts");
		manager.add(menu);
		IConfigurationElement[] config = Platform.getExtensionRegistry()
				.getConfigurationElementsFor(ExtensionID);
		for (IConfigurationElement element : config) {
			try {
				if (element.getName().contentEquals("category")) {
					LayoutCategory category = getCategory(element.getAttribute("id"));
					category.setName(element.getAttribute("name"));
					category.setOrder(element.getAttribute("orderID"));
				} else {
					String name = element.getAttribute("name");
					IViatraLayoutAlgorithm algorithm;
					algorithm = (IViatraLayoutAlgorithm) element
							.createExecutableExtension("layoutalgorithm");
					if (element.getAttribute("id").contentEquals(
							"org.eclipse.viatra2.visualisation.radiallayout")) {
						defaultAlgorithm = algorithm;
					}
					LayoutCategory category = getCategory(element.getAttribute("category"));
					
					SelectLayoutAction action = new SelectLayoutAction(name,
							algorithm);
					category.layouts.add(action);
				}
			} catch (CoreException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		SortedSet<LayoutCategory> sortedCategories = new TreeSet<LayoutCategory>(new CategoryComparator());
		for (LayoutCategory category : categories.values()) {
			if (category.layouts.size()>0) sortedCategories.add(category);
		}
		Iterator<LayoutCategory> iterator = sortedCategories.iterator();
		while (iterator.hasNext()) {
			LayoutCategory category = iterator.next();
			for (SelectLayoutAction action : category.layouts) {
				menu.add(action);
			}
			if (iterator.hasNext()) {
				menu.add(new Separator());
			}
		}
		/*
		 * int layoutStyle = LayoutStyles.NO_LAYOUT_NODE_RESIZING |
		 * LayoutStyles.ENFORCE_BOUNDS | ZestStyles.IGNORE_INVISIBLE_LAYOUT;
		 * radialAlgorithm = new RadialLayoutAlgorithm( layoutStyle);
		 * SelectLayoutAction radial = new SelectLayoutAction("Radial",
		 * radialAlgorithm); menu.add(radial); algorithms.add(radialAlgorithm);
		 * springAlgorithm = new SpringLayoutAlgorithm(layoutStyle);
		 * springAlgorithm.setRandom(true); SelectLayoutAction spring = new
		 * SelectLayoutAction("Spring", springAlgorithm); menu.add(spring);
		 * algorithms.add(springAlgorithm); GridLayoutAlgorithm gridAlgorithm =
		 * new GridLayoutAlgorithm(layoutStyle); SelectLayoutAction grid = new
		 * SelectLayoutAction("Grid", gridAlgorithm); menu.add(grid);
		 * algorithms.add(gridAlgorithm); CompositeLayoutAlgorithm
		 * compositeAlgorithm = new CompositeLayoutAlgorithm( layoutStyle, new
		 * LayoutAlgorithm[] { new DirectedGraphLayoutAlgorithm(layoutStyle),
		 * new HorizontalShift(layoutStyle) }); SelectLayoutAction composite =
		 * new SelectLayoutAction("Hierarchy", compositeAlgorithm);
		 * menu.add(composite); algorithms.add(compositeAlgorithm);
		 * SimulatedCooling coolingAlgorithm = new SimulatedCooling(
		 * layoutStyle, new ICriteria[] { new InBounds(), /*new
		 * PointDistribution(10),
		 */
		/*
		 * new ArcAngleDistribution(1000), new EdgeLength(0.0001), new
		 * EdgeIntersection(1000), new DistinctNodes(1000000) }, 15);
		 * SelectLayoutAction cooling = new SelectLayoutAction(
		 * "Simulated Cooling (Slow)", coolingAlgorithm); menu.add(cooling);
		 */
		/*
		 * CompositeLayoutAlgorithm compositeCooling = new
		 * CompositeLayoutAlgorithm( layoutStyle, new LayoutAlgorithm[] {
		 * gridAlgorithm, coolingAlgorithm }); SelectLayoutAction compCooling =
		 * new SelectLayoutAction("New Cooling", compositeCooling);
		 * menu.add(compCooling); algorithms.add(compositeCooling);
		 */
	}

	/**
	 * @return the algorithms
	 */
	public ArrayList<LayoutAlgorithm> getAlgorithms() {
		return algorithms;
	}
}
