/*******************************************************************************
 * Copyright (c) 2009 Zoltan Ujhelyi and Daniel Varro.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Zoltan Ujhelyi - initial API and implementation
 *******************************************************************************/
package org.eclipse.viatra2.visualisation.patterns.sources;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Machine;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.GTPatternCall;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTPattern;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTPatternBody;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.PatternVariable;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.Entity;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.ModelElement;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.Relation;
import org.eclipse.viatra2.visualisation.patterns.viewmodel.AbstractViewModelElement;
import org.eclipse.viatra2.visualisation.patterns.viewmodel.PatternTypes;
import org.eclipse.viatra2.visualisation.patterns.viewmodel.ViewParameterAssignment;
import org.eclipse.viatra2.visualisation.patterns.viewmodel.ViewPattern;
import org.eclipse.viatra2.visualisation.patterns.viewmodel.ViewPatternBody;
import org.eclipse.viatra2.visualisation.patterns.viewmodel.ViewPatternModelElement;
import org.eclipse.zest.core.viewers.IGraphEntityRelationshipContentProvider;
import org.eclipse.zest.core.viewers.INestedContentProvider;

/**
 * Content provider for pattern visualisation.
 * @author Zoltan Ujhelyi
 */
public abstract class PatternContentProvider implements
		IGraphEntityRelationshipContentProvider, INestedContentProvider {

	Machine machine;
	GTPattern pattern;
	Hashtable<String, ViewPatternModelElement> viewItems = new Hashtable<String, ViewPatternModelElement>();

	List<ViewParameterAssignment> loadPatternParameters(ViewPattern pattern,
			ViewPatternModelElement element) {
		ArrayList<ViewParameterAssignment> relations = new ArrayList<ViewParameterAssignment>();
		for (String result : pattern.getParametersName(element)) {
			ViewParameterAssignment assignment = new ViewParameterAssignment();
			assignment.setParamName(result);
			assignment.setParameter(element);
			assignment.setPattern(pattern);
			relations.add(assignment);
		}
		return relations;
	}

	List<AbstractViewModelElement> loadRelationsBetween(
			ViewPatternModelElement element1, ViewPatternModelElement element2) {
		ArrayList<AbstractViewModelElement> relations = new ArrayList<AbstractViewModelElement>();
		Entity entity1 = (Entity) element1.getModelElement();
		Entity entity2 = (Entity) element2.getModelElement();

		for (Relation relation : entity1.getRelationsFrom()) {
			if (relation.getTo().equals(entity2)) {
				ViewPatternModelElement viewRelation = new ViewPatternModelElement();
				viewRelation.setModelElement(relation);
				relations.add(viewRelation);
			}

		}
		return relations;
	}

	public Object[] getElements(Object inputElement) {
		if (machine != null && pattern == null) {
			// Building pattern list
			ArrayList<ViewPattern> patternList = new ArrayList<ViewPattern>();
			for (GTPattern pattern : machine.getGtPatternDefinitions()) {
				patternList.add(new ViewPattern(pattern,
						PatternTypes.CALLEDPATTERN));
			}
			return patternList.toArray();
		} else if (pattern != null) {
			// Building pattern bodies
			ArrayList<Object> elements = new ArrayList<Object>();
			int id = 1;
			for (GTPatternBody body : pattern.getPatternBodies()) {
				fillBody(body, id, elements);
				id++;
			}
			return elements.toArray();
		} else
			return null;
	}

	private void fillBody(GTPatternBody body, int id, List<Object> elements) {
		//viewItems = new Hashtable<String, ViewPatternModelElement>();
		ViewPatternBody viewModel = new ViewPatternBody();
		viewModel.setBody(body);
		viewModel.setId(id);
		elements.add(viewModel);
		fillPatternGraph(body.getPatternGraph(), viewModel, elements);
		fillPatternCalls(viewModel, body.getCalledPatterns(), elements,
				PatternTypes.CALLEDPATTERN);
		fillPatterns(viewModel, body.getLocalPatternDefinition(), elements,
				PatternTypes.LOCALPATTERN);
		fillPatternCalls(viewModel, body.getNegativePatterns(), elements,
				PatternTypes.NEGATIVEPATTERN);
	}

	abstract void fillPatternGraph(Entity patternGraph, ViewPatternBody body,
			List<Object> elements);

	void fillPatternCalls(ViewPatternBody body, EList<GTPatternCall> patternCalls,
			List<Object> elements, PatternTypes type) {
		for (GTPatternCall call : patternCalls) {
			fillPattern(body, elements, type, call.getCalledPattern());
		}
	}

	void fillPatterns(ViewPatternBody body, EList<GTPattern> patterns,
			List<Object> elements, PatternTypes type) {
		for (GTPattern pattern : patterns) {
			fillPattern(body, elements, type, pattern);
		}
	}

	/**
	 * @param body
	 * @param elements
	 * @param type
	 * @param pattern
	 */
	private void fillPattern(ViewPatternBody body, List<Object> elements,
			PatternTypes type, GTPattern pattern) {
		ViewPattern viewPattern = new ViewPattern(pattern, type);
		// Adding parameters
		for (PatternVariable symParam : pattern.getSymParameters()) {
			EList<ModelElement> patternElements = symParam
					.getElementInPattern();
			ViewPatternModelElement viewElement = viewItems
					.get((patternElements.get(0)).getName());
			if (viewElement != null) {
				viewPattern.addParameter(symParam.getName(), viewElement);
			}
		}
		body.addCalledPattern(viewPattern);
		elements.add(viewPattern);
	}

	public void dispose() {
	}

	public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
		if (newInput instanceof Machine) {
			this.machine = (Machine) newInput;
			this.pattern = null;
		} else if (newInput instanceof GTPattern) {
			this.pattern = (GTPattern) newInput;
		}

	}

	public Object[] getChildren(Object arg) {
		if (arg instanceof ViewPatternBody) {
			return ((ViewPatternBody) arg).getChildren();
		}
		return null;
	}

	public boolean hasChildren(Object arg) {
		return arg instanceof ViewPatternBody;
	}

}