/*******************************************************************************
 * Copyright (c) 2009 Zoltan Ujhelyi and Daniel Varro.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Zoltan Ujhelyi - initial API and implementation
 *******************************************************************************/
package org.eclipse.viatra2.visualisation.patterns.viewmodel;

import org.eclipse.emf.common.util.EList;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.Entity;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.ModelElement;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.Relation;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.Relationship;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.TypeOf;

/**
 * Represents a model element in the view model. There is a one-to-one
 * connection between ViewPatternModelElements and {@link ModelElement}
 * objects.q
 * @author Zoltan Ujhelyi
 */
public class ViewPatternModelElement extends AbstractViewModelElement {

	ModelElement element;
	ViewPatternBody body;

	/**
	 * @return the entity
	 */
	public ModelElement getModelElement() {
		return element;
	}

	/**
	 * @param entity
	 *            the entity to set
	 */
	public void setModelElement(ModelElement entity) {
		this.element = entity;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String name = element.getName();
		/*
		 * if (!element.getType().isEmpty()) { name = name.concat(" : "); name =
		 * name.concat(((ModelElement) element.getType().iterator()
		 * .next()).getTypeStr()); }
		 */
		return name;
	}

	/**
	 * Sets the pattern body that acts as a container of the model element.
	 * @param body
	 */
	public void setBody(ViewPatternBody body) {
		this.body = body;

	}

	/**
	 * Returns the pattern body that is the container of the model element.
	 * @return the body
	 */
	public ViewPatternBody getBody() {
		return body;
	}

	@Override
	public String getName() {
		return toString();
	}

	/**
	 * Decides whether the model element is a relation
	 * @return true if it is a relation
	 */
	public boolean isRelation() {
		return element instanceof Relation;
	}

	/**
	 * Decides whether the model element is an entity
	 * @return true if it is an entity
	 */
	public boolean isEntity() {
		return element instanceof Entity;
	}

	/**
	 * Returns the number of relations going from the current model element to a
	 * selected model element.
	 * @param target a target model element where the relations are calculated
	 * @return the number of outgoing relations
	 */
	public int getNumberOfRelationsFrom(ViewPatternModelElement target) {
		if (isEntity() && target.isRelation()) {
			EList relationsFrom = ((Entity) element).getRelationsFrom();
			if (relationsFrom.contains(target.element)) {
				return 1;
			}
		}
		return 0;
	}

	/**
	 * The number of incoming relations from a selected source model element.
	 * @param source the source model element
	 * @return the number of incoming relations
	 */
	public int getNumberOfRelationsTo(ViewPatternModelElement source) {
		if (isEntity() && source.isRelation()) {
			EList relationsFrom = ((Entity) element).getRelationsTo();
			if (relationsFrom.contains(source.element)) {
				return 1;
			}
		}
		return 0;
	}
	
	private String getTypeString(ModelElement element){
		for (Relationship rel : element.getSuperRelationships()) {
			if (rel instanceof TypeOf) {
				return rel.getSupplierStr();
			}
		}
		return "";
	}

	@Override
	public String getTypeString() {
		String string = getTypeString(element);
		if (string != "") {
			int pos = string.lastIndexOf(".");
			string = string.substring(pos + 1);
			return string;
		}
		if (isEntity())
			return "Entity";
		else if (isRelation()) return "Relation";
		return "Model Element";
	}

}
