/*******************************************************************************
 * Copyright (c) 2009 Zoltan Ujhelyi and Daniel Varro.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Zoltan Ujhelyi - initial API and implementation
 *******************************************************************************/
package org.eclipse.viatra2.visualisation.view;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.jface.viewers.IContentProvider;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.IViewPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.ViewPart;
import org.eclipse.viatra2.visualisation.Activator;
import org.eclipse.viatra2.visualisation.IVisualisationDescriptor;
import org.eclipse.viatra2.visualisation.ViatraGraphViewer;
import org.eclipse.viatra2.visualisation.common.action.SetAutoLayoutAction;
import org.eclipse.viatra2.visualisation.layouts.IViatraLayoutAlgorithm;
import org.eclipse.viatra2.visualisation.modelspace.actions.groups.LayoutGroup;
import org.eclipse.zest.core.viewers.AbstractZoomableViewer;
import org.eclipse.zest.core.viewers.IZoomableWorkbenchPart;
import org.eclipse.zest.core.viewers.ZoomContributionViewItem;

/**
 * This class acts as the common entry point of the visualisation, it handles
 * the user commands from the UI and alters the graph model respectively.
 * @author Zoltan Ujhelyi
 */
public class ViatraVisualisationView extends ViewPart implements
		IZoomableWorkbenchPart {

	/**
	 * The ID of the visualisation view
	 */
	public final static String viewId = "org.eclipse.viatra2.visualisation.view";
	ViatraGraphViewer gv;
	Composite parent;
	Object descriptorSource;
	IVisualisationDescriptor descriptor;
	Point gvSize;
	boolean autoLayout = true;

	// Action definitions
	private ZoomContributionViewItem zoom;
	private Action clearGraph = new Action("Clear Graph") {

		@Override
		public void run() {
			descriptor.clearGraphModel();
		}
	};
	private Action refreshGraph = new Action("Refresh Graph") {

		@Override
		public void run() {
			if (descriptor != null)
				gv.applyLayout();
		}

	};
	private Action saveFile = new Action("Save Graph as Image") {
		@Override
		public void run() {
			FileDialog save = new FileDialog(PlatformUI.getWorkbench()
					.getDisplay().getActiveShell(), SWT.SAVE);
			save.setFilterExtensions(new String[] { "*.png" });
			save.setFilterNames(new String[] { "PNG File" });
			save.setFileName("modelspace.png");
			String filename = save.open();
			gv.saveImage(filename, SWT.IMAGE_PNG);
		}
	};
	private SetAutoLayoutAction autoLayoutAction;
	
	/**
	 * If the view is resized refresh the graph
	 */
	private ControlListener resizeListener = new ControlAdapter() {

		/* (non-Javadoc)
		 * @see org.eclipse.swt.events.ControlAdapter#controlResized(org.eclipse.swt.events.ControlEvent)
		 */
		@Override
		public void controlResized(ControlEvent e) {
			if (descriptor != null && autoLayout) refresh();
		}
	};

	/**
	 * @param autoLayout the autoLayout to set
	 */
	public void setAutoLayout(boolean autoLayout) {
		this.autoLayout = autoLayout;
	}

	/**
	 * Searches and returns the descriptor related to its source. If no such
	 * descriptor is available, the return value is null.
	 * @param descriptorSource
	 *            the source object of the visualisation descriptor
	 * @return the related descriptor, or null if none is available
	 */
	public IVisualisationDescriptor getDescriptor(Object descriptorSource) {
		if (this.descriptorSource != null
				&& this.descriptorSource.equals(descriptorSource)) {
			return descriptor;
		} else
			return null;
	}

	/**
	 * Searches and returns the default descriptor.
	 * @return the default visualisation descriptor, or null if none is set.
	 */
	public IVisualisationDescriptor getDescriptor() {
		return descriptor;
	}

	/**
	 * Adds a descriptorSource-descriptor pair to the view.
	 * @param descriptorSource
	 *            the source object
	 * @param descriptor
	 *            the visualisation descriptor
	 */
	public void setDescriptor(Object descriptorSource,
			IVisualisationDescriptor descriptor) {
		this.descriptorSource = descriptorSource;
		if (this.descriptor != null) this.descriptor.dispose();
		this.descriptor = descriptor;
		loadVisualisationDescriptor(descriptor);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.part.WorkbenchPart#dispose()
	 */
	@Override
	public void dispose() {
		if (descriptor != null) descriptor.dispose();

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.ui.part.WorkbenchPart#createPartControl(org.eclipse.swt.widgets
	 * .Composite)
	 */
	@Override
	public void createPartControl(Composite parent) {
		this.parent = parent;
		if (descriptor != null) loadVisualisationDescriptor(descriptor);
		parent.addControlListener(resizeListener);
	}

	private void loadVisualisationDescriptor(IVisualisationDescriptor descriptor) {
		// removing old controls
		for (Control control : parent.getChildren()) {
			control.dispose();
		}
		if (zoom != null) zoom.dispose();
		this.descriptor = descriptor;
		initializeGraphViewer(parent);
		initializeActions();
		createMenu();
		createToolbar();
		parent.layout();
	}

	private void initializeActions() {
		clearGraph.setImageDescriptor(Activator.imageDescriptorFromPlugin(
				Activator.PLUGIN_ID, "icons/clear.gif"));
		refreshGraph.setImageDescriptor(Activator.imageDescriptorFromPlugin(
				Activator.PLUGIN_ID, "icons/refresh.gif"));
		saveFile.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages()
				.getImageDescriptor(ISharedImages.IMG_ETOOL_SAVE_EDIT));
		zoom = new ZoomContributionViewItem(this);
		autoLayoutAction = new SetAutoLayoutAction("Layout on resize");
		autoLayoutAction.setChecked(autoLayout);
	}

	/**
	 * Initializes the graph viewer component
	 * @param parent
	 */
	private void initializeGraphViewer(Composite parent) {
		descriptor.createFormContributions(parent);
		gv = descriptor.getViewer();
		// descriptor.clearGraphModel();
		getSite().setSelectionProvider(gv);
		gvSize = gv.getControl().getSize();
	}

	private void createToolbar() {
		IToolBarManager mgr = getViewSite().getActionBars().getToolBarManager();
		mgr.removeAll();
		descriptor.addToolbarContributions(mgr);
		mgr.add(refreshGraph);
		mgr.add(clearGraph);
		mgr.update(true);
	}

	/**
	 * Creates the menu of the view.
	 */
	private void createMenu() {
		IMenuManager mgr = getViewSite().getActionBars().getMenuManager();
		mgr.removeAll();
		descriptor.addMenuContributions(mgr);

		LayoutGroup layoutItems = new LayoutGroup();
		layoutItems.fillContextMenu(mgr);
		setLayoutAlgorithm(layoutItems.getDefaultLayout());

		// Adding common elements
		mgr.add(new Separator());
		mgr.add(saveFile);
		mgr.add(refreshGraph);
		mgr.add(autoLayoutAction);
		mgr.add(clearGraph);
		mgr.add(new Separator());

		mgr.add(zoom);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.part.WorkbenchPart#setFocus()
	 */
	@Override
	public void setFocus() {

	}

	/**
	 * Sets the model space provider for the graph visualisation
	 * @param provider
	 */
	public void setProvider(IContentProvider provider) {
		gv.setContentProvider(provider);
		// triggering a restructuration of the graph
		gv.setInput(new Object());
		gv.applyLayout();
		refresh();
	}

	/**
	 * Sets the layout algorithm for graph visualisation
	 * @param algorithm
	 */
	public void setLayoutAlgorithm(IViatraLayoutAlgorithm algorithm) {
		descriptor.setLayoutAlgorithm(algorithm);
		refresh();
	}

	/**
	 * Triggers a redraw of the visualisation with the currently set parameters
	 */
	public void refresh() {
		/*Point newSize = gv.getControl().getSize();
		if (gvSize.x == newSize.x && gvSize.y == newSize.y) {
			gv.refresh();
		} else {*/
			gv.applyLayout();
			//gv.refresh();
		//}
	}

	public AbstractZoomableViewer getZoomableViewer() {
		return gv;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.part.WorkbenchPart#getAdapter(java.lang.Class)
	 */
	@Override
	public Object getAdapter(Class adapter) {
		if (descriptor != null) {
			Object descriptorAdapter = descriptor.getAdapter(adapter);
			if (descriptorAdapter != null) return descriptorAdapter;
		}
		return super.getAdapter(adapter);
	}

	/**
	 * Opens or shows the instance of the Viatra Visualisation view
	 * @return a reference to the opened view allowing the caller to reach it.
	 */
	public static ViatraVisualisationView openView() {

		IWorkbenchPage page = PlatformUI.getWorkbench()
				.getActiveWorkbenchWindow().getActivePage();
		IViewPart form = page.findView(viewId);
		if (form == null) {
			try {
				page.showView(viewId);
			} catch (PartInitException e) {
			}
		} else
			page.bringToTop(form);

		return (ViatraVisualisationView) PlatformUI.getWorkbench()
				.getActiveWorkbenchWindow().getActivePage().findViewReference(
						viewId).getView(false);
	}

}
