/*******************************************************************************
 * Copyright (c) 2004-2008 Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Istvan Rath - initial API and implementation
 *******************************************************************************/

 package org.eclipse.viatra2.frameworkgui.content;

import org.eclipse.viatra2.frameworkgui.FrameworkGUIPlugin;
import org.eclipse.viatra2.frameworkgui.actions.AbstractFrameworkGUIAction;
import org.eclipse.viatra2.frameworkgui.actions.CheckAction;
import org.eclipse.viatra2.frameworkgui.actions.ClearTagsAction;
import org.eclipse.viatra2.frameworkgui.actions.LoaderAction;
import org.eclipse.viatra2.frameworkgui.actions.MergeFileAction;
import org.eclipse.viatra2.frameworkgui.actions.NativeImportAction;
import org.eclipse.viatra2.frameworkgui.actions.extend.ContributedActionContainer;
import org.eclipse.viatra2.frameworkgui.views.FrameworkTreeView;

import java.io.File;
import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.swt.graphics.Image;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.views.properties.IPropertySource2;
import org.eclipse.viatra2.framework.IFramework;
import org.eclipse.viatra2.framework.IFrameworkGlobalListener;
import org.eclipse.viatra2.imports.NativeImporterFactory;
import org.eclipse.viatra2.loaders.LoaderFactory;

public class FrameworkContainer implements ITreeParent, IFrameworkGlobalListener
{

	public static Image fwImage = FrameworkGUIPlugin.getImageDescriptor("icons/viatra_framework.png").createImage();

	
	
	public FrameworkContainer(IFramework f, FrameworkTreeView tv) 
	{
		treeview = tv;
		framework = f;
		f.addFrameworkGlobalListener(this);
		//runDialogEntries = new Vector<String>();
		extensions = new HashSet<IFrameworkContentProviderExtension>();
		native_import_actions = new HashSet<AbstractFrameworkGUIAction>();
		loader_actions = new HashSet<AbstractFrameworkGUIAction>();
		contributed_actions = new HashSet<AbstractFrameworkGUIAction>();
		makeActions();
		initializeExtensions();
	}
	
	protected Set<IFrameworkContentProviderExtension> extensions;
	
	protected IFramework framework;
	
	public IFramework getFramework()
	{
		return framework;
	}

	
	protected FrameworkTreeView treeview;
	
	
	
	protected void initializeExtensions()
	{
		for (IFrameworkContentProviderExtension ex : FrameworkGUIPlugin.getDefault().getContentProviders())
		{
			ex.init(treeview, framework, this);
			extensions.add(ex);
		}
	}
	
	public String getLabelText()
	{
		return framework.getId()+" ("+
			new File(framework.getCurrentFilename()).getName()+ ")";
	}
	
	public boolean hasChildren()
	{
		return !getChildren().isEmpty();
	}
	
	public Set<ITreeObject> getChildren()
	{
		HashSet<ITreeObject> ret = new HashSet<ITreeObject>();
		for (IFrameworkContentProviderExtension e : extensions)
		{
			ret.add(e.getTopLevelEntry());
		}
		return ret;
	}
	
	public void dispose()
	{
		for (IFrameworkContentProviderExtension e : extensions)
		{
			e.dispose();
		}
		framework.removeFrameworkGlobalListener(this);
	}

	public FrameworkContainer getFrameworkContainer() 
	{
		return this;
	}

	public Image getLabelImage() {
		return fwImage;
	}

	public ITreeParent getParent() { return null; }

	public IPropertySource2 getPropertySource() 
	{
		// TODO
		return null;
	}

	/* ACTIONS */
	
	private MergeFileAction mergeAction;

	private CheckAction checkAction;
	
	private ClearTagsAction clearAction;
	
	private void makeActions()
	{
		checkAction = new CheckAction(this.treeview);
		clearAction = new ClearTagsAction(this.treeview);
		mergeAction = new MergeFileAction(this.treeview);
		makeNativeImporterActions();
		makeLoaderActions();
		makeContributedActions();
	}
	
	private HashSet<AbstractFrameworkGUIAction> native_import_actions;
	
	private HashSet<AbstractFrameworkGUIAction> loader_actions;
	
	private HashSet<AbstractFrameworkGUIAction> contributed_actions;
	
	private void makeContributedActions()
	{
		for (ContributedActionContainer c : FrameworkGUIPlugin.getDefault().getContributedActions())
		{
			if (c.enablesFor.equals(IFramework.class))
			{
				c.action.setupInternals(this.treeview);
				contributed_actions.add(c.action);
			}
		}
	}
	
	private void makeNativeImporterActions() 
	{
		Map<String, NativeImporterFactory> hm = framework.getNativeImporters();
		AbstractFrameworkGUIAction a;
		for (NativeImporterFactory f : hm.values()) 
		{
			a = new NativeImportAction(this.treeview, f.getId());
			String ni = f.getImporterName();
			a.setText(ni);
			a.setToolTipText(ni);
			a.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages().getImageDescriptor(ISharedImages.IMG_OBJ_FILE));
			native_import_actions.add(a);
		}
	}

	private void makeLoaderActions() 
	{
		Map<String, LoaderFactory> hm = framework.getLoaders();
		AbstractFrameworkGUIAction a;
		for (LoaderFactory f : hm.values()) 
		{
			a = new LoaderAction(this.treeview, f.getId());
			String ni = f.getLoaderName();
			a.setText(ni);
			a.setToolTipText(ni);
			a.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages().getImageDescriptor(ISharedImages.IMG_OBJ_FILE));
			loader_actions.add(a);
		}
	}
	
	public Collection<IAction> getActions() 
	{
		Vector<IAction> v = new Vector<IAction>();
		v.add(mergeAction);
		v.add(checkAction);
		v.add(clearAction);
		return v;
	}
	
	public Collection<IContributionItem> getContributionItems() 
	{
		Vector<IContributionItem> v = new Vector<IContributionItem>();
		if (native_import_actions!=null && native_import_actions.size()>0)
		{
			MenuManager nativemgr = new MenuManager("Native importers");
			for (AbstractFrameworkGUIAction a : native_import_actions)
			{
				nativemgr.add(a);
			}
			v.add(nativemgr);
		}
		if (loader_actions!=null && loader_actions.size()>0)
		{
			MenuManager loadermgr = new MenuManager("Program loaders");
			for (AbstractFrameworkGUIAction a : loader_actions)
			{
				loadermgr.add(a);
			}
			v.add(loadermgr);
		}
		if (contributed_actions!=null && contributed_actions.size()>0)
		{
			MenuManager contrmgr = new MenuManager("Contributions");
			for (AbstractFrameworkGUIAction a : contributed_actions)
			{
				contrmgr.add(a);
			}
			v.add(contrmgr);
		}
		return v;
	}

	public IAction getDoubleClickAction() {
		return null;
	}

	public void frameworkNameChanged() {
		this.treeview.refreshViewer(this);
		
	}
	
	
	
}
