/*******************************************************************************
 * Copyright (c) 2004-2009 Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Istvan Rath - initial API and implementation
 *******************************************************************************/


package org.eclipse.viatra2.frameworkgui.content.transformation;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * Manager class to handle reference resolving management.
 * @author istvan rath
 *
 */
public class CrossReferenceManager {

	/**
	 * Queries whether a machine has any unresolved outgoing references.
	 * @param machineID
	 * @return
	 */
	public boolean hasUnresolvedReferences(String machineID) {
		return unresolved.keySet().contains(machineID);
	}
	
	/**
	 * Stores the map Machine->Set(unresolved_target_references).
	 */
	private Map<String, Set<String>> unresolved = new HashMap<String, Set<String>>();
	
	/**
	 * Stores machines referred by a machine.
	 */
	private Map<String, Set<String>> callgraph_forward = new HashMap<String, Set<String>>();
	
	/**
	 * Stores machines referring a machine.
	 */
	private Map<String, Set<String>> callgraph_backward = new HashMap<String, Set<String>>();

	
	/**
	 * Adds a machine reference to the registry.
	 * Also initializes the set of the caller. 
	 * @param callerID
	 * @param referredID
	 */
	public void addMachineReference(String callerID, String referredID)
	{
		addToGraph(callerID, referredID, callgraph_forward);
		addToGraph(referredID, callerID, callgraph_backward);
//		removeFromGraph(callerID, referredID, unresolved);
	}
	
	/*
	 * Lazily initializes the target sets.
	 * Performs reference edge addition.
	 */
	private void addToGraph(String key, String value, Map<String,Set<String>> map) {
		Set<String> target = map.get(key);
		if (target==null) {
			target = new HashSet<String>();
		}
		target.add(value);
		map.put(key, target);
	}
	
	/*
	 * Removes a reference from a graph.
	 */
//	private void removeFromGraph(String key, String value, Map<String, Set<String>> map) {
//		Set<String> _s = map.get(key);
//		if (_s!=null) {
//			_s.remove(value);
//		}
//	}
	
	/**
	 * Deletes all references belonging to the machine identified by callerID.
	 * @param callerID
	 */
	public void deleteMachineReferences(String callerID) {
//		Set<String> referenced_machines = callgraph_forward.get(callerID);
		// we're not allowed to do this!
//		if (referenced_machines!=null) {
//			for (String referenced_machine_id : referenced_machines) {
//				callgraph_backward.get(referenced_machine_id).remove(callerID);
//			}
//		}
		callgraph_forward.put(callerID, new HashSet<String>());
	}
	
	/**
	 * Registers all machines to be "unresolved" referenced the removed machine.
	 * @param removedMachineID
	 */
	public void checkUnresolvedReference(String removedMachineID) {
		Set<String> referencing_machines = callgraph_backward.get(removedMachineID);
		if (referencing_machines!=null) {
			for (String ref_m : referencing_machines) {
				Set<String> _t = unresolved.get(ref_m);
				if (_t == null) {
					_t = new HashSet<String>();
				}
				_t.add(removedMachineID);
				unresolved.put(ref_m, _t);
			}
		}
	}
	
	/**
	 * Updates the "resolved" status of machines which reference the newly added machine.
	 * @param machineAddedID
	 */
	public void checkResolvedReference(String machineAddedID) {
		Set<String> referencing_machines = callgraph_backward.get(machineAddedID);
		if (referencing_machines!=null) {
			for (String ref_m : referencing_machines) {
				// check whether ref_m can be cleared from "unresolved"
				Set<String> unresolved_targets = unresolved.get(ref_m);
				if (unresolved_targets!=null) {
					unresolved_targets.remove(machineAddedID);
					unresolved.put(ref_m, unresolved_targets);
				}
				if (unresolved_targets.size()==0) {
					unresolved.remove(ref_m);
				}
			}
		}
	}
}
