/*******************************************************************************
 * Copyright (c) 2008-2010, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *   Istvan Rath - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.frameworkgui.views.console.commands;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.viatra2.framework.IFramework;
import org.eclipse.viatra2.frameworkgui.FrameworkGUIPlugin;

/**
 * Utility class to execute VIATRA Console commands.
 * 
 * 
 * @author istvan rath
 *
 */
public class CommandExecutor {

	private static Map<String, List<String>> historyMap = new HashMap<String, List<String>>();
	
	public static List<String> getCommandHistory(IFramework fw){
		return historyMap.get(fw.getId());
	}
	
	public static void storeCommand(IFramework fw, String raw) {
		List<String> history = historyMap.get(fw.getId());
		if (history == null) {
			history = new ArrayList<String>();
			historyMap.put(fw.getId(), history);
		}
		history.add(raw);
	}
	
	public static String getCommandAtBackwardIndex(IFramework fw, int ix) {
		List<String> history = historyMap.get(fw.getId());
		if (history !=null) {
			try {
				return history.get(history.size()-ix);
			} catch (Throwable t) {
				return null;
			}
		}
		return null;
	}
	
	public static void clearCommandHistory(IFramework fw) {
		List<String> history = historyMap.get(fw.getId());
		if (history !=null) {
			history.clear();
		}
	}
	

	/**
	 * Gets a parsed command DTO for a raw string.
	 * 
	 * Accepts:
	 * 
	 * cmd_name(arg0, arg1, ...) type raw strings.
	 * 
	 * @param raw
	 * @return
	 */
	public static ParsedCommand getParsedCommand(String raw) {
		String cmd_name = "";
		ArrayList<String> params = new ArrayList<String>();
		
		if (raw.indexOf("(")>0 && raw.indexOf(")")>0) {
			// we have valid parametrization
			cmd_name = raw.substring(0, raw.indexOf("(")).trim().toLowerCase();
			
			try {
				Object[] raw_params = raw.substring((raw.indexOf("(") + 1), raw.lastIndexOf(")")).split(",");
				for (Object rp : raw_params) {
					params.add(rp.toString().trim());
				}
			}
			catch (Throwable t) { t.printStackTrace(); }
		}
		else {
			cmd_name = raw.trim().toLowerCase();
		}

		ParsedCommand c = new ParsedCommand();
		c.cmdName = cmd_name;
		c.params = params;
		c.isValid = true;
		
		// TODO add more parsing validation logic
		
		return c;
	}
	
	public static IVIATRAConsoleCommandProvider getCommandByName(IFramework fw, String name) {
		for (IVIATRAConsoleCommandProviderFactory f : FrameworkGUIPlugin.getDefault().getCommandProviderFactories())
		{
			for (IVIATRAConsoleCommandProvider cp : f.getProviders(fw)) {
				if (cp.getClass().getSimpleName().toLowerCase().equals(name)) {
					// found a match, invoke
					return cp;
				}
			}
		}
		return null;
	}
	
	public static List<IVIATRAConsoleCommandProvider> getCommandByNamePrefix(IFramework fw, String prefix) {
		ArrayList<IVIATRAConsoleCommandProvider> matches = new ArrayList<IVIATRAConsoleCommandProvider>();
		for (IVIATRAConsoleCommandProviderFactory f : FrameworkGUIPlugin.getDefault().getCommandProviderFactories())
		{
			for (IVIATRAConsoleCommandProvider cp : f.getProviders(fw)) {
				if (cp.getClass().getSimpleName().toLowerCase().startsWith(prefix)) {
					// found a match, push
					matches.add(cp);
				}
			}
		}
		return matches;
	}
	
	
	/**
	 * Try to parse the contents of cmd as a VIATRA Console command in the format
	 * command_name(argument0, argument1, ...) and execute it.
	 * 
	 * @return true if the execution was successful 
	 */
	public static boolean executeCommand(IFramework fw, ParsedCommand pcmd) {
			if (!pcmd.isValid) return false;
		// lookup by name
		boolean foundMatch = false;
		for (IVIATRAConsoleCommandProviderFactory f : FrameworkGUIPlugin.getDefault().getCommandProviderFactories())
		{
			for (IVIATRAConsoleCommandProvider cp : f.getProviders(fw)) {
				if (cp.getClass().getSimpleName().toLowerCase().equals(pcmd.cmdName)) {
					// found a match, invoke
					cp.executeCommand(fw, pcmd.params); 
					foundMatch = true;
					break;
				}
			}
			if (foundMatch) break;
		}
		return foundMatch;
	}
}
