/*******************************************************************************
 * Copyright (c) 2004-2008 Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Istvan Rath - initial API and implementation
 *******************************************************************************/

 package org.eclipse.viatra2.frameworkgui.wizards;

import java.lang.reflect.InvocationTargetException;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRunnable;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.codegen.jet.JETException;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.wizard.Wizard;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.INewWizard;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.wizards.newresource.BasicNewResourceWizard;
import org.eclipse.viatra2.frameworkgui.FrameworkGUIPlugin;
import org.eclipse.viatra2.frameworkgui.wizards.jet.JETGateway;
import org.eclipse.viatra2.frameworkgui.wizards.model.FileTemplateExtension;
import org.eclipse.viatra2.frameworkgui.wizards.ui.TemplateSelectionPage;
import org.eclipse.viatra2.frameworkgui.wizards.ui.VIATRANewFileCreationPage;

public abstract class NewVIATRAWizard extends Wizard implements INewWizard {

	String iKind;

	public NewVIATRAWizard(String aKind) {
		iKind = aKind;
	}

	@Override
	public boolean canFinish() {
		if (iMainPage!=null)
			return iMainPage.isPageComplete();
		else
			return false;
	}
	
	protected IStructuredSelection iSelection;

	private IWorkbench mWorkbench;


	public void init(IWorkbench workbench, IStructuredSelection selection) {
		iSelection = selection;
		mWorkbench = workbench;
		setWindowTitle("Create a new VIATRA2 " + iKind.toUpperCase() + " file");
	}

	protected VIATRANewFileCreationPage iMainPage;

	protected TemplateSelectionPage iSelectionPage;

	@Override
	public void addPages() {
		iMainPage = new VIATRANewFileCreationPage(iKind, iSelection);
		addPage(iMainPage);
		iSelectionPage = new TemplateSelectionPage(iKind);
		addPage(iSelectionPage);
	}
	
	IFile newFile;
	
	protected void finishPage(IProgressMonitor monitor)	throws InterruptedException, CoreException 
	{

		if (newFile != null)
		{
			if (iSelectionPage.getSelectedTemplate()!=null)
			{
				// get the selected template
				FileTemplateExtension template = iSelectionPage.getSelectedTemplate();
				JETGateway gw = new JETGateway(template);
				String result;
				try {
					result = gw.generate(monitor);
					gw.save(newFile, monitor, result.getBytes());
					
					// try to delete .JETEmitters project
					ResourcesPlugin.getWorkspace().getRoot().getProject(".JETEmitters").delete(true, true, monitor);
				} catch (JETException e) {
					FrameworkGUIPlugin.log(e);
				}
				catch (Exception e)
				{
					e.printStackTrace();
				}
			}
			selectAndReveal(newFile);
			openResource(newFile);
		}
	}

	/*
	 * @see Wizard#performFinish
	 */
	public boolean performFinish() {
		// Check main page completion
		if (!iMainPage.isPageComplete())
			return false;

		// Check selection page completion
		//if (!iSelectionPage.isPageComplete())
		//	return false;

		// Create new file
		String basefn = iMainPage.getFileName();
		if (!basefn.endsWith("." + iKind.toLowerCase()))
			iMainPage.setFileName(basefn + "." + iKind.toLowerCase());
		newFile = iMainPage.createNewFile();
		
		IWorkspaceRunnable op = new IWorkspaceRunnable() {
			public void run(IProgressMonitor monitor) throws CoreException,
					OperationCanceledException {
				try {
					finishPage(monitor);
				} catch (InterruptedException e) {
					throw new OperationCanceledException(e.getMessage());
				}
			}
		};

		try {
			getContainer().run(false, true, new WorkbenchRunnableAdapter(op));
		} catch (InvocationTargetException e) {
			handleFinishException(getShell(), e);
			return false;
		} catch (InterruptedException e) {
			return false;
		}
		return true;
	}

	/**
	 * An <code>IRunnableWithProgress</code> that adapts
	 * <code>IWorkspaceRunnable</code> so that it can be executed inside
	 * <code>IRunnableContext</code>.<code>OperationCanceledException</code>
	 * thrown by the apapted runnabled are caught and rethrown as a
	 * <code>InterruptedException</code>.
	 */
	private static class WorkbenchRunnableAdapter implements
			IRunnableWithProgress {

		private IWorkspaceRunnable fWorkspaceRunnable;

		public WorkbenchRunnableAdapter(IWorkspaceRunnable runnable) {
			fWorkspaceRunnable = runnable;
		}

		/*
		 * @see IRunnableWithProgress#run(IProgressMonitor)
		 */
		public void run(IProgressMonitor monitor)
				throws InvocationTargetException, InterruptedException {
			try {
				ResourcesPlugin.getWorkspace().run(fWorkspaceRunnable, monitor);
			} catch (OperationCanceledException e) {
				throw new InterruptedException(e.getMessage());
			} catch (CoreException e) {
				throw new InvocationTargetException(e);
			}
		}
	}

	protected void handleFinishException(Shell shell, InvocationTargetException e) {
		String exceptionMessage = "Error while generating the VIATRA2 file.";
		if (e.getCause() != null && e.getCause().getMessage() != null) {
			exceptionMessage = e.getCause().getMessage();
		} else if (e.getMessage() != null) {
			exceptionMessage = e.getMessage();
		}

		IStatus status = new Status(IStatus.ERROR, FrameworkGUIPlugin.getPluginId(), IStatus.ERROR, exceptionMessage, e);
		FrameworkGUIPlugin.log(status);

		String title = "Error"; //$NON-NLS-1$
		String message = "Error happened while invoking the JET engine."; //$NON-NLS-1$
		ErrorDialog.openError(shell, title, message, status);
	}

	protected void openResource(final IResource resource) {
		if (resource.getType() == IResource.FILE) {
			final IWorkbenchPage activePage = FrameworkGUIPlugin.getActivePage();
			if (activePage != null) {
				final Display display = getShell().getDisplay();
				if (display != null) {
					display.asyncExec(new Runnable() {
						public void run() {
							try {
								IDE.openEditor(activePage, (IFile) resource,
										true);
							} catch (PartInitException e) {
								FrameworkGUIPlugin.log(e);
							}
						}
					});
				}
			}
		}
	}

	protected void selectAndReveal(IResource newResource) {
		BasicNewResourceWizard.selectAndReveal(newResource, mWorkbench
				.getActiveWorkbenchWindow());
	}
}
