/*******************************************************************************
 * Copyright (c) 2004-2008 Peter Pasztor, Akos Horvath, Gergely Varro, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Peter Pasztor, Akos Horvath, Gergely Varro, Istvan Rath - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.gtasm.interpreter.impl.machine;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.viatra2.codegen.CodeOutputPlugin;
import org.eclipse.viatra2.errors.VPMRuntimeControlledException;
import org.eclipse.viatra2.errors.VPMRuntimeException;
import org.eclipse.viatra2.framework.IFramework;
import org.eclipse.viatra2.gtasm.interpreter.exception.ASMInterpreterErrorStrings;
import org.eclipse.viatra2.gtasm.interpreter.exception.ViatraTransformationException;
import org.eclipse.viatra2.gtasm.interpreter.executionEnvironment.ASMFunctionContent;
import org.eclipse.viatra2.gtasm.interpreter.executionEnvironment.IExecutionEnvironment;
import org.eclipse.viatra2.gtasm.interpreter.impl.executionEnvironment.ExecutionEnvironment;
import org.eclipse.viatra2.gtasm.interpreter.impl.rules.RuleInterpreter;
import org.eclipse.viatra2.gtasm.interpreter.term.rules.TermEvaluator;
import org.eclipse.viatra2.gtasm.patternmatcher.Scope;
import org.eclipse.viatra2.gtasm.patternmatcher.extension.ViatraGTASMPatternMatcherPlugin;
import org.eclipse.viatra2.gtasm.patternmatcher.patterns.PatternMatcherProvider;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.core.RuntimeAnnotation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.ASMFunction;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.InitialValue;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Machine;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Rule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Variable;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.enums.ValueKind;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.Term;
import org.eclipse.viatra2.interpreters.IProgressReport;
import org.eclipse.viatra2.logger.Logger;


/**
 * Initializes the interpreter. 
 * Will also initialize
 *  - the pattern matcher
 *  - the GT rule matcher
 * 
 * @author Peter Pasztor, Akos Horvath
 * 
 */
public class MachineInterpreter {

	protected Machine machine;
	// VIATRA Framework on which we are working
	protected IFramework frame;
	protected IExecutionEnvironment mainExecutionEnvironment;

	Logger log;
	CodeOutputPlugin codeout;

	public MachineInterpreter(IFramework f, Machine m) {
		Scope.DEFAULT_PARENT = f.getTopmodel().getModelManager().getRoot();
		frame = f;
		machine = m;
		log = frame.getLogger();
		codeout = frame.getCodeOutput();
		mainExecutionEnvironment = new ExecutionEnvironment(f);
		log.debug("GTASM Interpreter initialized.");

	}

	private void initASMFunctions(Machine input_machine) throws VPMRuntimeException{
		// TODO: In case of using multiple machines for a transformation it has to be
		// checked that the Module contains all invoked machine!
		List<Machine> machines = input_machine.getModule().getMachine();

		for (Object obj : machines) {
			Machine m = (Machine) obj;
			for (Object o : m.getAsmFunctionDefinitions()) {
				ASMFunction fun = (ASMFunction) o;
				HashMap<EList<Object>, Object> map = new HashMap<EList<Object>, Object>();
				for (Object o2 : fun.getInitialValues()) {
					InitialValue initialValue = (InitialValue) o2;
					try {
						EList<Object> locationsList = new BasicEList<Object>();
						for (Object o3 : initialValue.getLocations()) {
							Term locationsTerm = (Term) o3;
							locationsList.add(TermEvaluator.getInstance().evaluate(
											mainExecutionEnvironment,
											locationsTerm));
						}
						map.put(locationsList, TermEvaluator.getInstance().evaluate(
										mainExecutionEnvironment,
										initialValue.getValue()));

					} catch (ViatraTransformationException e) {
						// log.debug("Initialization of ASMFunctions failed!");
						VPMRuntimeException eNew = new VPMRuntimeException("Initialization of ASMFunctions failed: "+e.getMessage());
						throw eNew;		
					}
				}
				ASMFunctionContent.getInstance().put(fun, map);
			}
		}
	}

	private void initAllASMFunctions() throws VPMRuntimeException {
		for (Object obj : frame.getMachines()) {
			if (obj instanceof Machine)
				initASMFunctions((Machine) obj);
		}
	}

	public void initPatternMatcher() throws VPMRuntimeException {

		EList<RuntimeAnnotation> rtAnnotations = machine.getRuntimeAnnotations();
		if (rtAnnotations != null && rtAnnotations.size() > 0) {
			for (Object obj : rtAnnotations) {
				RuntimeAnnotation an = null;
				if (obj instanceof RuntimeAnnotation)
					an = (RuntimeAnnotation) obj;
				else
					throw new VPMRuntimeException(
							ASMInterpreterErrorStrings.ANNOTATION_ILLEGAL);
				if ("@incremental".equals(an.getAnnotationName().toLowerCase())) {
//					boolean parallel = false;
//					for (Object o2 : an.getElements()) {
//						if (o2 instanceof RuntimeAnnotationElement) {
//							RuntimeAnnotationElement rae = (RuntimeAnnotationElement) o2;
//							if ("parallel".equals(rae.getKey()))
//							{
//								String lcValue = rae.getValue().toLowerCase();
//								parallel = "true".equals(lcValue) || "parallel".equals(lcValue) || "1".equals(lcValue); 
//								break;
//							}
//						}
//					}					
					PatternMatcherProvider
						.setPatternMatcherIDToUse(ViatraGTASMPatternMatcherPlugin.INCREMENTAL_PM_FACTORY_ID);
					PatternMatcherProvider
						.setGTRuleMathcerIDToUse(ViatraGTASMPatternMatcherPlugin.ALTERNATIVE_PATTERN_MATCHER_FOR_GT_FACTORY_ID);
				} else if ("@localsearch".equals(an.getAnnotationName()
						.toLowerCase())) {
					PatternMatcherProvider
							.setPatternMatcherIDToUse(ViatraGTASMPatternMatcherPlugin.LOCAL_PM_FACTORY_ID);
					PatternMatcherProvider
							.setGTRuleMathcerIDToUse(ViatraGTASMPatternMatcherPlugin.LOCAL_GT_FACTORY_ID);
				}
			}
		}
	}

	public void operate(Map<String, Object> parameters, IProgressReport pr, Rule entryPoint)
			throws VPMRuntimeException {
//		if(machine.getMainRule() == null)	
//			throw new VPMRuntimeException(ASMInterpreterErrorStrings.MACHINE_WITHOUT_MAIN_RULE);
//		Istvan: this check has been moved to the {@link EMFInterpreter} class.
		
		log.debug("GTASM execution started");
		// Initialization
		try {
			// set logger and codeout for org.eclipse.viatra2.gtasm.interpreter.impl.interpreter classes
			initInterpreters();
			// Initialization: load the initial values of variables and ASM
			// functions in all Machines of the model space
			initAllASMFunctions();
			// modifies the default pattern matcher in case of specific
			// annotations
			initPatternMatcher();
			// Starting code output
			codeout.beginWork();
			// Call main rule

			try {
//				for (Object localVariable : machine.getMainRule()
				for (Object localVariable : entryPoint
						.getLocalVariables()) {
					if (parameters.containsKey(((Variable) localVariable)
							.getName())) {
						mainExecutionEnvironment.addVariable(
								(Variable) localVariable, parameters
										.get(((Variable) localVariable)
												.getName()));
					} else
						mainExecutionEnvironment.addVariable(
								(Variable) localVariable,
								ValueKind.UNDEF_LITERAL);
				}
			} catch (ViatraTransformationException e) {
				String[] context = {entryPoint.getFqn()};
				throw new ASMInterpreterException(
						ASMInterpreterErrorStrings.ENTRYPOINT_PARAMETERS_INITILAZIATION_FAILED
								+ e.getMessage()
								,context
								, entryPoint);
			}

			RuleInterpreter.getInstance().interpretRule(mainExecutionEnvironment, entryPoint, pr);
			clearASMFunctions();
			log.debug("GTASM execution finished.");
			
		} catch (ViatraTransformationException e) {
			log.message(Logger.ERROR, e.getMessage(), e);
			throw new VPMRuntimeControlledException("The transformation encountered an error: "
					+ e.getMessage(), e);
		} finally {
			PatternMatcherProvider.resetPatternMatcher();
			// Stopping code output
			codeout.endWork();
		}
	}

	private void clearASMFunctions() {
		ASMFunctionContent.getInstance().clear();
	}

	public void initInterpreters() {
		RuleInterpreter.initInterpreters(log, codeout);
	}
}
