/*******************************************************************************
 * Copyright (c) 2004-2008 Peter Pasztor, Akos Horvath, Gergely Varro, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Peter Pasztor, Akos Horvath, Gergely Varro, Istvan Rath - initial API and implementation
 *******************************************************************************/
package org.eclipse.viatra2.gtasm.interpreter.impl.rules;


import java.util.Hashtable;

import org.eclipse.emf.common.util.EList;
import org.eclipse.viatra2.gtasm.interpreter.exception.ASMInterpreterErrorStrings;
import org.eclipse.viatra2.gtasm.interpreter.exception.ViatraTransformationException;
import org.eclipse.viatra2.gtasm.interpreter.executionEnvironment.IExecutionEnvironment;
import org.eclipse.viatra2.gtasm.interpreter.extension.INativeASMRule;
import org.eclipse.viatra2.gtasm.interpreter.extension.NativeASMRuleProvider;
import org.eclipse.viatra2.gtasm.interpreter.impl.machine.ASMInterpreterException;
import org.eclipse.viatra2.gtasm.interpreter.term.rules.JavaNativeValue;
import org.eclipse.viatra2.gtasm.interpreter.term.rules.TermEvaluator;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.SymbolicRuleParameter;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Variable;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.enums.DirectionKind;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.simpleRules.CallRule;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.VariableReference;


public class NativeASMRuleInterpreter {
	
	private static NativeASMRuleInterpreter _instance = new NativeASMRuleInterpreter();
	
	public static NativeASMRuleInterpreter getInstance() {
		return _instance;
	}
	
	/** Executes the native ASM rule attached to the invoked ASM rule specified in the input parameter <i>callRule</i>
	 * @param executionEnvironment The execution environment of the invocation context
	 * @param symVarBindings The symmetric parameter binding
	 * @param callRule The call rule invoking an annotated AMS rule
	 * @return Returns true if the rule is executed without error
	 * @throws ViatraTransformationException All exceptions are boxed into a subtype of the ViatraTransfomrationException.
	 */
	public boolean executeNativeASMRule(
			IExecutionEnvironment executionEnvironment,
			Hashtable<Variable, Object> symVarBindings,
			CallRule callRule)
			throws ViatraTransformationException{
		
		//checks if the native rule is available at the org.eclipse.viatra2.gtasm.patternmatcher.extension point 
		INativeASMRule nativeRule = NativeASMRuleProvider.getNativeASMRule(callRule.getRule().getFqn());
		//checks if the rule is available
		if(nativeRule == null)
			{
			String[] context = {callRule.getRule().getFqn()};
			throw new ASMInterpreterException(
				ASMInterpreterErrorStrings.NATIVE_ASMRULE_DOES_NOT_EXIST
				,context 	
				,callRule);
			}
		
		boolean result = false;
		Hashtable<Variable, Variable> outVarBindings=new Hashtable<Variable, Variable>();
		EList<SymbolicRuleParameter> symParams = callRule.getRule().getSymParameters();
		
		try {
			result = nativeRule.invoke(executionEnvironment.getFramework().getTopmodel(), symVarBindings);
		} catch (ViatraTransformationException e) {
			throw e.addNewStackElement(callRule);
		} catch (Throwable t){
			String[] context = {callRule.getAsmRule().getNamespace().getFqn()+"."+callRule.getAsmRule().getFqn(),t.getMessage()};
			throw new ASMInterpreterException(ASMInterpreterErrorStrings.NATIVE_ASMRULE_EXECUTION_ERROR
					,context
					,callRule);
		}

		// If the rule failed, we have to restore the previous state. - so we don't have to write the values of the 
		// variables back to the current execution environment, and we have to (?) restore the ASMFunctions too.
		if(result)
		{
		//	symVarBindings=(Hashtable<Variable, Object>)callRuleExecutionEnvironment.onTerminate();
		
			// Getting the values back from the execution environment. Only for parameters with OUT or INOUT directions.
			for (int i=0;i<callRule.getActualParameters().size();i++) {
				if(!symParams.get(i).getDirection().equals(DirectionKind.IN_LITERAL))
				{
					Variable sym=symParams.get(i).getVariable();
					outVarBindings.put(sym,((VariableReference)callRule.getActualParameters().get(i)).getVariable());
				}	
			}	
		}
		
		// Putting the changed variables back to the original environment
		for (Object symbolicVariable : outVarBindings.keySet()) {
			try {
				executionEnvironment.setVariableValue(outVarBindings.get(symbolicVariable),
						TermEvaluator.getInstance().isASMNativeType(symVarBindings.get(symbolicVariable))?
								symVarBindings.get(symbolicVariable):
									new JavaNativeValue(symVarBindings.get(symbolicVariable)));
			} catch (ViatraTransformationException e) {
				String[] context = {((Variable)symbolicVariable).getName(),callRule.getAsmRule().getNamespace().getFqn()+"."+callRule.getAsmRule().getFqn(),e.getMessage()};
				throw new ASMInterpreterException(
						ASMInterpreterErrorStrings.SET_VAR_VALUE_AFTER_NATIVE_CALL
						,context
						,callRule);
			}
		}
		return result;
	}
}
