/*******************************************************************************
 * Copyright (c) 2004-2008 Peter Pasztor, Akos Horvath, Gergely Varro, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Peter Pasztor, Akos Horvath, Gergely Varro, Istvan Rath - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.gtasm.interpreter.term.rules;


import org.eclipse.viatra2.gtasm.interpreter.exception.ViatraTransformationException;
import org.eclipse.viatra2.gtasm.interpreter.executionEnvironment.IExecutionEnvironment;
import org.eclipse.viatra2.gtasm.interpreter.term.internal.TermInterpreterErrorString;
import org.eclipse.viatra2.gtasm.interpreter.term.internal.TermInterpreterException;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.enums.MultiplicityKind;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.enums.ValueKind;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.Term;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.builtInFunctions.ArithmeticOperation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.builtInFunctions.Division;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.builtInFunctions.Minus;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.builtInFunctions.Multiply;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.builtInFunctions.Plus;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.builtInFunctions.Remainder;


public class ArithmeticOperationEvaluator extends TermEvaluator {
	private static ArithmeticOperationEvaluator _instance = new ArithmeticOperationEvaluator();

	private ArithmeticOperationEvaluator() {
		super();
	}

	public static ArithmeticOperationEvaluator getInstance() {
		return _instance;
	}

	@Override
	public Object evaluate(IExecutionEnvironment executionEnvironment,
			Term termToBeEvaluated) throws ViatraTransformationException  {
		ArithmeticOperation arithmeticOperation = (ArithmeticOperation) termToBeEvaluated;

		if (arithmeticOperation.getActualParameters().size() != 2
				&& (!(arithmeticOperation.getActualParameters().size() == 1 && arithmeticOperation instanceof Minus)))
			{
			String[] context = {arithmeticOperation.getName()};
			throw new TermInterpreterException(TermInterpreterErrorString.PARAM_NUM_ARITHMETIC,
					context,arithmeticOperation);
			}

		// First we have to evaluate both parameters, and then perform the
		// operation.
		TermEvaluator termEvaluator = TermEvaluator.getInstance();
		Object op1 = termEvaluator.evaluate(executionEnvironment,
				(arithmeticOperation.getActualParameters().get(0)));
		if (ValueKind.UNDEF_LITERAL.equals(op1))
			return ValueKind.UNDEF_LITERAL;

		Object op2 = null;
		if (arithmeticOperation.getActualParameters().size() == 2) {
			op2 = termEvaluator.evaluate(executionEnvironment,
					(arithmeticOperation.getActualParameters().get(1)));
			if (ValueKind.UNDEF_LITERAL.equals(op2))
				return ValueKind.UNDEF_LITERAL;
		}

		if (arithmeticOperation instanceof Plus) {
			if (op1 instanceof String || op2 instanceof String) {
				return convertToString(op1) + convertToString(op2);
				// Only String allowed by now
				/*
				 * if(op2 instanceof String) return ((String) op1) +
				 * (op2.toString()); else throw new
				 * TermInterpreterException(TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_NO+"2",termToBeEvaluated);
				 */
			} else if (op1 instanceof Double) {
				// Double and Integer permitted
				if (op2 instanceof Double)
					return ((Double) op1) + ((Double) op2);
				else if (op2 instanceof Integer)
					return ((Double) op1)
							+ (Double.valueOf(((Integer) op2).doubleValue()));
				else
					{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
					throw new TermInterpreterException(
							TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_2,
							context,
							termToBeEvaluated);
					}
			} else if (op1 instanceof Integer) {
				if (op2 instanceof Double)
					return (Double.valueOf(((Integer) op1).doubleValue()
							+ ((Double) op2)));
				else if (op2 instanceof Integer)
					return ((Integer) op1) + ((Integer) op2);
				else
					{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
					throw new TermInterpreterException(
							TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_2,
							context,
							termToBeEvaluated);
					}
			} else if (op1 instanceof MultiplicityKind) {
				if (((MultiplicityKind.MANY_TO_MANY_LITERAL.equals(op1)) && (op2 instanceof MultiplicityKind))
						|| ((MultiplicityKind.MANY_TO_MANY_LITERAL.equals(op2)) && (op1 instanceof MultiplicityKind))
						|| ((MultiplicityKind.ONE_TO_MANY_LITERAL.equals(op1)) && (MultiplicityKind.MANY_TO_ONE_LITERAL
								.equals(op2)))
						|| ((MultiplicityKind.ONE_TO_MANY_LITERAL.equals(op2)) && (MultiplicityKind.MANY_TO_ONE_LITERAL
								.equals(op1))))
					return MultiplicityKind.MANY_TO_MANY_LITERAL;

				else if (((MultiplicityKind.ONE_TO_ONE_LITERAL.equals(op1)) && (MultiplicityKind.MANY_TO_ONE_LITERAL
						.equals(op2)))
						|| ((MultiplicityKind.ONE_TO_ONE_LITERAL.equals(op2)) && (MultiplicityKind.MANY_TO_ONE_LITERAL
								.equals(op1)))
						|| ((MultiplicityKind.MANY_TO_ONE_LITERAL.equals(op1)) && (MultiplicityKind.MANY_TO_ONE_LITERAL
								.equals(op2))))
					return MultiplicityKind.MANY_TO_ONE_LITERAL;
				else if (((MultiplicityKind.ONE_TO_ONE_LITERAL.equals(op1)) && (MultiplicityKind.ONE_TO_MANY_LITERAL
						.equals(op2)))
						|| ((MultiplicityKind.ONE_TO_ONE_LITERAL.equals(op2)) && (MultiplicityKind.ONE_TO_MANY_LITERAL
								.equals(op1)))
						|| ((MultiplicityKind.ONE_TO_MANY_LITERAL.equals(op1)) && (MultiplicityKind.ONE_TO_MANY_LITERAL
								.equals(op2))))
					return MultiplicityKind.ONE_TO_MANY_LITERAL;
				else if (((MultiplicityKind.ONE_TO_ONE_LITERAL.equals(op1)) && (MultiplicityKind.ONE_TO_ONE_LITERAL
						.equals(op2))))
					return MultiplicityKind.ONE_TO_ONE_LITERAL;
				else
					{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
					throw new TermInterpreterException(
							TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_2,
							context,
							termToBeEvaluated);
					}
			} else if (op1 instanceof Boolean)
				if (op2 instanceof Boolean)
					return (Boolean) op1 || (Boolean) op2;
				else
					{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
					throw new TermInterpreterException(
							TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_2,
							context,
							termToBeEvaluated);
					}
			else
				{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
				throw new TermInterpreterException(
						TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_1,
						context,
						termToBeEvaluated);
				}
		} else if (arithmeticOperation instanceof Minus) {
			if (op1 instanceof Double) { // a simple negative operator
				if (arithmeticOperation.getActualParameters().size() == 1)
					return ((Double) op1) * -1;
				// Double and Integer permitted
				if (op2 instanceof Double)
					return ((Double) op1) - ((Double) op2);
				else if (op2 instanceof Integer)
					return ((Double) op1)
							- (Double.valueOf(((Integer) op2).doubleValue()));
				else
					{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
					throw new TermInterpreterException(
							TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_2,
							context,
							termToBeEvaluated);
					}
			} else if (op1 instanceof Integer) { // a simple negative
													// operator
				if (arithmeticOperation.getActualParameters().size() == 1)
					return ((Integer) op1) * -1;
				if (op2 instanceof Double)
					return (Double.valueOf(((Integer) op1).doubleValue()
							- ((Double) op2)));
				else if (op2 instanceof Integer)
					return ((Integer) op1) - ((Integer) op2);
				else
					{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
					throw new TermInterpreterException(
							TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_2,
							context,
							termToBeEvaluated);
					}
			}
			// else if(op1 instanceof String)
			// if(op2 instanceof String)
			// {
			// // If it ends with the String op2, the operation cuts off the
			// ending.
			// if(((String)op1).endsWith((String)op2))
			// {
			// return
			// ((String)op1).substring(0,((String)op1).length()-((String)op2).length());
			// }
			// else return ValueKind.UNDEF_LITERAL;
			// }
			// else throw new
			// TermInterpreterException(TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_NO+"2",termToBeEvaluated);
			else
				{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
				throw new TermInterpreterException(
						TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_1,
						context,
						termToBeEvaluated);
				}
		} else if (arithmeticOperation instanceof Multiply) {

			if (op1 instanceof Double) {
				// Double and Integer permitted
				if (op2 instanceof Double)
					return ((Double) op1) * ((Double) op2);
				else if (op2 instanceof Integer)
					return ((Double) op1)
							* (Double.valueOf(((Integer) op2).doubleValue()));
				else
					{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
					throw new TermInterpreterException(
							TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_2,
							context,
							termToBeEvaluated);
					}
			} else if (op1 instanceof Integer) {
				if (op2 instanceof Double)
					return (Double.valueOf(((Integer) op1).doubleValue()
							* ((Double) op2)));
				else if (op2 instanceof Integer)
					return ((Integer) op1) * ((Integer) op2);
				else
					{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
					throw new TermInterpreterException(
							TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_2,
							context,
							termToBeEvaluated);
					}
			} else if (op1 instanceof MultiplicityKind) {
				if (((MultiplicityKind.ONE_TO_ONE_LITERAL.equals(op1)) && (op2 instanceof MultiplicityKind))
						|| ((MultiplicityKind.ONE_TO_ONE_LITERAL.equals(op2)) && (op1 instanceof MultiplicityKind))
						|| ((MultiplicityKind.ONE_TO_MANY_LITERAL.equals(op1)) && (MultiplicityKind.MANY_TO_ONE_LITERAL
								.equals(op2)))
						|| ((MultiplicityKind.ONE_TO_MANY_LITERAL.equals(op2)) && (MultiplicityKind.MANY_TO_ONE_LITERAL
								.equals(op1))))
					return MultiplicityKind.ONE_TO_ONE_LITERAL;

				else if (((MultiplicityKind.MANY_TO_MANY_LITERAL.equals(op1)) && (MultiplicityKind.ONE_TO_MANY_LITERAL
						.equals(op2)))
						|| ((MultiplicityKind.MANY_TO_MANY_LITERAL.equals(op2)) && (MultiplicityKind.ONE_TO_MANY_LITERAL
								.equals(op1)))
						|| ((MultiplicityKind.ONE_TO_MANY_LITERAL.equals(op1)) && (MultiplicityKind.ONE_TO_MANY_LITERAL
								.equals(op2))))
					return MultiplicityKind.ONE_TO_MANY_LITERAL;
				else if (((MultiplicityKind.MANY_TO_MANY_LITERAL.equals(op1)) && (MultiplicityKind.MANY_TO_ONE_LITERAL
						.equals(op2)))
						|| ((MultiplicityKind.MANY_TO_MANY_LITERAL.equals(op2)) && (MultiplicityKind.MANY_TO_ONE_LITERAL
								.equals(op1)))
						|| ((MultiplicityKind.MANY_TO_ONE_LITERAL.equals(op1)) && (MultiplicityKind.MANY_TO_ONE_LITERAL
								.equals(op2))))
					return MultiplicityKind.MANY_TO_ONE_LITERAL;
				else if (((MultiplicityKind.MANY_TO_MANY_LITERAL.equals(op1)) && (MultiplicityKind.MANY_TO_MANY_LITERAL
						.equals(op2))))
					return MultiplicityKind.MANY_TO_MANY_LITERAL;
				else
					{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
					throw new TermInterpreterException(
							TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_2,
							context,
							termToBeEvaluated);
					}
			} else if (op1 instanceof Boolean)
				if (op2 instanceof Boolean)
					return (Boolean) op1 && (Boolean) op2;
				else
					{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
					throw new TermInterpreterException(
							TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_2,
							context,
							termToBeEvaluated);
					}
			else
				{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
				throw new TermInterpreterException(
						TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_1,
						context,
						termToBeEvaluated);
				}
		} else if (arithmeticOperation instanceof Division) { // Not
																// implemented
																// for strings
			if (op1 instanceof Double) {
				// Double and Integer permitted
				if (op2 instanceof Double)
					return ((Double) op1) / ((Double) op2);
				else if (op2 instanceof Integer)
					return ((Double) op1)
							/ (Double.valueOf(((Integer) op2).doubleValue()));
				else
					{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
					throw new TermInterpreterException(
							TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_2,
							context,
							termToBeEvaluated);
					}
			} else if (op1 instanceof Integer) {
				if (op2 instanceof Double)
					return (Double.valueOf(((Integer) op1).doubleValue()
							/ ((Double) op2)));
				else if (op2 instanceof Integer)
					return ((Integer) op1) / ((Integer) op2);
				else
					{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
					throw new TermInterpreterException(
							TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_2,
							context,
							termToBeEvaluated);
					}
			} else
				{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
				throw new TermInterpreterException(
						TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_1,
						context,
						termToBeEvaluated);
				}

		} else if (arithmeticOperation instanceof Remainder) { // Only for
																// Integer!
			if (op1 instanceof Integer) {
				if (op2 instanceof Integer)
					return ((Integer) op1) % ((Integer) op2);
				else
				{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
				throw new TermInterpreterException(
						TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_2,
						context,
						termToBeEvaluated);
				}
			}
			// else if(op1 instanceof String)
			// if(op2 instanceof String)
			// {
			// // If it begins with the String op2, the operation cuts off the
			// beginning.
			//						
			// if(((String)op1).substring(0,
			// ((String)op2).length()).equals(op2))
			// {
			// return ((String)op1).substring(((String)op2).length());
			// }
			// else return op1;
			// }
			// else throw new
			// TermInterpreterException(TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_NO+"2",termToBeEvaluated);
			else
				{String[] context = {arithmeticOperation.getName(),convertToJavaType(op1), convertToJavaType(op2)};
				throw new TermInterpreterException(
						TermInterpreterErrorString.PARAM_NOT_SUITABLE_WITH_OPERATION_1,
						context,
						termToBeEvaluated);
				}

		} else
			{String[] context = {arithmeticOperation.getName()};
			throw new TermInterpreterException(
					TermInterpreterErrorString.UNIMP_ARITHMETIC,
					context,
					termToBeEvaluated);
			}
	}

	public String convertToString(Object o) {
		if (o instanceof MultiplicityKind) {
			if (MultiplicityKind.MANY_TO_MANY_LITERAL.equals(o))
				return "many to many";
			else if (MultiplicityKind.ONE_TO_MANY_LITERAL.equals(o))
				return "one to many";
			else if (MultiplicityKind.MANY_TO_ONE_LITERAL.equals(o))
				return "many to one";
			else if (MultiplicityKind.ONE_TO_ONE_LITERAL.equals(o))
				return "one to one";
		}

		return o.toString();
	}
}
