/*******************************************************************************
 * Copyright (c) 2004-2008 Akos Horvath, Gergely Varro and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Akos Horvath, Gergely Varro - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.gtasm.patternmatcher.impl.patternmatcher.internal.callgraph;

import org.eclipse.viatra2.gtasm.patternmatcher.exceptions.PatternMatcherCompileTimeException;
import org.eclipse.viatra2.gtasm.patternmatcher.exceptions.PatternMatcherRuntimeException;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.patternmatcher.internal.MatchingFrame;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.patternmatcher.internal.PatternMatcher;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.patternmatcher.internal.PatternMatcherErrorStrings;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.patternmatcher.internal.operation.SearchPlanOperation;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.patternmatcher.internal.rgg.RemoteGoal;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.patternmatcher.internal.rgg.Rule;
import org.eclipse.viatra2.gtasm.patternmatcher.impl.patternmatcher.term.ITermHandler;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.GTPatternCall;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTPattern;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTPatternBody;
import org.eclipse.viatra2.core.IModelManager;
import org.eclipse.viatra2.logger.Logger;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;


public class PatternNode extends OddLevelNode 
	implements Iterable<List<IFlattenedPatternElement>> {
    private boolean isRoot;
    protected PatternMatcher patternMatcher;
    protected EvenLevelNode[] children;
    // private boolean[] adornmentRestriction;
    
    protected PatternNode(BodyNode parent, GTPatternCall call){
    	super(parent,call);
    	isRoot = false;
    }
    
    public PatternNode(PatternMatcher patternMatcher, 
    		BodyNode parent, GTPatternCall call) throws PatternMatcherCompileTimeException {
        super(parent,call);
        this.isRoot = false;
        this.patternMatcher = patternMatcher;
        this.children = new BodyNode[getPattern().getPatternBodies().size()];
        buildCallGraph();
    }
    
    public PatternNode(PatternMatcher patternMatcher, GTPattern pattern) 
    		throws PatternMatcherCompileTimeException {
        super(null, pattern);
        this.isRoot = true;
        this.patternMatcher = patternMatcher;
        this.children = new BodyNode[pattern.getPatternBodies().size()];
        buildCallGraph();
    }
    
  private void buildCallGraph() throws PatternMatcherCompileTimeException {
    	// Call graph building
        List<GTPatternBody> bodies = getPattern().getPatternBodies();
    	for (int i = 0; i < bodies.size(); i++) {
			children[i] = new BodyNode(patternMatcher, this, bodies.get(i));
		}
    	// Flattening process
        if (isRoot) {
        	List<FlattenedPattern> flattenedPatterns = 
        		new ArrayList<FlattenedPattern>();
            int index = 0;
            for (List<IFlattenedPatternElement> patternVariant : this) {
                printPatternVariant(patternVariant, patternMatcher.getLogger());

                FlattenedPattern fp = 
                	new FlattenedPattern(this, patternVariant);
                flattenedPatterns.add(fp);
                index++;
            }
            // Sort the flattened patterns.
            // Collections.sort(flattenedPatterns);
            children = new FlattenedPattern[flattenedPatterns.size()];
            children = flattenedPatterns.toArray(children);
        }
    }
    
    protected PatternNode causesRecursion(GTPattern patternToTest) {
        if (getPattern() == patternToTest) {
            isRoot = true;
            return this;
        } else {
            return (getParent() == null ? null : getParent().causesRecursion(patternToTest));
        }
    }

    @Override
    protected boolean traverse(PatternVariantIterator token) {
        int size = children.length;
        assert 0 <= index && index <= size;
        if (index <= 0) {
            BodyNode body = (BodyNode) children[index];
            token.addLast(body);
            index++;
            body.index = 0;
            return body.traverse(token);
        } else {
            BodyNode body = (BodyNode) children[index-1];
            boolean successful = body.traverse(token);
            if (successful) {
                return true;
            } else {
                token.removeLast();
                if (index < size) {
                    body = (BodyNode) children[index];
                    index++;
                    token.addLast(body);
                    body.index = 0;
                    return body.traverse(token);
                } else {
                    return false;
                }
            }
        }
    }

    boolean isRoot() {
        return isRoot;
    }
    
    public Iterator<List<IFlattenedPatternElement>> iterator() {
        return new PatternVariantIterator(this);
    }

    private static void printPatternVariant(List<IFlattenedPatternElement> patternVariant, Logger logger) {
        for (IFlattenedPatternElement element : patternVariant) {
            logger.debug(element.toString());
        }
    }
    
    public ITermHandler getTermHandler() {
        return patternMatcher.getTermHandler();
    }
    
    public Logger getLogger() {
        return patternMatcher.getLogger();
    }

    public IModelManager getModelManager() {
    	return patternMatcher.getModelManager();
    }
    
    // Can be be moved to a descendant of PatternNode
    public FlattenedPattern getSingleFlattenedPattern() throws PatternMatcherCompileTimeException {
    	if (children.length == 1 && children[0] instanceof FlattenedPattern) {
    		return (FlattenedPattern) children[0];
    	} else {
    		String[] context = {};
    		throw new PatternMatcherCompileTimeException(
    				PatternMatcherErrorStrings.MULTIBODY_OR_RECURSIVE_PATTERN_ON_RHS
    				,context
    				,getPattern());
    	}
    }
   
    public FlattenedPattern[] getFlattenedPatterns() throws PatternMatcherCompileTimeException {
    	FlattenedPattern[] result = new FlattenedPattern[children.length];
    	for(int i = 0; i< children.length; i++){
    		if(children[i] instanceof FlattenedPattern)
    			result[i] = (FlattenedPattern) children[i];
//    		else
//    			{
//    			String[] context = {};
//    			throw new PatternMatcherCompileTimeException(
//    					PatternMatcherErrorStrings.);
//    			
//    			}
    	}
    	return result;
    }
    
    
    
    public String toString() {
    	return getPattern().getName();
    }
    
	public RemoteGoal buildRuleGoalGraph(Boolean[] adornment,
			Map<String, RemoteGoal> rggRoots) throws PatternMatcherRuntimeException {
		RemoteGoal result = new RemoteGoal(this, adornment);
		rggRoots.put(RemoteGoal.generateID(this, adornment), result);
		for (int i = 0; i < children.length; i++) {
			FlattenedPattern fp = (FlattenedPattern) children[i];
			List<SearchPlanOperation> finalSearchPlan = fp.generateSearchPlan(adornment,patternMatcher.getModelManager());
	    	SortedSet<Integer> preCallInterfaceVariables = new TreeSet<Integer>();
	    	for (int j = 0; j < adornment.length; j++) {
				preCallInterfaceVariables.add(j);
			}
			Rule rule = (Rule) fp.buildSubtree(rggRoots, result.getMagicSet(), finalSearchPlan, finalSearchPlan.size() - 1, preCallInterfaceVariables);
			// Rule rule = fp.buildRuleGoalGraph(adornment,rggRoots,result.getMagicSet());
			result.addChild(rule, new LinkedList<MatchingFrame>());
		}
		return result;
	}
}
