/*******************************************************************************
 * Copyright (c) 2004-2008 Gabor Bergmann and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Gabor Bergmann - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.gtasm.patternmatcher.incremental.adapters;

import java.lang.ref.WeakReference;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.eclipse.viatra2.core.ICoreNotificationListener;
import org.eclipse.viatra2.core.IEntity;
import org.eclipse.viatra2.core.IModelElement;
import org.eclipse.viatra2.core.IRelation;
import org.eclipse.viatra2.core.notification.ICoreNotificationObject;
import org.eclipse.viatra2.core.notification.ICoreNotificationObjectCreateEntity;
import org.eclipse.viatra2.core.notification.ICoreNotificationObjectCreateInstanceOf;
import org.eclipse.viatra2.core.notification.ICoreNotificationObjectCreateRelation;
import org.eclipse.viatra2.core.notification.ICoreNotificationObjectCreateSupertypeOf;
import org.eclipse.viatra2.core.notification.ICoreNotificationObjectDeleteContainment;
import org.eclipse.viatra2.core.notification.ICoreNotificationObjectDeleteEntity;
import org.eclipse.viatra2.core.notification.ICoreNotificationObjectDeleteInstanceOf;
import org.eclipse.viatra2.core.notification.ICoreNotificationObjectDeleteRelation;
import org.eclipse.viatra2.core.notification.ICoreNotificationObjectDeleteSupertypeOf;
import org.eclipse.viatra2.core.notification.ICoreNotificationObjectMoveTo;
import org.eclipse.viatra2.core.notification.ICoreNotificationObjectSetName;
import org.eclipse.viatra2.core.notification.ICoreNotificationObjectSetRelationFrom;
import org.eclipse.viatra2.core.notification.ICoreNotificationObjectSetRelationTo;
import org.eclipse.viatra2.core.notification.ICoreNotificationObjectSetValue;
import org.eclipse.viatra2.framework.IFramework;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.boundary.IManipulationListener;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.boundary.PredicateEvaluatorNode;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.boundary.ReteBoundary;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.matcher.ReteEngine;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.network.Direction;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.network.ReteContainer;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.network.Tunnel;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.remote.Address;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.tuple.FlatTuple;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.tuple.Tuple;

/**
 * Registering/unregistering is synchronized, as well as notification
 * processing. Reason: thread safety of Registering/unregistering and message
 * sending and action notifications
 * 
 * @author Gabor Bergmann
 * 
 */
public class GlobalModelspaceListener implements ICoreNotificationListener,
		IManipulationListener {

	protected WeakReference<IFramework> framework;
	protected ReteEngine<?> engine;
	protected ReteBoundary<?> boundary;
	protected ReteContainer headContainer;

	// TermEvaluatorNodes that have to be notified on changes affecting a given ModelElement
	protected Map<Object, Set<PredicateEvaluatorNode>> sensitiveTerms;

	/**
	 * Prerequisite: engine has its network, framework and boundary fields
	 * initialized
	 * 
	 * @param engine
	 */
	public GlobalModelspaceListener(ReteEngine<?> engine, IFramework framework) {
		super();
		this.engine = engine;
		this.boundary = engine.getBoundary();
		this.headContainer = engine.getReteNet().getHeadContainer();

		this.sensitiveTerms = new HashMap<Object, Set<PredicateEvaluatorNode>>();

		this.framework = new WeakReference<IFramework>(framework);
		framework.getTopmodel().getNotificationManager()
				.addAllListener(this);
		engine.addDisconnectable(this);
	}

	public synchronized void actionPerformed(
			ICoreNotificationObject notification) {
		//System.out.println("Core notification: "+notification.getActionType())
		// ;

		// if (notification.getActionType().equals(ICoreNotificationObject.
		// TA_TRANSACTION_BEGIN))
		// {
		// inTransaction = true;
		// return;
		// }
		// else if (notification.getActionType().equals(ICoreNotificationObject.
		// TA_UNDOABLE_TRANSACTION_BEGIN))
		// {
		// return;
		// }
		// else if (notification.getActionType().equals(ICoreNotificationObject.
		// TA_TRANSACTION_END))
		// {
		// inTransaction = false;
		// iTreeViewer.refresh();
		// iContentProvider.setDirty(); // notify editor that the model has
		// changed
		// return;
		// }
		// if (inTransaction)
		// return;
		//		 

		
		/*else*/ if (notification.getActionType().equals(
				ICoreNotificationObject.ACTION_SET_NAME)) {
			ICoreNotificationObjectSetName n = (ICoreNotificationObjectSetName) notification;
			notifyElementChange(n.getElement());

		} else if (notification.getActionType().equals(
				ICoreNotificationObject.ACTION_SET_VALUE)) {
			ICoreNotificationObjectSetValue n = (ICoreNotificationObjectSetValue) notification;
			notifyElementChange(n.getEntity());
		} else if (notification.getActionType().equals(
				ICoreNotificationObject.ACTION_CREATE_INSTANCEOF)) {
			ICoreNotificationObjectCreateInstanceOf n = (ICoreNotificationObjectCreateInstanceOf) notification;

			IModelElement instance = n.getInstance();
			if (instance instanceof IEntity)
				boundary.updateUnary(Direction.INSERT, instance, n.getType());
			else if (instance instanceof IRelation) {
				IRelation relation = (IRelation) instance;
				boundary.updateTernaryEdge(Direction.INSERT, relation, relation
						.getFrom(), relation.getTo(), n.getType());
			}
			boundary.updateInstantiation(Direction.INSERT, n.getType(), n
					.getInstance());
		} else if (notification.getActionType().equals(
				ICoreNotificationObject.ACTION_DELETE_INSTANCEOF)) {
			ICoreNotificationObjectDeleteInstanceOf n = (ICoreNotificationObjectDeleteInstanceOf) notification;

			IModelElement instance = n.getInstance();
			if (instance instanceof IEntity)
				boundary.updateUnary(Direction.REVOKE, instance, n.getType());
			else if (instance instanceof IRelation) {
				IRelation relation = (IRelation) instance;
				boundary.updateTernaryEdge(Direction.REVOKE, relation, relation
						.getFrom(), relation.getTo(), n.getType());
			}
			boundary.updateInstantiation(Direction.REVOKE, n.getType(), n
					.getInstance());
		} else if (notification.getActionType().equals(
				ICoreNotificationObject.ACTION_CREATE_SUPERTYPEOF)) {
			ICoreNotificationObjectCreateSupertypeOf n = (ICoreNotificationObjectCreateSupertypeOf) notification;

			IModelElement superElement = n.getSuper();
			IModelElement subElement = n.getSub();
			boundary.updateGeneralization(Direction.INSERT, superElement,
					subElement);
			if (superElement instanceof IEntity) {
				Address<? extends Tunnel> superRoot = boundary
						.getUnaryRoot(superElement);
				if (superRoot != null) // only if we have to keep track of the
										// supertype
				{
					Address<? extends Tunnel> subRoot = boundary
							.accessUnaryRoot(subElement);
					engine.getReteNet().connectRemoteNodes(subRoot, superRoot,
							true);
				}
			} else if (superElement instanceof IRelation) {
				Address<? extends Tunnel> superRoot = boundary
						.getTernaryEdgeRoot(superElement);
				if (superRoot != null) // only if we have to keep track of the
										// supertype
				{
					Address<? extends Tunnel> subRoot = boundary
							.accessTernaryEdgeRoot(subElement);
					engine.getReteNet().connectRemoteNodes(subRoot, superRoot,
							true);
				}
			}
		} else if (notification.getActionType().equals(
				ICoreNotificationObject.ACTION_DELETE_SUPERTYPEOF)) {
			ICoreNotificationObjectDeleteSupertypeOf n = (ICoreNotificationObjectDeleteSupertypeOf) notification;

			IModelElement superElement = n.getSuper();
			IModelElement subElement = n.getSub();
			boundary.updateGeneralization(Direction.REVOKE, superElement,
					subElement);
			if (superElement instanceof IEntity) {
				Address<? extends Tunnel> superRoot = boundary
						.getUnaryRoot(superElement);
				if (superRoot != null) // only if we have to keep track of the
										// supertype
				{
					Address<? extends Tunnel> subRoot = boundary
							.accessUnaryRoot(subElement);
					engine.getReteNet().disconnectRemoteNodes(subRoot,
							superRoot, true);
				}
			} else if (superElement instanceof IRelation) {
				Address<? extends Tunnel> superRoot = boundary
						.getTernaryEdgeRoot(superElement);
				if (superRoot != null) // only if we have to keep track of the
										// supertype
				{
					Address<? extends Tunnel> subRoot = boundary
							.accessTernaryEdgeRoot(subElement);
					engine.getReteNet().disconnectRemoteNodes(subRoot,
							superRoot, true);
				}
			}
		}
		// else if (notification.getActionType().equals(ICoreNotificationObject.
		// ACTION_SET_VIEW_INFO)) {
		// ICoreNotificationObjectSetViewInfo n =
		// (ICoreNotificationObjectSetViewInfo)notification;
		// iTreeViewer.update(n.getElement(), new String[]{"name"});
		// }
		else if (notification.getActionType().equals(
				ICoreNotificationObject.ACTION_CREATE_ENTITY)) {
			ICoreNotificationObjectCreateEntity n = (ICoreNotificationObjectCreateEntity) notification;

			IEntity entity = n.getCreated();
			// for (IModelElement type : entity.getTypes())
			// boundary.updateEntity(Direction.INSERT, entity,
			// type.getFullyQualifiedName());
			boundary.updateUnary(Direction.INSERT, entity, null);
			boundary.updateContainment(Direction.INSERT, n.getContainer(),
					entity);
		} else if (notification.getActionType().equals(
				ICoreNotificationObject.ACTION_CREATE_RELATION)) {
			ICoreNotificationObjectCreateRelation n = (ICoreNotificationObjectCreateRelation) notification;

			IRelation relation = n.getNewRelation();
			// for (IModelElement type : relation.getTypes())
			// boundary.updateRelation(Direction.INSERT, relation,
			// relation.getFrom(), relation.getTo(),
			// type.getFullyQualifiedName());
			boundary.updateTernaryEdge(Direction.INSERT, relation, n.getFrom(), n
					.getTo(), null);
		} else if (notification.getActionType().equals(
				ICoreNotificationObject.ACTION_MOVE_ELEMENT_TO)) {
			ICoreNotificationObjectMoveTo n = (ICoreNotificationObjectMoveTo) notification;

			IEntity element = n.getElement();
			boundary.updateContainment(Direction.REVOKE, n.getOldContainer(),
					element);
			boundary.updateContainment(Direction.INSERT, n.getNewContainer(),
					element);

			notifyElementChange(n.getElement());
		} else if (notification.getActionType().equals(
				ICoreNotificationObject.ACTION_DELETE_ENTITY)) {
			ICoreNotificationObjectDeleteEntity n = (ICoreNotificationObjectDeleteEntity) notification;

			IEntity entity = n.getDeleted();
			// for (IModelElement type : entity.getTypes())
			// boundary.updateEntity(Direction.REVOKE, entity,
			// type.getFullyQualifiedName());
			boundary.updateUnary(Direction.REVOKE, entity, null);
			// boundary.revokeRelation(n.getParent(), entity,
			// ReteBoundary.containmentLabel);
		} else if (notification.getActionType().equals(
				ICoreNotificationObject.ACTION_DELETE_RELATION)) {
			ICoreNotificationObjectDeleteRelation n = (ICoreNotificationObjectDeleteRelation) notification;

			IRelation relation = n.getDeleted();
			// for (IModelElement type : relation.getTypes())
			// boundary.updateRelation(Direction.REVOKE, relation,
			// relation.getFrom(), relation.getTo(),
			// type.getFullyQualifiedName());
			boundary.updateTernaryEdge(Direction.REVOKE, relation, n.getFrom(), n
					.getTo(), null);
		} else if (notification.getActionType().equals(
				ICoreNotificationObject.ACTION_DELETE_CONTAINMENT)) {
			ICoreNotificationObjectDeleteContainment n = (ICoreNotificationObjectDeleteContainment) notification;
			boundary.updateContainment(Direction.REVOKE, n.getParent(), n
					.getChild());
		} else if (notification.getActionType().equals(
				ICoreNotificationObject.ACTION_SET_RELATION_FROM)) {
			ICoreNotificationObjectSetRelationFrom n = (ICoreNotificationObjectSetRelationFrom) notification;

			IRelation relation = n.getRelation();
			for (IModelElement type : relation.getTypes()) {
				boundary.updateTernaryEdge(Direction.REVOKE, relation, n
						.getOldFrom(), relation.getTo(), type);
				boundary.updateTernaryEdge(Direction.INSERT, relation, n
						.getNewFrom(), relation.getTo(), type);
			}
			boundary.updateTernaryEdge(Direction.REVOKE, relation, n.getOldFrom(),
					relation.getTo(), null);
			boundary.updateTernaryEdge(Direction.INSERT, relation, n.getNewFrom(),
					relation.getTo(), null);
			notifyElementChange(relation);
		} else if (notification.getActionType().equals(
				ICoreNotificationObject.ACTION_SET_RELATION_TO)) {
			ICoreNotificationObjectSetRelationTo n = (ICoreNotificationObjectSetRelationTo) notification;

			IRelation relation = n.getRelation();
			for (IModelElement type : relation.getTypes()) {
				boundary.updateTernaryEdge(Direction.REVOKE, relation, relation
						.getFrom(), n.getOldTo(), type);
				boundary.updateTernaryEdge(Direction.INSERT, relation, relation
						.getFrom(), n.getNewTo(), type);
			}
			boundary.updateTernaryEdge(Direction.REVOKE, relation, relation
					.getFrom(), n.getOldTo(), null);
			boundary.updateTernaryEdge(Direction.INSERT, relation, relation
					.getFrom(), n.getNewTo(), null);
			notifyElementChange(relation);
		}
		// iContentProvider.setDirty(); // notify editor that the model has
		// changed

	}

	protected void notifyElementChange(Object element) {
		synchronized (sensitiveTerms) {
			Set<PredicateEvaluatorNode> nodes = sensitiveTerms.get(element);
			if (nodes != null) {
				Tuple tuple = new FlatTuple(element);
				for (PredicateEvaluatorNode node : nodes)
					boundary.notifyEvaluator(node.getElementChangeNotifier(), tuple);
			}
		}
	}

	public void registerSensitiveTerm(Object element, PredicateEvaluatorNode node) {
		synchronized (sensitiveTerms) {
			Set<PredicateEvaluatorNode> nodes = sensitiveTerms.get(element);
			if (nodes == null) {
				nodes = new HashSet<PredicateEvaluatorNode>();
				sensitiveTerms.put(element, nodes);
			}
			nodes.add(node);
		}
	}

	public void unregisterSensitiveTerm(Object element, PredicateEvaluatorNode node) {
		synchronized (sensitiveTerms) {
			Set<PredicateEvaluatorNode> nodes = sensitiveTerms.get(element);
			nodes.remove(node);
			if (nodes.isEmpty()) sensitiveTerms.remove(element);
		}
	}

	public int getListenerCategory() {
		// TODO Auto-generated method stub
		return 0;
	}

	public void disconnect() {
		framework.get().getTopmodel().getNotificationManager()
				.removeAllListener(this);
	}

}
