/*******************************************************************************
 * Copyright (c) 2004-2008 Gabor Bergmann and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Gabor Bergmann - initial API and implementation
 *******************************************************************************/

package org.eclipse.viatra2.gtasm.patternmatcher.incremental.simple;

import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.construction.RetePatternBuildException;
import org.eclipse.viatra2.gtasm.patternmatcher.incremental.rete.matcher.IPatternMatcherContext.EdgeInterpretation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.definitions.Variable;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.Constant;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.FunctionInvocation;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.GTPatternCall;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.ModelElementQuery;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.Term;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.asm.terms.VariableReference;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.gt.GTPatternBody;
import org.eclipse.viatra2.gtasmmodel.gtasm.metamodel.modelmanagement.queryFunctions.ElementReference;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.Entity;
import org.eclipse.viatra2.gtasmmodel.vpm.editmodel.Relation;



enum HyperEdgeClass {
	TYPE_UNARY, 
	TYPE_TERNARY_EDGE, // has an identity of itself
	TYPE_BINARY_EDGE, // no identity, arity=2
	CONTAINMENT_DIRECT, 
	CONTAINMENT_TRANSITIVE, 
	INSTANTIATION_DIRECT, 
	INSTANTIATION_TRANSITIVE, 
	GENERALIZATION_DIRECT, 
	GENERALIZATION_TRANSITIVE, 
	PATTERN_CALL, 
	NEGATIVE_PATTERN_CALL, 
	COUNT_PATTERN_CALL, 
	CHECK_EXPRESSION;

	boolean allTypeInfoRequired() // check only after all type information is asserted on the variables
	{
		return (this == CHECK_EXPRESSION);
	}
	
	boolean isJoinBased() // true if joins a Supplier specified under
							// supplierLabel
	{
		return (this != CHECK_EXPRESSION);
	}
	
	boolean touchSensitive() // true if ignores untouched nodes
	{
		return (this == NEGATIVE_PATTERN_CALL || this == COUNT_PATTERN_CALL);
	}
	
	static int[] iEmpty = {};
	static int[] iFirst = {0};
	static int[] iSecond = {1};
	static int[] iFirstTwo = {0, 1};
	static int[] iFirstThree = {0, 1, 2};
	/**
	 * @return index array containing all indices up to length
	 */
	public static int[] iAll(int length) {
		int[] all = new int[length];
		for (int i=0; i<length; i++) all[i] = i;
		return all;
	}
	public static int[] iOne(int position) {
		int[] one = {position};
		return one;
	}	
	public static int[] iAllButOne(int length, int position) {
		int[] all = new int[length-1];
		int k=0;
		for (int i=0; i<length; i++) if (i!=position) all[k++] = i;
		return all;
	}
}

/**
 * Hyperedge, actually.
 * 
 * @author Gabor Bergmann
 */
class PatternEdge implements Comparable<PatternEdge> {
	public PatternGraph pGraph;

	public Object origin;
	public Object extra;

	public PatternNode[] nodes;
	// public int weight = 0; // heuristical estimate of the join cost
	public Map<PatternNode, Integer> multiplicity; /*
													 * NOT PART OF KEY, will not
													 * be compared or hashed
													 */
	public Set<PatternNode> boundNodes; /*
										 * NOT PART OF KEY, will not be compared
										 * or hashed
										 */
	public Set<PatternNode> unboundNodes; /*
										 * NOT PART OF KEY, will not be compared
										 * or hashed
										 */
	public int degreeOfFreedom; /*
								 * pNodes having multiple instances in nodes[]
								 * are counted only once
								 *//*
									 * NOT PART OF KEY, will not be compared or
									 * hashed
									 */
	public int degreeOfFreedomWithMultiplicity; /*
												 * NOT PART OF KEY, will not be
												 * compared or hashed
												 */
	public int degreeOfBinding; /*
								 * NOT PART OF KEY, will not be compared or
								 * hashed
								 */
	public int degreeOfBindingWithMultiplicity; /*
												 * NOT PART OF KEY, will not be
												 * compared or hashed
												 */
	public int totalLingeringTypeInfoCounter; 	/*
												 * NOT PART OF KEY, will not be
												 * compared or hashed
												 */
	
	public Map<PatternNode, Integer> invertedIndex; /*
													 * NOT PART OF KEY, will not
													 * be compared or hashed
													 */

	/*
	 * note: for negative calls, untouched variables are NOT to be bound and
	 * will not be considered unbound, will not contribute to DoF
	 */

	public HyperEdgeClass edgeClass;
	public Object supplierKey;


	public PatternEdge(PatternGraph pGraph, Entity entity, Object typeObj) {
		this.pGraph = pGraph;
		origin = entity;
		extra = null;		
		this.supplierKey = typeObj;


		edgeClass = HyperEdgeClass.TYPE_UNARY;

		PatternNode[] nodes = { pGraph.getPNode(entity) };
		this.nodes = nodes;
	}

	public PatternEdge(PatternGraph pGraph, Relation rel, Object typeObj) {
		this.pGraph = pGraph;
		origin = rel;		
		extra = null;
		
		this.supplierKey = typeObj;
		if (pGraph.builder.context.edgeInterpretation() == EdgeInterpretation.TERNARY)
		{
			edgeClass = HyperEdgeClass.TYPE_TERNARY_EDGE;
			PatternNode[] nodes = { pGraph.getPNode(rel),
					pGraph.getPNode(rel.getFromStr()), pGraph.getPNode(rel.getToStr()) };
			this.nodes = nodes;
		} else {
			edgeClass = HyperEdgeClass.TYPE_BINARY_EDGE;
			PatternNode[] nodes = { 
					pGraph.getPNode(rel.getFromStr()), pGraph.getPNode(rel.getToStr()) };
			this.nodes = nodes;			
		}

	}

	// use only for Relationship constraints
	public PatternEdge(PatternGraph pGraph, PatternNode parent,
			PatternNode child, HyperEdgeClass relationshipClass) {
		this.pGraph = pGraph;
		origin = null;
		extra = null;

		edgeClass = relationshipClass;
		supplierKey = ""; /*
							 * transitive?
							 * ReteBoundary.containmentTransitiveLabel :
							 * ReteBoundary.containmentLabel;
							 */

		PatternNode[] nodes = { parent, child };
		this.nodes = nodes;

	}

	public PatternEdge(PatternGraph pGraph, GTPatternCall call, boolean negative)
			throws RetePatternBuildException { 
		this.pGraph = pGraph;
		origin = call;
		boolean count = call.getMatchCounter()!=null;
		if (!count)
		{
			edgeClass = negative ? HyperEdgeClass.NEGATIVE_PATTERN_CALL
				: HyperEdgeClass.PATTERN_CALL;
			extra = null;
		}
		else if (!negative) {
			edgeClass = HyperEdgeClass.COUNT_PATTERN_CALL;
			extra = call.getMatchCounter().getVariableReference().getVariable();
		}
		else {
			String[] context = {call.getCalledPattern().getName(), pGraph.gtPattern.getName(), 
					call.getMatchCounter().getVariableReference().getVariable().getName()};
			throw new RetePatternBuildException("Negative pattern calls cannot count matches: " 
					+ "incomprehensible negative call to pattern {1} in the body of pattern {2} with matches accumulated in {3}", 
					context, pGraph.gtPattern);
		}
		
		supplierKey = call.getCalledPattern();

		int paramNum = call.getActualParameters().size();
		this.nodes = new PatternNode[count? paramNum+1 : paramNum];
		for (int i = 0; i < paramNum; ++i) {
			Object actualParam = call.getActualParameters().get(i);
			if (actualParam instanceof VariableReference) {
				VariableReference paramVar = (VariableReference) actualParam;
				PatternNode paramNode = pGraph.getPNode(paramVar.getVariable());

				nodes[i] = paramNode;
			} else if (actualParam instanceof Constant) {
				Constant constant = (Constant) actualParam;
				PatternNode paramNode = pGraph.newConstant(constant);

				nodes[i] = paramNode;
			} else if (actualParam instanceof ElementReference) {
				ElementReference reference = (ElementReference) actualParam;
				PatternNode paramNode = pGraph.newConstant(reference);

				nodes[i] = paramNode;
			} else {
				Term term = (Term) actualParam;
				PatternNode paramNode = pGraph.substituteTerm(term);

				nodes[i] = paramNode;
			}
		}
		if (count) nodes[paramNum] = pGraph.getPNode((Variable)extra);

		// skip injectivity check if already performed in positive, injective pattern call
		if (!negative && !count && call.getCalledPattern().isDistinctMatching()) {
			boolean safe = true;
			for (Object o: call.getCalledPattern().getPatternBodies()) {
				safe = safe && ((GTPatternBody)o).getVariableAssignments().isEmpty();
			}
			if (safe) for (PatternNode pNode : nodes)
				pGraph.permittableEquality.get(pNode).addAll(
						Arrays.asList(nodes));
		}
		// skip injectivity check for scalars
		if (count) {
			nodes[nodes.length-1].makeExemptFromInjectivity();
		}

	}
	

	
	/**
	 * if equalsTo!=null check that eval(expression) = equalsTo, if
	 * equalsTo==null check that eval(expression) = true
	 * 
	 * @throws RetePatternBuildException
	 */
	public PatternEdge(PatternGraph pGraph, Term expression,
			PatternNode equalsTo) throws RetePatternBuildException {
		this.pGraph = pGraph;
		origin = expression;
		extra = equalsTo;

		edgeClass = HyperEdgeClass.CHECK_EXPRESSION;
		supplierKey = expression;

		// extracting affected variables with a depth-first-search method
		HashSet<PatternNode> affected = new HashSet<PatternNode>();
		LinkedList<Term> termQueue = new LinkedList<Term>();
		termQueue.add(expression);
		while (termQueue.size() > 0) {
			Term term = termQueue.removeFirst();

			if (term instanceof VariableReference) {
				VariableReference termVar = (VariableReference) term;
				PatternNode termNode = pGraph.getPNode(termVar.getVariable());
				affected.add(termNode);
			} else if (term instanceof FunctionInvocation) {
				FunctionInvocation fi = (FunctionInvocation) term;
				for (Object o : fi.getActualParameters()) {
					termQueue.addFirst((Term) o);
				}
			} else if (term instanceof ModelElementQuery) {
				ModelElementQuery meq = (ModelElementQuery) term;
				termQueue.addFirst(meq.getArgument());
			} else if (term instanceof GTPatternCall) {
				GTPatternCall pc = (GTPatternCall) term;
				
				String[] errorData = {pc.getCalledPattern().getFqn(), pGraph.gtPattern.getFqn()};
				String message = "Offending call to pattern {1} in enclosing pattern {2}. The incremental pattern matcher does not support GT Pattern calls in terms embedded into GT Patterns; please use a 'find' clause directly in the pattern body.";
				//pGraph.builder.logger.error(message);
				throw new RetePatternBuildException(message, errorData, term);

//				If we did want to handle this case
//				for (Object o : pc.getActualParameters()) {
//					termQueue.addFirst((Term) o);
//				}
			}
		}
		if (equalsTo != null)
			affected.add(equalsTo);
		int nodeNum = affected.size();
		this.nodes = new PatternNode[nodeNum];
		for (PatternNode pNode : affected) {
			this.nodes[--nodeNum] = pNode;
		}
	}

	// void findWeight() throws RetePatternBuildException {
	// Supplier mySupplier = extractSupplier();
	// if (mySupplier != null) {
	// Bag b = new Bag();
	// mySupplier.pullInto(b);
	// pGraph.builder.reteNet.flushUpdates();
	//			
	// weight = b.contents.size();
	// }
	// }

	public void touchQuantifiedNodes() {
		for (int i : touchingIndices())
			nodes[i].isTouched = true;	
	}
	public void connectNodes() {
		boundNodes = new HashSet<PatternNode>();
		unboundNodes = new HashSet<PatternNode>();
		multiplicity = new HashMap<PatternNode, Integer>();
		degreeOfFreedom = 0; // will be counted soon
		degreeOfFreedomWithMultiplicity = 0; // will be counted soon
		degreeOfBinding = 0;
		degreeOfBindingWithMultiplicity = 0;
		invertedIndex = new HashMap<PatternNode, Integer>();

		for (int i = 0; i < nodes.length; ++i) {
			PatternNode node = nodes[i];
			invertedIndex.put(node, i);

			pGraph.edgeList.get(node).add(this);

			if (!multiplicity.containsKey(node))
				multiplicity.put(node, 0);
			multiplicity.put(node, multiplicity.get(node) + 1);

			// if (!pGraph.bindedNodes.contains(node))
			// {
			// unboundNodes.add(node);
			// degreeOfFreedom++;
			// }
			if (!edgeClass.touchSensitive() || node.isTouched) {
				degreeOfFreedomWithMultiplicity++;
				if (unboundNodes.add(node))
					degreeOfFreedom++;
			}
		}
	}

	public void bindNode(PatternNode node) {
		if (unboundNodes.contains(node)) {
			boundNodes.add(node);
			unboundNodes.remove(node);
			degreeOfFreedom--;
			degreeOfFreedomWithMultiplicity -= multiplicity.get(node);
			degreeOfBinding++;
			degreeOfBindingWithMultiplicity += multiplicity.get(node);
		}
	}

	/**
	 * an edge is bound if all its node variables are bound
	 */
	public boolean isBound() {
		return (degreeOfFreedom == 0);
	}
	/**
	 * @return true if all type information on all nodes has been expressed
	 */
	private boolean isFullyTyped() {
		return (totalLingeringTypeInfoCounter == 0);
	}
	
	/**
	 * unsafe to be checked for some reason
	 */
	public boolean isUnsafe() {
		return isUnsafeQuantification() || isUnsafeTypeless();
	}
	
	/**
	 * negative patterns or check expressions that are not completely bound are
	 * not to be checked
	 */
	public boolean isUnsafeQuantification() {
		for (int i : bindRequiredIndices())
			if (unboundNodes.contains(nodes[i]))
				return true;
		return false;
	}
	/**
	 * check expressions are not to be checked before without matching the types of nodes 
	 */
	public boolean isUnsafeTypeless() {
		return (edgeClass.allTypeInfoRequired() && !isFullyTyped());
	}


	/**
	 * true if a standard DualInputNode-based checker is required
	 */
	public boolean isJoinBased() {
		return edgeClass.isJoinBased();
	}
	
	// returns the indices of touched (quantifies) nodes
	int[] touchingIndices() 
	{
		switch(edgeClass) {
		case CHECK_EXPRESSION: // fall-through
		case NEGATIVE_PATTERN_CALL: return HyperEdgeClass.iEmpty;
		case COUNT_PATTERN_CALL: return HyperEdgeClass.iOne(nodes.length - 1);
		default: return HyperEdgeClass.iAll(nodes.length);
		}
	}
	
	int[] bindRequiredIndices()
	{
		switch(edgeClass) {
		case CHECK_EXPRESSION: // fall-through
		case NEGATIVE_PATTERN_CALL: return HyperEdgeClass.iAll(nodes.length);
		case COUNT_PATTERN_CALL: return HyperEdgeClass.iAll(nodes.length - 1);
		default: return HyperEdgeClass.iEmpty;
		}	
	}
	
	/**
	 * Indices for which type information is conveyed
	 * @return array containing indices of type-restricted nodes in this.nodes; null means ALL
	 */
	public int[] conveysTypeInformationIndices() 
	{
		switch(edgeClass) {
		case TYPE_UNARY: 
		case TYPE_BINARY_EDGE: 
		case TYPE_TERNARY_EDGE: 
		case PATTERN_CALL: return HyperEdgeClass.iAll(nodes.length);
		case INSTANTIATION_DIRECT:
		case INSTANTIATION_TRANSITIVE: return HyperEdgeClass.iSecond;
		default: return HyperEdgeClass.iEmpty;
		}
	}

	@Override
	public boolean equals(Object arg0) {
		if (arg0 == null)
			return false;
		if (arg0 instanceof PatternEdge) {
			PatternEdge other = (PatternEdge) arg0;
			if (edgeClass != other.edgeClass)
				return false;
			if (origin != null)
				return origin.equals(other.origin);
			return (nodes.equals(other.nodes) && supplierKey
					.equals(other.supplierKey));
		} else
			return false;
	}

	@SuppressWarnings("unchecked")
	public int compareTo(PatternEdge other) {

		if (isUnsafe() ^ other.isUnsafe()) {
			// exactly one of them is unsafe, safe edge is preferred
			if (isUnsafe())
				return 1;
			else
				return -1;
		}
		// both edges safe or both unsafe
		if (isBound() ^ other.isBound()) {
			// exactly one of them is bound, bound edge is preferred
			if (isBound())
				return -1;
			else
				return +1;
		}
		// both edges bound or both unbound
		if (degreeOfBinding != other.degreeOfBinding)
			return -1 * (degreeOfBinding - other.degreeOfBinding); // edges with
																	// more
																	// bound
																	// variables
																	// are
																	// preferred
			// if (weight != other.weight)
			// return (weight - other.weight); // edges with smaller weight are
			// preferred
		if (degreeOfFreedom != other.degreeOfFreedom)
			return degreeOfFreedom - other.degreeOfFreedom; // lower DOF
															// preferred
		if (nodes.length != other.nodes.length)
			return -1 * (nodes.length - other.nodes.length);
		for (int i = 0; i < nodes.length; ++i)
			if (!nodes[i].equals(other.nodes[i]))
				return nodes[i].compareTo(other.nodes[i]);
		if (edgeClass != other.edgeClass)
			return edgeClass.ordinal() - other.edgeClass.ordinal();
		if ((extra == null) ^ (other.extra == null)) {
			// exactly one of them is null, null is preferred
			if (extra == null)
				return -1;
			else
				return +1;
		}
		if (extra != null) {
			if (!extra.toString().equals(other.extra.toString()))
				return extra.toString().compareTo(other.extra.toString());
		}
		if ((supplierKey == null) ^ (other.supplierKey == null)) {
			// exactly one of them is null, null is preferred
			if (supplierKey == null)
				return -1;
			else
				return +1;
		}
		if (supplierKey != null) {
			if (supplierKey instanceof Comparable)
				return ((Comparable<Object>) supplierKey)
						.compareTo(other.supplierKey);
			else
				return supplierKey.toString().compareTo(
						other.supplierKey.toString());
		}
		return 0; // both are null

	}




//	public Address<? extends Supplier> extractSupplier()
//			throws RetePatternBuildException {
//		Address<? extends Supplier> mySupplier;
//
//		switch (edgeClass) {
//
//		case TYPE_UNARY:
//			mySupplier = pGraph.builder.boundary
//					.accessEntityRoot(supplierKey);
//			break;
//
//		case TYPE_TERNARY_EDGE:
//			mySupplier = pGraph.builder.boundary
//					.accessRelationRoot(supplierKey);
//			break;
//		case TYPE_BINARY_EDGE:
//			mySupplier = pGraph.builder.boundary
//					.accessReferenceRoot(supplierKey);
//			break;			
//		case CONTAINMENT_DIRECT:
//			mySupplier = pGraph.builder.boundary.accessContainmentRoot();
//			break;
//		case CONTAINMENT_TRANSITIVE:
//			mySupplier = pGraph.builder.boundary
//					.accessContainmentTransitiveRoot();
//			break;
//		case GENERALIZATION_DIRECT:
//			mySupplier = pGraph.builder.boundary.accessGeneralizationRoot();
//			break;
//		case GENERALIZATION_TRANSITIVE:
//			mySupplier = pGraph.builder.boundary
//					.accessGeneralizationTransitiveRoot();
//			break;
//		case INSTANTIATION_DIRECT:
//			mySupplier = pGraph.builder.boundary.accessInstantiationRoot();
//			break;
//		case INSTANTIATION_TRANSITIVE:
//			mySupplier = pGraph.builder.boundary
//					.accessInstantiationTransitiveRoot();
//			break;
//
//		case NEGATIVE_PATTERN_CALL: // fall-through intentional
//		case PATTERN_CALL:
//			mySupplier = pGraph.builder.boundary
//					.accessProduction((GTPattern) supplierKey);
//			break;
//
//		case CHECK_EXPRESSION:
//			mySupplier = null;
//			break;
//		default:
//			pGraph.builder.logger.error("No such HyperEdgeClass"
//					+ edgeClass.toString());
//			throw new RetePatternBuildException();
//		}
//
//		return mySupplier;
//	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String s = "PatternEdge[" + edgeClass + "](";
		for (PatternNode n : nodes)
			s += n.toString() + ";";
		s += ")";
		return s;
	}

}
